/**
 * A Mega644P Local DHCP Base Node Final
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2A/PCINT31)
 *							-----------
 */

/* Global/System Constants */
#define TRUE 					1
#define FALSE 					0

#define CPU_FREQUENCY_1_MHZ		1000000				// 1.8-5.5V
#define CPU_FREQUENCY_2_MHZ		2000000				// 1.8-5.5V
#define CPU_FREQUENCY_4_MHZ		4000000				// 1.8-5.5V
#define CPU_FREQUENCY_8_MHZ		8000000				// 2.7-5.5V
#define CPU_FREQUENCY_16_MHZ	16000000			// 4.0-5.5V
#define CPU_FREQUENCY_HZ		CPU_FREQUENCY_8_MHZ

#define sbi(data,bit)			(data |= (1<<bit))	// Set Bit
#define cbi(data,bit) 			(data &= ~(1<<bit))	// Clear Bit
#define gbi(data,bit)			(data & (1<<bit))	// Get Bit
#define tbi(data,bit)			(data ^= (1<<bit))	// Toggle Bit

/* Node Constants */
#define SENTINEL_BASE_NODE		0x00
#define SENTINEL_RELAY_NODE		0x01
#define SENTINEL_SENSOR_NODE	0x02
#define SENTINEL_NODE_TYPE		SENTINEL_BASE_NODE	// Identify this node

/* Debug Constants */
#define PROTOBOARD_LED_DEBUG			TRUE	// Whether to enable prototype LED debug (availabe on rev 1 and 3)

#define UART0_ZIGBEE_SPEC_DEBUG			FALSE	// Whehther to enable UART hardware specification debug 
#define UART0_ZIGBEE_IRQ_STATUS_DEBUG	FALSE	// Whehther to enable UART IRQ status debug 
#define UART0_ZIGBEE_TRAC_STATUS_DEBUG	FALSE	// Whehther to enable UART TRAC status debug 
#define UART0_ZIGBEE_RX_PHY_DATA_DEBUG	FALSE	// Whehther to enable UART RX PHY data debug 
#define UART0_ZIGBEE_RX_PHY_LQI_DEBUG	FALSE	// Whehther to enable UART RX PHY LQI debug 
#define UART0_ZIGBEE_RX_MAC_DATA_DEBUG	FALSE	// Whehther to enable UART RX MAC data debug 
#define UART0_ZIGBEE_RX_MAC_FRAME_DEBUG	FALSE	// Whehther to enable UART RX MAC frame debug 
#define UART0_ZIGBEE_ED_LEVEL_DEBUG		FALSE	// Whehther to enable UART ED level debug 

#define UART0_SENTINEL_STATUS_DEBUG		TRUE	// Whehther to enable UART Sentinel status debug 
#define UART0_SENTINEL_LOCAL_DHCP_DEBUG	TRUE	// Whehther to enable UART local DHCP debug 

/* Protoboard Constants */
#define PROTOBOARD_LED_DDR		DDRD	// Define LED DDR
#define PROTOBOARD_LED_PORT		PORTD	// Define LED PORT
#define PROTOBOARD_LED_0		PD4		// Define LED 2 PORT pin
#define PROTOBOARD_LED_1		PD5		// Define LED 1 PORT pin

/* Includes */
#include "sentinel.h"					// Include common Sentinel library

/* Constants for Timer */
#define TX_NO_ACK_LATENCY_MS	500		// Fixed latency delay after no ack before retransmission

/* Volatile Variables */

/* Global Variables */

/* Function Prototypes */

/**
 * External Interrupt 0 ISR function
 * Initialized by the MCU hardware whenever the ZigBee transmitter makes an interrupt request
 */
ISR (INT0_vect) {
	zigbee_set_irq_flag();				// Sets IRQ flag to indicate an interrupt request has occurred
}

/**
 * Main thread/function for the base node
 */
int main(void) {
	/* Local variable Declaration and Initialization */
	uint8_t rxtx_data[ZIGBEE_PHY_DATA_LENGTH];	// RXTX data buffer
	uint8_t rxtx_data_length;					// RXTX data length
	
	uint8_t rx_status = ZIGBEE_RX_SUCCESS;		// RX status message
	uint8_t rx_ed_level = 0;					// RX ED level
	uint8_t rx_busy_status_flag = FALSE;		// RX busy status flag
	uint8_t rx_listen_status_flag = FALSE;		// RX listen status flag
	
	uint8_t tx_status = ZIGBEE_TX_SUCCESS;		// TX status message
	uint8_t tx_on_the_fly_enable_flag = TRUE;	// TX on-the-fly enable flag
	
	/* Global Variable Initialization */
	mcu_status = MCU_STATUS_P_ON;				// Initialize MCU status to Power On status after a Power On event
	
	/* Set PORT for LED Functionality */
	#if PROTOBOARD_LED_DEBUG
		sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_0);
		sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_1);
		sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
		sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);
	#endif

	/* MCU Initialization */
	init_mcu(MCU_PRR_ALL);											// Initialize MCU
	init_uart0(UART_BR_4800, UART_RXTX_MODE, UART_NO_ISR_ENABLE);	// Initialize UART
	init_spi_master(SPI_CLK_DIV_2);									// Initialize SPI Master
	init_ext_int0(TRUE, EXT_INT_SC_RISING_EDGE);					// Initialize External Int0
	
	sei();	// Enable Interrupt

    /* ZigBee Hardware Initialization */
	zigbee_init();					// Initialize zigbee transmitter after power-on event
   	zigbee_clear_irq_status();		// Clear all IRQ status
	zigbee_clear_irq_flag();		// Clear IRQ flag
	zigbee_set_extended_mode(TRUE);	// Enable extended (MAC layer) mode
	zigbee_set_pad_io_clkm(ZIGBEE_PAD_IO_CLKM_2MA);	// Minimize IO current output
	
	/* Set CPU CLK frequency from ZigBee CLKM pin */
	#if CPU_FREQUENCY_HZ == CPU_FREQUENCY_1_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_1MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_2_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_2MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_4_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_4MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_8_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_8MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_16_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_16MHZ);
	#endif

	/* UART0 Hardware Spec Debug */
	#if UART0_ZIGBEE_SPEC_DEBUG
		uart0_send_byte(UART_TX_ZIGBEE_PART_NUM);
		uart0_send_byte(zigbee_hw_part_num);
		uart0_send_byte(UART_TX_ZIGBEE_VER_NUM);
		uart0_send_byte(zigbee_hw_ver_num);
		uart0_send_byte(UART_TX_ZIGBEE_MAN_ID_0);
		uart0_send_byte(zigbee_hw_man_id_0);
		uart0_send_byte(UART_TX_ZIGBEE_MAN_ID_1);
		uart0_send_byte(zigbee_hw_man_id_1);
	#endif
	
	#if PROTOBOARD_LED_DEBUG
		cbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
	#endif

	while(1) {
		/* Parse MCU status */
		switch(mcu_status) {
			case MCU_STATUS_P_ON:	// Power on state
				/* Output status information */
				#if UART0_SENTINEL_STATUS_DEBUG
					uart0_send_byte(UART_TX_SENTINEL_STATUS);
					uart0_send_byte(SENTINEL_STATUS_P_ON);
					uart0_send_byte(SENTINEL_NODE_TYPE);
					uart0_send_byte(eeprom_read_byte(&eeprom_local_dhcp_local_depth));
					uart0_send_short(SENTINEL_BASE_SHORT_ADDR);
					uart0_send_short(SENTINEL_BASE_PAN_ID);
				#endif
				
				/* Set local address to transmitter */
				zigbee_set_tx_src_short_addr(SENTINEL_BASE_SHORT_ADDR);	// Source short addr during TX operation
				zigbee_set_tx_src_pan_id(SENTINEL_BASE_PAN_ID);			// Source pan ID during TX operation
				zigbee_set_addr_filter_short_addr(SENTINEL_BASE_SHORT_ADDR);	// Addr filter (destination short addr) during RX Addr filter 
				zigbee_set_addr_filter_pan_id(SENTINEL_BASE_PAN_ID);	// Addr filter (destination pan ID) during RX operation
				
				/* Set basic transmitter operation */
				zigbee_set_tx_sequence_number((uint8_t) (SENTINEL_BASE_SHORT_ADDR & 0xFF));	// Initialize TX framw sequencen umber
				zigbee_enable_tx_ack_request();		// Enable TX ack request (for transmission confirmation)
				mcu_status = MCU_STATUS_IDLE;		// Set MCU status to idle state
				break;
			case MCU_STATUS_IDLE:	// Idle state
				#if UART0_SENTINEL_STATUS_DEBUG
					uart0_send_byte(UART_TX_SENTINEL_STATUS);
					uart0_send_byte(SENTINEL_STATUS_IDLE);
				#endif
				mcu_status = MCU_STATUS_RX;			// Set MCU status to RX state
				break;
			case MCU_STATUS_TX_IDLE:	// TX idle state
				/* Parse TX status */
				switch (tx_status) {
					case ZIGBEE_TX_SUCCESS:				// TX successful
						mcu_status = MCU_STATUS_IDLE;	// Set MCU status to idle state after successful transmission
						break;
					case ZIGBEE_TX_INVALID_ARGUMENT:	// TX error due to invalid argument/parameter
						rxtx_data_length--;				// Reduce data length to correct for invalid argument error
						mcu_status = MCU_STATUS_TX;		// Set MCU status to TX state to retry transmission
						break;
					case ZIGBEE_TX_BUFFER_UNDERRUN:		// TX error due to buffer underrun (SPI uploads too slow)
						tx_on_the_fly_enable_flag = FALSE;	// Disable on-the-fly transmission mode
						mcu_status = MCU_STATUS_TX;		// Set MCU status to TX state to retry transmission
						break;
					case ZIGBEE_TX_CHANNEL_ACCESS_FAILURE:	// TX error due to channel access failure (too much channel traffic)
						mcu_status = MCU_STATUS_TX;		// Set MCU status to TX state to retry transmission
						break;
					case ZIGBEE_TX_NO_ACK:				// TX error due to no ack
						_delay_ms(TX_NO_ACK_LATENCY_MS);// Delay latency for local DHCP due to massive response from all nearby servers
						mcu_status = MCU_STATUS_TX;		// Set MCU status to TX state to retry transmission
						break;
					case ZIGBEE_TX_INVALID:				// TX error due to some invalid problem
						zigbee_sw_rst();				// Perform software reset to transmitter
						mcu_status = MCU_STATUS_TX;		// Set MCU status to TX state to retry transmission
						break;
				}
				break;
			case MCU_STATUS_TX:			// TX state
				/* Parse transmitter status */
				switch(zigbee_get_trx_status()) {
					case ZIGBEE_TRX_STATUS_TX_ARET_ON:		// Extended TX ready state
						zigbee_inc_tx_sequence_number();	// Increment TX frame sequence number
						tx_status = zigbee_extended_tx(rxtx_data, rxtx_data_length, tx_on_the_fly_enable_flag);	// Transmit data
						zigbee_enable_trx_off();			// Set transmitter to TRX off state after waiting for TX of data and RX of ack
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_TRX_OFF);
						#endif
						mcu_status = MCU_STATUS_TX_IDLE;	// Set transmitter state to TX idle state for parsing the TX status
						break;
					case ZIGBEE_TRX_STATUS_BUSY_TX_ARET:// Extended TX busy state
						break;
					default:							// All other states
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_TX);
							uart0_send_byte(rxtx_data_length);
							uart0_send_byte_array(rxtx_data, rxtx_data_length);
							uart0_send_short(zigbee_get_tx_dest_short_addr());
							uart0_send_short(zigbee_get_tx_dest_pan_id());
						#endif
						zigbee_enable_extended_tx();	// Set transmitter to extended TX ready state
						break;
				}
				break;
			case MCU_STATUS_RX_IDLE:	// RX idle state
				/* Parse RX status */
				switch (rx_status) {
					case ZIGBEE_RX_SUCCESS:	// RX successful
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_RX_SUCCESS);
							uart0_send_byte(rxtx_data_length);
							uart0_send_byte_array(rxtx_data, rxtx_data_length);
							uart0_send_short(zigbee_get_rx_src_short_addr());
							uart0_send_short(zigbee_get_rx_src_pan_id());
						#endif
						/* Parse RX data */
						switch(sentinel_rf_rx_parse(rxtx_data, &rxtx_data_length)) {
							case SENTINEL_RF_PARSE_IDLE:		// Idle
								mcu_status = MCU_STATUS_IDLE;	// Set MCU status to Idle state
								break;
							case SENTINEL_RF_PARSE_TRANSMIT:	// Transmit parsed data
								tx_on_the_fly_enable_flag = TRUE;	// Enable on-the-fly transmission
								mcu_status = MCU_STATUS_TX;		// Set MCU status to TX state
								break;
							case SENTINEL_RF_PARSE_SENSOR_DETECTION:	// Sensor detection
								/* Output sensor detection report to the Sentinel UI */
								uart0_send_byte(UART_TX_SENTINEL_SENSOR_DETECTION_REPORT);	// Sensor Detection Report header
								uart0_send_byte(rxtx_data[1]);	// SensorAddrL
								uart0_send_byte(rxtx_data[2]);	// SesnorAddrH
								mcu_status = MCU_STATUS_RX;		// Set MCU status to RX state
								break;
						}
						break;
					case ZIGBEE_RX_LOW_LQI:	// RX error due to low LQI
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_RX_LOW_LQI);
						#endif
						mcu_status = MCU_STATUS_RX;	// Set transmitter to RX state
						break;
					case ZIGBEE_RX_DUPLICATE_FRAME:	// RX error due to duplicate frame
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_RX_DUPLICATE_FRAME);
						#endif
						mcu_status = MCU_STATUS_RX;	// Set transmitter to RX state
						break;
					default:				// RX error due to some unknown error
						mcu_status = MCU_STATUS_RX;	// Set transmitter to RX state
						break;
				}
				break;
			case MCU_STATUS_RX:
				/* Parse transmitter status */
				switch(zigbee_get_trx_status()) {
					case ZIGBEE_TRX_STATUS_RX_AACK_NOCLK:	// Extended RX listening with CLKM disabled
					case ZIGBEE_TRX_STATUS_RX_AACK_ON:		// Extended RX listening
						rx_busy_status_flag = FALSE;	// Clear RX busy status flag
						#if UART0_SENTINEL_STATUS_DEBUG
							if (!rx_listen_status_flag) {	// Output RX listening state to Sentinel UI only once
								uart0_send_byte(UART_TX_SENTINEL_STATUS);
								uart0_send_byte(SENTINEL_STATUS_RX_LISTEN);
							}
						#endif
						rx_listen_status_flag = TRUE;	// Set RX listening status flag
						
						/* Process any available UART data */
						if (UCSR0A & _BV(RXC0)) {
							zigbee_enable_trx_off();		// Set transmitter to TRX OFF to make sure no RX data is missed
							#if UART0_SENTINEL_STATUS_DEBUG
								uart0_send_byte(UART_TX_SENTINEL_STATUS);
								uart0_send_byte(SENTINEL_STATUS_TRX_OFF);
							#endif
							sentinel_uart_rx_parse();		// Parse UART RX data
						}
						break;
					case ZIGBEE_TRX_STATUS_BUSY_RX_AACK_NOCLK:	// Extended RX busy with CLKM disabled
					case ZIGBEE_TRX_STATUS_BUSY_RX_AACK:		// Extended RX busy
						/* This state is reached ~300us before TRX END interrupt */
						rx_listen_status_flag = FALSE;	// Clear RX listening status flag
						#if UART0_SENTINEL_STATUS_DEBUG
							if (!rx_busy_status_flag) {	// Output RX busy state to Sentinel UI only once
								uart0_send_byte(UART_TX_SENTINEL_STATUS);
								uart0_send_byte(SENTINEL_STATUS_RX_BUSY);
							}
						#endif
						rx_busy_status_flag = TRUE;		// Set RX busy status flag
						
						if (zigbee_get_irq_flag()) {	// Determine whether the IRQ flag has been raised
							zigbee_clear_irq_flag();	// Clear IRQ flag to allow the next one to be detected
							zigbee_update_irq_status();	// Parse IRQ status
														// TRX END interrupt should occure
						}
						if (zigbee_get_rx_ready()) {	// Determine whether RX is ready with the data
							zigbee_clear_rx_ready();	// Clear RX ready flag
							rx_status = zigbee_extended_rx(rxtx_data, &rxtx_data_length);	// Retrieve RX data
							rx_ed_level = zigbee_get_frame_ed_level();	// Read ED level of RX frame
																		// Needs to be read within 224 s after the TRX END
							zigbee_enable_trx_off();					// Turn off TRX after ACK to ensure no successive frames will be missed
							#if UART0_SENTINEL_STATUS_DEBUG
								uart0_send_byte(UART_TX_SENTINEL_STATUS);
								uart0_send_byte(SENTINEL_STATUS_TRX_OFF);
							#endif
							mcu_status = MCU_STATUS_RX_IDLE;	// Set transmitter to RX idle state for parsing the RX status
						}
						break;
					default:	// Any other state
						rx_listen_status_flag = FALSE;	// Clear RX listening status flag
						rx_busy_status_flag = FALSE;	// Clear RX busy status flag
						zigbee_enable_extended_rx();	// Set transmitter to extended RX listening state
						break;
				}
				break;
			default:	// Unknown state
				mcu_status = MCU_STATUS_IDLE;	// Set MCU to idle state
				break;
		}
	}
	
	return 0;	// This should never be reached
}
