/**
 * A Mega644P Local DHCP Relay Node Final
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2A/PCINT31)
 *							-----------
 */

/* Global/System Constants */
#define TRUE 					1
#define FALSE 					0

#define CPU_FREQUENCY_1_MHZ		1000000				// 1.8-5.5V
#define CPU_FREQUENCY_2_MHZ		2000000				// 1.8-5.5V
#define CPU_FREQUENCY_4_MHZ		4000000				// 1.8-5.5V
#define CPU_FREQUENCY_8_MHZ		8000000				// 2.7-5.5V
#define CPU_FREQUENCY_16_MHZ	16000000			// 4.0-5.5V
#define CPU_FREQUENCY_HZ		CPU_FREQUENCY_1_MHZ

#define sbi(data,bit)			(data |= (1<<bit))	// Set Bit
#define cbi(data,bit) 			(data &= ~(1<<bit))	// Clear Bit
#define gbi(data,bit)			(data & (1<<bit))	// Get Bit
#define tbi(data,bit)			(data ^= (1<<bit))	// Toggle Bit

/* Node Constants */
#define SENTINEL_BASE_NODE		0x00
#define SENTINEL_RELAY_NODE		0x01
#define SENTINEL_SENSOR_NODE	0x02
#define SENTINEL_NODE_TYPE		SENTINEL_RELAY_NODE	// Identify this node

/* Debug Constants */
#define PROTOBOARD_LED_DEBUG			TRUE	// Whether to enable prototype LED debug (availabe on rev 1 and 3)

#define UART0_ZIGBEE_SPEC_DEBUG			FALSE	// Whehther to enable UART hardware specification debug 
#define UART0_ZIGBEE_IRQ_STATUS_DEBUG	FALSE	// Whehther to enable UART IRQ status debug 
#define UART0_ZIGBEE_TRAC_STATUS_DEBUG	FALSE	// Whehther to enable UART TRAC status debug 
#define UART0_ZIGBEE_RX_PHY_DATA_DEBUG	FALSE	// Whehther to enable UART RX PHY data debug 
#define UART0_ZIGBEE_RX_PHY_LQI_DEBUG	FALSE	// Whehther to enable UART RX PHY LQI debug 
#define UART0_ZIGBEE_RX_MAC_DATA_DEBUG	FALSE	// Whehther to enable UART RX MAC data debug 
#define UART0_ZIGBEE_RX_MAC_FRAME_DEBUG	FALSE	// Whehther to enable UART RX MAC frame debug 
#define UART0_ZIGBEE_ED_LEVEL_DEBUG		FALSE	// Whehther to enable UART ED level debug 

#define UART0_SENTINEL_STATUS_DEBUG		FALSE	// Whehther to enable UART Sentinel status debug 
#define UART0_SENTINEL_LOCAL_DHCP_DEBUG	FALSE	// Whehther to enable UART local DHCP debug 

/* Protoboard Constants */
#define PROTOBOARD_LED_DDR				DDRD	// Define LED DDR
#define PROTOBOARD_LED_PORT				PORTD	// Define LED PORT
#define PROTOBOARD_LED_0				PD4		// Define LED 2 PORT pin
#define PROTOBOARD_LED_1				PD5		// Define LED 1 PORT pin

/* Includes */
#include "sentinel.h"					// Include common Sentinel library

/* Constants for Timer */
#define TIMER2A_RES_MS									200
#define LOCAL_DHCP_RX_OFFER_TIMEOUT_LATENCY_MS			5000	// Delay latency to wait for offers from DHCP servers
#define LOCAL_DHCP_RX_OFFER_TIMEOUT_LATENCY_THRESHOLD	(uint16_t) (LOCAL_DHCP_RX_OFFER_TIMEOUT_LATENCY_MS / TIMER2A_RES_MS)
#define LOCAL_DHCP_TX_NO_ACK_LATENCY_MS					500		// Delay latency before retransmission to DHCP server after no ack during local DHCP operation
#define LOCAL_DHCP_TX_NO_ACK_LATENCY_THRESHOLD			(uint16_t) (LOCAL_DHCP_TX_NO_ACK_LATENCY_MS / TIMER2A_RES_MS)
#define LOCAL_DHCP_RX_ACK_TIMEOUT_LATENCY_MS			1000	// Delay latency to wait for an expected response
#define LOCAL_DHCP_RX_ACK_TIMEOUT_LATENCY_THRESHOLD		(uint16_t) (LOCAL_DHCP_RX_ACK_TIMEOUT_LATENCY_MS / TIMER2A_RES_MS)
#define LOCAL_DHCP_RETRY_LATENCY_MS						10000	// Delay latency before retrying local DHCP after a series of no acks during local DHCP operation
#define LOCAL_DHCP_RETRY_LATENCY_THRESHOLD				(uint16_t) (LOCAL_DHCP_RETRY_LATENCY_MS / TIMER2A_RES_MS)
#define LOCAL_DHCP_RESET_LATENCY_MS						60000	// Delay latency before reseting local DHCP after a series of retries during local DHCP operation
#define LOCAL_DHCP_RESET_LATENCY_THRESHOLD				(uint16_t) (LOCAL_DHCP_RESET_LATENCY_MS / TIMER2A_RES_MS)
#define TX_NO_ACK_LATENCY_MS							500		// Delay latency before retransmission after no ack during normal operation
#define TX_NO_ACK_LATENCY_THRESHOLD						(uint16_t) (TX_NO_ACK_LATENCY_MS / TIMER2A_RES_MS)
#define TX_NO_ACK_EXTENDED_LATENCY_MS					60000	// Extended delay latency before retransmission after no ack during normal operation
#define TX_NO_ACK_EXTENDED_LATENCY_THRESHOLD			(uint16_t) (TX_NO_ACK_EXTENDED_LATENCY_MS / TIMER2A_RES_MS)

#define LOCAL_DHCP_TX_NO_ACK_MAX_COUNT					2	// Maximum no acks before retrying during local DHCP operation
#define LOCAL_DHCP_RETRY_MAX_COUNT						4	// Maximum retries before resetting during local DHCP operation
#define TX_NO_ACK_MAX_COUNT								5	// Maximum no acks before extended delay during normal operation

/* Volatile Variables */
volatile uint16_t local_dhcp_rx_offer_timeout_latency_tick_count;	// Timer tick count for RX offer timeout during local DHCP operation
volatile uint16_t local_dhcp_rx_ack_timeout_latency_tick_count;		// Timer tick count for RX ack timeout during local DHCP operation
volatile uint16_t tx_no_ack_latency_tick_count;		// Timer tick count for TX latency after no ack

/* Global Variables */

/* Function Prototypes */

/**
 * Timer2 Compare A ISR function
 * Performs timing functionality during MCU active, idle, and power save states by asynchronously
 * incrementing the volatile tick count variables
 */
ISR(TIMER2_COMPA_vect) {
	local_dhcp_rx_offer_timeout_latency_tick_count++;
	local_dhcp_rx_ack_timeout_latency_tick_count++;
	tx_no_ack_latency_tick_count++;
}

/**
 * External Interrupt 0 ISR function
 * Initialized by the MCU hardware whenever the ZigBee transmitter makes an interrupt request
 */
ISR (INT0_vect) {
	zigbee_set_irq_flag();				// Sets IRQ flag to indicate an interrupt request has occurred
}

/**
 * Main thread/function for the relay node
 */
int main(void) {
	/* Local variable Declaration and Initialization */
	uint8_t rxtx_data[ZIGBEE_PHY_DATA_LENGTH];	// RXTX data buffer
	uint8_t rxtx_data_length;					// RXTX data length
	
	uint8_t rx_status = ZIGBEE_RX_SUCCESS;		// RX status message
	uint8_t rx_ed_level = 0;					// RX ED level
	uint8_t rx_busy_status_flag = FALSE;		// RX busy status flag
	uint8_t rx_listen_status_flag = FALSE;		// RX listen status flag
	
	uint8_t tx_status = ZIGBEE_TX_SUCCESS;		// TX status message
	uint8_t tx_on_the_fly_enable_flag = TRUE;	// TX on-the-flag enable flag
	
	uint8_t tx_no_ack_latency_enable_flag = FALSE;	// TX no ack latency enable flag
	uint8_t tx_no_ack_counter = 0;					// TX no ack counter
	
	uint8_t local_dhcp_offer_counter = 0;		// Local DHCP offer counter
	uint8_t local_dhcp_offer_ed_level[16];		// Local DHCP offer ed level buffer
	uint16_t local_dhcp_offer_client_short_addr[16];	// Local DHCP client short addr buffer
	uint16_t local_dhcp_offer_server_short_addr[16];	// Local DHCP server short addr buffer
	uint8_t local_dhcp_offer_server_node_type[16];		// Local DHCP offer server node type buffer
	uint8_t local_dhcp_offer_server_depth[16];	// Local DHCP offer server depth buffer
	uint8_t local_dhcp_offer_request_index = 0;	// Local DHCP offer request index
	
	uint8_t local_dhcp_tx_no_ack_counter = 0;	// Local DHCP tx no ack acounter
	uint8_t local_dhcp_retry_counter = 0;		// Local DHCP retry counter
	
	uint8_t local_dhcp_local_depth = 0;			// Local DHCP local depth
	uint16_t local_dhcp_local_short_addr = 0;	// Local DHCP local short addr
	uint16_t local_dhcp_server_short_addr = 0;	// Local DHCP server short addr
	uint16_t local_dhcp_server_pan_id = 0;		// Local DHCP server short addr
	uint8_t local_dhcp_server_node_type = 0;	// Local DHCP server node type
	uint8_t local_dhcp_server_depth = 0;		// Local DHCP server depth
	
	/* Global Variable Initialization */
	mcu_status = MCU_STATUS_P_ON;				// Initialize MCU status to Power On status after a Power On event
	
	/* Set PORT for LED Functionality */
	#if PROTOBOARD_LED_DEBUG
		sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_0);
		sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_1);
		sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
		sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);
	#endif

	/* MCU Initialization */
	init_mcu(MCU_PRR_ALL);							// Initialize MCU
	init_uart0(UART_BR_4800, UART_RXTX_MODE, UART_NO_ISR_ENABLE);	// Initialize UART
	init_spi_master(SPI_CLK_DIV_2);					// Initialize SPI Master
	init_ext_int0(TRUE, EXT_INT_SC_RISING_EDGE);	// Initialize External Int0
	init_timer2();									// Initialize Timer2
	
	sei();	// Enable Interrupt

    /* ZigBee Hardware Initialization */
	zigbee_init();					// Hardware initialization after power-on event
   	zigbee_clear_irq_status();		// Clear all IRQ status
	zigbee_clear_irq_flag();		// Clear IRQ flag
	zigbee_set_extended_mode(TRUE);	// Enable extended (MAC layer) mode
	zigbee_set_pad_io_clkm(ZIGBEE_PAD_IO_CLKM_2MA);	// Minimize IO current output
	
	/* Set CPU CLK frequency from ZigBee CLKM pin */
	#if CPU_FREQUENCY_HZ == CPU_FREQUENCY_1_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_1MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_2_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_2MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_4_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_4MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_8_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_8MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_16_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_16MHZ);
	#endif
	
	/* UART0 Hardware Spec Debug */
	#if UART0_ZIGBEE_SPEC_DEBUG
		uart0_send_byte(UART_TX_ZIGBEE_PART_NUM);
		uart0_send_byte(zigbee_hw_part_num);
		uart0_send_byte(UART_TX_ZIGBEE_VER_NUM);
		uart0_send_byte(zigbee_hw_ver_num);
		uart0_send_byte(UART_TX_ZIGBEE_MAN_ID_0);
		uart0_send_byte(zigbee_hw_man_id_0);
		uart0_send_byte(UART_TX_ZIGBEE_MAN_ID_1);
		uart0_send_byte(zigbee_hw_man_id_1);
	#endif
	
	#if PROTOBOARD_LED_DEBUG
		cbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
	#endif

	while(1) {
		/* Parse MCU status */
		switch(mcu_status) {
			case MCU_STATUS_P_ON:	// Power on state
				#if UART0_SENTINEL_STATUS_DEBUG
					uart0_send_byte(UART_TX_SENTINEL_STATUS);
					uart0_send_byte(SENTINEL_STATUS_P_ON);
					uart0_send_byte(SENTINEL_NODE_TYPE);
				#endif
				
				/* Determine whether local DHCP has been completed */
				if (eeprom_read_byte(&eeprom_local_dhcp_complete_flag)) {	// Read eeprom memory to determine
					local_dhcp_local_depth = eeprom_read_byte(&eeprom_local_dhcp_local_depth);	// Retrieve local DHCP local depth
					local_dhcp_local_short_addr = eeprom_read_word(&eeprom_local_dhcp_local_short_addr);	// Retrieve local DHCP local short addr
					local_dhcp_server_short_addr = eeprom_read_word(&eeprom_local_dhcp_server_short_addr);	// Retrieve local DHCP server short addr
					local_dhcp_server_pan_id = eeprom_read_word(&eeprom_local_dhcp_server_pan_id);			// Retrieve local DHCP server pan ID
					local_dhcp_server_node_type = eeprom_read_byte(&eeprom_local_dhcp_server_node_type);	// Retrieve local DHCP server node type
					local_dhcp_server_depth = eeprom_read_byte(&eeprom_local_dhcp_server_depth);	// Retrieve local DHCP server depth
					
					/* Set local address to transmitter */
					zigbee_set_tx_src_short_addr(local_dhcp_local_short_addr);
					zigbee_set_tx_src_pan_id(SENTINEL_RELAY_PAN_ID);
					zigbee_set_addr_filter_short_addr(local_dhcp_local_short_addr);
					zigbee_set_addr_filter_pan_id(SENTINEL_RELAY_PAN_ID);
					
					/* Set server address to transmitter */
					zigbee_set_tx_dest_short_addr(local_dhcp_server_short_addr);
					zigbee_set_tx_dest_pan_id(local_dhcp_server_pan_id);
					
					/* Initialize basic transmitter operation */
					zigbee_set_tx_sequence_number((uint8_t) (local_dhcp_local_short_addr & 0xFF));
					
					local_dhcp_status = LOCAL_DHCP_STATUS_COMPLETE;	// Set local DHCP status to complete
				}
				else {
					local_dhcp_status = LOCAL_DHCP_STATUS_INIT;		// Set local DHCP status to initialization
				}
				zigbee_enable_tx_ack_request();		// Enable TX ack request (for transmission confirmation)
				mcu_status = MCU_STATUS_DHCP;		// Set MCU status to MCU status 
				break;
			case MCU_STATUS_DHCP:	// DHCP state
				/* Parse local DHCP status */
				switch(local_dhcp_status) {
					case LOCAL_DHCP_STATUS_INIT:	// Init state
						local_dhcp_retry_counter = 0;		// Reset local DHCP retry counter
						local_dhcp_status = LOCAL_DHCP_STATUS_TX_DISCOVERY;	 // Set local DHCP status to TX discovery state
						break;
					case LOCAL_DHCP_STATUS_TX_DISCOVERY:	// TX Discovery state
						local_dhcp_offer_counter = 0;		// Reset local DHCP offer counter
						local_dhcp_tx_no_ack_counter = 0;	// Reset TX no ack counter
						
						/* Set local address to transmitter */
						zigbee_set_tx_src_short_addr(SENTINEL_RELAY_INIT_SHORT_ADDR);
						zigbee_set_tx_src_pan_id(SENTINEL_RELAY_INIT_PAN_ID);
						zigbee_set_addr_filter_short_addr(SENTINEL_RELAY_INIT_SHORT_ADDR);
						zigbee_set_addr_filter_pan_id(SENTINEL_RELAY_INIT_PAN_ID);
						
						/* Transmit local DHCP discover request to different pan IDs based on retry count */
						zigbee_set_tx_dest_short_addr(SENTINEL_GLOBAL_SHORT_ADDR);	// Transmit to all nodes in the pan ID
						if ((local_dhcp_retry_counter % 2) == 0) {
							#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
								uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
								uart0_send_byte(SENTINEL_LOCAL_DHCP_TX_DISCOVERY);
								uart0_send_byte(SENTINEL_BASE_NODE);
							#endif
							zigbee_set_tx_dest_pan_id(SENTINEL_BASE_PAN_ID);		// Transmit to base node pan ID
						}
						else {
							#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
								uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
								uart0_send_byte(SENTINEL_LOCAL_DHCP_TX_DISCOVERY);
								uart0_send_byte(SENTINEL_RELAY_NODE);
							#endif
							zigbee_set_tx_dest_pan_id(SENTINEL_RELAY_PAN_ID);		// Transmit to relay node pan ID
						}
						local_dhcp_retry_counter++;		// Increment retry counter
						
						zigbee_set_tx_sequence_number((uint8_t) (SENTINEL_RELAY_INIT_SHORT_ADDR & 0xFF));	
						
						/* TX Discovery: Header | NodeType */
						rxtx_data[0] = SENTINEL_RF_LOCAL_DHCP_DISCOVERY;
						rxtx_data[1] = SENTINEL_RELAY_NODE;
						rxtx_data_length = 2;
						
						local_dhcp_status = LOCAL_DHCP_STATUS_RX_OFFER;	// Set local DHCP status to RX offer state
						mcu_status = MCU_STATUS_TX;	// Set MCU status to TX state
						break;
					case LOCAL_DHCP_STATUS_RX_OFFER:	// RX Offer state
						#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
							uart0_send_byte(SENTINEL_LOCAL_DHCP_RX_OFFER_MODE);
						#endif
						timer_enable_timer2a_isr();	// Enable Timer2 ISR by enabling its IRQ mask
						local_dhcp_rx_offer_timeout_latency_tick_count = 0;	// Reset the RX offer timer tick count
						mcu_status = MCU_STATUS_RX;
						break;
					case LOCAL_DHCP_STATUS_TX_REQUEST:	// TX Request state
						#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
							uart0_send_byte(SENTINEL_LOCAL_DHCP_TX_REQUEST);
							uart0_send_byte(local_dhcp_offer_request_index);
							uart0_send_byte(local_dhcp_offer_server_node_type[local_dhcp_offer_request_index]);
							uart0_send_byte(local_dhcp_offer_server_depth[local_dhcp_offer_request_index]);
							uart0_send_byte(local_dhcp_offer_ed_level[local_dhcp_offer_request_index]);
							uart0_send_short(local_dhcp_offer_server_short_addr[local_dhcp_offer_request_index]);
							uart0_send_short(local_dhcp_offer_client_short_addr[local_dhcp_offer_request_index]);
						#endif
						
						/* TX Request: Header | NodeType | ServerShortAddrL | ServerShortAddrH | ClientShortAddrL | ClientShortAddrH */
						rxtx_data[0] = SENTINEL_RF_LOCAL_DHCP_REQUEST;
						rxtx_data[1] = SENTINEL_RELAY_NODE;
						rxtx_data[2] = (uint8_t) (local_dhcp_offer_server_short_addr[local_dhcp_offer_request_index] & 0xFF);
						rxtx_data[3] = (uint8_t) ((local_dhcp_offer_server_short_addr[local_dhcp_offer_request_index] >> 8) & 0xFF);
						rxtx_data[4] = (uint8_t) (local_dhcp_offer_client_short_addr[local_dhcp_offer_request_index] & 0xFF);
						rxtx_data[5] = (uint8_t) ((local_dhcp_offer_client_short_addr[local_dhcp_offer_request_index] >> 8) & 0xFF);
						rxtx_data_length = 6;
						
						local_dhcp_status = LOCAL_DHCP_STATUS_RX_ACK;	// Set local DHCP status to RX ack state
						mcu_status = MCU_STATUS_TX;	// Set MCU status to TX state
						break;
					case LOCAL_DHCP_STATUS_RX_ACK:	// RX ack state
						timer_enable_timer2a_isr();	// Enable Timer2 ISR by enabling its IRQ mask
						local_dhcp_rx_ack_timeout_latency_tick_count = 0;	// Reset the RX ack timer tick count
						local_dhcp_status = LOCAL_DHCP_STATUS_CONFIG;	// Set local DHCP status to config state
						mcu_status = MCU_STATUS_RX;	// Set MCU status to RX state
						break;
					case LOCAL_DHCP_STATUS_CONFIG:	// Config state
						#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
							uart0_send_byte(SENTINEL_LOCAL_DHCP_CONFIG);
						#endif
						
						/* Write completed local DHCP data to EEPROM */
						eeprom_write_byte(&eeprom_local_dhcp_local_depth, local_dhcp_local_depth);
						eeprom_write_word(&eeprom_local_dhcp_local_short_addr, local_dhcp_local_short_addr);
						eeprom_write_word(&eeprom_local_dhcp_server_short_addr, local_dhcp_server_short_addr);
						eeprom_write_word(&eeprom_local_dhcp_server_pan_id, local_dhcp_server_pan_id);
						eeprom_write_byte(&eeprom_local_dhcp_server_node_type, local_dhcp_server_node_type);
						eeprom_write_byte(&eeprom_local_dhcp_server_depth, local_dhcp_server_depth);
						eeprom_write_byte(&eeprom_local_dhcp_complete_flag, TRUE);
						
						/* Set local address to transmitter */
						zigbee_set_tx_src_short_addr(local_dhcp_local_short_addr);
						zigbee_set_tx_src_pan_id(SENTINEL_RELAY_PAN_ID);
						zigbee_set_addr_filter_short_addr(local_dhcp_local_short_addr);
						zigbee_set_addr_filter_pan_id(SENTINEL_RELAY_PAN_ID);
						
						/* Set server address to transmitter */
						zigbee_set_tx_dest_short_addr(local_dhcp_server_short_addr);
						zigbee_set_tx_dest_pan_id(local_dhcp_server_pan_id);
						
						/* Set basic transmitter operation */
						zigbee_set_tx_sequence_number((uint8_t) (local_dhcp_local_short_addr & 0xFF));
						
						local_dhcp_status = LOCAL_DHCP_STATUS_COMPLETE;	// Set local DHCP status to complete state
						break;
					case LOCAL_DHCP_STATUS_COMPLETE:	// Complete state
						#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
							uart0_send_byte(SENTINEL_LOCAL_DHCP_COMPLETE);
							uart0_send_byte(local_dhcp_local_depth);
							uart0_send_short(local_dhcp_local_short_addr);
							uart0_send_short(SENTINEL_RELAY_PAN_ID);
							uart0_send_short(local_dhcp_server_short_addr);
							uart0_send_short(local_dhcp_server_pan_id);
						#endif
						mcu_status = MCU_STATUS_IDLE;	// Set MCU status to idle state
						break;
				}
				break;
			case MCU_STATUS_IDLE:	// Idle state
				#if UART0_SENTINEL_STATUS_DEBUG
					uart0_send_byte(UART_TX_SENTINEL_STATUS);
					uart0_send_byte(SENTINEL_STATUS_IDLE);
				#endif
				if (local_dhcp_status == LOCAL_DHCP_STATUS_COMPLETE) {
					mcu_status = MCU_STATUS_RX;		// Set MCU status to RX state if local DHCP is completed
				}
				else {
					mcu_status = MCU_STATUS_DHCP;	// Set MCU status to RX state if local DHCP is still incomplete
				}
				break;
			case MCU_STATUS_SLEEP:	// Sleep state
				zigbee_enable_extended_rx_sleep();	// Enable extended RX sleep state
				set_sleep_mode(SLEEP_MODE_PWR_DOWN);	// Enable MCU Power Down sleep state
				sleep_mode();		// Enable sleep mode
				/* Wait until an interrupt request occurs from the transmitter */
				mcu_status = MCU_STATUS_RX;	// Set MCU status to RX state
				break;
			case MCU_STATUS_TX_IDLE:	// TX idle state
				/* Parse TX status */
				switch (tx_status) {
					case ZIGBEE_TX_SUCCESS:	// TX sucessful
						/* Determine whether local DHCP has been completed */
						if (local_dhcp_status == LOCAL_DHCP_STATUS_COMPLETE) {
							tx_no_ack_counter = 0;			// Clear TX no ack counter
							mcu_status = MCU_STATUS_IDLE;	// Set MCU status to Idle state
						}
						else {
							local_dhcp_tx_no_ack_counter = 0;	// Clear local DHCP tx no ack counter
							mcu_status = MCU_STATUS_DHCP;	// Set MCU status to DHCP state
						}
						break;
					case ZIGBEE_TX_INVALID_ARGUMENT:	// TX error due to invalid argument
						rxtx_data_length--;		// Reduce data length to correct for invalid argument error
						mcu_status = MCU_STATUS_TX;	// Set MCU status to TX state to retry transmission
						break;
					case ZIGBEE_TX_BUFFER_UNDERRUN:	// TX error due to buffer underrun
						tx_on_the_fly_enable_flag = FALSE;	// Disable on-the-fly transmission mode
						mcu_status = MCU_STATUS_TX;	// Set MCU status to TX state to retry transmission
						break;
					case ZIGBEE_TX_CHANNEL_ACCESS_FAILURE:	// TX error due to channel access failure
						mcu_status = MCU_STATUS_TX;	// Set MCU status to TX state to retry transmission
						break;
					case ZIGBEE_TX_NO_ACK:	// TX error due to no ack
						/* Determine whether transmission no ack latency functionality has been enabled */
						if (tx_no_ack_latency_enable_flag) {
							tx_no_ack_latency_enable_flag = FALSE;	// Clear tx no ack latency enable flag
							
							tx_no_ack_latency_tick_count = 0;	// Reset tx no ack latency tick count
							local_dhcp_tx_no_ack_counter++;		// Increment local DHCP tx no ack counter
							tx_no_ack_counter++;				// Increment tx no ack counter
							
							timer_enable_timer2a_isr();			// Enable Timer2 ISR by enable IRQ mask
							set_sleep_mode(SLEEP_MODE_PWR_SAVE);	// Enable Power Save sleep mode
						}
						sleep_mode();	// Enter Sleep mode
						
						/* Determine the course of action and latency before retransmission after no ack for normal operation */
						if (local_dhcp_status == LOCAL_DHCP_STATUS_COMPLETE) {
							/* Determine whether tx no ack counter has exceeded threshold during normal operation */
							if (tx_no_ack_counter > TX_NO_ACK_MAX_COUNT) {
								/* Wait for an extended delay before retrying transmission */
								if (tx_no_ack_latency_tick_count > TX_NO_ACK_EXTENDED_LATENCY_THRESHOLD) {
									tx_no_ack_counter = 0;			// Clear TX no ack counter
									timer_disable_timer2a_isr();	// Disable Timer2 ISR through IRQ mask
									mcu_status = MCU_STATUS_TX;		// Set MCU status to TX state
								}
							}
							else {
								/* Wait for a normal delay before retrying transmission */
								if (tx_no_ack_latency_tick_count > TX_NO_ACK_LATENCY_THRESHOLD) {
									timer_disable_timer2a_isr();	// Disable Timer2 ISR through IRQ mask
									mcu_status = MCU_STATUS_TX;		// Set MCU status to TX state
								}
							}
						}
						/* Determine the course of action and latency before retransmission after no ack for local DHCP operation by tx no ack counter*/
						else if (local_dhcp_tx_no_ack_counter > LOCAL_DHCP_TX_NO_ACK_MAX_COUNT) {
							/* Determine whether the number of local DHCP retries has exceeded the threshold */
							if (local_dhcp_retry_counter > LOCAL_DHCP_RETRY_MAX_COUNT) {
								/* Wait for a reset delay before retrying local DHCP */
								if (tx_no_ack_latency_tick_count > LOCAL_DHCP_RESET_LATENCY_THRESHOLD) {
									timer_disable_timer2a_isr();	// Disable Timer2 ISR through IRQ mask
									sentinel_reset_local_dhcp();	// Reset local DHCP
								}
							}
							else {
								/* Wait for a retry delay before retrying local DHCP */
								if (tx_no_ack_latency_tick_count > LOCAL_DHCP_RETRY_LATENCY_THRESHOLD) {
									timer_disable_timer2a_isr();	// Disable Timer2 ISR through IRQ mask
									sentinel_retry_local_dhcp();	// Retry local DHCP
								}
							}
						}
						else {
							/* Wait for a normal delay before retrying transmission */
							if (tx_no_ack_latency_tick_count > LOCAL_DHCP_TX_NO_ACK_LATENCY_THRESHOLD) {
								timer_disable_timer2a_isr();	// Disable Timer2 ISR through IRQ mask
								mcu_status = MCU_STATUS_TX;		// Set MCU status to TX state
							}
						}
						break;
					case ZIGBEE_TX_INVALID:	// TX error due to unknown error
						zigbee_sw_rst();	// Perform software reset to transmitter
						mcu_status = MCU_STATUS_TX;	// Set MCU status to TX state to retry transmission
						break;
				}
				break;
			case MCU_STATUS_TX:	// TX state
				/* Parse transmitter status */
				switch(zigbee_get_trx_status()) {
					case ZIGBEE_TRX_STATUS_TX_ARET_ON:	// Extended TX ready state
						zigbee_inc_tx_sequence_number();	// Increment TX frame sequence number
						tx_status = zigbee_extended_tx(rxtx_data, rxtx_data_length, tx_on_the_fly_enable_flag);	// Transmit data
						zigbee_enable_trx_off();		// Set transmitter to TRX off state after waiting for TX of data and RX of ack
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_TRX_OFF);
						#endif
						tx_no_ack_latency_enable_flag = TRUE;	// Enable tx no latency enable flag in case of no ack
						mcu_status = MCU_STATUS_TX_IDLE;	// Set MCU status to TX Idle state
						break;
					case ZIGBEE_TRX_STATUS_BUSY_TX_ARET:	// Extended TX busy state
						break;
					default:	// All other states
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_TX);
							uart0_send_byte(rxtx_data_length);
							uart0_send_byte_array(rxtx_data, rxtx_data_length);
							uart0_send_short(zigbee_get_tx_dest_short_addr());
							uart0_send_short(zigbee_get_tx_dest_pan_id());
						#endif
						zigbee_enable_extended_tx();	// Set transmitter to extended TX ready state
						break;
				}
				break;
			case MCU_STATUS_RX_IDLE:	// RX Idle state
				/* Parse RX status */
				switch (rx_status) {
					case ZIGBEE_RX_SUCCESS:	// RX successful
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_RX_SUCCESS);
							uart0_send_byte(rxtx_data_length);
							uart0_send_byte_array(rxtx_data, rxtx_data_length);
							uart0_send_short(zigbee_get_rx_src_short_addr());
							uart0_send_short(zigbee_get_rx_src_pan_id());
						#endif
						/* Parse RX data */
						switch(sentinel_rf_rx_parse(rxtx_data, &rxtx_data_length)) {
							case SENTINEL_RF_PARSE_IDLE:	// Idle
								mcu_status = MCU_STATUS_IDLE;	// Set transmitter to RX state
								break;
							case SENTINEL_RF_PARSE_TRANSMIT:	// Transmit parsed data
								tx_on_the_fly_enable_flag = TRUE;	// Enable on-the-fly transmission
								mcu_status = MCU_STATUS_TX;	// Set MCU status to TX state
								break;
							case SENTINEL_RF_PARSE_DHCP_OFFER:	// Parse offer from local DHCP server
								local_dhcp_offer_ed_level[local_dhcp_offer_counter] = rx_ed_level;	// Store ED level
								local_dhcp_offer_server_node_type[local_dhcp_offer_counter] = rxtx_data[0];// Store server node type
								local_dhcp_offer_server_depth[local_dhcp_offer_counter] = rxtx_data[1];	// Store server depth
								local_dhcp_offer_client_short_addr[local_dhcp_offer_counter] = 
									(((uint16_t) rxtx_data[3]) << 8) | ((uint16_t) rxtx_data[2]);	// Store client short addr
								local_dhcp_offer_server_short_addr[local_dhcp_offer_counter] = 
									(((uint16_t) rxtx_data[5]) << 8) | ((uint16_t) rxtx_data[4]);	// Store server short addr
								#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
									uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
									uart0_send_byte(SENTINEL_LOCAL_DHCP_RX_OFFER);
									uart0_send_byte(local_dhcp_offer_counter);
									uart0_send_byte(local_dhcp_offer_server_node_type[local_dhcp_offer_counter]);
									uart0_send_byte(local_dhcp_offer_server_depth[local_dhcp_offer_counter]);
									uart0_send_byte(local_dhcp_offer_ed_level[local_dhcp_offer_counter]);
									uart0_send_short(local_dhcp_offer_client_short_addr[local_dhcp_offer_counter]);
									uart0_send_short(local_dhcp_offer_server_short_addr[local_dhcp_offer_counter]);
								#endif
								local_dhcp_offer_counter++;	// Increment local DHCP offer counter
								mcu_status = MCU_STATUS_RX;	// Set MCU status to RX state
								break;
							case SENTINEL_RF_PARSE_DHCP_ACK:	// Parse ack from local DHCP server
								local_dhcp_server_node_type = rxtx_data[0];	// Store server node type
								local_dhcp_server_depth = rxtx_data[1];		// Store server depth
								local_dhcp_local_depth = rxtx_data[2];	// Store local depth
								local_dhcp_local_short_addr = 
									(((uint16_t) rxtx_data[4]) << 8) | ((uint16_t) rxtx_data[3]);	// Store local short addr
								local_dhcp_server_short_addr = 
									(((uint16_t) rxtx_data[6]) << 8) | ((uint16_t) rxtx_data[5]);	// Store server short addr
								local_dhcp_server_pan_id = 
									(((uint16_t) rxtx_data[8]) << 8) | ((uint16_t) rxtx_data[7]);	// Store server pan ID
								mcu_status = MCU_STATUS_DHCP;	// Set MCU status to DHCP state
								break;
							case SENTINEL_RF_PARSE_SENSOR_DETECTION:	// Sensor detection
								/* Relay sensor detection data to local DHCP server */
								zigbee_set_tx_dest_short_addr(local_dhcp_server_short_addr);
								zigbee_set_tx_dest_pan_id(local_dhcp_server_pan_id);
								
								#if UART0_SENTINEL_STATUS_DEBUG
									uart0_send_byte(UART_TX_SENTINEL_SENSOR_DETECTION_RELAY);
									uart0_send_byte(rxtx_data[1]);	// SensorAddrL
									uart0_send_byte(rxtx_data[2]);	// SensorAddrH
									uart0_send_short(zigbee_get_rx_src_short_addr());
									uart0_send_short(zigbee_get_rx_src_pan_id());
								#endif
								
								tx_on_the_fly_enable_flag = TRUE;	// Enable on-the-fly transmission mode
								mcu_status = MCU_STATUS_TX;	// Set MCU status to DHCP state
								break;
						}
						break;
					case ZIGBEE_RX_LOW_LQI:	// RX error due to low LQI
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_RX_LOW_LQI);
						#endif
						mcu_status = MCU_STATUS_RX;	// Set MCU status to RX state
						break;
					case ZIGBEE_RX_DUPLICATE_FRAME:	// RX error due to duplicate frame detection
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_RX_DUPLICATE_FRAME);
						#endif
						mcu_status = MCU_STATUS_RX;	// Set MCU status to RX state
						break;
					default:		// RX error due to unknown error
						mcu_status = MCU_STATUS_RX;	// Set MCU status to RX state
						break;
				}
				break;
			case MCU_STATUS_RX:	// RX state
				/* Parse transmitter status */
				switch(zigbee_get_trx_status()) {
					case ZIGBEE_TRX_STATUS_RX_AACK_NOCLK:	// Extended RX listening with CLKM disabled
					case ZIGBEE_TRX_STATUS_RX_AACK_ON:		// Extended RX listening
						rx_busy_status_flag = FALSE;	// Clear RX busy status flag
						#if UART0_SENTINEL_STATUS_DEBUG
							if (!rx_listen_status_flag) {	// Output RX listening state to Sentinel UI only once
								uart0_send_byte(UART_TX_SENTINEL_STATUS);
								uart0_send_byte(SENTINEL_STATUS_RX_LISTEN);
							}
						#endif
						rx_listen_status_flag = TRUE;	// Set RX listening status flag
						
						/* Determine whether RX offer timeout functionality should be enabled */
						if ((local_dhcp_status == LOCAL_DHCP_STATUS_RX_OFFER) && 
							(local_dhcp_rx_offer_timeout_latency_tick_count > LOCAL_DHCP_RX_OFFER_TIMEOUT_LATENCY_THRESHOLD)) {
							zigbee_enable_trx_off();	// Turn off transmitter to ensure no successive frames will be missed
							timer_disable_timer2a_isr();	// Disable Timer2 ISR through the IRQ mask
							
							#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
								uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
								uart0_send_byte(SENTINEL_LOCAL_DHCP_RX_OFFER_TIMEOUT);
							#endif
							
							local_dhcp_offer_request_index = 0;	// Reset the local DHCP offer request index
							/* Determine whether any offers were received */
							if (local_dhcp_offer_counter > 0) {
								/* Loop through all offers to find the one with the best attributes */
								for (uint8_t i=0; i<local_dhcp_offer_counter; i++) {
									/* Determine whether ED level of the current offer is higher */
									if (local_dhcp_offer_ed_level[i] > local_dhcp_offer_ed_level[local_dhcp_offer_request_index]) {
										local_dhcp_offer_request_index = i;	// Set the current offer index as the offer request index
									}
									/* Determine whether the depth is closer to base if the ED level is the same */ 
									else if ((local_dhcp_offer_ed_level[i] == local_dhcp_offer_ed_level[local_dhcp_offer_request_index]) 
										&& (local_dhcp_offer_server_depth[i] < local_dhcp_offer_server_depth[local_dhcp_offer_request_index])) {
										local_dhcp_offer_request_index = i;	// Set the current offer index as the offer request index
									}
								}
								local_dhcp_status = LOCAL_DHCP_STATUS_TX_REQUEST;	// Set the local DHCP status to the TX Request state
							}
							else {
								sentinel_retry_local_dhcp();	// Retry local DHCP
							}
							mcu_status = MCU_STATUS_DHCP;	// Set MCU status to DHCP state
						}
						/* Determine whether RX ack timeout functionality should be enabled */
						if (local_dhcp_status == LOCAL_DHCP_STATUS_RX_ACK && local_dhcp_rx_ack_timeout_latency_tick_count > LOCAL_DHCP_RX_ACK_TIMEOUT_LATENCY_THRESHOLD) {
							zigbee_enable_trx_off();	// Turn off transmitter to ensure no successive frames will be missed
							timer_disable_timer2a_isr();
							
							#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
								uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
								uart0_send_byte(SENTINEL_LOCAL_DHCP_RX_ACK_TIMEOUT);
							#endif
							
							sentinel_retry_local_dhcp();	// Retry local DHCP
						}
						/* Determine whether to move to RX sleep state during normal operation */
						if (local_dhcp_status == LOCAL_DHCP_STATUS_COMPLETE) {
							#if UART0_SENTINEL_STATUS_DEBUG
								uart0_send_byte(UART_TX_SENTINEL_STATUS);
								uart0_send_byte(SENTINEL_STATUS_SLEEP);
							#endif
							mcu_status = MCU_STATUS_SLEEP;	// Set MCU status to Sleep state
						}
						break;
					case ZIGBEE_TRX_STATUS_BUSY_RX_AACK_NOCLK:	// Extended RX busy with CLKM disabled
					case ZIGBEE_TRX_STATUS_BUSY_RX_AACK:	// Extended RX busy
						/* This state is reached ~300us before TRX END interrupt */
						rx_listen_status_flag = FALSE;	// Clear RX listening status flag
						#if UART0_SENTINEL_STATUS_DEBUG
							if (!rx_busy_status_flag) {	// Output RX busy state to Sentinel UI only once
								uart0_send_byte(UART_TX_SENTINEL_STATUS);
								uart0_send_byte(SENTINEL_STATUS_RX_BUSY);
							}
						#endif
						rx_busy_status_flag = TRUE;	// Set RX busy status flag
						
						if (zigbee_get_irq_flag()) {	// Determine whether the IRQ flag has been raised
							zigbee_clear_irq_flag();	// Clear IRQ flag to allow the next one to be detected
							zigbee_update_irq_status(); // Parse IRQ status	
														// TRX_END interrupt
						}
						if (zigbee_get_rx_ready()) {	// Determine whether RX is ready with the data
							zigbee_clear_rx_ready();	// Clear RX ready flag
							rx_status = zigbee_extended_rx(rxtx_data, &rxtx_data_length);	// Retrieve RX data
							rx_ed_level = zigbee_get_frame_ed_level();	// Read ED level of RX frame
																		// Needs to be read within 224 s after the TRX_END
							zigbee_enable_trx_off();	// Turn off transmitter to ensure no successive frames will be missed
							#if UART0_SENTINEL_STATUS_DEBUG
								uart0_send_byte(UART_TX_SENTINEL_STATUS);
								uart0_send_byte(SENTINEL_STATUS_TRX_OFF);
							#endif
							mcu_status = MCU_STATUS_RX_IDLE;	// Set transmitter to RX idle state for parsing the RX status
						}
						break;
					default:	// Any other state
						rx_listen_status_flag = FALSE;	// Clear RX listening status flag
						rx_busy_status_flag = FALSE;	// Clear RX busy status flag
						zigbee_enable_extended_rx();	// Set transmitter to extended RX listening state
						break;
				}
				break;
			default:	// Unknown state
				mcu_status = MCU_STATUS_IDLE;	// Set MCU to idle state
				break;
		}
	}
	
	return 0;	// This should never be reached
}
