/**
 * A Mega644P Timer Library
 *
 * Ruibing Wang (rw98@cornell.edu)
 */

/**
 * Interrupt vectors:
 *
 * TIMER0_COMPA_vect
 * TIMER0_COMPB_vect
 * TIMER0_OVF_vect
 * TIMER1_CAPT_vect
 * TIMER1_COMPA_vect
 * TIMER1_COMPB_vect
 * TIMER1_OVF_vect
 * TIMER2_COMPA_vect
 * TIMER2_COMPB_vect
 * TIMER2_OVF_vect
 */

/* Timer Constants */
#define TIMER_DIV_0			0
#define TIMER_DIV_1			1
#define TIMER_DIV_8			8
#define TIMER_DIV_32		32	// Timer2 Only
#define TIMER_DIV_64		64
#define TIMER_DIV_128		128	// Timer2 Only
#define TIMER_DIV_256		256
#define TIMER_DIV_1024		1024

/* Timer0 Constants */
#define TIMER0A_ISR_ENABLE		TRUE
#define TIMER0B_ISR_ENABLE		FALSE
#define TIMER0A_DEBUG		FALSE
#define TIMER0A_RES_US		200000
#define TIMER0B_RES_US		50
#define TIMER0_DIV			TIMER_DIV_1024
#define TIMER0_DIV_MASK		0b00000111
#define TIMER0_OCR0A		(uint8_t) (((float)CPU_FREQUENCY_HZ/(float)TIMER0_DIV) / ((float)1000000/(float)TIMER0A_RES_US) + 0.5)
#define TIMER0_OCR0B		(uint8_t) (((float)CPU_FREQUENCY_HZ/(float)TIMER0_DIV) / ((float)1000000/(float)TIMER0B_RES_US) + 0.5)
#define TIMER0_DDR			DDRB
#define TIMER0_PORT			PORTB
#define TIMER0_PIN			PINB
#define TIMER0A_OUTPUT		PB3
#define TIMER0B_OUTPUT		PB4

/* Timer1 Constants */
#define OCR1A_FREQUENCY_HZ	153600	// Set Timer1 clock oscillator output frequency
#define TIMER1_DIV			TIMER_DIV_1		// Set Timer1 divider (lowest has the highest acurracy)
#define TIMER1_DIV_MASK		0b00000111	// Define Timer1 divider register mask
#define TIMER1_OCR1A		(uint16_t) ((float)CPU_FREQUENCY_HZ / ((float)TIMER1_DIV*(float)OCR1A_FREQUENCY_HZ*2) + 0.5)	// Caculate Timer1 OCR1A register value (for clock oscillator)
#define TIMER1_DDR			DDRD	// Set Timer1 OCR1A DDR port
#define TIMER1_PORT			PORTD
#define TIMER1_PIN			PIND
#define TIMER1A_OUTPUT		PD5		// Set Timer1 OCR1A pin
#define TIMER1B_OUTPUT		PD4

/* Timer2 Constants */
#define TIMER2A_ISR_ENABLE	FALSE
#define TIMER2B_ISR_ENABLE	FALSE
#define TIMER2A_DEBUG		FALSE
#define TIMER2A_RES_US		200000
#define TIMER2B_RES_US		32
#define TIMER2_DIV			TIMER_DIV_1024	// Set Timer2 divider
#define TIMER2_DIV_MASK		0b00000111	// Define Timer2 divider register mask
#define TIMER2_OCR2A		(uint8_t) (((float)CPU_FREQUENCY_HZ/(float)TIMER2_DIV) / ((float)1000000/(float)TIMER2A_RES_US) + 0.5)	// Calculate Timer2 OCR2A register compare match value
#define TIMER2_OCR2B		(uint8_t) (((float)CPU_FREQUENCY_HZ/(float)TIMER2_DIV) / ((float)1000000/(float)TIMER2B_RES_US) + 0.5)	// Calculate Timer2 OCR2A register compare match value
#define TIMER2_DDR			DDRD
#define TIMER2_PIN			PIND
#define TIMER2A_OUTPUT		PD7
#define TIMER2B_OUTPUT		PD6

/* Timer0 Function Prototypes */
void init_timer0(void);
void timer_enable_timer0a_isr(void);
void timer_disable_timer0a_isr(void);
void timer_enable_timer0b_isr(void);
void timer_disable_timer0b_isr(void);

/* Timer1 Function Prototypes */
void init_timer1(void);
void timer_enable_timer1a_isr(void);
void timer_disable_timer1a_isr(void);
void timer_enable_timer1b_isr(void);
void timer_disable_timer1b_isr(void);

/* Timer2 Function Prototypes */
void init_timer2(void);
void timer_enable_timer2a_isr(void);
void timer_disable_timer2a_isr(void);
void timer_enable_timer2b_isr(void);
void timer_disable_timer2b_isr(void);

/**
 * Enable Timer0A trough the IRQ mask
 */
void timer_enable_timer0a_isr(void) {
	sbi(TIMSK0, OCIE0A);	// Set IRQ mask
}

/**
 * Disable Timer0A through the IRQ mask
 */
void timer_disable_timer0a_isr(void) {
	cbi(TIMSK0, OCIE0A);	// Clear IRQ mask
}

/**
 * Enable Timer0B trough the IRQ mask
 */
void timer_enable_timer0b_isr(void) {
	sbi(TIMSK0, OCIE0B);	// Set IRQ mask
}

/**
 * Disable Timer0B trough the IRQ mask
 */
void timer_disable_timer0b_isr(void) {
	cbi(TIMSK0, OCIE0B);	// Clear IRQ mask
}

/**
 * Enable Timer1A trough the IRQ mask
 */
void timer_enable_timer1a_isr(void) {
	sbi(TIMSK1, OCIE1A);	// Set IRQ mask
}

/**
 * Disable Timer1A trough the IRQ mask
 */
void timer_disable_timer1a_isr(void) {
	cbi(TIMSK1, OCIE1A);	// Clear IRQ mask
}

/**
 * Enable Timer1B trough the IRQ mask
 */
void timer_enable_timer1b_isr(void) {
	sbi(TIMSK1, OCIE1B);	// Set IRQ mask
}

/**
 * Disable Timer1B trough the IRQ mask
 */
void timer_disable_timer1b_isr(void) {
	cbi(TIMSK1, OCIE1B);	// Clear IRQ mask
}

/**
 * Enable Timer2A trough the IRQ mask
 */
void timer_enable_timer2a_isr(void) {
	sbi(TIMSK2, OCIE2A);	// Set IRQ mask
}

/**
 * Disable Timer2A trough the IRQ mask
 */
void timer_disable_timer2a_isr(void) {
	cbi(TIMSK2, OCIE2A);	// Clear IRQ mask
}

/**
 * Enable Timer2B trough the IRQ mask
 */
void timer_enable_timer2b_isr(void) {
	sbi(TIMSK2, OCIE2B);	// Set IRQ mask
}

/**
 * Disable Timer2B trough the IRQ mask
 */
void timer_disable_timer2b_isr(void) {
	cbi(TIMSK2, OCIE2B);	// Clear IRQ mask
}

/**
 * Initialize Timer1
 *
 * For interrupt-based CTC mode
 */
void init_timer0(void) {
	cbi(PRR, PRTIM0);		// Clear the Timer0 power reduction register bit
	
	TCCR0A = 0b00000010;	// 7:6 - COM0A1:0: Compare Match Output A Mode
							// 5:4 - COM0B1:0: Compare Match Output B Mode
							// 3:2 - Reserved Bits
							// 1:0 - WGM01:0: Waveform Generation Mode
							//	000 - Normal (0xFF)
							//	001 - PWM, Phase Correct (0xFF)
							//	010 - CTC (OCRA)
							//	011 - Fast PWM (0xFF)
							//	100 - Reserved
							//	101 - PWM, Phase Correct (OCRA)
							//	110 - Reserved
							//	111 - Fast PWM (OCRA)
	
	/* Enable OCR0A output for digital debug */
	#if TIMER0A_DEBUG
		sbi(TIMER0_DDR, TIMER0A_OUTPUT);	
		TCCR0A |= (0b01 << 6);
	#endif
	
	TCCR0B = 0b00000000;	// 7 - FOC0A: Force Output Compare A
							// 6 - FOC0B: Force Output Compare B
							// 5:4 - Reserved Bits
							// 3 - WGM02: Waveform Generation Mode
							// 2:0 - CS02:0: Clock select
							//	000 - Timer stopped
							//	001 - 1
							//	010 - 8
							//	011 - 64
							//	100 - 256
							//	101 - 1024
							//	110 - External clock source on T0 pin, falling edge
							//	111 - ", rising edge
	
	/* Determine timer clock divider for Timer0 */
	#if TIMER1_DIV == TIMER_DIV_0
		TCCR0B = (TCCR0B & ~TIMER1_DIV_MASK) | 0b000;
	#elif TIMER1_DIV == TIMER_DIV_1
		TCCR0B = (TCCR0B & ~TIMER1_DIV_MASK) | 0b001;
	#elif TIMER1_DIV == TIMER_DIV_8
		TCCR0B = (TCCR0B & ~TIMER1_DIV_MASK) | 0b010;
	#elif TIMER1_DIV == TIMER_DIV_64
		TCCR0B = (TCCR0B & ~TIMER1_DIV_MASK) | 0b011;
	#elif TIMER1_DIV == TIMER_DIV_256
		TCCR0B = (TCCR0B & ~TIMER1_DIV_MASK) | 0b100;
	#elif TIMER1_DIV == TIMER_DIV_1024
		TCCR0B = (TCCR0B & ~TIMER1_DIV_MASK) | 0b101;
	#endif
	
	TIMSK0 = 0b00000000;	// 7:3 - Reserved bits for other timer interrupts
							// 2 - OCIE0B: Timer0 Timer/Counter Output Compare Match B Interrupt Enable
							// 1 - OCIE0A: Timer0 Timer/Counter Output Compare Match A Interrupt Enable
							// 0 - TOIE0: Timer/Counter0 Overflow Interrupt Enable
	
	/* Determine whether to enable Timer0A ISR */
	#if TIMER0A_ISR_ENABLE
		OCR0A = TIMER0_OCR0A;	// Set Timer0 OCR0A value
		TIMSK0 |= _BV(OCIE0A);	// Set IRQ mask
	#endif
	
	/* Determine whether to enable Timer0B ISR */
	#if TIMER0B_ISR_ENABLE
		OCR0B = TIMER0_OCR0B;	// Set Timer0 OCR0B value
		TIMSK0 |= _BV(OCIE0B);	// Set IRQ mask
	#endif
}

/**
 * Initialize Timer1
 *
 * For compare output mode on OCR1A
 */
void init_timer1(void) {
	cbi(PRR, PRTIM1);		// Clear the Timer1 power reduction register bit
	
	sbi(TIMER1_DDR, TIMER1A_OUTPUT);	// Enable OCR1A output for digital debug

	TCCR1A = 0b01000000;	// 7:6  COMnA1:0: Compare Output Mode for Channel A
							// 5:4 - COMnB1:0: Compare Output Mode for Channel B
							// 3:2 - Reserved
							// 1:0 - WGMn1:0: Waveform Generation Mode
							//	0000 - Normal (0xFFFF)
							//	0001 - PWM, Phase Correct, 8-bit (0x00FF)
							//	0010 - PWM, Phase Correct, 9-bit (0x01FF)
							//	0011 - PWM, Phase Correct, 10-bit (0x03FF)
							//	0100 - CTC (OCRnA)
							//	0101 - Fast PWM, 8-bit (0x00FF)
							//	0110 - Fast PWM, 9-bit (0x01FF)
							//	0111 - Fast PWM, 10-bit (0x03FF)
							//	1000 - PWM, Phase and Frequency Correct (ICRn)
							//	1001 - PWM, Phase and Frequency Correct (OCRnA)
							//	1010 - PWM, Phase Correct (ICRn)
							//	1011 - PWM, Phase Correct (OCRnA)
							//	1100 - CTC (ICRn)
							//	1101 - Reserved
							//	1110 - Fast PWM (ICRn)
							//	1111 - Fast PWM (OCRnA)
							
	TCCR1B = 0b00001000;	// 7 - ICNCn: Input Capture Noise Canceler
							// 6 - ICESn: Input Capture Edge Select
							// 5 - Reserved
							// 4:3 - WGMn3:2: Waveform Generation Mode
							// 2:0 - CSn2:0: Clock Select
							//	000 - Timer stopped
							//	001 - 1
							//	010 - 8
							//	011 - 64
							//	100 - 256
							//	101 - 1024
							//	110 - External clock source on T0 pin, falling edge
							//	111 - ", rising edge
	
	/* Determine timer clock divider for Timer1 */
	#if TIMER1_DIV == TIMER_DIV_0
		TCCR1B = (TCCR1B & ~TIMER1_DIV_MASK) | 0;
	#elif TIMER1_DIV == TIMER_DIV_1
		TCCR1B = (TCCR1B & ~TIMER1_DIV_MASK) | 1;
	#elif TIMER1_DIV == TIMER_DIV_8
		TCCR1B = (TCCR1B & ~TIMER1_DIV_MASK) | 2;
	#elif TIMER1_DIV == TIMER_DIV_64
		TCCR1B = (TCCR1B & ~TIMER1_DIV_MASK) | 3;
	#elif TIMER1_DIV == TIMER_DIV_256
		TCCR1B = (TCCR1B & ~TIMER1_DIV_MASK) | 4;
	#elif TIMER1_DIV == TIMER_DIV_1024
		TCCR1B = (TCCR1B & ~TIMER1_DIV_MASK) | 5;
	#endif
						
	TCCR1C = 0b00000000;	// 7 - FOCnA: Force Output Compare for Channel A
							// 6 - FOCnB: Force Output Compare for Channel B
							// 5:0 - Reserved
	
	/* Set Timer1 OCR1A value */
	OCR1AH = (TIMER1_OCR1A) >> 8;
	OCR1AL = (TIMER1_OCR1A << 8) >> 8;
	
	TIMSK1 = 0b00000000;	// 7:6 - Reserved
							// 5 - ICIE1: Timer/Counter1, Input Capture Interrupt Enable
							// 4:3 - Reserved
							// 2 - OCIE1B: Timer/Counter1, Output Compare B Match Interrupt Enable
							// 1 - OCIE1A: Timer/Counter1, Output Compare A Match Interrupt Enable
							// 0 - TOIE1: Timer/Counter1, Overflow Interrupt Enable
	
}

/**
 * Initialize Timer2
 *
 * For interrupt-based CTC mode
 */
void init_timer2(void) {
	cbi(PRR, PRTIM2);		// Clear the Timer2 power reduction register bit

	TCCR2A = 0b00000010;	// 7:6  COM2A1:0: Compare Match Output A Mode
							// 5:4  COM2B1:0: Compare Match Output B Mode
							// 3:2  Reserved Bits
							// 1:0  WGM21:0: Waveform Generation Mode
	
	/* Determine whether to enable OCR2A output for digital debug */
	#if TIMER2A_DEBUG
		sbi(TIMER2_DDR,TIMER2A_OUTPUT);
		sbi(TCCR2A,6);
	#endif

	TCCR2B = 0b00000000;	// 7  FOC2A: Force Output Compare A
							// 6  FOC2B: Force Output Compare B
							// 5:4  Reserved Bits
							// 3  WGM22: Waveform Generation Mode
							// 2:0  CS22:0: Clock Select
							//	000 - No clk
							//	001 - 1
							//	010 - 8
							//	011 - 32
							//	100 - 64
							//	101 - 128
							//	110 - 256
							//	111 - 1024
	
	/* Determine timer clock divider for Timer2 */
	#if TIMER2_DIV == TIMER_DIV_0
		TCCR2B = (TCCR2B & ~TIMER2_DIV_MASK) | 0;
	#elif TIMER2_DIV == TIMER_DIV_1
		TCCR2B = (TCCR2B & ~TIMER2_DIV_MASK) | 1;
	#elif TIMER2_DIV == TIMER_DIV_8
		TCCR2B = (TCCR2B & ~TIMER2_DIV_MASK) | 2;
	#elif TIMER2_DIV == TIMER_DIV_32
		TCCR2B = (TCCR2B & ~TIMER2_DIV_MASK) | 3;
	#elif TIMER2_DIV == TIMER_DIV_64
		TCCR2B = (TCCR2B & ~TIMER2_DIV_MASK) | 4;
	#elif TIMER2_DIV == TIMER_DIV_128
		TCCR2B = (TCCR2B & ~TIMER2_DIV_MASK) | 5;
	#elif TIMER2_DIV == TIMER_DIV_256
		TCCR2B = (TCCR2B & ~TIMER2_DIV_MASK) | 6;
	#elif TIMER2_DIV == TIMER_DIV_1024
		TCCR2B = (TCCR2B & ~TIMER2_DIV_MASK) | 7;
	#endif
	
	
	ASSR = 0b00000000;		// 6  EXCLK: Enable External Clock Input
							// 5  AS2: Asynchronous Timer/Counter2
							// 4  TCN2UB: Timer/Counter2 Update Busy
							// 3  OCR2AUB: Output Compare Register2 Update Busy
							// 2  OCR2BUB: Output Compare Register2 Update Busy
							// 1  TCR2AUB: Timer/Counter Control Register2 Update Busy
							// 0  TCR2BUB: Timer/Counter Control Register2 Update Busy
	
	TIMSK2 = 0b00000000;	// 7:3 - Reserved Bits
							// 2 - OCIE2B: Timer/Counter2 Output Compare Match B Interrupt Enable
							// 1  OCIE2A: Timer/Counter2 Output Compare Match A Interrupt Enable
							// 0  TOIE2: Timer/Counter2 Overflow Interrupt Enable
	
	
	
	
	/* Determine whether to enable Timer2A ISR */
	#if TIMER2A_ISR_ENABLE
		OCR2A = TIMER2_OCR2A;	// Set Timer2 OCR2A value
		TIMSK2 |= _BV(OCIE2A);	// Set IRQ mask
	#endif
	
	/* Determine whether to enable Timer2A ISR */
	#if TIMER2B_ISR_ENABLE
		OCR2B = TIMER2_OCR2B;	// Set Timer2 OCR2A value
		TIMSK2 |= _BV(OCIE2B);	// Set IRQ mask
	#endif		
}