/**
 * A Mega644P UART Library
 *
 * Ruibing Wang (rw98@cornell.edu)
 */

/**
 * CPU_FREQ (MHz)	BAUD_RATE (bps)		Error (%)
 * 1				2400				0.2
 *					4800				0.2
 *					9600				-7.0 (Errors)
 *					14400				8.5 (Errors)
 *					19200				8.5 (Errors)
 *					28800				8.5 (Errors)
 *					38400				-18.6
 * 2				2400				0.2
 *					4800				0.2
 *					9600				0.2
 *					14400				-3.5
 *					19200				-7.0
 *					28800				8.5
 *					38400				8.5	
 * 4				2400				0.2
 *					4800				0.2
 *					9600				0.2
 *					14400				2.1
 *					19200				0.2
 *					28800				-3.5
 *					38400				-7.0					
 * 8				2400				0.2
 *					4800				0.2
 *					9600				0.2
 *					14400				-0.8
 *					19200				0.2
 *					28800				2.1
 *					38400				0.2					
 */

/* UART Constants */
#define UART_BR_2400			2400
#define UART_BR_4800			4800
#define UART_BR_9600			9600
#define UART_BR_14400			14400
#define UART_BR_19200			19200
#define UART_BR_28800			28800
#define UART_BR_38400			38400

#define UART_RXTX_MODE			0
#define UART_RX_MODE			1
#define UART_TX_MODE			2

#define UART_RXTX_ISR_ENABLE	0
#define UART_RX_ISR_ENABLE		1
#define UART_TX_ISR_ENABLE		2
#define UART_NO_ISR_ENABLE		3

/* Function Prototypes */
void uart0_send_byte(uint8_t data);
void uart0_send_byte_array(uint8_t *data, uint8_t data_length);
void uart0_send_short(uint16_t data);
void uart0_send_long(uint32_t data);
void uart0_send_extended(uint64_t data);
void init_uart0(uint16_t baud_rate, uint8_t rxtx_mode, uint8_t rxtx_isr_enable);

void uart1_send_byte(uint8_t data);
void uart1_send_byte_array(uint8_t *data, uint8_t data_length);
void uart1_send_short(uint16_t data);
void uart1_send_long(uint32_t data);
void uart1_send_extended(uint64_t data);
void init_uart1(uint16_t baud_rate, uint8_t rxtx_mode, uint8_t rxtx_isr_enable);

/**
 * Transmit a byte through UART0
 */
void uart0_send_byte(uint8_t data) {
	while ( !( UCSR0A & _BV(UDRE0)) );	// Wait for empty buffer
	UDR0 = data;						// Put data into buffer, sends the data 
}

/**
 * Transmit byte array through UART0
 *
 * Can be used to transmitt strings
 * Length terminated (versus null terminated)
 */
void uart0_send_byte_array(uint8_t *data, uint8_t data_length) {
	for (uint8_t i=0; i<data_length; i++) {
		uart0_send_byte(data[i]);
	}
}

/**
 * Transmit a short through UART0
 */
void uart0_send_short(uint16_t data) {
	uart0_send_byte((uint8_t) (data & 0xFF));			// data[0:7]
	uart0_send_byte((uint8_t) ((data >> 8) & 0xFF));	// data[8:15]
}

/**
 * Transmit a long through UART0
 */
void uart0_send_long(uint32_t data) {
	uart0_send_byte((uint8_t) (data & 0xFF));			// data[0:7]
	uart0_send_byte((uint8_t) ((data >> 8) & 0xFF));	// data[8:15]
	uart0_send_byte((uint8_t) ((data >> 16) & 0xFF));	// data[16:23]
	uart0_send_byte((uint8_t) ((data >> 24) & 0xFF));	// data[24:31]
}

/**
 * Transmit an extended through UART0
 */
void uart0_send_extended(uint64_t data) {
	uart0_send_byte((uint8_t) (data & 0xFF));			// data[0:7]
	uart0_send_byte((uint8_t) ((data >> 8) & 0xFF));	// data[8:15]
	uart0_send_byte((uint8_t) ((data >> 16) & 0xFF));	// data[16:23]
	uart0_send_byte((uint8_t) ((data >> 24) & 0xFF));	// data[24:31]
	uart0_send_byte((uint8_t) ((data >> 32) & 0xFF));	// data[32:39]
	uart0_send_byte((uint8_t) ((data >> 40) & 0xFF));	// data[40:47]
	uart0_send_byte((uint8_t) ((data >> 48) & 0xFF));	// data[48:55]
	uart0_send_byte((uint8_t) ((data >> 56) & 0xFF));	// data[56:63]
}

/**
 * Transmit byte array through UART1
 *
 * Can be used to transmitt strings
 * Length terminated (versus null terminated)
 */
void uart1_send_byte_array(uint8_t *data, uint8_t data_length) {
	for (uint8_t i=0; i<data_length; i++) {
		uart0_send_byte(data[i]);
	}
}

/**
 * Transmit a byte through UART1
 */
void uart1_send_byte(uint8_t data) {
	while ( !( UCSR1A & _BV(UDRE1)) );	// Wait for empty buffer
	UDR0 = data;						// Put data into buffer, sends the data 
}

/**
 * Transit a short through UART1
 */
void uart1_send_short(uint16_t data) {
	uart1_send_byte((uint8_t) (data & 0xFF));			// data[0:7]
	uart1_send_byte((uint8_t) ((data >> 8) & 0xFF));	// data[8:15]
}

/**
 * Transmit a long through UART1
 */
void uart1_send_long(uint32_t data) {
	uart0_send_byte((uint8_t) (data & 0xFF));			// data[0:7]
	uart0_send_byte((uint8_t) ((data >> 8) & 0xFF));	// data[8:15]
	uart0_send_byte((uint8_t) ((data >> 16) & 0xFF));	// data[16:23]
	uart0_send_byte((uint8_t) ((data >> 24) & 0xFF));	// data[24:31]
}

/**
 * Transmit an extended through UART1
 */
void uart1_send_extended(uint64_t data) {
	uart0_send_byte((uint8_t) (data & 0xFF));			// data[0:7]
	uart0_send_byte((uint8_t) ((data >> 8) & 0xFF));	// data[8:15]
	uart0_send_byte((uint8_t) ((data >> 16) & 0xFF));	// data[16:23]
	uart0_send_byte((uint8_t) ((data >> 24) & 0xFF));	// data[24:31]
	uart0_send_byte((uint8_t) ((data >> 32) & 0xFF));	// data[32:39]
	uart0_send_byte((uint8_t) ((data >> 40) & 0xFF));	// data[40:47]
	uart0_send_byte((uint8_t) ((data >> 48) & 0xFF));	// data[48:55]
	uart0_send_byte((uint8_t) ((data >> 56) & 0xFF));	// data[56:63]
}

/**
 * Initialize UART0
 *
 * For thread-based operation
 */
void init_uart0(uint16_t baud_rate, uint8_t rxtx_mode, uint8_t rxtx_isr_enable) {	
	cbi(PRR, PRUSART0);		// Clear UART0 power reduction register bit

	uint16_t uart0_ubbr = ((uint16_t) (((float)CPU_FREQUENCY_HZ) / 
		(16*(float)baud_rate) - 0.5));	// Calculate UART0 UBBR

	/* Set UBBR0 register, high before low */
	UBRR0H = (uart0_ubbr >> 8) & 0xFF;
	UBRR0L = uart0_ubbr & 0xFF;

	UCSR0A = 0b00000000;	// 7 - RXCn: UART RX complete flag
							// 6 - TXCn: UART TX complete flag
							// 5 - UDREn: UART DRE flag
							// 4 - FEn: Frame error flag
							// 3 - DORn: Data overrun flag
							// 2 - UPEn: Parity error flag
							// 1 - U2Xn: Double UART speed mode
							// 0 - MPCMn: Multi processor communication mode

	UCSR0B = 0b00000000; 	// 7 - RXCIEn: RX Complete Interrupt Enable
							// 6 - TXCIEn: TX Complete Interrupt Enable
							// 5 - UDRIEn: UART DRE Interrupt Enable
							// 4 - RXENn: Enable receiver
							// 3 - TXENn: Enable transmitter
							// 2 - UCSZn2: Character size
							// 1 - RXB8n: Receive Data Bit 8
							// 0 - TXB8n: Transmit Data Bit 8
	
	/* Determine RXTX mode */
	switch(rxtx_mode) {
		case UART_RXTX_MODE:	// RXTX mode
			sbi(UCSR0B, RXEN0);	// Enable RX for UART0
			sbi(UCSR0B, TXEN0);	// Enable TX for UART0
			break;
		case UART_RX_MODE:		// RX only mode
			sbi(UCSR0B, RXEN0);	// Enable RX for UART0
			break;
		case UART_TX_MODE:		// TX only mode
			sbi(UCSR0B, TXEN0);	// Enable TX for UART0
			break;
	}
	
	/* Determine ISR mode */
	switch(rxtx_isr_enable) {
		case UART_RXTX_ISR_ENABLE:	// RXTX ISR mode
			sbi(UCSR0B, RXCIE0);	// Enable RX ISR for UART0
			sbi(UCSR0B, TXCIE0);	// Enable TX ISR for UART0
			break;
		case UART_RX_ISR_ENABLE:	// RX only ISR mode
			sbi(UCSR0B, RXCIE0);	// Enable RX ISR for UART0
			break;
		case UART_TX_ISR_ENABLE:	// TX only ISR mode
			sbi(UCSR0B, TXCIE0);	// Enable TX ISR for UART0
			break;
		case UART_NO_ISR_ENABLE:	// No ISR mode
			break;
	}

	UCSR0C = 0b00000110;	// 7:6 - UMSELn1:0: UART Mode Select
							//		00 - Asynchronous UART
							//		01 - Synchronous UART
							//		10 - Reserved
							//		11 - Master SPI
							// 5:4 - UPMn1:0: Parity mode
							//		00 - Disabled
							//		01 - Reserved
							//		10 - Enabled, Even
							//		11 - Enabled, Odd
							// 3 - USBSn: Select stop bit length
							//		0 - 1-bit
							//		1 - 2-bit
							// 2:1 - UCSZn1:0: Character size, with UCSRB.2
							//		000 - 5 bit
							//		001 - 6 bit
							//		010 - 7 bit
							//		011 - 8 bit
							//		100 - Reserved
							//		101 - Reserved
							//		110 - Reserved
							//		111 - 9 bit
							// 0 - UCPOLn: Select clock polarity
							//		0 - Rising XCK edge
							//		1 - Falling XCK edge 						
}

/**
 * Initialize UART1
 *
 * For thread-based operation
 */
void init_uart1(uint16_t baud_rate, uint8_t rxtx_mode, uint8_t rxtx_isr_enable) {	
	cbi(PRR, PRUSART1);		// Clear UART1 power reduction register bit

	uint16_t uart1_ubbr = ((uint16_t) (((float)CPU_FREQUENCY_HZ) / 
		(16*(float)baud_rate) - 0.5));		// Calculate UART1 UBBR
	
	/* Set UBBR1 register, high before low */
	UBRR1H = (uart1_ubbr >> 8) & 0xFF;
	UBRR1L = uart1_ubbr & 0xFF;	

	UCSR1A = 0b00000000;	// 7 - RXCn: UART RX complete flag
							// 6 - TXCn: UART TX complete flag
							// 5 - UDREn: UART DRE flag
							// 4 - FEn: Frame error flag
							// 3 - DORn: Data overrun flag
							// 2 - UPEn: Parity error flag
							// 1 - U2Xn: Double UART speed mode
							// 0 - MPCMn: Multi processor communication mode

	UCSR1B = 0b00000000; 	// 7 - RXCIEn: RX Complete Interrupt Enable
							// 6 - TXCIEn: TX Complete Interrupt Enable
							// 5 - UDRIEn: UART DRE Interrupt Enable
							// 4 - RXENn: Enable receiver
							// 3 - TXENn: Enable transmitter
							// 2 - UCSZn2: Character size
							// 1 - RXB8n: Receive Data Bit 8
							// 0 - TXB8n: Transmit Data Bit 8
	
	/* Determine RXTX mode */
	switch(rxtx_mode) {
		case UART_RXTX_MODE:	// RXTX mode
			sbi(UCSR1B, RXEN1);	// Enable RX for UART1
			sbi(UCSR1B, TXEN1);	// Enable TX for UART1
			break;
		case UART_RX_MODE:		// RX only mode
			sbi(UCSR1B, RXEN1);	// Enable RX for UART1
			break;
		case UART_TX_MODE:		// TX only mode
			sbi(UCSR1B, TXEN1);	// Enable TX for UART1
			break;
	}
	
	/* Determine RXTX ISR mode */
	switch(rxtx_isr_enable) {
		case UART_RXTX_ISR_ENABLE:	// RXTX ISR mode
			sbi(UCSR1B, RXCIE1);	// Enable RX for UART1
			sbi(UCSR1B, TXCIE1);	// Enable TX for UART1
			break;
		case UART_RX_ISR_ENABLE:	// RX only ISR mode
			sbi(UCSR1B, RXCIE1);	// Enable RX for UART1
			break;
		case UART_TX_ISR_ENABLE:	// TX only ISR mode
			sbi(UCSR1B, TXCIE1);	// Enable TX for UART1
			break;
		case UART_NO_ISR_ENABLE:	// No ISR mode
			break;
	}

	UCSR1C = 0b00000110;	// 7:6 - UMSELn1:0: UART Mode Select
							//		00 - Asynchronous UART
							//		01 - Synchronous UART
							//		10 - Reserved
							//		11 - Master SPI
							// 5:4 - UPMn1:0: Parity mode
							//		00 - Disabled
							//		01 - Reserved
							//		10 - Enabled, Even
							//		11 - Enabled, Odd
							// 3 - USBSn: Select stop bit length
							//		0 - 1-bit
							//		1 - 2-bit
							// 2:1 - UCSZn1:0: Character size, with UCSRB.2
							//		000 - 5 bit
							//		001 - 6 bit
							//		010 - 7 bit
							//		011 - 8 bit
							//		100 - Reserved
							//		101 - Reserved
							//		110 - Reserved
							//		111 - 9 bit
							// 0 - UCPOLn: Select clock polarity
							//		0 - Rising XCK edge
							//		1 - Falling XCK edge 						
}