/*** Mathematic library for the TET System
 *
 * Ruibing Wang
 * Abiomed
 *
 * Based on codes by Prof Bruce Land, Cornell University
 */

/* Includes */
#include <math.h>

/* Global/System Constants */
#define enable	1
#define disable	0

/* Program Setting for which filter to enable, only one can be enabled at a time */
#define IIR2_general	disable
#define IIR2_butter		disable
#define IIR1			enable 

/* Type conversions to and from 8:8 fixed point representation macros */
#define int2fix(a)	(((int)(a))<<8)            //Convert char to fix. a is a char
#define fix2int(a)	((signed char)((a)>>8))    //Convert fix to char. a is an int
#define float2fix(a)	((int)((a)*256.0))         //Convert float to fix. a is a float
#define fix2float(a)	((float)(a)/256.0)         //Convert fix to float. a is an int 

/* Multiplication and divisions for fixed point macros, twice as fast as floating mul/div */
#define multfix_slow(a,b)	((int)((((long)(a))*((long)(b)))>>8)) //multiply two fixed #
#define divfix(a,b)  ((int)((((long)(a))<<8)/((long)(b)))) //divide two fixed #

/* Square root for fixed point macros, may be slower than floating version */
#define sqrtfix_slow(a)		(lsqrt(((long)(a))<<8))

void init_IIR2_general(void);
int IIR2_general_slow(int xx);
int IIR2_general_fast(int xx);
int IIR1_low(int xx);
void init_IIR1_low(void);
int IIR2_butter_low(int xx);
int Butter2high(int xx);
int Butter2band(int xx);
void init_IIR2_butter_low(void);
void init_IIR2_butter_high(void);
void init_IIR2_butter_band(void); 

int sinefix(int TargetAngle);
int cosfix(int TargetAngle);
int sqrtfix_fast(int a);
int multfix_fast(int a, int b);
int macfix(int a, int b, int c);
int Butter2band(int xx)

// if IIR2 == enable
// elif IIR1 == enable
// endif

/* IIR state variables -- must be in RAM */
#ifdef IIR2_general
	#pragma regalloc-
	int b1, b2, b3, a2, a3, xn_1, xn_2, yn, yn_1, yn_2;
	#pragma regalloc+
#endif

#ifdef IIR2_butterworth
	#pragma regalloc-
	int b1,a2,a3,xn, xn_1, xn_2, yn, yn_1, yn_2;
	#pragma regalloc+
#endif

#ifdef IIR1
	#pragma regalloc-
	int alpha, yn_1;
	#pragma regalloc+
#endif

/* Sine/Cosine Lookup Table */
flash int Angles[]={
	float2fix(45.0), 
	float2fix(26.565), 
	float2fix(14.0362), 
	float2fix(7.12502),	
	float2fix(3.57633),	
	float2fix(1.78991), 
	float2fix(0.895174), 
	float2fix(0.447614), 
	float2fix(0.223811), 
	float2fix(0.111906), 
	float2fix(0.055953), 
	float2fix(0.027977)
	};

/** 
 * Compute sine of a fixed angle from -90 to +90 degrees 
 */
int sinefix(int TargetAngle) {
	int X, Y, CurrAngle;
	unsigned char Step;
	
	Y = 0;
	CurrAngle = 0;
	X = 0x009b;    //0.607
	
	for(Step = 0; Step < 12; Step++) {	
		int NewX;

		if(TargetAngle > CurrAngle) {	
			NewX = X - (Y >> Step);
			Y = (X >> Step) + Y;
			X = NewX;
			CurrAngle += Angles[Step]; 
		}
		else {	
			NewX = X + (Y >> Step);
			Y = -(X >> Step) + Y;
			X = NewX;
			CurrAngle -= Angles[Step]; 
		}
	} 
	return Y;
}

/**
 * Compute cosine of a fixed angle from -90 to +90 degrees 
 */
int cosfix(int TargetAngle) {
	int X, Y, CurrAngle;
	unsigned char Step;
	
	Y = 0;
	CurrAngle = 0;
	X = 0x009b;    //0.607;
	
	for(Step = 0; Step < 12; Step++) {
		int NewX;

		if(TargetAngle > CurrAngle) {	
			NewX = X - (Y >> Step);
			Y = (X >> Step) + Y;
			X = NewX;
			CurrAngle += Angles[Step]; 
		}
		else {	
			NewX=X + (Y >> Step);
			Y = -(X >> Step) + Y;
			X = NewX;
			CurrAngle -= Angles[Step]; 
		}
	} 
	return X;
}

/**
 * Fast square root operation for fixed 
 */
int sqrtfix_fast(int a) {
	int hi, lo, guess, slp, in, d; 
	
	// Assign linear least squares values 
	if (a >= 0x7900)		{slp=11;in=1489;d=4;}	// 121 to 127
	else if (a >= 0x6400)	{slp=12;in=1369;d=9;}	// 100 
	else if (a >= 0x5100)	{slp=13;in=1254;d=7;}	// 81
	else if (a >= 0x4000)	{slp=15;in=1087;d=5;}	// 64
	else if (a >= 0x3100)	{slp=17;in=959;d=5;}	// 49
	else if (a >= 0x2400)	{slp=20;in=857;d=6;}	// 36
	else if (a >= 0x1900)	{slp=23;in=703;d=9;}	// 25
	else if (a >= 0x1000)	{slp=28;in=575;d=10;}	// 16 
	else if (a >= 0x0900)	{slp=36;in=447;d=9;}	// 9
	else if (a >= 0x0400)	{slp=51;in=318;d=10;}	// 4
	else if (a >= 0x0200)	{slp=75;in=217;d=5;}	// 2
	else if (a >= 0x0100)	{slp=106;in=154;d=3;}	// 1
	else if (a >= 0x0040)	{slp=168;in=94;d=8;}	// 1/4 to 1
	else					{slp=502;in=22;d=10;}	// 1/4
	
	// Do the linear fit
	guess = multfix(a, slp)+in;
	lo = guess - d;
	hi = guess + d;
	
	// Iterate to get the last few bits
	while ((hi-lo) > 0x0002) {
		//TCNT1=0; TCCR1B=1;
		if (multfix(guess, guess) >= a)
			hi = guess;
		else
			lo = guess;
		guess = ((hi + lo) >> 1); 
		//TCCR1B=0; printf("%d ",TCNT1);
	}
	return guess;
}

/*
 * Fast multiplication for fixed point representation in assembly
 * Takes about 52 cycles to complete, 3 times faster than macro version
 */
int multfix_fast(int a, int b) {
	#asm
	push r20
	push r21
   
	LDD  R22, Y+2			;load a
	LDD  R23, Y+3 
	
	LD   R20, Y				;load b
	LDD  R21, Y+1
	
	muls	r23, r21		; (signed)ah * (signed)bh
	mov		r31, r0			; r18, r0
	mul		r22, r20		; al * bl
	mov     r30, r1			; movw	r17:r16, r1:r0
							; mov     r16, r0
	mulsu	r23, r20		; (signed)ah * bl
	add		r30, r0			; r17, r0
	adc		r31, r1			; r18, r1
	mulsu	r21, r22		; (signed)bh * al
	add		r30, r0			; r17, r0
	adc		r31, r1			; r18, r1
	
	pop r21
	pop r20
    #endasm
}

/**
 * Fast fixed multiplication with accumulate
 * Return = a*b + c
 */
int macfix(int a, int b, int c) {
	//r31:r30:r24 += r23:r22 * r21:r20 
	//
	#asm      
	push r20
	push r21
    
	LDD  R22, Y+4		; load a
	LDD  R23, Y+5 
	LDD  R20, Y+2		; load b
	LDD  R21, Y+3
	LD   R30, Y			; load c to lsb result
	LDD  R31, Y+1		; then msb result
	clr  r24			; low order byte
	clr  r27			; permanent zero
	
						; mac operation 
    muls r23, r21		; (signed)ah * (signed)bh
	add	r31, r0
	mul	r22, r20		; al * bl
	add	r24, r0
	adc	r30, r1
	adc	r31, r27
	mulsu r23, r20		; (signed)ah * bl
	add	r30, r0
	adc	r31, r1
	mulsu	r21, r22	; (signed)bh * al
	add	r30, r0
	adc	r31, r1
						; end mac operation
	
	pop r21
	pop r20
	#endasm
}

void init_IIR2_general(void) {
	//2-pole butterworth lowpass, cutoff=0.25 (1/4 Nyquist freq), gain=1 
	b1 = float2fix(0.098) ;
	b2 = float2fix(0.195) ;
	b3 = float2fix(0.098) ;                
	a2 = float2fix(-0.943) ;
	a3 = float2fix(0.333) ;
}

/**
 * Second order IIR Filter
 * Unoptimized, each sample takes 400 cycles to filter
 *
 * y(n) = b(1)*x(n) + b(2)*x(n-1) +  b(3)*x(n-2)
 * 	- a(2)*y(n-1) -  a(3)*y(n-2) 
 * Assumes a(1)=1
 * a's and b's need to be global, in RAM and set to fixed point values  
 * also input and output history values
 * example: 
 *	#pragma regalloc-
 *	int b1,b2,b3,a2,a3, xn_1, xn_2, yn_1, yn_2 ;
 *	#pragma regalloc+ 
 *	b1=0x0010;  
 */
int IIR2_general_slow(int xx) {
    int yy; 
    yy = 0;
    //sum the 5 terms 
    yy = macfix(b1, xx, yy); 
    yy = macfix(b2, xn_1, yy);
    yy = macfix(b3, xn_2, yy);
    yy = macfix(-a2, yn_1, yy);
    yy = macfix(-a3, yn_2, yy); 
    //update the state variables
    xn_2 = xn_1;
    xn_1 = xx;
    yn_2 = yn_1;
    yn_1 = yy;
    return yy;
}

/**
 * Second order IIR Filter
 * Optimized, each sample takes 182 cycles to filter
 * Uses assembly to replace unoptimized, slow version
 *
 * It would not be reasonable to use this fixed-point algorithm with normalized cutoff frequency 
 * below about 0.10 because of the lack of relative precision in the b coefficients. 
 * With a normalized cutoff frequency=0.25, the impulse response is accurate to within 1% of peak response. 
 * With a normalized cutoff frequency=0.10, the impulse response is accurate within 5% of peak response.
 *
 * #pragma regalloc-
 * int b1,b2,b3,a2,a3, xn_1, xn_2, yn_1, yn_2 ;
 * #pragma regalloc+ 
 * b1=0x0010;
 * a1=float2fix(-(value from matlab))   
 */
 #pragma warn-
int IIR2_general_fast(int xx) {
    #asm
    .macro mult_acc			; r31:r30:r24 += r23:r22 * r21:r20
        muls r23, r21		; (signed)ah * (signed)bh
    	add	r31, r0
    	mul	r22, r20		; al * bl
    	add	r24, r0
    	adc r30, r1
    	adc r31, r27
    	mulsu	r23, r20	; (signed)ah * bl
    	add	r30, r0
    	adc	r31, r1
    	mulsu	r21, r22	; (signed)bh * al
    	add	r30, r0
    	adc	r31, r1
    .endm 
    
    push r20			; save parameter regs
    push r21
    
    clr r27				; permanent zero
    clr r24				; clear 24 bit result reg; msb to lsb => r31:r30:r24
    clr r30      
    clr r31
    
    lds  R22, _b1		; load b1 from RAM
    lds  R23, _b1+1 
    ld   R20, Y			; load input parameter xx from stack
    ldd  R21, Y+1
    mult_acc			; b1*xx  
    
    lds  R22, _b2		; load b2 from RAM
    lds  R23, _b2+1 
    lds  R20, _xn_1		; load x(n-1) from RAM
    lds  R21, _xn_1+1
    mult_acc			; b2*x(n-1) 
    
    lds  R22, _b3		; load b3 from RAM
    lds  R23, _b3+1 
    lds  R20, _xn_2		; load x(n-2) from RAM
    lds  R21, _xn_2+1
    mult_acc			; b3*x(n-2)
    
    lds  R22, _a2		; load -a2 from RAM
    lds  R23, _a2+1 
    lds  R20, _yn_1		; load y(n-1) from RAM
    lds  R21, _yn_1+1
    mult_acc			; -a2*y(n-1)  
    
    lds  R22, _a3		; load -a3 from RAM
    lds  R23, _a3+1 
    lds  R20, _yn_2		; load y(n-2) from RAM
    lds  R21, _yn_2+1
    mult_acc			; -a3*y(n-2)
    
    lds  R20, _xn_1		; load x(n-1) from RAM
    lds  R21, _xn_1+1
    sts  _xn_2, r20		; store x(n-2) to RAM
    sts  _xn_2+1, R21 
    ld   R20, Y			; load input parameter xx from stack
    ldd  R21, Y+1
    sts  _xn_1, r20		; store x(n-1) to RAM
    sts  _xn_1+1, R21 
    lds  R20, _yn_1		; load y(n-1) from RAM
    lds  R21, _yn_1+1
    sts  _yn_2, R20		; store y(n-2) to RAM
    sts  _yn_2+1, R21 
    sts  _yn_1, r30		; store new output as y(n-1) to  RAM
    sts  _yn_1+1, r31 
	
    pop r21				; restore parameter regs
    pop r20
    #endasm
}
#pragma warn+

/**
 * A first order IIR lowpass filter
 * Each sample takes about 69 cycles to execute
 * Use this filter for normalized cutoff frequencies in the 
 * range of 0.01 to 0.4 with reasonable accuracy.
 *
 * y(n+1) = y(n)*alpha + input*(1-alpha) 
 * => y(n) = (y(n-1)-input)*alpha + input
 * where alpha = exp(- pi * normalized freq) 
 * and normalized freq is cutoff/(Fs/2)
 */
int IIR1_low(int xx) {
	#asm
	.macro mult_acc			; r31:r30:r24 += r23:r22 * r21:r20
		muls	r23, r21	; (signed)ah * (signed)bh
		add	r31, r0
		mul	r22, r20		; al * bl
		add	r24, r0
		adc	r30, r1
		adc	r31, r27
		mulsu r23, r20		; (signed)ah * bl
		add	r30, r0
		adc	r31, r1
		mulsu	r21, r22	; (signed)bh * al
		add	r30, r0
		adc	r31, r1
	.endm 
    
	push r20			; save parameter regs
	push r21
    
	clr r27				; permanent zero
	clr r24				; clear 24 bit result reg; msb to lsb => r31:r30:r24
    
	ld	R30, Y			; load input parameter xx from stack
	ldd	R31, Y+1		; then put in result location
	
	lds	R20, _yn_1		; load y(n-1) from  RAM
	lds	R21, _yn_1+1
	sub	r20, r30		; form y(n-1) - xx
	sbc	r21, r31       
	lds	R22, _alpha		; load alpha from RAM
	lds	R23, _alpha+1 
	mult_acc			; alpha*(yn_1 - xx)) + xx
    
	sts	_yn_1, r30		; store new output as y(n-1) to  RAM
	sts	_yn_1+1, r31 
	
	pop	r21				; restore parameter regs
	pop	r20
	
	#endasm 
}

void init_IIR1_low(void) {
	//1-pole RC lowpass, cutoff=0.0424 (1/4 Nyquist freq), gain=1 
	alpha = float2fix(0.875) ;
}

/**
 * Butterworth second order, lowpass filter
 * Each sample takes 148 cycles to filter
 *
 * Optimized so only b1, a2, and a3 need to be specified
 * Since b(1)=b(3)=0.5*b(2), b(1)*x(n)+b(2)*x(n-1)+b(3)*x(n-2)
 *	=> b(1)*x(n)+2*b(1)*x(n-1)+b(1)*x(n-2)
 *	=> b(1)* (x(n)+(x(n-1)<<1)+x(n-2))
 *   
 * Since adding and shifting are much faster than multiplying, it
 * is faster than the general IIR2 filter.
 */
int IIR2_butter_low(int xx) {
	#asm
	.macro mult_acc			; r31:r30:r24 += r23:r22 * r21:r20
		muls r23, r21		; (signed)ah * (signed)bh
		add	r31, r0
		mul	r22, r20		; al * bl
		add	r24, r0
		adc	r30, r1
		adc	r31, r27
		mulsu r23, r20		; (signed)ah * bl
		add	r30, r0
		adc	r31, r1
		mulsu	r21, r22	; (signed)bh * al
		add	r30, r0
		adc	r31, r1
	.endm 
    
	push r20				; save parameter regs
	push r21
    
	clr r27					; permanent zero
	clr r24 				; clear 24 bit result reg; msb to lsb => r31:r30:r24
	clr r30      
	clr r31
    
	lds  R22, _xn_1		; load x(n-1) from  RAM
	lds  R23, _xn_1+1
	lsl  r22			; mult by two
	rol  r23
	ld   R20, Y			; load input parameter xx from stack
	ldd  R21, Y+1
	add  r20, r22		; form xx+2x(n-1)
	adc  r21, r23       
	lds  R22, _xn_2		; load x(n-2) from  RAM
	lds  R23, _xn_2+1
	add  r20, r22		; form xx+2x(n-1)+x(n-2)
	adc  r21, r23 
	lds  R22, _b1		; load b1 from RAM
	lds  R23, _b1+1 
	mult_acc			; b1*(xx+2*x(n-1)+x(n-2)) 
    
	lds  R22, _a2		; load -a2 from RAM
	lds  R23, _a2+1 
	lds  R20, _yn_1		; load y(n-1) from  RAM
	lds  R21, _yn_1+1
	mult_acc			; -a2*y(n-1)  
    
	lds  R22, _a3		; load -a3 from RAM
	lds  R23, _a3+1 
	lds  R20, _yn_2		; load y(n-2) from  RAM
	lds  R21, _yn_2+1
	mult_acc			; -a3*y(n-2)
    
    lds  R20, _xn_1		; load x(n-1) from  RAM
	lds  R21, _xn_1+1
	sts  _xn_2, r20		; store x(n-2) to  RAM
	sts  _xn_2+1, R21 
	ld   R20, Y			; load input parameter xx from stack
	ldd  R21, Y+1
	sts  _xn_1, r20		; store x(n-1) to  RAM
	sts  _xn_1+1, R21 
	lds  R20, _yn_1		; load y(n-1) from  RAM
	lds  R21, _yn_1+1
	sts  _yn_2, R20		; store y(n-2) to  RAM
	sts  _yn_2+1, R21 
	sts  _yn_1, r30		; store new output as y(n-1) to  RAM
	sts  _yn_1+1, r31 
	
	pop r21				;restore parameter regs
	pop r20	
	#endasm 
}

/**
 * Butterworth second order, highpass filter
 * Each sample takes 148 cycles to filter
 *
 * Optimized so only b1, a2, and a3 need to be specified
 * Since b(1)=b(3)=-0.5*b(2), b(1)*x(n)+b(2)*x(n-1)+b(3)*x(n-2)
 *	=> b(1)*x(n)-2*b(1)*x(n-1)+b(1)*x(n-2) 
 *	=> b(1)* (x(n)-(x(n-1)<<1)+x(n-2))
 */
int Butter2high(int xx) {
    #asm
    .macro mult_acc         ; r31:r30:r24 += r23:r22 * r21:r20
        muls r23, r21		; (signed)ah * (signed)bh
    	add	r31, r0
    	mul	r22, r20		; al * bl
    	add	r24, r0
    	adc	r30, r1
    	adc	r31, r27
    	mulsu r23, r20		; (signed)ah * bl
    	add	r30, r0
    	adc	r31, r1
    	mulsu	r21, r22	; (signed)bh * al
    	add	r30, r0
    	adc	r31, r1
    .endm
 
    push r20    		; save parameter regs
    push r21
    
	clr r27             ; permanent zero
    clr r24             ; clear 24 bit result reg; msb to lsb => r31:r30:r24
    clr r30      
    clr r31
    
    lds  R22, _xn_1     ; load x(n-1) from  RAM
	lds  R23, _xn_1+1
	lsl  r22            ; mult by two
	rol  r23
	ld   R20, Y         ; load input parameter xx from stack
	ldd  R21, Y+1
	sub  r20, r22       ; form xx-2x(n-1)
	sbc  r21, r23       
	lds  R22, _xn_2     ; load x(n-2) from  RAM
	lds  R23, _xn_2+1
	add  r20, r22       ; form xx-2x(n-1)+x(n-2)
	adc  r21, r23 
	lds  R22, _b1       ; load b1 from RAM
	lds  R23, _b1+1 
    mult_acc            ; b1*(xx-2*x(n-1)+x(n-2)) 
    
    lds  R22, _a2       ; load -a2 from RAM
	lds  R23, _a2+1 
	lds  R20, _yn_1     ; load y(n-1) from  RAM
	lds  R21, _yn_1+1
    mult_acc            ; -a2*y(n-1)  
    
    lds  R22, _a3       ; load -a3 from RAM
	lds  R23, _a3+1 
	lds  R20, _yn_2     ; load y(n-2) from  RAM
	lds  R21, _yn_2+1
    mult_acc            ; -a3*y(n-2)
    
    lds  R20, _xn_1     ; load x(n-1) from  RAM
	lds  R21, _xn_1+1
	sts  _xn_2, r20     ; store x(n-2) to  RAM
	sts  _xn_2+1, R21 
	ld   R20, Y         ; load input parameter xx from stack
	ldd  R21, Y+1
	sts  _xn_1, r20     ; store x(n-1) to  RAM
	sts  _xn_1+1, R21 
	lds  R20, _yn_1     ; load y(n-1) from  RAM
	lds  R21, _yn_1+1
	sts  _yn_2, R20     ; store y(n-2) to  RAM
	sts  _yn_2+1, R21 
	sts  _yn_1, r30     ; store new output as y(n-1) to  RAM
	sts  _yn_1+1, r31 
	
    pop r21             ; restore parameter regs
	pop r20	
	#endasm 
}

/**
 * Butterworth second order, bandpass filter
 * Each sample takes 140 cycles to filter
 *
 * Optimized so only b1, a2, and a3 need to be specified
 * Since b(1)=-b(3) and b(2)=0, b(1)*x(n)+b(2)*x(n-1)+b(3)*x(n-2) 
 *	=> b(1)*x(n)+0*b(1)*x(n-1)-b(1)*x(n-2) 
 *	=> b(1)*(x(n)-x(n-2))
 */
int Butter2band(int xx) {
    #asm
    .macro mult_acc         ; r31:r30:r24 += r23:r22 * r21:r20
        muls r23, r21		; (signed)ah * (signed)bh
    	add	r31, r0
    	mul	r22, r20		; al * bl
    	add	r24, r0
    	adc	r30, r1
    	adc	r31, r27
    	mulsu r23, r20		; (signed)ah * bl
    	add	r30, r0
    	adc	r31, r1
    	mulsu	r21, r22	; (signed)bh * al
    	add	r30, r0
    	adc	r31, r1
    .endm
 
    push r20    		; save parameter regs
    push r21
    
	clr r27             ; permanent zero
    clr r24             ; clear 24 bit result reg; msb to lsb => r31:r30:r24
    clr r30      
    clr r31
    
    lds  R22, _xn_2     ; load x(n-2) from  RAM
	lds  R23, _xn_2+1
	ld   R20, Y         ; load input parameter xx from stack
	ldd  R21, Y+1
	sub  r20, r22       ; form xx-x(n-2)
	sbc  r21, r23       
	lds  R22, _b1       ; load b1 from RAM
	lds  R23, _b1+1 
    mult_acc            ; b1*(xx-x(n-2)) 
    
    lds  R22, _a2       ; load -a2 from RAM
	lds  R23, _a2+1 
	lds  R20, _yn_1     ; load y(n-1) from  RAM
	lds  R21, _yn_1+1
    mult_acc            ; -a2*y(n-1)  
    
    lds  R22, _a3       ; load -a3 from RAM
	lds  R23, _a3+1 
	lds  R20, _yn_2     ; load y(n-2) from  RAM
	lds  R21, _yn_2+1
    mult_acc            ; -a3*y(n-2)
    
    lds  R20, _xn_1     ; load x(n-1) from  RAM
	lds  R21, _xn_1+1
	sts  _xn_2, r20     ; store x(n-2) to  RAM
	sts  _xn_2+1, R21 
	ld   R20, Y         ; load input parameter xx from stack
	ldd  R21, Y+1
	sts  _xn_1, r20     ; store x(n-1) to  RAM
	sts  _xn_1+1, R21 
	lds  R20, _yn_1     ; load y(n-1) from  RAM
	lds  R21, _yn_1+1
	sts  _yn_2, R20     ; store y(n-2) to  RAM
	sts  _yn_2+1, R21 
	sts  _yn_1, r30     ; store new output as y(n-1) to  RAM
	sts  _yn_1+1, r31 
	
    pop r21             ; restore parameter regs
	pop r20
	#endasm 
}

void init_IIR2_butter_low(void) {
	//2-pole butterworth lowpass, cutoff=0.25 (1/4 Nyquist freq), gain=1 
	b1 = float2fix(0.098) ;                
	a2 = float2fix(0.943) ;   //note that sign is negated from design input
	a3 = float2fix(-0.333) ;  //note that sign is negated from design input 
}

void init_IIR2_butter_high(void) {
	//2-pole butterworth highpass, cutoff=0.25
	b1 = float2fix(0.569) ;               
	a2 = float2fix(0.943) ;   //note that sign is negated from design input
	a3 = float2fix(-.333) ;  //note that sign is negated from design input
}

void init_IIR2_butter_band(void) {
	//2-pole butterworth bandpass, cutoff=[0.25 0.35]
	b1 = float2fix(0.137) ;               
	a2 = float2fix(1.028) ;   //note that sign is negated from design input
	a3 = float2fix(-0.727) ;  //note that sign is negated from design input
}
