/**
 * Cornell University 
 * ECE 476 Spring 2007
 * Final Project
 *
 * Authors: Ruibing Wang (rw98), Ali Goheer (mag97)
 * 
 * Based upon wireless protocal library by Meghan Desai
 */

// Includes
#include <mega32.h>
#include <stdio.h>
#include <stdlib.h>  
#include <string.h>
#include <delay.h>

// Common Baud Rate for the UART
// Value 	- Speed
// 416 		- 2400			
// 249 		- 4000
// 207 		- 4800 (max) 
#define BAUD_RATE	249	

// Max expected packet and data size
#define MAX_RX_PACKET_LENGTH	32
#define MAX_RX_DATA_LENGTH	16
#define MAX_TX_PACKET_LENGTH	32 
#define MAX_TX_DATA_LENGTH	16

/**
 * Reservations of specific chars for all transmissions:
 * 
 * 0b10XXXXXX	- code lookup table
 * 0b11000011	- 
 * 0b11000110	- next
 * 0b11000101	- a
 * 0b11001100	- d
 * 0b11001010	- start
 * 0b11001001	-
 * 0b11011000	- c
 * 0b11010100	- end
 * 0b11010010	- b
 * 0b11010001	- resend
 */

// Int value
#define sync_char		0b10101010	// 170

#define garbage_char_a	0b11000101	// 197
#define garbage_char_b	0b11010010	// 210
#define garbage_char_c	0b11011000	// 216
#define garbage_char_d	0b11001100	// 204

#define start_char	0b11001010	// 202
#define end_char		0b11010100	// 212

#define next_char		0b11000110	// 198
#define resend_char	0b11001001	// 201

// Decode Lookup Table
flash unsigned char code[16] =  {
			// Actual nibble	- Int value
	0b10001011,	// 0000			- 139
	0b10001101,	// 0001			- 141
	0b10010011,	// 0010			- 147
	0b10010101,	// 0011			- 149
	0b10010110,	// 0100			- 150
	0b10011001,	// 0101			- 153
	0b10011010,	// 0110			- 154
	0b10011100,	// 0111			- 156
	0b10100011,	// 1000			- 163
	0b10100101,	// 1001			- 165
	0b10100110,	// 1010			- 166
	0b10101001,	// 1011			- 169
	0b10101100,	// 1100			- 172
	0b10110001,	// 1101			- 177
	0b10110010,	// 1110			- 178
	0b10110100	// 1111			- 180
};

// General function prototypes
void tx_init(void);
void rx_init(void);
unsigned char decode_byte(unsigned char msb, unsigned char lsb);
unsigned char decode(unsigned char msbyte);

// RX variables
unsigned char rx_is_sync, rx_is_start, rx_is_end, rx_is_ready;	// RX state
unsigned char rx_temp, rx_temp_hash, rx_temp_response;	// RX's temp variables
unsigned char rx_data_id, rx_data_index, rx_data_length, rx_data_hash, rx_data[MAX_RX_DATA_LENGTH];	// RX data
unsigned char rx_packet_length, rx_packet_index, rx_packet_data[MAX_RX_PACKET_LENGTH];	// RX packet, holds data
unsigned char rx_request_resend, rx_request_next;	// RX two-way communciation

// RX function prototypes
unsigned char rx_data_verification(void);
void rx_reset(void);
unsigned char is_rx_ready(void);
unsigned char get_rx_data_length(void);
unsigned char get_rx_packet_length(void);
void get_rx_data(unsigned char *data);
void get_rx_packet(unsigned char *data);

// TX variables
enum {TX_INIT, TX_BUSY};	// TX states
unsigned char tx_state;
unsigned char tx_data[MAX_TX_DATA_LENGTH], tx_data_length, tx_data_hash;	// TX data
unsigned char tx_packet_data[MAX_TX_PACKET_LENGTH], tx_packet_data_length, tx_packet_data_index;	// TX packet, holds data
unsigned char tx_request_resend, tx_request_next;	// TX two-way communcation

// TX function prototypes
void tx_response(unsigned char response, unsigned char fast_mode);
void tx_byte_array(unsigned char *data, unsigned char length, unsigned char id, unsigned char fast_mode);
unsigned char tx_is_ready(void);
void tx_reset(void);


/**
 * General Function
 *
 * This function sets the UCSRB and UBRRL register appropriatly.
 * Can be configures to initiate any universal features (i.e. timer)
 */
void rxtx_init(void) {
	UBRRL = BAUD_RATE;
}

/**
 * General Function
 *
 * This function enables the rx feature of the module,
 * and it disables the tx feature.
 * It will put PORTD.1 on high and PORTD.0 to improve the signal
 */
void rx_enable(void) {
	UCSRB.0 = 0;	//Bit 0 - TXB8
	UCSRB.1 = 0;	//Bit 1 - RXB8
	UCSRB.2 = 0;	//Bit 2 - UCSZ2
	UCSRB.3 = 0;	//Bit 3 - TXEN
	UCSRB.4 = 1;	//Bit 4 - RXEN
	UCSRB.5 = 0;	//Bit 5 - UDRIE
	UCSRB.6 = 0;	//Bit 6 - TXCIE
	UCSRB.7 = 1;	//Bit 7 - RXCIE
	
	PORTD.1 = 1;	// Must put TX pin on high during RX
			// or it will attenuate incoming signals
	PORTD.0 = 0;
}

/**
 * General Function
 *
 * This function enables the tx feature of the module,
 * and it disables the rx feature.
 * It will put PORTD.0 on high and PORTD.1 to improve the signal
 */
void tx_enable(void) {
	UCSRB.0 = 0;	//Bit 0 - TXB8
	UCSRB.1 = 0;	//Bit 1 - RXB8
	UCSRB.2 = 0;	//Bit 2 - UCSZ2
	UCSRB.3 = 1;	//Bit 3 - TXEN
	UCSRB.4 = 0;	//Bit 4 - RXEN
	UCSRB.5 = 1;	//Bit 5 - UDRIE
	UCSRB.6 = 0;	//Bit 6 - TXCIE
	UCSRB.7 = 0;	//Bit 7 - RXCIE
	
	PORTD.0 = 1;	// Must put RX pin on high during TX
			// or it will attenuate outgoing signals 
	PORTD.1 = 0;	
}

/**
 * General Function
 *
 * This function returns the decoded character formed from the two input bytes.
 * It looks through the lookup table for the msbyte (first nibble)
 * and the lsbyte (second nibble).
 *
 * When decoding an RX packet, 
 * the earlier (leftmost) byte should be the msb,
 * and the later (rightmost) byte should be the lsb. 
 */
unsigned char decode_byte(unsigned char msb, unsigned char lsb) {
	unsigned char i, buffer;
	
	// Look through the code table for msb
	for (i = 0; i < 16; i++) {
		if(msb==code[i]) {
			buffer = i;
		}
	}
	
	// Look through the code table for lsb
	for (i = 0; i < 16; i++) {
		if(lsb==code[i]) {
			buffer = (buffer<<4);
			buffer = buffer | i; 
		}
	}
	
	return buffer;
}

/**
 * General Function
 *
 * This function returns the decoded nibble formed from a single byte.
 *
 * Used only with nibble data, like the id.
 */
unsigned char decode(unsigned char msbyte) {
	unsigned char i, buffer;
	
	// Look through the code table for the byte
	buffer = start_char;
	for (i = 0; i < 16; i++) {
		if(msbyte==code[i])
			buffer=i;
	}
	return buffer;
}

/**
 * RX Function
 *
 * The RXC ISR will retrieve, parse, and error correct all incoming transmissions.
 * It is written so that the heavy calculations only happen when the end char is received,
 * meaning we can take as much time as we want.
 * 
 * Basic format of raw packets:
 * 	Data:	sync | start | id | length | data | hash | end
 *	Resend:	sync | start | resend | end
 *	Next:	sync | start | next | end
 *
 * Basic format of a parsed data packet:
 *	id[0] | length[1:2] | data[3:x] | hash[x+1:x+3]
 *	 1    |     2       |  2*length |       2
 */
interrupt [USART_RXC] void UART_RX(void) {
	unsigned char i;
	
	// Retrieve the byte
	rx_temp = UDR;
	
	// Only received when we have not received the end char
	if (!rx_is_ready) {	
		// This switch will parse the new character retrieved from the ISR,
		// it will only accept known raw bytes
		switch (rx_temp) {
			case sync_char:
				// A sync char,
				// may be a new message when last stalled,
				// reset everything
				rx_is_sync = 1;		
				rx_is_start = 0;	
				rx_is_end = 0;
				rx_packet_index = 0;
				rx_packet_length = 0;
				rx_temp_response = 0;
				break;
			case garbage_char_a:		
			case garbage_char_b:
			case garbage_char_c:
			case garbage_char_d:
				// Garbage chars, account for but ignore
				break;
			case next_char:			
				// A next char, comes after a start char
				// used for slaves to send to masters to inform them 
				// whether they have received something properly
				rx_temp_response = rx_temp;
				break;
			case start_char:
				// A start char, comes after a sync char
				// used to syncing the uart,
				// ended by a end char
				if (rx_is_sync) {
					rx_is_sync = 1;
					rx_is_start = 1;
					rx_is_end = 0;
				}
				break;	
			case end_char:
				// An end char, comes after data
				// but may be used to indicate abruptness/corrupt data,
				// indicates the end of a message
				if (rx_is_sync && rx_is_start) {
					rx_is_sync = 1;
					rx_is_start = 1;
					rx_is_end = 1;
					
					// Not dealing with a response message (i.e. no resend/next char),
					// must be the slave
					// Parse, convert, and ready the data packet
					// Then perform data integrity verification,
					// depending on the result will either flag a resend or next to master
                    			if (rx_temp_response == 0) {
						// Encoded packet length
						rx_packet_length = rx_packet_index;
					    	
						// Decoded packet id
						rx_data_id = decode(rx_packet_data[0]);
						
						// Decoded data length
						rx_data_length = decode_byte(rx_packet_data[1],rx_packet_data[2]);
						
						// Decoded data buffer
						rx_temp_hash = 0;
						for (i = 0; i < (rx_packet_length-2); i++) {
							rx_data[i] = decode_byte(rx_packet_data[(unsigned char)(i*2+3)],rx_packet_data[(unsigned char)(i*2+4)]);
							//rx_temp_hash ^= rx_data[i];
						}
					
						// Hash of data buffer
						for (i = 0; i < ((rx_packet_length-5)/2); i++) {
							rx_temp_hash ^= rx_data[i];
						}
						
						// Decoded expected data hash
						rx_data_hash = decode_byte(rx_packet_data[rx_packet_length-2],rx_packet_data[rx_packet_length-1]);
	                    			
						// Data verifcation test
						if (rx_data_verification()) {
							tx_request_resend = 0;
							tx_request_next = 1;
							rx_is_ready = 1;
						}
						else {
							tx_request_resend = 1;
							tx_request_next = 0;
							rx_reset();
						}
	     				}
					// Dealing with a response message (i.e. resend/next char),
					// must be the master
					// Respond accordingly
	     				else {
	     					if (rx_temp_response == next_char) {
	     				    		rx_request_next = 1;
	     						rx_request_resend = 0;
	     					}
	     					else if (rx_temp_response == resend_char) {
	     						rx_request_resend = 1;
	     						rx_request_next = 0; 
	     					}
	     					rx_reset();
	     				}            
				}
				break;
			case 139:
			case 141:
			case 147:
			case 149:
			case 150:
			case 153:
			case 154:
			case 156:
			case 163:
			case 165:
			case 166:
			case 169:
			case 172:
			case 177:
			case 178:
			case 180:
				// A byte in the code lookup table,
				// can only come after sync and start char,
				// just append to the data packet buffer for now
				if (rx_is_sync && rx_is_start && !rx_is_end) {
					rx_packet_data[rx_packet_index++] = rx_temp;
					
				}
				break;
		}
	}	
} 

/**
 * RX Function
 * 
 * Verifies the received data according to the following rules:
 *	1) Packet length must be an odd number (id + length + data count + hash = 1 + 2 + 2*x + 2 = odd)
 *	2) Must pass hash verification
 *	3) Actual data length must equal expected data length
 * If all three are met, return 1
 * If any of the three fails, return 0
 */
unsigned char rx_data_verification(void) {
	unsigned char oddness_check, hash_check, expectness_check, length_check;
	oddness_check = 0;
	hash_check = 0;
	expectness_check = 0;
	length_check = 0;
	
	// Test 1 - oddness
	if (rx_packet_length%2 == 1) {
		oddness_check = 1;
	}
	// Test 2 - hash
	if (rx_temp_hash == rx_data_hash) {
		hash_check = 1;
	}
	// Test 3 - expect
	if ((rx_packet_length - 5) == (rx_data_length*2)) {
		expectness_check = 1;
	}
	
	return oddness_check && hash_check && hash_check;
} 

/**
* RX Function
*
* This function is used to reset the receiver to allow it to receive the next byte. 
* This function must be called inorder to receive the next transmission after one has been received.
* But we should hold off calling this after a RX has been completed so we can get the data off 
*/
void rx_reset(void) {
	rx_is_sync = 0;
	rx_is_start = 0;
	rx_is_end = 0;

	rx_data_index = 0;
	rx_packet_index = 0;
	rx_is_ready = 0;	// Becareful of this
}

/*
 * RX Function
 *
 * This function returns whether the data from the RX is ready for use
 * It will be a verified data
 * Use this function to determine whether we can take the data,
 * take the data, and then reset everything
 */
unsigned char is_rx_ready(void) {
	return rx_is_ready;
}

/**
 * RX Function
 *
 * This function returns the length of the current data buffer
 * Should only be called after checking whether the data is ready
 */
unsigned char get_rx_data_length(void) {
	return rx_data_length;
}

/**
 * RX Function
 *
 * This function returns the length of the current packet buffer
 * Should only be called after checking whether the packet is ready
 */
unsigned char get_rx_packet_length(void) {
	return rx_packet_length;
}

/**
 * This function returns the owner id of the current data
 * Should only be called after checking whether the data is ready
 */
unsigned char get_rx_data_id(void) {
	return rx_data_id;
}

/**
 * RX Function
 *
 * Make a copy of the data buffer,
 * hence retrieving it, using a for-loop copy
 * to a buffer passed by pointer
 */
void get_rx_data(unsigned char *data) {
	unsigned char i;
	for (i = 0; i < rx_data_length; i++) {
		data[i] = rx_data[i];
	}
}

/**
 * RX Function
 *
 * Makes a copy of the packet buffer,
 * hence retrieving it, using a for-loop copy
 * to a buffer passed by poiint
 */
void get_rx_packet(unsigned char *data) {
	unsigned char i;
	for (i = 0; i < rx_packet_length; i++) {
		data[i] = rx_packet_data[i];
	}

}

/**
 * TX Function
 *
 * The UDRE ISR is taske with sending out data through the UART,
 * The interrupt occurs when the UDR is empty,
 * so we place a byte in the UDR whenever this interrupt occurs,
 * and keep on going until we finish the packet.
 */
interrupt [15] void UDR_empty(void) {
	// Not finished with sending out the encoded packet yet
	if (tx_packet_data_index < tx_packet_data_length) {
		UDR = tx_packet_data[tx_packet_data_index];
		tx_packet_data_index++;
	}
	// Finished, reset it
	if (tx_packet_data_index == tx_packet_data_length) {
		tx_reset();
	}
}

/**
 * TX Function
 *
 * Send out a response (resend/next char), 
 * through either fast mode or slow mode (recommended)
 */
void tx_response(unsigned char response, unsigned char fast_mode) {
	unsigned char offset;
	
	// Make sure the UART TX is not busy
	if(tx_state == TX_INIT) {
		// Take control of the UART TX	
		tx_state = TX_BUSY;
	
		// Fast mode using four syncs
	    	if (fast_mode) {	
			tx_packet_data[0] = sync_char;
			tx_packet_data[1] = sync_char;
			tx_packet_data[2] = sync_char;
			tx_packet_data[3] = sync_char;
			offset = 4;
		}
		// Slow mode using mixture of predetermined
		// garbage data and syncs
		else {			
			tx_packet_data[0] = garbage_char_a;
			tx_packet_data[1] = garbage_char_b;
			tx_packet_data[2] = garbage_char_c;
			tx_packet_data[3] = garbage_char_d;
			tx_packet_data[4] = garbage_char_a;
			tx_packet_data[5] = garbage_char_b;
			tx_packet_data[6] = sync_char;
			tx_packet_data[7] = sync_char;
			offset = 8;
		}
		
		// Start char
		tx_packet_data[offset] = start_char;
		
		// Response
		tx_packet_data[(unsigned char)(offset + 1)] = response;
		
		// End character
		tx_packet_data[(unsigned char)(offset + 2)] = end_char;
		
		// Packet length
		tx_packet_data_length = offset + 3;
		
		// Initiate TX
		tx_packet_data_index = 0;
	}
}
	

/**
 * TX Function
 * 
 * This function sends out an array of bytes
 * as the data encoded in a packet,
 * provided with the array length,
 * the id of the owner, and the transfer mode.
 */
void tx_byte_array(unsigned char *data, unsigned char length, unsigned char id, unsigned char fast_mode) {
	unsigned char i, offset;

	// The UART TX must be free
	if(tx_state == TX_INIT) {
		// Take control of the UART TX
		tx_state = TX_BUSY;
		
		// Synchronization
		if (fast_mode) {	
			// Fast mode using four syncs
			tx_packet_data[0] = sync_char;
			tx_packet_data[1] = sync_char;
			tx_packet_data[2] = sync_char;
			tx_packet_data[3] = sync_char;
			offset = 8;
		}
		else {			
			// Slow mode using a combination of predetermined garbage
			// data and syncs
			tx_packet_data[0] = garbage_char_a;
			tx_packet_data[1] = garbage_char_b;
			tx_packet_data[2] = garbage_char_c;
			tx_packet_data[3] = garbage_char_d;
			tx_packet_data[4] = garbage_char_a;
			tx_packet_data[5] = garbage_char_b;
			tx_packet_data[6] = sync_char;
			tx_packet_data[7] = sync_char;
			offset = 12;
		}

		// Start char
		tx_packet_data[8] = start_char;

		// ID char
		tx_packet_data[9] = code[id];

		// Length nibbes
		tx_data_length = length;
		tx_packet_data[10] = code[tx_data_length>>4];	// High byte
		tx_packet_data[11] = code[(tx_data_length<<4)>>4];	// Low byte
        
		// Data nibbles
		// Also hashes the data
		tx_data_hash = 0;
		for (i = 0; i < tx_data_length; i++) {
			tx_data_hash ^= data[i];
			tx_packet_data[(unsigned char)(offset + i*2)] = code[data[i]>>4];
			tx_packet_data[(unsigned char)(offset + i*2 + 1)] = code[(data[i]<<4)>>4];	
		}

		// Hash nibbles
		tx_packet_data[(unsigned char)(offset + tx_data_length*2)] = code[tx_data_hash>>4]; // High byte
		tx_packet_data[(unsigned char)(offset + tx_data_length*2 + 1)] = code[(tx_data_hash<<4)>>4];	// Low byte

		// End char
		tx_packet_data[(unsigned char)(offset + tx_data_length*2 + 2)] = end_char;
		
		// Packet length
		tx_packet_data_length = offset + tx_data_length*2 + 3;
		
		// Start TX
		tx_packet_data_index = 0;
	}
}

/**
 * TX Function
 *
 * Returns whether the UART TX is ready for another transfer
 */
unsigned char tx_is_ready(void) {
	if (tx_state == TX_INIT)
		return 1;
	else
		return 0;
}

/**
 * TX Function
 *
 * Resets the UART TX settings to make it available for another transmission
 * Handled automatically by end of a transmission by the ISR,
 * should not need to call this manually
 */
void tx_reset(void) {
	tx_state = TX_INIT;
	tx_packet_data_length = 0;
	tx_packet_data_index = MAX_TX_PACKET_LENGTH;
}
