/* SPI Constants */
#define SPI_SEL				PORTB.4
#define SPI_GARBAGE_DATA	0xAA
#define SPIF				7

/**
 * SPI Protocol (for the ZigBee Transceiver)
 * Command Byte
 *	B7	B6	B5	B4	B3	B2	B1	B0	Mode
 *	1	0	Register address [5:0]	Register Access Mode  Read Access
 *	1	1	Register address [5:0]	Register Access Mode  Write Access
 *	0	0	1	Reserved			Frame Buffer Access Mode  Read Access
 *	0	1	1	Reserved			Frame Buffer Access Mode  Write Access
 *	0	0	0	Reserved			SRAM Access Mode  Read Access
 *	0	1	0	Reserved			SRAM Access Mode  Write Access
 * 
 *			|--------------------------	|--------------	|----------	|----------	|----------	|----------
 * Register Write Access
 *	MOSI	1	1	address[5:0]		write data[7:0]
 *	MISO	XX							XX
 * Register Read Access
 *	MOSI	1	0	address[5:0]		XX
 *	MISO	XX							read data[7:0]
 * Frame Buffer Write Access
 *	MOSI	0	1	1	reserved[4:0]	PHR[7:0]		PSDU[7:0]	PSDU[7:0]	...			PSDU[7:0]
 *	MISO	XX							XX				XX			XX			...			XX
 * Frame Buffer Read Access
 *  MOSI	0	0	1	reserved[4:0]	XX				XX			...			XX			XX
 *	MISO	XX							PHR[7:0]		PSDU[7:0]	...			PSDU[7:0]	LQI[7:0]
 * SRAM Write Access
 *	MOSI	0	1	0	reserved[4:0]	address[7:0]	data[7:0]	...			data[7:0]
 *	MISO	XX							XX				XX			...			XX	
 * SRAM Read Access
 *	MOSI	0	0	0	reserved[4:0]	address[7:0]	XX			...			XX
 *	MISO	XX							XX				data[7:0]	...			data[7:0]	 
 *
 * Note: Logic values stated with XX on MOSI are ignored by the radio
 * transceiver, but need to have a valid level. Return values on MISO 
 * stated as XX shall be ignored by the microcontroller.
 * Note: Maximum frame length (# of PSDU bytes) is 127.
 */
void init_spi_master(void);
unsigned char spi_transfer_byte(unsigned char tx_data);
void spi_write_register(unsigned char address, unsigned char data);
void spi_read_register(unsigned char address, unsigned char *data); 
void spi_write_frame_buffer(unsigned char phr, unsigned char *psdu, unsigned char length);
void spi_read_frame_buffer(unsigned char *phr, unsigned char *lqi, unsigned char *psdu, unsigned char length);
void spi_write_sram(unsigned char address, unsigned char *data, unsigned char length);
void spi_read_sram(unsigned char address, unsigned char *data, unsigned char length);

/* Global Variables */
unsigned char spi_complete;

/**
 *
 */
interrupt [SPI_STC] void spi_int(void) {
	spi_complete = TRUE;
}

/**
 *
 */
unsigned char spi_transfer_byte(unsigned char tx_data) {
	unsigned char rx_data;
	
	/* Send data */
	SPDR = tx_data;
	
	/* Wait for transmission complete */
	while(!spi_complete);
	spi_complete = FALSE;
	
	// return the received data
	rx_data = SPDR;
	return rx_data;
}

/**
 *
 */
void spi_write_register(unsigned char address, unsigned char data) {
	SPI_SEL = 0;
	spi_transfer_byte((0b11000000 | address));
	spi_transfer_byte(data);
	SPI_SEL = 1;
}

/**
 *
 */
void spi_read_register(unsigned char address, unsigned char *data) {
	SPI_SEL = 0;
	spi_transfer_byte((0b10000000 | address));
	*data = spi_transfer_byte(SPI_GARBAGE_DATA);
	SPI_SEL = 1;
}

void spi_write_frame_buffer(unsigned char phr, unsigned char *psdu, unsigned char length) {
	unsigned char i;
	SPI_SEL = 0;
    spi_transfer_byte(0b01100000);
    spi_transfer_byte(phr);
    for (i=0; i<length; i++) {
    	spi_transfer_byte(psdu[i]);
    } 
	SPI_SEL = 1;
}

void spi_read_frame_buffer(unsigned char *phr, unsigned char *lqi, unsigned char *psdu, unsigned char length) {
	unsigned char i;
	SPI_SEL = 0;
    spi_transfer_byte(0b00100000);
    *phr = spi_transfer_byte(SPI_GARBAGE_DATA);
    for (i=0; i<length; i++) {
    	psdu[i] = spi_transfer_byte(SPI_GARBAGE_DATA);
    }
    *lqi = spi_transfer_byte(SPI_GARBAGE_DATA); 
	SPI_SEL = 1;
}


void spi_write_sram(unsigned char address, unsigned char *data, unsigned char length) {
	unsigned char i;
	SPI_SEL = 0;
    spi_transfer_byte(0b01000000);
    spi_transfer_byte(address);
    for (i=0; i<length; i++) {
    	spi_transfer_byte(data[i]);
    } 
	SPI_SEL = 1;
}


void spi_read_sram(unsigned char address, unsigned char *data, unsigned char length) {
	unsigned char i;
	SPI_SEL = 0;
    spi_transfer_byte(0b00000000);
    spi_transfer_byte(address);
    for (i=0; i<length; i++) {
    	data[i] = spi_transfer_byte(SPI_GARBAGE_DATA);
    } 
	SPI_SEL = 1;
}

/**
 *
 */
void init_spi_master(void) {
	/**
	 * If SS is configured as an output, the pin is a general 
	 * output pin which does not affect the SPI system. 
	 * Typically, the pin will be driving the SS pin of the SPI Slave. 
	 *
	 * If SS is configured as an input, it must be held high 
	 * to ensure Master SPI operation. If the SS pin is driven 
	 * low by peripheral circuitry when the SPI is configured 
	 * as a Master with the SS pin defined as an input, the SPI 
	 * system interprets this as another master selecting the SPI as a
	 * slave and starting to send data to it.
	 */

	DDRB.4 = 1; // SS as Output
	DDRB.5 = 1;	// MOSI as Output
	DDRB.6 = 0;	// MISO as Input
	DDRB.7 = 1;	// SCK as Output
	SPI_SEL = 1;	// Set SS High
	
	SPCR = 0b11010000;	// 7 - SPI Interrupt Enable
						// 6 - SPI Enable
						// 5 - Data Order
						// 4 - Master/Slave Select
						// 3 - Clock Polarity
						// 2 - Clock Phase
						// 1:0 - SPI Clock Rate
						//	SPI2X = 0
						//		00 - 1/4
						//		01 - 1/16
						//		10 - 1/64
						//		11 - 1/128
						//	SPI2X = 1
						//		00 - 1/2
						//		01 - 1/8
						//		10 - 1/32
						//		11 - 1/64
						
	SPSR = 0b00000000;	// 7 - SPI Interrupt Flag
						// 6 - Write COLlision Flag
						// 5:1 - Reserved Bits
						// 0 - Double SPI Speed Bit
						
	spi_complete = TRUE;
} 