/* Global Constants */
#define TW_MAX_RETRY 5

/* General Constants for TWI */
#define TWI_FREQUENCY_HZ	100000
#define TWI_READ 			1
#define TWI_WRITE 			0
#define TWI_GENERAL_ADDRESS	0
#define TWSR_MASK 			0xF8

/* Constants for TWCR Register */
#define TWINT		0b10000000
#define TWEA		0b01000000
#define TWSTA		0b00100000
#define TWSTO		0b00010000
#define TWWC		0b00001000
#define TWEN		0b00000100
#define TWIE		0b00000001

/* General TWI Master staus codes */                      
#define TWI_START					0x08	// START has been transmitted  
#define TWI_REP_START				0x10	// Repeated START has been transmitted
#define TWI_ARB_LOST				0x38	// Arbitration lost

/* TWI Master Transmitter staus codes */                      
#define TWI_MT_ADR_ACK				0x18  	// SLA+W has been tramsmitted and ACK received
#define TWI_MT_ADR_NACK				0x20  	// SLA+W has been tramsmitted and NACK received 
#define TWI_MT_DATA_ACK				0x28  	// Data byte has been tramsmitted and ACK received
#define TWI_MT_DATA_NACK			0x30  	// Data byte has been tramsmitted and NACK received 

/* TWI Master Receiver staus codes */  
#define TWI_MR_ADR_ACK				0x40  	// SLA+R has been tramsmitted and ACK received
#define TWI_MR_ADR_NACK				0x48  	// SLA+R has been tramsmitted and NACK received
#define TWI_MR_DATA_ACK				0x50  	// Data byte has been received and ACK tramsmitted
#define TWI_MR_DATA_NACK			0x58  	// Data byte has been received and NACK tramsmitted

/* TWI Slave Transmitter staus codes */
#define TWI_ST_ADR_ACK				0xA8  	// Own SLA+R has been received; ACK has been returned
#define TWI_ST_ADR_ACK_M_ARB_LOST	0xB0  	// Arbitration lost in SLA+R/W as Master; own SLA+R has been received; ACK has been returned
#define TWI_ST_DATA_ACK				0xB8  	// Data byte in TWDR has been transmitted; ACK has been received
#define TWI_ST_DATA_NACK			0xC0  	// Data byte in TWDR has been transmitted; NOT ACK has been received
#define TWI_ST_DATA_ACK_LAST_BYTE	0xC8  	// Last data byte in TWDR has been transmitted (TWEA = 0); ACK has been received

/* TWI Slave Receiver staus codes */
#define TWI_SR_ADR_ACK				0x60  	// Own SLA+W has been received ACK has been returned
#define TWI_SR_ADR_ACK_M_ARB_LOST 	0x68  	// Arbitration lost in SLA+R/W as Master; own SLA+W has been received; ACK has been returned
#define TWI_SR_GEN_ACK				0x70  	// General call address has been received; ACK has been returned
#define TWI_SR_GEN_ACK_M_ARB_LOST 	0x78  	// Arbitration lost in SLA+R/W as Master; General call address has been received; ACK has been returned
#define TWI_SR_ADR_DATA_ACK			0x80  	// Previously addressed with own SLA+W; data has been received; ACK has been returned
#define TWI_SR_ADR_DATA_NACK		0x88  	// Previously addressed with own SLA+W; data has been received; NOT ACK has been returned
#define TWI_SR_GEN_DATA_ACK			0x90  	// Previously addressed with general call; data has been received; ACK has been returned
#define TWI_SR_GEN_DATA_NACK		0x98  	// Previously addressed with general call; data has been received; NOT ACK has been returned
#define TWI_SR_STOP_RESTART			0xA0  	// A STOP condition or repeated START condition has been received while still addressed as Slave

/* TWI Miscellaneous status codes */
#define TWI_NO_STATE				0xF8	// No relevant state information available; TWINT = 0
#define TWI_BUS_ERROR				0x00	// Bus error due to an illegal START or STOP condition

/* Constants for TWI Functions' Return Status */
#define TWI_SUCCESS					0
#define TWI_INVALID_STATUS_ERROR	1
#define TWI_NO_ACK_ERROR			2
#define TWI_INVALID_SLA_ERROR		3
#define TWI_ARB_LOST_ERROR			4
#define TWI_NO_DATA_ERROR			5

/**
 *	READ
 *		twi_start(i2cAddress,I2C_READ);
 *		twi_read(serialBuffer,rxLength);
 *		twi_stop();
 *					
 *	WRITE
 *		twi_start(i2cAddress,I2C_WRITE);
 *		twi_write(serialBuffer,txLength);
 *		twi_stop();
 *
 *	WRITE & READ				
 *		twi_start(i2cAddress,I2C_WRITE);
 *		twi_write(serialBuffer,txLength);                 
 *		twi_start(i2cAddress,I2C_READ); 
 *		twi_read(serialBuffer,rxLength);
 *		twi_stop();
 *
 *	READ REGISTER
 *		twi_read_register(i2cAddress, regAddress, rxLength, serialBuffer);
 *
 *	WRITE REGISTER
 *		twi_read_register(i2cAddress, regAddress, rxLength, serialBuffer);
 *	
 *	WRITE & READ REGISTER 	
 *		twi_write_register(i2cAddress, regAddress, txLength, serialBuffer);
 *		twi_read_register(i2cAddress, regAddress, rxLength, serialBuffer);
 */

/* Prototype Functions for TWI */
void init_twi_master(unsigned long cpu_freq, unsigned char pullup_resistor);
unsigned char twi_master_start(unsigned char sla, unsigned char mode);
unsigned char twi_write_byte(unsigned char data);
unsigned char twi_write(unsigned char *data, unsigned char length);
unsigned char twi_read_byte(unsigned char *data);
unsigned char twi_read(unsigned char *data, unsigned char length);
unsigned char twi_read_register(unsigned char slaveWriteAddress, unsigned char slaveReadAddress, unsigned int regAddress, unsigned char length, unsigned char *data);
unsigned char twi_write_register(unsigned char slaveWriteAddress, unsigned int regAddress, unsigned char length, unsigned char *data);
void twi_stop(void);

void init_twi_master(unsigned long cpu_freq, unsigned char pullup_resistor) {
	float temp;
	
	if (pullup_resistor) {
		DDRC.0 = 0;
		DDRC.1 = 0;
		PORTC.0 = 1;
		PORTC.1 = 1;
	}		
	
	temp = (float) cpu_freq / (float) TWI_FREQUENCY_HZ;
    TWBR = (temp - 16) / 2 ;	// TWI Bit Rate Register		
								// I2C Freq = CPU Freq / (16 + 2 * TWBR * 4 ^ TWPS)
								
	TWSR = 0x00;		// TWI Status Register
						// 7:3 - TWI Status
						// 2 - Reserved
						// 1:0 - TWI Prescaler Bits
						//		00 - 1
						//		01 - 4
						//		10 - 16
						//		11 - 64	
}
		
unsigned char twi_master_start(unsigned char sla, unsigned char mode) {
	if (0x80 & sla)					// Check SLA is 7 bits
		return TWI_INVALID_SLA_ERROR;	
	
	TWCR = TWINT | TWSTA | TWEN;	// Send START condition
	
	while(!(TWCR & TWINT));			// Wait for transmission
	
	switch (TWSR & TWSR_MASK) {		// Check status
		case TWI_REP_START:			// OK
		case TWI_START:				// OK
			break;
		case TWI_ARB_LOST:			// Lost arbitration, can try again
			return TWI_ARB_LOST_ERROR;
		default:					// Not in START condition, do not send STOP condition
			return TWI_INVALID_STATUS_ERROR;
	}
    
	TWDR = (sla << 1) | mode;		// Send SLA+R/W
	TWCR = TWINT | TWEN;			// Clear interrupt to start transmission
	
	while(!(TWCR & TWINT));			// Wait for transmission
	
	if (mode == TWI_WRITE) {		// For SLA+W
		switch (TWSR & TWSR_MASK) {	// Check status
			case TWI_MT_ADR_ACK:	// OK
				break;
			case TWI_MT_ADR_NACK:	// Device busy during write, can try again
				return TWI_NO_ACK_ERROR;
			case TWI_ARB_LOST:		// Lost arbitration, can try again
				return TWI_ARB_LOST_ERROR;
			default:				// Error, must send STOP condition 
				return TWI_INVALID_STATUS_ERROR;
		}
	}
	else {							// For SLA+R
		switch (TWSR & TWSR_MASK) {	// Check status
			case TWI_MR_ADR_ACK:	// OK
				break;
			case TWI_MR_ADR_NACK:	// Device busy during write
				return TWI_NO_ACK_ERROR;
			case TWI_ARB_LOST:	// Lost arbitration, can try again
				return TWI_ARB_LOST_ERROR;
			default:				// Error, must send STOP condition 
				return TWI_INVALID_STATUS_ERROR;
		}
	}
	
	return TWI_SUCCESS;	
}
	
unsigned char twi_write_byte(unsigned char data) {
	TWDR = data;
	TWCR = TWINT | TWEN;			// Clear interrupt to start transmission
	while (!(TWCR & TWINT));		// Wait for transmission
	switch (TWSR & TWSR_MASK) {		// Check status
		case TWI_MT_DATA_ACK:		// OK
			return TWI_SUCCESS;
		case TWI_MT_DATA_NACK:		// Device write protected	
			return TWI_NO_ACK_ERROR;
    	case TWI_ARB_LOST:		// Lost arbitration
			return TWI_ARB_LOST_ERROR;
		default:					// Error, must send STOP condition		
			return TWI_INVALID_STATUS_ERROR;
	}
}

unsigned char twi_write(unsigned char *data, unsigned char length) {
	unsigned char i;

	for (i = 0; i < length; i++) {
		TWDR = *data++;
		TWCR = TWINT | TWEN;		// Clear interrupt to start transmission
		while (!(TWCR & TWINT));	// Wait for transmission
		switch (TWSR & TWSR_MASK) {	// Check status
			case TWI_MT_DATA_ACK:	// OK		
				break;
			case TWI_MT_DATA_NACK:	// Device write protected
				return TWI_MT_DATA_NACK; 
    		case TWI_ARB_LOST:	// Lost arbitration, re-arbitrate (unlimited)
				return TWI_ARB_LOST_ERROR;
			default:		// Error, must send stop condition		
				return TWI_INVALID_STATUS_ERROR;
		}
	}
	
	return TWI_SUCCESS;
}

unsigned char twi_read_byte(unsigned char *data) {
	TWCR = TWINT | TWEN;

	while(!(TWCR & TWINT));
	switch (TWSR & TWSR_MASK) {
		case TWI_MR_DATA_NACK:
		case TWI_MR_DATA_ACK:		
			break;
		default:
			return TWI_INVALID_STATUS_ERROR;
	}
	*data = TWDR;
	return TWI_SUCCESS;

}

unsigned char twi_read(unsigned char *data, unsigned char length) {
	unsigned int i;

	for (i = 0; i < length; i++) {
		if (i == (length-1))
			TWCR = TWINT | TWEN;
		else
			TWCR = TWINT | TWEN | TWEA;
		while (!(TWCR & TWINT));
		switch (TWSR & TWSR_MASK) {
			case TWI_MR_DATA_NACK:
			case TWI_MR_DATA_ACK:
				*data++ = TWDR;
				break;
			default:
				return TWI_INVALID_STATUS_ERROR;
		}
	}
	return TWI_SUCCESS;
}

unsigned char twi_read_register(unsigned char slaveWriteAddress, unsigned char slaveReadAddress, unsigned int regAddress, unsigned char length, unsigned char *data) {
	unsigned char sla, i, n = 0;
	sla = slaveWriteAddress | (((regAddress >> 8) & 0x07) << 1);

	retry:
	if (n++ >= TW_MAX_RETRY)
		return TWI_NO_ACK_ERROR;
	begin:

	TWCR = TWINT | TWSTA | TWEN;	// Send Start condition
	while (!(TWCR & TWINT));		// Wait for transmission (TWINT)
	switch (TWSR & TWSR_MASK) {		// Check status
		case TWI_REP_START:
		case TWI_START:				// Ok
			break;
		case TWI_ARB_LOST:			// Lost arbitration, re-arbitration (unlimited)
			goto begin;
		default:					// Not in start condition, do no send stop condition
			return TWI_INVALID_STATUS_ERROR;
	}
	
	TWDR = sla | TWI_WRITE;		// Send SLA+W
	TWCR = TWINT | TWEN;		// Clear interrupt to start transmission
	while (!(TWCR & TWINT));	// Wait for transmission (TWINT)
	switch (TWSR & TWSR_MASK) {	// Check status
		case TWI_MT_ADR_ACK:	// OK
			break;
		case TWI_MT_ADR_NACK:	// NACK during select: device busy writing, try again (limited)
			goto retry;
   		case TWI_ARB_LOST:		// Lost arbitration, re-arbitrate (unlimited)
			goto begin;
		default:				// Error, must send stop condition
			goto error;
	}

	TWDR = regAddress;			// Send low 8 bits of registry address
	TWCR = TWINT | TWEN;		// Clear interrupt to start transmission
	while (!(TWCR & TWINT));	// Wait for transmission (TWINT)
	switch (TWSR & TWSR_MASK) {	// Check status
		case TWI_MT_DATA_ACK:	// OK		
			break;
		case TWI_MT_DATA_NACK:	// NACK during select: device busy writing, quit	
			goto quit;
   		case TWI_ARB_LOST:		// Lost arbitration, re-arbitrate (unlimited)
			goto begin;
		default:				// Error, must send stop condition		
			goto error;
	}
	
	//Stop/pause TWI interface
	TWCR = 0; 
	delay_us(2);

	TWCR = TWINT | TWSTA | TWEN;	// Send repeat start condition
	while (!(TWCR & TWINT));		// Wait for transmission
	switch (TWSR & TWSR_MASK) {				// OK, but should not happen
		case TWI_START:
		case TWI_REP_START:
			break;
		case TWI_ARB_LOST:
			goto begin;
		default:
			goto error;
	}
    
	sla = slaveReadAddress |(((regAddress >> 8) & 0x07) << 1); 
	
	TWDR = sla | TWI_READ;		// Send SLA+R
	TWCR = TWINT | TWEN;		// Clear interrupt to start transmission
	while (!(TWCR & TWINT));	// Wait for transmission
	switch (TWSR & TWSR_MASK) {
		case TWI_MR_ADR_ACK:
			break;
		case TWI_MR_ADR_NACK:
			goto quit;
		case TWI_ARB_LOST:
			goto begin;
		default:
			goto error;
	}
    
	for (i = 0; i < length; i++) {
		if (i == (length-1))
			TWCR = TWINT | TWEN;
		else
			TWCR = TWINT | TWEN | TWEA;

		while (!(TWCR & TWINT));

		switch (TWSR & TWSR_MASK) {
			case TWI_MR_DATA_NACK:
			case TWI_MR_DATA_ACK:
				*data++ = TWDR;
				break;
			default:
				goto error;
		}
	}
	
	TWCR = TWINT | TWSTO | TWEN;
	while(TWCR & TWSTO);
	return TWI_SUCCESS;
	
	error:
	quit:
	TWCR = TWINT | TWSTO | TWEN;
	while(TWCR & TWSTO);
	return TWI_INVALID_STATUS_ERROR;
}

unsigned char twi_write_register(unsigned char slaveWriteAddress, unsigned int regAddress, unsigned char length, unsigned char *data) {
	unsigned char sla, i, n = 0;
	sla = slaveWriteAddress |(((regAddress >> 8) & 0x07) << 1);

	retry:
	if (n++ >= TW_MAX_RETRY)
		return TWI_NO_ACK_ERROR;
	begin:

	TWCR = TWINT | TWSTA | TWEN;	// Send Start condition
	while (!(TWCR & TWINT));		// Wait for transmission (TWINT)
	switch (TWSR & TWSR_MASK) {		// Check status
		case TWI_REP_START:
		case TWI_START:				// Ok
			break;
		case TWI_ARB_LOST:			// Lost arbitration, re-arbitration (unlimited)
			goto begin;
		default:					// Not in start condition, do no send stop condition
		return TWI_INVALID_STATUS_ERROR;
	}

	TWDR = sla | TWI_WRITE;			// Send SLA+W
	TWCR = TWINT | TWEN;			// Clear interrupt to start transmission
	while (!(TWCR & TWINT));		// Wait for transmission (TWINT)
	switch (TWSR & TWSR_MASK) {		// Check status
		case TWI_MT_ADR_ACK:		// OK
			break;
		case TWI_MT_ADR_NACK:		// NACK during select: device busy writing, try again (limited)
			goto retry;
   		case TWI_ARB_LOST:			// Lost arbitration, re-arbitrate (unlimited)
			goto begin;
		default:					// Error, must send stop condition
			goto error;
	}
	
	TWDR = regAddress;				// Send low 8 bits of registry address
	TWCR = TWINT | TWEN;			// Clear interrupt to start transmission
	while (!(TWCR & TWINT));		// Wait for transmission (TWINT)
	switch (TWSR & TWSR_MASK) {		// Check status
		case TWI_MT_DATA_ACK:		// OK		
			break;
		case TWI_MT_DATA_NACK:		// NACK during select: device busy writing, quit	
			goto quit;
   		case TWI_ARB_LOST:			// Lost arbitration, re-arbitrate (unlimited)
			goto begin;
		default:					// Error, must send stop condition		
			goto error;
	}
	
	for (i = 0; i < length; i++) {
		TWDR = *data++;
		TWCR = TWINT | TWEN;
		while (!(TWCR & TWINT));
		switch (TWSR & TWSR_MASK) {	// Check status
			case TWI_MT_DATA_ACK:	// OK		
				break;
			case TWI_MT_DATA_NACK:	// Device write protected, quit	
				goto quit;
    		case TWI_ARB_LOST:	// Lost arbitration, re-arbitrate (unlimited)
				goto begin;
			default:		// Error, must send stop condition		
				goto error;
		}
	}


	TWCR = TWINT | TWSTO | TWEN;
	while(TWCR & TWSTO);
	return TWI_SUCCESS;
	
	error:
	quit:
	TWCR = TWINT | TWSTO | TWEN;
	while(TWCR & TWSTO);
	return TWI_INVALID_SLA_ERROR;
}

void twi_stop(void) {
	TWCR = TWINT | TWEN | TWSTO;	// Send STOP condition 
    while(TWCR & TWSTO);			// Wait for transmission    
}
