/* General Constants for TWI */
#define TWI_READ 			1
#define TWI_WRITE 			0
#define TWI_GENERAL_ADDRESS	0
#define TWI_BUFFER_SIZE		8
#define TWSR_MASK 		0xF8

/* Constants for TWCR */
#define TWINT		0b10000000
#define TWEA		0b01000000
#define TWSTA		0b00100000
#define TWSTO		0b00010000
#define TWWC		0b00001000
#define TWEN		0b00000100
#define TWIE		0b00000001

/* General TWI Master staus codes */                      
#define TWI_START					0x08	// START has been transmitted  
#define TWI_REP_START				0x10	// Repeated START has been transmitted
#define TWI_ARB_LOST				0x38	// Arbitration lost

/* TWI Master Transmitter staus codes */                      
#define TWI_MT_ADR_ACK				0x18  	// SLA+W has been tramsmitted and ACK received
#define TWI_MT_ADR_NACK				0x20  	// SLA+W has been tramsmitted and NACK received 
#define TWI_MT_DATA_ACK				0x28  	// Data byte has been tramsmitted and ACK received
#define TWI_MT_DATA_NACK			0x30  	// Data byte has been tramsmitted and NACK received 

/* TWI Master Receiver staus codes */  
#define TWI_MR_ADR_ACK				0x40  	// SLA+R has been tramsmitted and ACK received
#define TWI_MR_ADR_NACK				0x48  	// SLA+R has been tramsmitted and NACK received
#define TWI_MR_DATA_ACK				0x50  	// Data byte has been received and ACK tramsmitted
#define TWI_MR_DATA_NACK			0x58  	// Data byte has been received and NACK tramsmitted

/* TWI Slave Transmitter staus codes */
#define TWI_ST_ADR_ACK				0xA8  	// Own SLA+R has been received; ACK has been returned
#define TWI_ST_ADR_ACK_M_ARB_LOST	0xB0  	// Arbitration lost in SLA+R/W as Master; own SLA+R has been received; ACK has been returned
#define TWI_ST_DATA_ACK				0xB8  	// Data byte in TWDR has been transmitted; ACK has been received
#define TWI_ST_DATA_NACK			0xC0  	// Data byte in TWDR has been transmitted; NOT ACK has been received
#define TWI_ST_DATA_ACK_LAST_BYTE	0xC8  	// Last data byte in TWDR has been transmitted (TWEA = 0); ACK has been received

/* TWI Slave Receiver staus codes */
#define TWI_SR_ADR_ACK				0x60  	// Own SLA+W has been received ACK has been returned
#define TWI_SR_ADR_ACK_M_ARB_LOST 	0x68  	// Arbitration lost in SLA+R/W as Master; own SLA+W has been received; ACK has been returned
#define TWI_SR_GEN_ACK				0x70  	// General call address has been received; ACK has been returned
#define TWI_SR_GEN_ACK_M_ARB_LOST 	0x78  	// Arbitration lost in SLA+R/W as Master; General call address has been received; ACK has been returned
#define TWI_SR_ADR_DATA_ACK			0x80  	// Previously addressed with own SLA+W; data has been received; ACK has been returned
#define TWI_SR_ADR_DATA_NACK		0x88  	// Previously addressed with own SLA+W; data has been received; NOT ACK has been returned
#define TWI_SR_GEN_DATA_ACK			0x90  	// Previously addressed with general call; data has been received; ACK has been returned
#define TWI_SR_GEN_DATA_NACK		0x98  	// Previously addressed with general call; data has been received; NOT ACK has been returned
#define TWI_SR_STOP_RESTART			0xA0  	// A STOP condition or repeated START condition has been received while still addressed as Slave

/* TWI Miscellaneous status codes */
#define TWI_NO_STATE				0xF8	// No relevant state information available; TWINT = 0
#define TWI_BUS_ERROR				0x00	// Bus error due to an illegal START or STOP condition

/* Constants for TWI Function Return Status */
#define TWI_SUCCESS 				0
#define TWI_ERROR_INVALID_STATUS 	1
#define TWI_ERROR_NO_ACK 			2
#define TWI_ERROR_INVALID_ADDRESS 	3
#define TWI_ERROR_NO_DATA 			4
#define TWI_ERROR_BUFFER_FULL 		5

/* Global Variables for TWI */
unsigned char twi_recieve_pos, twi_transmit_pos, twi_recieve_lost;
unsigned char twi_recieve_buffer[TWI_BUFFER_SIZE], twi_transmit_buffer[TWI_BUFFER_SIZE];
unsigned char twi_ready_flag;

/* Function Prototype for TWI */
unsigned char init_twi_slave(unsigned char sla, unsigned char gce);
void twi_process_status(void);
unsigned char twi_write_byte(unsigned char data);
unsigned char twi_write(unsigned char *data, unsigned char length);
unsigned char twi_writef(unsigned char flash *data, unsigned char length);
unsigned char twi_read_byte(unsigned char *byte);
unsigned char twi_read(unsigned char *data, unsigned char length);
unsigned char twi_data_lost(void);
unsigned char twi_receive_ready(void);
unsigned char twi_read(unsigned char *data, unsigned char length);

interrupt [TWI] void twi_isr(void) {
   twi_ready_flag = 1;
}

void twi_process_status(void) { 
    unsigned char i;
    twi_ready_flag = 0;
	
    switch(TWSR & TWSR_MASK) {
        case TWI_SR_ADR_ACK:
        case TWI_SR_ADR_ACK_M_ARB_LOST:
        case TWI_SR_GEN_ACK:
        case TWI_SR_GEN_ACK_M_ARB_LOST:
            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;
            
        case TWI_SR_ADR_DATA_ACK:
        case TWI_SR_ADR_DATA_NACK:
        case TWI_SR_GEN_DATA_ACK:
        case TWI_SR_GEN_DATA_NACK:
            if(twi_recieve_pos == (TWI_BUFFER_SIZE)) {
               twi_recieve_lost = 1;
            } else {
               twi_recieve_buffer[twi_recieve_pos++] = TWDR;
            }            
            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;
                        
        case TWI_SR_STOP_RESTART:
            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;
            
        case TWI_ST_ADR_ACK:
   	      if(twi_transmit_pos) {
   	         TWDR = twi_transmit_buffer[0];
   	         twi_transmit_pos--;
             for(i=0; i<twi_transmit_pos; i++)
                  twi_transmit_buffer[i] = twi_transmit_buffer[(unsigned char) (i+1)];
   	      } 
   	      else {
   	            TWDR = 0xFF;
   	      }

            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;
            
        case TWI_ST_ADR_ACK_M_ARB_LOST:
   	      if(twi_transmit_pos) {
   	         TWDR = twi_transmit_buffer[0];
   	         twi_transmit_pos--;
               for(i=0; i<twi_transmit_pos; i++)
                  twi_transmit_buffer[i] = twi_transmit_buffer[(unsigned char) (i+1)];
   	      } else {
   	            TWDR = 0xFF;
   	      }

            TWCR = TWINT | TWEN | TWIE;
            break;
            
        case TWI_ST_DATA_ACK:
   	      if(twi_transmit_pos) {
   	         TWDR = twi_transmit_buffer[0];
   	         twi_transmit_pos--;
               for(i=0; i<twi_transmit_pos; i++)
                  twi_transmit_buffer[i] = twi_transmit_buffer[(unsigned char) (i+1)];
   	      } else {
   	            TWDR = 0xFF;
   	      }

            TWCR = TWINT | TWEA| TWEN | TWIE;
            break;
            
        case TWI_ST_DATA_NACK:
            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;
        
        case TWI_ST_DATA_ACK_LAST_BYTE:
            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;

        case TWI_NO_STATE:
            break;
            
        case TWI_BUS_ERROR:
            TWCR = TWINT | TWSTO | TWEA | TWEN | TWIE;
            break;
            
        default:
    }
}

unsigned char twi_write_byte(unsigned char data) {
   
   if(twi_transmit_pos==TWI_BUFFER_SIZE) {
      return TWI_ERROR_BUFFER_FULL;
   }
   
   twi_transmit_buffer[twi_transmit_pos++] = data;
   return TWI_SUCCESS;
}

unsigned char twi_write(unsigned char *data, unsigned char length) {
   unsigned char err, i;

   for(i=0;i<length;i++) {
      err = twi_write_byte(*data);
      if(err) return err;
      data++;
   }
   return TWI_SUCCESS;   
}

unsigned char twi_writef(unsigned char flash *data, unsigned char length) {
   unsigned char err, i;

   for(i=0;i<length;i++) {
      err = twi_write_byte(*data);
      if(err) return err;
      data++;
   }
   return TWI_SUCCESS;
}

unsigned char twi_transmit_room() {
    return (TWI_BUFFER_SIZE - twi_transmit_pos);
}

unsigned char twi_read_byte(unsigned char *byte) {
    unsigned char i;
    
    if(twi_recieve_pos) {
        *byte = twi_recieve_buffer[0];
        twi_recieve_pos--;
        for(i=0; i<twi_recieve_pos; i++)
            twi_recieve_buffer[i] = twi_recieve_buffer[(unsigned char) (i+1)];
        return TWI_SUCCESS;
    } else {
        return TWI_ERROR_NO_DATA;
    }
}

unsigned char twi_read(unsigned char *data, unsigned char length) {
    unsigned char err, i;
    
    for(i = 0; i < length; i++) {
        err = twi_read_byte(data++);
        if (err) return err;
    }
    
    return TWI_SUCCESS; 
}

unsigned char twi_receive_ready(void) {
    return twi_recieve_pos;
}

unsigned char twi_data_lost(void) {
   unsigned char tmp;
   tmp = twi_recieve_lost;
   twi_recieve_lost = 0;
   return tmp;
}

unsigned char init_twi_slave(unsigned char sla, unsigned char gce){
    unsigned char i;

    if(0x80 & sla) 
    	return TWI_ERROR_INVALID_ADDRESS;

    TWAR = (sla<<1) | gce;
    TWCR = TWEA | TWEN | TWIE;

    twi_recieve_pos = 0;
    twi_transmit_pos = 0;
    twi_recieve_lost = 0;
    for(i = 0; i<TWI_BUFFER_SIZE; i++) {
        twi_transmit_buffer[i] = 0;
        twi_recieve_buffer[i] = 0;
    }
    
    //turn on ISRs
    #asm("sei");
    
    return TWI_SUCCESS;
}
