/* Global Constants */
#define TRUE 	1
#define FALSE 	0

/* General Constants for TWI */
#define TWI_READ 			1
#define TWI_WRITE 			0
#define TWI_GENERAL_ADDRESS	0
#define TWI_BUFFER_SIZE 	8
#define TWSR_MASK 			0xF8

/* Constants for TWCR */
#define TWINT		0b10000000
#define TWEA		0b01000000
#define TWSTA		0b00100000
#define TWSTO		0b00010000
#define TWWC		0b00001000
#define TWEN		0b00000100
#define TWIE		0b00000001

/* General TWI Master staus codes */                      
#define TWI_START					0x08	// START has been transmitted  
#define TWI_REP_START				0x10	// Repeated START has been transmitted
#define TWI_ARB_LOST				0x38	// Arbitration lost

/* TWI Master Transmitter staus codes */                      
#define TWI_MT_ADR_ACK				0x18  	// SLA+W has been tramsmitted and ACK received
#define TWI_MT_ADR_NACK				0x20  	// SLA+W has been tramsmitted and NACK received 
#define TWI_MT_DATA_ACK				0x28  	// Data byte has been tramsmitted and ACK received
#define TWI_MT_DATA_NACK			0x30  	// Data byte has been tramsmitted and NACK received 

/* TWI Master Receiver staus codes */  
#define TWI_MR_ADR_ACK				0x40  	// SLA+R has been tramsmitted and ACK received
#define TWI_MR_ADR_NACK				0x48  	// SLA+R has been tramsmitted and NACK received
#define TWI_MR_DATA_ACK				0x50  	// Data byte has been received and ACK tramsmitted
#define TWI_MR_DATA_NACK			0x58  	// Data byte has been received and NACK tramsmitted

/* TWI Slave Transmitter staus codes */
#define TWI_ST_ADR_ACK				0xA8  	// Own SLA+R has been received; ACK has been returned
#define TWI_ST_ADR_ACK_M_ARB_LOST	0xB0  	// Arbitration lost in SLA+R/W as Master; own SLA+R has been received; ACK has been returned
#define TWI_ST_DATA_ACK				0xB8  	// Data byte in TWDR has been transmitted; ACK has been received
#define TWI_ST_DATA_NACK			0xC0  	// Data byte in TWDR has been transmitted; NOT ACK has been received
#define TWI_ST_DATA_ACK_LAST_BYTE	0xC8  	// Last data byte in TWDR has been transmitted (TWEA = 0); ACK has been received

/* TWI Slave Receiver staus codes */
#define TWI_SR_ADR_ACK				0x60  	// Own SLA+W has been received ACK has been returned
#define TWI_SR_ADR_ACK_M_ARB_LOST 	0x68  	// Arbitration lost in SLA+R/W as Master; own SLA+W has been received; ACK has been returned
#define TWI_SR_GEN_ACK				0x70  	// General call address has been received; ACK has been returned
#define TWI_SR_GEN_ACK_M_ARB_LOST 	0x78  	// Arbitration lost in SLA+R/W as Master; General call address has been received; ACK has been returned
#define TWI_SR_ADR_DATA_ACK			0x80  	// Previously addressed with own SLA+W; data has been received; ACK has been returned
#define TWI_SR_ADR_DATA_NACK		0x88  	// Previously addressed with own SLA+W; data has been received; NOT ACK has been returned
#define TWI_SR_GEN_DATA_ACK			0x90  	// Previously addressed with general call; data has been received; ACK has been returned
#define TWI_SR_GEN_DATA_NACK		0x98  	// Previously addressed with general call; data has been received; NOT ACK has been returned
#define TWI_SR_STOP_RESTART			0xA0  	// A STOP condition or repeated START condition has been received while still addressed as Slave

/* TWI Miscellaneous status codes */
#define TWI_NO_STATE				0xF8	// No relevant state information available; TWINT = 0
#define TWI_BUS_ERROR				0x00	// Bus error due to an illegal START or STOP condition

/* Global Variables for TWI */
unsigned char twi_data_buffer[TWI_BUFFER_SIZE], twi_buffer_index, twi_data_length;
unsigned char twi_gce_flag, twi_rx_flag, twi_tx_flag;

/* Function Prototypes */
void init_twi_slave(unsigned char sla, unsigned char gce);
unsigned char twi_busy(void);
void twi_start_w_data(unsigned char *data, unsigned char length);
void twi_start(void);
unsigned char twi_read_data(unsigned char *data, unsigned char length); 

interrupt [TWI] void twi_isr(void) {
	switch (TWSR & TWSR_MASK) {
		case TWI_ST_ADR_ACK:		// Set buffer pointer to first data location
			twi_buffer_index = 0;	
		case TWI_ST_DATA_ACK:
			TWDR = twi_data_buffer[twi_buffer_index++];
			TWCR = TWEN | TWIE | TWINT | TWEA;  
			break;
		case TWI_ST_DATA_NACK:		// This could be the end of the transmission.
			if (twi_buffer_index == twi_data_length) {	// Have we transceived all expected data?
				twi_tx_flag = 1;						// Set status bits to completed successfully. 
			}   
			TWCR = TWEN;
			break;     
		case TWI_SR_GEN_ACK:
			twi_gce_flag = 0;
		case TWI_SR_ADR_ACK:						// Dont need to clear gce flag due to that it is the default state.
			twi_rx_flag = 1;      
			twi_buffer_index = 0;					// Set buffer pointer to first data location
            TWCR = TWEN | TWIE | TWINT | TWEA;
			break;
		case TWI_SR_ADR_DATA_ACK:
		case TWI_SR_GEN_DATA_ACK:					// Set flag transmission successfull.
			twi_data_buffer[twi_buffer_index++] = TWDR;
			twi_tx_flag = 1;
			TWCR = TWCR = TWEN | TWIE | TWINT | TWEA;
			break;
		case TWI_SR_STOP_RESTART:					// Put TWI Transceiver in passive mode.
			TWCR = TWEN;
			break;           
		case TWI_SR_ADR_DATA_NACK:    
		case TWI_SR_GEN_DATA_NACK:
		case TWI_ST_DATA_ACK_LAST_BYTE:
		case TWI_NO_STATE:
		case TWI_BUS_ERROR:
		default:     
			TWCR = TWEN;
  }
}

void init_twi_slave(unsigned char sla, unsigned char gce) {
	TWAR = (sla << 1) | gce;	// Set own TWI slave address. Accept TWI General Calls.
	TWDR = 0xFF;				// Default content = SDA released.
	TWCR = TWEN;
}

unsigned char twi_busy(void) {
	return (TWCR & TWIE);	// IF TWI interrupt is enabled then the Transceiver is busy
}

void twi_start_w_data(unsigned char *data, unsigned char length) {
	unsigned char i;

	while (twi_busy());				// Wait until TWI is ready for next transmission.

	twi_data_length = length;		// Number of data to transmit.
	for (i = 0; i < length; i++) {	// Copy data that may be transmitted if the TWI Master requests data.
		twi_data_buffer[i] = data[i];
	}
	twi_tx_flag = 0;
	twi_rx_flag = 0;
	twi_gce_flag = 0;       
	TWCR = TWEN | TWIE | TWINT | TWEA;
}

void twi_start(void) {
	while (twi_busy());				// Wait until TWI is ready for next transmission.
	twi_tx_flag = 0;
	twi_rx_flag = 0;
	twi_gce_flag = 0;      
	TWCR = TWEN | TWIE | TWINT | TWEA;
}

unsigned char twi_read_data(unsigned char *data, unsigned char length) {
	unsigned char i;

	while (twi_busy());			// Wait until TWI is ready for next transmission

	if (twi_tx_flag) {                                             
    	for (i = 0; i < length; i++) {
			data[i] = twi_data_buffer[i];
    	}
		twi_rx_flag = FALSE;          // Slave Receive data has been read from buffer
	}
	return(twi_tx_flag);                                   
}
