/**
 * A Mega644P Local DHCP Base Node Prototype
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2A/PCINT31)
 *							-----------
 */

/* Global/System Constants */
#define TRUE 					1
#define FALSE 					0

#define CPU_FREQUENCY_1_MHZ		1000000				// 1.8-5.5V
#define CPU_FREQUENCY_2_MHZ		2000000				// 1.8-5.5V
#define CPU_FREQUENCY_4_MHZ		4000000				// 1.8-5.5V
#define CPU_FREQUENCY_8_MHZ		8000000				// 2.7-5.5V
#define CPU_FREQUENCY_16_MHZ	16000000			// 4.0-5.5V
#define CPU_FREQUENCY_HZ		CPU_FREQUENCY_1_MHZ

#define sbi(data,bit)			(data |= (1<<bit))	// Set Bit
#define cbi(data,bit) 			(data &= ~(1<<bit))	// Clear Bit
#define gbi(data,bit)			(data & (1<<bit))	// Get Bit
#define tbi(data,bit)			(data ^= (1<<bit))	// Toggle Bit

/* Node Constants */
#define SENTINEL_BASE_NODE		0x00
#define SENTINEL_RELAY_NODE		0x01
#define SENTINEL_SENSOR_NODE	0x02
#define SENTINEL_NODE_TYPE		SENTINEL_BASE_NODE

/* Debug Constants */
#define PROTOBOARD_LED_DEBUG			TRUE

#define UART0_ZIGBEE_SPEC_DEBUG			FALSE
#define UART0_ZIGBEE_IRQ_STATUS_DEBUG	FALSE
#define UART0_ZIGBEE_TRAC_STATUS_DEBUG	FALSE
#define UART0_ZIGBEE_STATE_DEBUG		FALSE	// Heavy traffic
#define UART0_ZIGBEE_RX_PHY_DATA_DEBUG	FALSE
#define UART0_ZIGBEE_RX_PHY_LQI_DEBUG	FALSE
#define UART0_ZIGBEE_RX_MAC_DATA_DEBUG	FALSE
#define UART0_ZIGBEE_RX_MAC_FRAME_DEBUG	FALSE
#define UART0_ZIGBEE_ED_LEVEL_DEBUG		FALSE
#define UART0_ZIGBEE_CCA_DEBUG			FALSE

#define UART0_SENTINEL_STATUS_DEBUG		TRUE
#define UART0_SENTINEL_LOCAL_DHCP_DEBUG	TRUE

/* Protoboard Constants */
#define PROTOBOARD_LED_DDR				DDRD
#define PROTOBOARD_LED_PORT				PORTD
#define PROTOBOARD_LED_0				PD4
#define PROTOBOARD_LED_1				PD5

/* Includes */
#include "sentinel.h"

#define TX_NO_ACK_LATENCY_MS			500

/* Volatile Variables */

/* Global Variables */

/* Function Prototypes */

/**
 *
 */
ISR (INT0_vect) {
	zigbee_set_irq_flag();
}

/**
 *
 */
int main(void) {
	/* Local Variables */
	uint8_t rxtx_data[ZIGBEE_PHY_DATA_LENGTH];
	uint8_t rxtx_data_length;
	
	uint8_t rx_msg = ZIGBEE_RX_SUCCESS;
	uint8_t rx_ed_level = 0;
	uint8_t rx_busy_flag = FALSE;
	uint8_t rx_listen_flag = FALSE;
	
	uint8_t tx_msg = ZIGBEE_TX_SUCCESS;
	uint8_t tx_on_the_fly_flag = TRUE;
	
	/* Global Variable Initialization */
	mcu_status = MCU_STATUS_P_ON;
	
	/* Set PORT for LED Functionality */
	#if PROTOBOARD_LED_DEBUG
		sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_0);
		sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_1);
		sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
		sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);
	#endif

	/* MCU Initialization */
	init_mcu(MCU_PRR_ALL);			// Initialize MCU
	init_uart0(UART_BR_4800, UART_RXTX_MODE, UART_NO_ISR_ENABLE);		// Initialize UART
	init_spi_master(SPI_CLK_DIV_2);	// Initialize SPI Master
	init_ext_int0(TRUE, EXT_INT_SC_RISING_EDGE);	// Initialize External Int0

    /* ZigBee Hardware Initialization */
	zigbee_init();							// Hardware initialization after power-on event
   	zigbee_clear_irq_status();
	zigbee_clear_irq_flag();
	zigbee_set_extended_mode(TRUE);
	zigbee_set_pad_io_clkm(ZIGBEE_PAD_IO_CLKM_2MA);
	
	#if CPU_FREQUENCY_HZ == CPU_FREQUENCY_1_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_1MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_2_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_2MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_4_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_4MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_8_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_8MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_16_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_16MHZ);
	#endif
	
	sei();

	/* UART0 Init Debug */
	#if UART0_ZIGBEE_SPEC_DEBUG
		uart0_send_byte(UART_TX_ZIGBEE_PART_NUM);
		uart0_send_byte(zigbee_hw_part_num);
		uart0_send_byte(UART_TX_ZIGBEE_VER_NUM);
		uart0_send_byte(zigbee_hw_ver_num);
		uart0_send_byte(UART_TX_ZIGBEE_MAN_ID_0);
		uart0_send_byte(zigbee_hw_man_id_0);
		uart0_send_byte(UART_TX_ZIGBEE_MAN_ID_1);
		uart0_send_byte(zigbee_hw_man_id_1);
	#endif
	
	#if PROTOBOARD_LED_DEBUG
		cbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
	#endif

	while(1) {
		switch(mcu_status) {
			case MCU_STATUS_P_ON:
				#if UART0_SENTINEL_STATUS_DEBUG
					uart0_send_byte(UART_TX_SENTINEL_STATUS);
					uart0_send_byte(SENTINEL_STATUS_P_ON);
					uart0_send_byte(SENTINEL_NODE_TYPE);
					uart0_send_byte(eeprom_read_byte(&eeprom_local_dhcp_local_depth));
					uart0_send_int(SENTINEL_BASE_SHORT_ADDR);
					uart0_send_int(SENTINEL_BASE_PAN_ID);
				#endif
				
				zigbee_set_tx_src_short_addr(SENTINEL_BASE_SHORT_ADDR);
				zigbee_set_tx_src_pan_id(SENTINEL_BASE_PAN_ID);
				zigbee_set_addr_filter_short_addr(SENTINEL_BASE_SHORT_ADDR);
				zigbee_set_addr_filter_pan_id(SENTINEL_BASE_PAN_ID);
				
				zigbee_set_tx_sequence_number((uint8_t) (SENTINEL_BASE_SHORT_ADDR & 0xFF));
				zigbee_enable_tx_ack_request();
				mcu_status = MCU_STATUS_IDLE;
				break;
			case MCU_STATUS_IDLE:
				#if UART0_SENTINEL_STATUS_DEBUG
					uart0_send_byte(UART_TX_SENTINEL_STATUS);
					uart0_send_byte(SENTINEL_STATUS_IDLE);
				#endif
				mcu_status = MCU_STATUS_RX;
				break;
			case MCU_STATUS_TX_IDLE:
				switch (tx_msg) {
					case ZIGBEE_TX_SUCCESS:
						mcu_status = MCU_STATUS_IDLE;
						break;
					case ZIGBEE_TX_INVALID_ARGUMENT:
						rxtx_data_length--;
						mcu_status = MCU_STATUS_TX;
						break;
					case ZIGBEE_TX_BUFFER_UNDERRUN:
						tx_on_the_fly_flag = FALSE;
						mcu_status = MCU_STATUS_TX;
						break;
					case ZIGBEE_TX_CHANNEL_ACCESS_FAILURE:	// TODO: Resolve somehow
						mcu_status = MCU_STATUS_TX;
						break;
					case ZIGBEE_TX_NO_ACK:
						_delay_ms(TX_NO_ACK_LATENCY_MS);
						mcu_status = MCU_STATUS_TX;
						break;
					case ZIGBEE_TX_INVALID:
						zigbee_sw_rst();
						mcu_status = MCU_STATUS_TX;
						break;
				}
				break;
			case MCU_STATUS_TX:
				switch(zigbee_get_trx_status()) {
					case ZIGBEE_TRX_STATUS_TX_ARET_ON:
						zigbee_inc_tx_sequence_number();
						tx_msg = zigbee_extended_tx(rxtx_data, rxtx_data_length, tx_on_the_fly_flag);
						zigbee_enable_trx_off();
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_TRX_OFF);
						#endif
						mcu_status = MCU_STATUS_IDLE;
						break;
					case ZIGBEE_TRX_STATUS_BUSY_TX_ARET:
						break;
					default:
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_TX);
							uart0_send_byte(rxtx_data_length);
							uart0_send_byte_array(rxtx_data, rxtx_data_length);
							uart0_send_int(zigbee_get_tx_dest_short_addr());
							uart0_send_int(zigbee_get_tx_dest_pan_id());
						#endif
						zigbee_enable_extended_tx();
						break;
				}
				break;
			case MCU_STATUS_RX_IDLE:
				switch (rx_msg) {
					case ZIGBEE_RX_SUCCESS:
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_RX_SUCCESS);
							uart0_send_byte(rxtx_data_length);
							uart0_send_byte_array(rxtx_data, rxtx_data_length);
							uart0_send_int(zigbee_get_rx_src_short_addr());
							uart0_send_int(zigbee_get_rx_src_pan_id());
						#endif
						switch(sentinel_rf_rx_parse(rxtx_data, &rxtx_data_length)) {
							case SENTINEL_RF_PARSE_IDLE:
								mcu_status = MCU_STATUS_RX;
								break;
							case SENTINEL_RF_PARSE_TRANSMIT:
								tx_on_the_fly_flag = TRUE;
								mcu_status = MCU_STATUS_TX;
								break;
							case SENTINEL_RF_PARSE_SENSOR_DETECTION:
								uart0_send_byte(UART_TX_SENTINEL_SENSOR_DETECTION_REPORT);
								uart0_send_byte(rxtx_data[1]);	// SensorAddrL
								uart0_send_byte(rxtx_data[2]);	// SesnorAddrH
								mcu_status = MCU_STATUS_RX;
								break;
						}
						break;
					case ZIGBEE_RX_LOW_LQI:
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_RX_LOW_LQI);
						#endif
						mcu_status = MCU_STATUS_RX;
						break;
					case ZIGBEE_RX_DUPLICATE_FRAME:
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_RX_DUPLICATE_FRAME);
						#endif
						mcu_status = MCU_STATUS_RX;
						break;
					default:
						mcu_status = MCU_STATUS_RX;
						break;
				}
				break;
			case MCU_STATUS_RX:
				switch(zigbee_get_trx_status()) {
					case ZIGBEE_TRX_STATUS_RX_AACK_NOCLK:
					case ZIGBEE_TRX_STATUS_RX_AACK_ON:
						rx_busy_flag = FALSE;
						#if UART0_SENTINEL_STATUS_DEBUG
							if (!rx_listen_flag) {
								uart0_send_byte(UART_TX_SENTINEL_STATUS);
								uart0_send_byte(SENTINEL_STATUS_RX_LISTEN);
							}
						#endif
						rx_listen_flag = TRUE;
						
						if (UCSR0A & _BV(RXC0)) { 						// Process any UART data
							zigbee_enable_trx_off();
							#if UART0_SENTINEL_STATUS_DEBUG
								uart0_send_byte(UART_TX_SENTINEL_STATUS);
								uart0_send_byte(SENTINEL_STATUS_TRX_OFF);
							#endif
							sentinel_uart_rx_parse();
						}
						break;
					case ZIGBEE_TRX_STATUS_BUSY_RX_AACK_NOCLK:
					case ZIGBEE_TRX_STATUS_BUSY_RX_AACK:				// Reaches BUSY_RX ~300us before TRX_END interrupt
						rx_listen_flag = FALSE;
						#if UART0_SENTINEL_STATUS_DEBUG
							if (!rx_busy_flag) {
								uart0_send_byte(UART_TX_SENTINEL_STATUS);
								uart0_send_byte(SENTINEL_STATUS_RX_BUSY);
							}
						#endif
						rx_busy_flag = TRUE;
						
						if (zigbee_get_irq_flag()) {
							
							zigbee_clear_irq_flag();
							zigbee_update_irq_status(); 				// TRX_END interrupt
						}
						if (zigbee_get_rx_ready()) {
							zigbee_clear_rx_ready();
							rx_msg = zigbee_extended_rx(rxtx_data, &rxtx_data_length);
							rx_ed_level = zigbee_get_frame_ed_level();	// Needs to be read within 224 s after the TRX_END
							zigbee_enable_trx_off();									// Turn off TRX after ACK
							rx_busy_flag = FALSE;
							#if UART0_SENTINEL_STATUS_DEBUG
								uart0_send_byte(UART_TX_SENTINEL_STATUS);
								uart0_send_byte(SENTINEL_STATUS_TRX_OFF);
							#endif
							mcu_status = MCU_STATUS_RX_IDLE;
						}
						break;
					default:
						rx_listen_flag = FALSE;
						rx_busy_flag = FALSE;
						zigbee_enable_extended_rx();
						break;
				}
				break;
			default:
				mcu_status = MCU_STATUS_IDLE;
				break;
		}
	}
	
	return 0;
}
