/**
 * A Mega644P Local DHCP Relay Node Prototype
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2A/PCINT31)
 *							-----------
 */

/* Global/System Constants */
#define TRUE 					1
#define FALSE 					0

#define CPU_FREQUENCY_1_MHZ		1000000				// 1.8-5.5V
#define CPU_FREQUENCY_2_MHZ		2000000				// 1.8-5.5V
#define CPU_FREQUENCY_4_MHZ		4000000				// 1.8-5.5V
#define CPU_FREQUENCY_8_MHZ		8000000				// 2.7-5.5V
#define CPU_FREQUENCY_16_MHZ	16000000			// 4.0-5.5V
#define CPU_FREQUENCY_HZ		CPU_FREQUENCY_1_MHZ

#define sbi(data,bit)			(data |= (1<<bit))	// Set Bit
#define cbi(data,bit) 			(data &= ~(1<<bit))	// Clear Bit
#define gbi(data,bit)			(data & (1<<bit))	// Get Bit
#define tbi(data,bit)			(data ^= (1<<bit))	// Toggle Bit

/* Node Constants */
#define SENTINEL_BASE_NODE		0x00
#define SENTINEL_RELAY_NODE		0x01
#define SENTINEL_SENSOR_NODE	0x02
#define SENTINEL_NODE_TYPE		SENTINEL_RELAY_NODE

/* Debug Constants */
#define PROTOBOARD_LED_DEBUG			TRUE

#define UART0_ZIGBEE_SPEC_DEBUG			FALSE
#define UART0_ZIGBEE_IRQ_STATUS_DEBUG	FALSE
#define UART0_ZIGBEE_TRAC_STATUS_DEBUG	FALSE
#define UART0_ZIGBEE_STATE_DEBUG		FALSE	// Heavy traffic
#define UART0_ZIGBEE_RX_PHY_DATA_DEBUG	FALSE
#define UART0_ZIGBEE_RX_PHY_LQI_DEBUG	FALSE
#define UART0_ZIGBEE_RX_MAC_DATA_DEBUG	FALSE
#define UART0_ZIGBEE_RX_MAC_FRAME_DEBUG	FALSE
#define UART0_ZIGBEE_ED_LEVEL_DEBUG		FALSE
#define UART0_ZIGBEE_CCA_DEBUG			FALSE

#define UART0_SENTINEL_STATUS_DEBUG		TRUE
#define UART0_SENTINEL_LOCAL_DHCP_DEBUG	TRUE

/* Protoboard Constants */
#define PROTOBOARD_LED_DDR				DDRD
#define PROTOBOARD_LED_PORT				PORTD
#define PROTOBOARD_LED_0				PD4
#define PROTOBOARD_LED_1				PD5

/* Includes */
#include "sentinel.h"

/* Constants for Timer */
#define TIMER2A_RES_MS								200
#define TIMER_SEC_LATENCY_MS						1000				
#define TIMER_SEC_LATENCY_THRESHOLD					(uint16_t) (TIMER_SEC_LATENCY_MS / TIMER2A_RES_MS)
#define TIMER_MIN_LATENCY_MS						60000				
#define TIMER_MIN_LATENCY_THRESHOLD					(uint16_t) (TIMER_MIN_LATENCY_MS / TIMER2A_RES_MS)
#define SENSOR_LATENCY_MS							10000
#define SENSOR_LATENCY_THRESHOLD					(uint16_t) (SENSOR_LATENCY_MS / TIMER2A_RES_MS)
#define LOCAL_DHCP_RX_TIMEOUT_LATENCY_MS			1000
#define LOCAL_DHCP_RX_TIMEOUT_LATENCY_THRESHOLD		(uint16_t) (LOCAL_DHCP_RX_TIMEOUT_LATENCY_MS / TIMER2A_RES_MS)
#define LOCAL_DHCP_OFFER_TIMEOUT_LATENCY_MS			5000
#define LOCAL_DHCP_OFFER_TIMEOUT_LATENCY_THRESHOLD	(uint16_t) (LOCAL_DHCP_OFFER_TIMEOUT_LATENCY_MS / TIMER2A_RES_MS)
#define LOCAL_DHCP_NO_ACK_LATENCY_MS				500
#define LOCAL_DHCP_NO_ACK_LATENCY_THRESHOLD			(uint16_t) (LOCAL_DHCP_NO_ACK_LATENCY_MS / TIMER2A_RES_MS)
#define LOCAL_DHCP_NO_ACK_RETRY_LATENCY_MS			5000
#define LOCAL_DHCP_NO_ACK_RETRY_LATENCY_THRESHOLD	(uint16_t) (LOCAL_DHCP_NO_ACK_RETRY_LATENCY_MS / TIMER2A_RES_MS)
#define LOCAL_DHCP_NO_ACK_TIMEOUT_LATENCY_MS		10000
#define LOCAL_DHCP_NO_ACK_TIMEOUT_LATENCY_THRESHOLD	(uint16_t) (LOCAL_DHCP_NO_ACK_TIMEOUT_LATENCY_MS / TIMER2A_RES_MS)
#define TX_NO_ACK_LATENCY_MS						500
#define TX_NO_ACK_LATENCY_THRESHOLD					(uint16_t) (TX_NO_ACK_LATENCY_MS / TIMER2A_RES_MS)

/* Volatile Variables */
volatile uint16_t local_dhcp_rx_timeout_latency_tick_count;
volatile uint16_t local_dhcp_offer_timeout_latency_tick_count;
volatile uint16_t local_dhcp_no_ack_latency_tick_count;
volatile uint16_t tx_no_ack_latency_tick_count;

/* Global Variables */

/* Function Prototypes */

/**
 *
 */
ISR(TIMER2_COMPA_vect) {
	local_dhcp_rx_timeout_latency_tick_count++;
	local_dhcp_offer_timeout_latency_tick_count++;
	local_dhcp_no_ack_latency_tick_count++;
	tx_no_ack_latency_tick_count++;
}

/**
 *
 */
ISR (INT0_vect) {
	zigbee_set_irq_flag();
}

/**
 *
 */
int main(void) {
	/* Local Variables */
	uint8_t rxtx_data[ZIGBEE_PHY_DATA_LENGTH];
	uint8_t rxtx_data_length;
	
	uint8_t rx_msg = ZIGBEE_RX_SUCCESS;
	uint8_t rx_ed_level = 0;
	uint8_t rx_busy_flag = FALSE;
	uint8_t rx_listen_flag = FALSE;
	
	uint8_t tx_msg = ZIGBEE_TX_SUCCESS;
	uint8_t tx_on_the_fly_flag = TRUE;
	uint8_t tx_no_ack_latency_enable_flag = FALSE;
	uint8_t tx_no_ack_counter = 0;
	
	uint8_t local_dhcp_offer_counter = 0;
	uint8_t local_dhcp_offer_ed_level[16];
	uint16_t local_dhcp_offer_client_short_addr[16];
	uint16_t local_dhcp_offer_server_short_addr[16];
	uint8_t local_dhcp_offer_server_node_type[16];
	uint8_t local_dhcp_offer_server_depth[16];
	uint8_t local_dhcp_offer_request_index = 0;
	
	uint8_t local_dhcp_local_depth = 0;
	uint16_t local_dhcp_local_short_addr = 0;
	uint16_t local_dhcp_server_short_addr = 0;
	uint16_t local_dhcp_server_pan_id = 0;
	uint8_t local_dhcp_server_node_type = 0;
	uint8_t local_dhcp_server_depth = 0;
	
	uint8_t local_dhcp_retry_counter = 0;
	
	/* Global Variable Initialization */
	mcu_status = MCU_STATUS_P_ON;
	
	/* Set PORT for LED Functionality */
	#if PROTOBOARD_LED_DEBUG
		sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_0);
		sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_1);
		sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
		sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);
	#endif

	/* MCU Initialization */
	init_mcu(MCU_PRR_ALL);			// Initialize MCU
	init_uart0(UART_BR_4800, UART_RXTX_MODE, UART_NO_ISR_ENABLE);		// Initialize UART
	init_spi_master(SPI_CLK_DIV_2);	// Initialize SPI Master
	init_ext_int0(TRUE, EXT_INT_SC_RISING_EDGE);	// Initialize External Int0
	init_timer2();
	
	sei();

    /* ZigBee Hardware Initialization */
	zigbee_init();							// Hardware initialization after power-on event
   	zigbee_clear_irq_status();
	zigbee_set_extended_mode(TRUE);
	zigbee_set_pad_io_clkm(ZIGBEE_PAD_IO_CLKM_2MA);
	zigbee_clear_irq_flag();
	
	#if CPU_FREQUENCY_HZ == CPU_FREQUENCY_1_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_1MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_2_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_2MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_4_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_4MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_8_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_8MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_16_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_16MHZ);
	#endif
	
	/* UART0 Init Debug */
	#if UART0_ZIGBEE_SPEC_DEBUG
		uart0_send_byte(UART_TX_ZIGBEE_PART_NUM);
		uart0_send_byte(zigbee_hw_part_num);
		uart0_send_byte(UART_TX_ZIGBEE_VER_NUM);
		uart0_send_byte(zigbee_hw_ver_num);
		uart0_send_byte(UART_TX_ZIGBEE_MAN_ID_0);
		uart0_send_byte(zigbee_hw_man_id_0);
		uart0_send_byte(UART_TX_ZIGBEE_MAN_ID_1);
		uart0_send_byte(zigbee_hw_man_id_1);
	#endif
	
	#if PROTOBOARD_LED_DEBUG
		cbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
	#endif

	while(1) {
		switch(mcu_status) {
			case MCU_STATUS_P_ON:
				#if UART0_SENTINEL_STATUS_DEBUG
					uart0_send_byte(UART_TX_SENTINEL_STATUS);
					uart0_send_byte(SENTINEL_STATUS_P_ON);
					uart0_send_byte(SENTINEL_NODE_TYPE);
				#endif
				
				if (eeprom_read_byte(&eeprom_local_dhcp_complete_flag)) {
					local_dhcp_local_depth = eeprom_read_byte(&eeprom_local_dhcp_local_depth);
					local_dhcp_local_short_addr = eeprom_read_word(&eeprom_local_dhcp_local_short_addr);
					local_dhcp_server_short_addr = eeprom_read_word(&eeprom_local_dhcp_server_short_addr);
					local_dhcp_server_pan_id = eeprom_read_word(&eeprom_local_dhcp_server_pan_id);
					local_dhcp_server_node_type = eeprom_read_byte(&eeprom_local_dhcp_server_node_type);
					local_dhcp_server_depth = eeprom_read_byte(&eeprom_local_dhcp_server_depth);
					
					zigbee_set_tx_src_short_addr(local_dhcp_local_short_addr);
					zigbee_set_tx_src_pan_id(SENTINEL_RELAY_PAN_ID);
					zigbee_set_addr_filter_short_addr(local_dhcp_local_short_addr);
					zigbee_set_addr_filter_pan_id(SENTINEL_RELAY_PAN_ID);
					
					zigbee_set_tx_dest_short_addr(local_dhcp_server_short_addr);
					zigbee_set_tx_dest_pan_id(local_dhcp_server_pan_id);
					
					zigbee_set_tx_sequence_number((uint8_t) (local_dhcp_local_short_addr & 0xFF));
					local_dhcp_status = LOCAL_DHCP_STATUS_COMPLETE;
				}
				else {
					local_dhcp_status = LOCAL_DHCP_STATUS_INIT;
				}
				zigbee_enable_tx_ack_request();
				mcu_status = MCU_STATUS_DHCP;
				break;
			case MCU_STATUS_DHCP:
				switch(local_dhcp_status) {
					case LOCAL_DHCP_STATUS_INIT:
						local_dhcp_retry_counter = 0;
						local_dhcp_status = LOCAL_DHCP_STATUS_TX_DISCOVERY;
						break;
					case LOCAL_DHCP_STATUS_TX_DISCOVERY:
						local_dhcp_offer_counter = 0;
						tx_no_ack_counter = 0;
						zigbee_set_tx_src_short_addr(SENTINEL_RELAY_INIT_SHORT_ADDR);	// Need to transmit PAN ID
						zigbee_set_tx_src_pan_id(SENTINEL_RELAY_INIT_PAN_ID);
						zigbee_set_addr_filter_short_addr(SENTINEL_RELAY_INIT_SHORT_ADDR);
						zigbee_set_addr_filter_pan_id(SENTINEL_RELAY_INIT_PAN_ID);
						
						zigbee_set_tx_dest_short_addr(SENTINEL_GLOBAL_SHORT_ADDR);
						if ((local_dhcp_retry_counter % 2) == 0) {
							#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
								uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
								uart0_send_byte(SENTINEL_LOCAL_DHCP_TX_DISCOVERY);
								uart0_send_byte(SENTINEL_BASE_NODE);
							#endif
							zigbee_set_tx_dest_pan_id(SENTINEL_BASE_PAN_ID);
						}
						else {
							#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
								uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
								uart0_send_byte(SENTINEL_LOCAL_DHCP_TX_DISCOVERY);
								uart0_send_byte(SENTINEL_RELAY_NODE);
							#endif
							zigbee_set_tx_dest_pan_id(SENTINEL_RELAY_PAN_ID);		
						}
						local_dhcp_retry_counter++;
						
						zigbee_set_tx_sequence_number((uint8_t) (SENTINEL_RELAY_INIT_SHORT_ADDR & 0xFF));
						
						rxtx_data[0] = SENTINEL_RF_LOCAL_DHCP_DISCOVERY;
						rxtx_data[1] = SENTINEL_RELAY_NODE;
						rxtx_data_length = 2;
						timer_enable_timer2a_isr();
						local_dhcp_rx_timeout_latency_tick_count = 0;
						local_dhcp_status = LOCAL_DHCP_STATUS_RX_OFFER;
						mcu_status = MCU_STATUS_TX;
						break;
					case LOCAL_DHCP_STATUS_RX_OFFER:
						#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
							uart0_send_byte(SENTINEL_LOCAL_DHCP_RX_OFFER_MODE);
						#endif
						local_dhcp_offer_timeout_latency_tick_count = 0;
						mcu_status = MCU_STATUS_RX;
						break;
					case LOCAL_DHCP_STATUS_TX_REQUEST:
						#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
							uart0_send_byte(SENTINEL_LOCAL_DHCP_TX_REQUEST);
							uart0_send_byte(local_dhcp_offer_request_index);
							uart0_send_byte(local_dhcp_offer_server_node_type[local_dhcp_offer_request_index]);
							uart0_send_byte(local_dhcp_offer_server_depth[local_dhcp_offer_request_index]);
							uart0_send_byte(local_dhcp_offer_ed_level[local_dhcp_offer_request_index]);
							uart0_send_int(local_dhcp_offer_server_short_addr[local_dhcp_offer_request_index]);
							uart0_send_int(local_dhcp_offer_client_short_addr[local_dhcp_offer_request_index]);
						#endif
						rxtx_data[0] = SENTINEL_RF_LOCAL_DHCP_REQUEST;
						rxtx_data[1] = SENTINEL_RELAY_NODE;
						rxtx_data[2] = (uint8_t) (local_dhcp_offer_server_short_addr[local_dhcp_offer_request_index] & 0xFF);
						rxtx_data[3] = (uint8_t) ((local_dhcp_offer_server_short_addr[local_dhcp_offer_request_index] >> 8) & 0xFF);
						rxtx_data[4] = (uint8_t) (local_dhcp_offer_client_short_addr[local_dhcp_offer_request_index] & 0xFF);
						rxtx_data[5] = (uint8_t) ((local_dhcp_offer_client_short_addr[local_dhcp_offer_request_index] >> 8) & 0xFF);
						rxtx_data_length = 6;
						timer_enable_timer2a_isr();
						local_dhcp_rx_timeout_latency_tick_count = 0;
						local_dhcp_status = LOCAL_DHCP_STATUS_RX_ACK;
						mcu_status = MCU_STATUS_TX;
						break;
					case LOCAL_DHCP_STATUS_RX_ACK:
						local_dhcp_status = LOCAL_DHCP_STATUS_CONFIG;
						mcu_status = MCU_STATUS_RX;
						break;
					case LOCAL_DHCP_STATUS_CONFIG:
						#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
							uart0_send_byte(SENTINEL_LOCAL_DHCP_CONFIG);
						#endif
						
						eeprom_write_byte(&eeprom_local_dhcp_local_depth, local_dhcp_local_depth);
						eeprom_write_word(&eeprom_local_dhcp_local_short_addr, local_dhcp_local_short_addr);
						eeprom_write_word(&eeprom_local_dhcp_server_short_addr, local_dhcp_server_short_addr);
						eeprom_write_word(&eeprom_local_dhcp_server_pan_id, local_dhcp_server_pan_id);
						eeprom_write_byte(&eeprom_local_dhcp_server_node_type, local_dhcp_server_node_type);
						eeprom_write_byte(&eeprom_local_dhcp_server_depth, local_dhcp_server_depth);
						eeprom_write_byte(&eeprom_local_dhcp_complete_flag, TRUE);
						
						zigbee_set_tx_src_short_addr(local_dhcp_local_short_addr);
						zigbee_set_tx_src_pan_id(SENTINEL_RELAY_PAN_ID);
						zigbee_set_addr_filter_short_addr(local_dhcp_local_short_addr);
						zigbee_set_addr_filter_pan_id(SENTINEL_RELAY_PAN_ID);
						
						zigbee_set_tx_dest_short_addr(local_dhcp_server_short_addr);
						zigbee_set_tx_dest_pan_id(local_dhcp_server_pan_id);
						
						zigbee_set_tx_sequence_number((uint8_t) (local_dhcp_local_short_addr & 0xFF));
						local_dhcp_status = LOCAL_DHCP_STATUS_COMPLETE;
						break;
					case LOCAL_DHCP_STATUS_COMPLETE:
						#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
							uart0_send_byte(SENTINEL_LOCAL_DHCP_COMPLETE);
							uart0_send_byte(local_dhcp_local_depth);
							uart0_send_int(local_dhcp_local_short_addr);
							uart0_send_int(SENTINEL_RELAY_PAN_ID);
							uart0_send_int(local_dhcp_server_short_addr);
							uart0_send_int(local_dhcp_server_pan_id);
						#endif
						mcu_status = MCU_STATUS_IDLE;
						break;
				}
				break;
			case MCU_STATUS_IDLE:
				#if UART0_SENTINEL_STATUS_DEBUG
					uart0_send_byte(UART_TX_SENTINEL_STATUS);
					uart0_send_byte(SENTINEL_STATUS_IDLE);
				#endif
				if (local_dhcp_status == LOCAL_DHCP_STATUS_COMPLETE) {
					mcu_status = MCU_STATUS_RX;
				}
				else {
					mcu_status = MCU_STATUS_DHCP;
				}
				break;
			case MCU_STATUS_SLEEP:
				zigbee_enable_extended_rx_sleep();
				set_sleep_mode(SLEEP_MODE_PWR_DOWN);
				sleep_mode();
				mcu_status = MCU_STATUS_RX;
				break;
			case MCU_STATUS_TX_IDLE:
				switch (tx_msg) {
					case ZIGBEE_TX_SUCCESS:
						tx_no_ack_counter = 0;
						
						if (local_dhcp_status == LOCAL_DHCP_STATUS_COMPLETE) {
							mcu_status = MCU_STATUS_IDLE;
						}
						else {
							mcu_status = MCU_STATUS_DHCP;
						}
						break;
					case ZIGBEE_TX_INVALID_ARGUMENT:
						rxtx_data_length--;
						mcu_status = MCU_STATUS_TX;
						break;
					case ZIGBEE_TX_BUFFER_UNDERRUN:
						tx_on_the_fly_flag = FALSE;
						mcu_status = MCU_STATUS_TX;
						break;
					case ZIGBEE_TX_CHANNEL_ACCESS_FAILURE:
						mcu_status = MCU_STATUS_TX;
						break;
					case ZIGBEE_TX_NO_ACK:
						if (tx_no_ack_latency_enable_flag) {
							tx_no_ack_latency_enable_flag = FALSE;
							
							local_dhcp_no_ack_latency_tick_count = 0;
							tx_no_ack_latency_tick_count = 0;
							
							tx_no_ack_counter++;
							
							timer_enable_timer2a_isr();
							set_sleep_mode(SLEEP_MODE_PWR_SAVE);
						}
						sleep_mode();
						
						if (local_dhcp_status == LOCAL_DHCP_STATUS_COMPLETE) {
							#if SENTINEL_TX_NO_ACK_LOCAL_DHCP_RESET_ENABLE
								if (tx_no_ack_counter > SENTINEL_TX_NO_ACK_LOCAL_DHCP_RESET_MAX_COUNT) {
									timer_disable_timer2a_isr();
									sentinel_reset_local_dhcp();
								}
								else if (tx_no_ack_latency_tick_count > TX_NO_ACK_LATENCY_THRESHOLD) {
									timer_disable_timer2a_isr();
									mcu_status = MCU_STATUS_TX;
								}
							#else
								if (tx_no_ack_latency_tick_count > TX_NO_ACK_LATENCY_THRESHOLD) {
									timer_disable_timer2a_isr();
									mcu_status = MCU_STATUS_TX;
								}
							#endif
						}
						else if (local_dhcp_retry_counter > SENTINEL_LOCAL_DHCP_NO_ACK_RETRY_MAX_COUNT) {
							if (local_dhcp_no_ack_latency_tick_count > LOCAL_DHCP_NO_ACK_TIMEOUT_LATENCY_THRESHOLD) {
								timer_disable_timer2a_isr();
								sentinel_reset_local_dhcp();
							}
						}
						else if (tx_no_ack_counter > SENTINEL_LOCAL_DHCP_NO_ACK_TIMEOUT_MAX_COUNT) {
							if (local_dhcp_no_ack_latency_tick_count > LOCAL_DHCP_NO_ACK_RETRY_LATENCY_THRESHOLD) {
								timer_disable_timer2a_isr();
								sentinel_retry_local_dhcp();
							}
						}
						else {
							if (local_dhcp_no_ack_latency_tick_count > LOCAL_DHCP_NO_ACK_LATENCY_THRESHOLD) {
								timer_disable_timer2a_isr();
								mcu_status = MCU_STATUS_TX;
							}
						}
						break;
					case ZIGBEE_TX_INVALID:
						zigbee_sw_rst();
						mcu_status = MCU_STATUS_TX;
						break;
				}
				break;
			case MCU_STATUS_TX:
				switch(zigbee_get_trx_status()) {
					case ZIGBEE_TRX_STATUS_TX_ARET_ON:
						zigbee_inc_tx_sequence_number();
						tx_msg = zigbee_extended_tx(rxtx_data, rxtx_data_length, tx_on_the_fly_flag);
						zigbee_enable_trx_off();
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_TRX_OFF);
						#endif
						tx_no_ack_latency_enable_flag = TRUE;
						mcu_status = MCU_STATUS_TX_IDLE;
						break;
					case ZIGBEE_TRX_STATUS_BUSY_TX_ARET:
						break;
					default:
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_TX);
							uart0_send_byte(rxtx_data_length);
							uart0_send_byte_array(rxtx_data, rxtx_data_length);
							uart0_send_int(zigbee_get_tx_dest_short_addr());
							uart0_send_int(zigbee_get_tx_dest_pan_id());
						#endif
						zigbee_enable_extended_tx();
						break;
				}
				break;
			case MCU_STATUS_RX_IDLE:
				switch (rx_msg) {
					case ZIGBEE_RX_SUCCESS:
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_RX_SUCCESS);
							uart0_send_byte(rxtx_data_length);
							uart0_send_byte_array(rxtx_data, rxtx_data_length);
							uart0_send_int(zigbee_get_rx_src_short_addr());
							uart0_send_int(zigbee_get_rx_src_pan_id());
						#endif
						switch(sentinel_rf_rx_parse(rxtx_data, &rxtx_data_length)) {
							case SENTINEL_RF_PARSE_IDLE:
								mcu_status = MCU_STATUS_IDLE;
								break;
							case SENTINEL_RF_PARSE_TRANSMIT:
								tx_on_the_fly_flag = TRUE;
								mcu_status = MCU_STATUS_TX;
								break;
							case SENTINEL_RF_PARSE_DHCP_OFFER:
								local_dhcp_offer_server_node_type[local_dhcp_offer_counter] = rxtx_data[0];
								local_dhcp_offer_server_depth[local_dhcp_offer_counter] = rxtx_data[1];
								local_dhcp_offer_ed_level[local_dhcp_offer_counter] = rx_ed_level;
								local_dhcp_offer_client_short_addr[local_dhcp_offer_counter] = (((uint16_t) rxtx_data[3]) << 8) | ((uint16_t) rxtx_data[2]);
								local_dhcp_offer_server_short_addr[local_dhcp_offer_counter] = (((uint16_t) rxtx_data[5]) << 8) | ((uint16_t) rxtx_data[4]);
								#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
									uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
									uart0_send_byte(SENTINEL_LOCAL_DHCP_RX_OFFER);
									uart0_send_byte(local_dhcp_offer_counter);
									uart0_send_byte(local_dhcp_offer_server_node_type[local_dhcp_offer_counter]);
									uart0_send_byte(local_dhcp_offer_server_depth[local_dhcp_offer_counter]);
									uart0_send_byte(local_dhcp_offer_ed_level[local_dhcp_offer_counter]);
									uart0_send_int(local_dhcp_offer_client_short_addr[local_dhcp_offer_counter]);
									uart0_send_int(local_dhcp_offer_server_short_addr[local_dhcp_offer_counter]);
								#endif
								local_dhcp_offer_counter++;
								mcu_status = MCU_STATUS_RX;
								break;
							case SENTINEL_RF_PARSE_DHCP_ACK:
								local_dhcp_server_node_type = rxtx_data[0];
								local_dhcp_server_depth = rxtx_data[1];
								local_dhcp_local_depth = rxtx_data[2];
								local_dhcp_local_short_addr = (((uint16_t) rxtx_data[4]) << 8) | ((uint16_t) rxtx_data[3]);
								local_dhcp_server_short_addr = (((uint16_t) rxtx_data[6]) << 8) | ((uint16_t) rxtx_data[5]);
								local_dhcp_server_pan_id = (((uint16_t) rxtx_data[8]) << 8) | ((uint16_t) rxtx_data[7]);
								mcu_status = MCU_STATUS_DHCP;
								break;
							case SENTINEL_RF_PARSE_SENSOR_DETECTION:
								zigbee_set_tx_dest_short_addr(local_dhcp_server_short_addr);
								zigbee_set_tx_dest_pan_id(local_dhcp_server_pan_id);

								uart0_send_byte(UART_TX_SENTINEL_SENSOR_DETECTION_RELAY);
								uart0_send_byte(rxtx_data[1]);	// SensorAddrL
								uart0_send_byte(rxtx_data[2]);	// SensorAddrH
								uart0_send_int(zigbee_get_rx_src_short_addr());
								uart0_send_int(zigbee_get_rx_src_pan_id());
								
								tx_on_the_fly_flag = TRUE;
								mcu_status = MCU_STATUS_TX;
								break;
						}
						break;
					case ZIGBEE_RX_LOW_LQI:
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_RX_LOW_LQI);
						#endif
						mcu_status = MCU_STATUS_RX;
						break;
					case ZIGBEE_RX_DUPLICATE_FRAME:
						#if UART0_SENTINEL_STATUS_DEBUG
							uart0_send_byte(UART_TX_SENTINEL_STATUS);
							uart0_send_byte(SENTINEL_STATUS_RX_DUPLICATE_FRAME);
						#endif
						mcu_status = MCU_STATUS_RX;
						break;
					default:
						mcu_status = MCU_STATUS_RX;
						break;
				}
				break;
			case MCU_STATUS_RX:
				switch(zigbee_get_trx_status()) {
					case ZIGBEE_TRX_STATUS_RX_AACK_NOCLK:
					case ZIGBEE_TRX_STATUS_RX_AACK_ON:
						rx_busy_flag = FALSE;
						#if UART0_SENTINEL_STATUS_DEBUG
							if (!rx_listen_flag) {
								uart0_send_byte(UART_TX_SENTINEL_STATUS);
								uart0_send_byte(SENTINEL_STATUS_RX_LISTEN);
							}
						#endif
						rx_listen_flag = TRUE;
						
						if (local_dhcp_status == LOCAL_DHCP_STATUS_RX_ACK && local_dhcp_rx_timeout_latency_tick_count > LOCAL_DHCP_RX_TIMEOUT_LATENCY_THRESHOLD) {
							zigbee_enable_trx_off();
							timer_disable_timer2a_isr();
							
							#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
								uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
								uart0_send_byte(SENTINEL_LOCAL_DHCP_RETRY_TIMEOUT);
							#endif
							
							local_dhcp_status = LOCAL_DHCP_STATUS_TX_DISCOVERY;
							mcu_status = MCU_STATUS_DHCP;
						}
						if (local_dhcp_status == LOCAL_DHCP_STATUS_RX_OFFER && local_dhcp_offer_timeout_latency_tick_count > LOCAL_DHCP_OFFER_TIMEOUT_LATENCY_THRESHOLD) {
							zigbee_enable_trx_off();
							timer_disable_timer2a_isr();
							
							#if UART0_SENTINEL_LOCAL_DHCP_DEBUG
								uart0_send_byte(UART_TX_SENTINEL_LOCAL_DHCP);
								uart0_send_byte(SENTINEL_LOCAL_DHCP_OFFER_TIMEOUT);
							#endif
							
							local_dhcp_offer_request_index = 0;
							if (local_dhcp_offer_counter > 0) {
								for (uint8_t i=0; i<local_dhcp_offer_counter; i++) {
									if (local_dhcp_offer_ed_level[i] > local_dhcp_offer_ed_level[local_dhcp_offer_request_index]) {
										local_dhcp_offer_request_index = i;
									}
									else if ((local_dhcp_offer_ed_level[i] == local_dhcp_offer_ed_level[local_dhcp_offer_request_index]) && (local_dhcp_offer_server_depth[i] = local_dhcp_offer_server_depth[local_dhcp_offer_request_index])) {
										local_dhcp_offer_request_index = i;
									}
								}
								local_dhcp_status = LOCAL_DHCP_STATUS_TX_REQUEST;
							}
							else {
								sentinel_retry_local_dhcp();
							}
							mcu_status = MCU_STATUS_DHCP;
						}
						/*
						if (local_dhcp_status == LOCAL_DHCP_STATUS_COMPLETE) {
							#if UART0_SENTINEL_STATUS_DEBUG
								uart0_send_byte(UART_TX_SENTINEL_STATUS);
								uart0_send_byte(SENTINEL_STATUS_SLEEP);
							#endif
							mcu_status = MCU_STATUS_SLEEP;
						}
						*/
						break;
					case ZIGBEE_TRX_STATUS_BUSY_RX_AACK_NOCLK:
					case ZIGBEE_TRX_STATUS_BUSY_RX_AACK:
						rx_listen_flag = FALSE;
						#if UART0_SENTINEL_STATUS_DEBUG
							if (!rx_busy_flag) {
								uart0_send_byte(UART_TX_SENTINEL_STATUS);
								uart0_send_byte(SENTINEL_STATUS_RX_BUSY);
							}
						#endif
						rx_busy_flag = TRUE;
						
						if (zigbee_get_irq_flag()) {
							zigbee_clear_irq_flag();
							zigbee_update_irq_status(); 					// TRX_END interrupt
						}
						if (zigbee_get_rx_ready()) {
							zigbee_clear_rx_ready();
							if (local_dhcp_status != LOCAL_DHCP_STATUS_RX_OFFER) {
								timer_disable_timer2a_isr();
							}
							rx_msg = zigbee_extended_rx(rxtx_data, &rxtx_data_length);
							rx_ed_level = zigbee_get_frame_ed_level();	// Needs to be read within 224 s after the TRX_END
							zigbee_enable_trx_off();		// Turn off TRX after ACK
							rx_busy_flag = FALSE;
							#if UART0_SENTINEL_STATUS_DEBUG
								uart0_send_byte(UART_TX_SENTINEL_STATUS);
								uart0_send_byte(SENTINEL_STATUS_TRX_OFF);
							#endif
							mcu_status = MCU_STATUS_RX_IDLE;
						}
						break;
					default:
						rx_listen_flag = FALSE;
						rx_busy_flag = FALSE;
						zigbee_enable_extended_rx();
						break;
				}
				break;
			default:
				mcu_status = MCU_STATUS_IDLE;
				break;
		}
	}
	
	return 0;
}
