/**
 * A Mega644P Base Node Prototype
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2A/PCINT31)
 *							-----------
 */

/* Global/System Constants */
#define TRUE 				1
#define FALSE 				0

#define CPU_FREQUENCY_1_MHZ		1000000
#define CPU_FREQUENCY_2_MHZ		2000000
#define CPU_FREQUENCY_4_MHZ		4000000
#define CPU_FREQUENCY_8_MHZ		8000000
#define CPU_FREQUENCY_16_MHZ	16000000
#define CPU_FREQUENCY_HZ		CPU_FREQUENCY_1_MHZ

#define sbi(data,bit)		(data |= (1<<bit))	// Set Bit
#define cbi(data,bit) 		(data &= ~(1<<bit))	// Clear Bit
#define gbi(data,bit)		(data & (1<<bit))	// Get Bit
#define tbi(data,bit)		(data ^= (1<<bit))	// Toggle Bit

/* Zigbee Constants */
#define BASE_NODE_SHORT_ADDR	0x0000
#define BASE_NODE_PAN_ID		0x0000
#define RELAY_NODE_SHORT_ADDR	0x0000
#define RELAY_NODE_PAN_ID		0x0001
#define SENSOR_NODE_SHORT_ADDR	0x0000
#define SENSOR_NODE_PAN_ID		0x0010

/* Debug Constants */
#define PROTOBOARD_LED_DEBUG	TRUE

#define UART0_ZIGBEE_SPEC_DEBUG			FALSE
#define UART0_ZIGBEE_IRQ_STATUS_DEBUG	TRUE
#define UART0_ZIGBEE_TRAC_STATUS_DEBUG	TRUE
#define UART0_ZIGBEE_STATE_DEBUG		FALSE	// Heavy traffic
#define UART0_ZIGBEE_RX_PHY_DATA_DEBUG	TRUE
#define UART0_ZIGBEE_RX_PHY_LQI_DEBUG	TRUE
#define UART0_ZIGBEE_RX_MAC_DATA_DEBUG	TRUE
#define UART0_ZIGBEE_RX_MAC_FRAME_DEBUG	TRUE
#define UART0_ZIGBEE_CCA_DEBUG			FALSE

/* Protoboard Constants */
#define PROTOBOARD_LED_DDR	DDRD
#define PROTOBOARD_LED_PORT	PORTD
#define PROTOBOARD_LED_0	PD4
#define PROTOBOARD_LED_1	PD5

/* Includes */
#include "sentinel.h"

/* EEPROM Variables */
/**
 * uint8_t 	eeprom_read_byte (const uint8_t *addr)
 * uint16_t eeprom_read_word (const uint16_t *addr)
 * void 	eeprom_read_block (void *pointer_ram, const void *pointer_eeprom, size_t n)
 * void 	eeprom_write_byte (uint8_t *addr, uint8_t value)
 * void 	eeprom_write_word (uint16_t *addr, uint16_t value)
 * void 	eeprom_write_block (const void *pointer_ram, void *pointer_eeprom, size_t n)
 */
//#define EEPROM_DATA_BUFFER_LENGTH	5
//uint8_t EEMEM eeprom_data_index = 0;
//uint8_t EEMEM eeprom_data_buffer[EEPROM_DATA_BUFFER_LENGTH] = {0, 0, 0, 0, 0};
//uint16_t EEMEM eeprom_data = 0;

/* Constants for Timer */
#define TIMER2A_RES_MS					200
#define TIMER_SEC_LATENCY_MS			1000				
#define TIMER_SEC_LATENCY_THRESHOLD		(uint16_t) (TIMER_SEC_LATENCY_MS / TIMER2A_RES_MS)
#define TIMER_MIN_LATENCY_MS			60000				
#define TIMER_MIN_LATENCY_THRESHOLD		(uint16_t) (TIMER_MIN_LATENCY_MS / TIMER2A_RES_MS)
#define SENSOR_LATENCY_MS				10000
#define SENSOR_LATENCY_THRESHOLD		(uint16_t) (SENSOR_LATENCY_MS / TIMER2A_RES_MS)

/* Volatile Variables */
volatile uint8_t sensor_detection_flag;
volatile uint16_t sensor_latency_tick_count;

/* Global Variables */
uint8_t sensor_latency_enable_flag;

/* Function Prototypes */


ISR(TIMER2_COMPA_vect) {
	sensor_latency_tick_count++;
}

ISR (INT0_vect) {
	zigbee_set_irq_flag();
}

ISR (INT1_vect) {
	sensor_detection_flag = TRUE;
	ext_int_disable_int1();			// Must remain in ISR
}

/**
 *
 */
int main(void) {
	/* Local Variables */
	uint8_t rxtx_data[ZIGBEE_PHY_DATA_LENGTH];
	uint8_t rxtx_data_length;
	uint8_t rxtx_msg = ZIGBEE_RXTX_SUCCESS;
	uint8_t tx_on_the_fly_flag = TRUE;
	
	/* Global Variable Initialization */
	mcu_status = MCU_STATUS_P_ON;
	sensor_detection_flag = FALSE;
	sensor_latency_tick_count = 0;
	sensor_latency_enable_flag = FALSE;

	/* Set PORT for LED Functionality */
	#if PROTOBOARD_LED_DEBUG
		sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_0);
		sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_1);
		sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
		sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);
	#endif

	/* MCU Initialization */
	init_mcu(MCU_PRR_ALL);			// Initialize MCU
	init_uart0(4800, UART_TX_MODE, UART_NO_ISR_ENABLE);		// Initialize UART
	init_spi_master(SPI_CLK_DIV_2);	// Initialize SPI Master
	init_ext_int0(TRUE, EXT_INT_SC_RISING_EDGE);	// Initialize External Int0

    /* ZigBee Hardware Initialization */
	zigbee_init();							// Hardware initialization after power-on event
   	zigbee_clear_irq_status();
	zigbee_set_extended_mode(TRUE);
	zigbee_set_pad_io_clkm(ZIGBEE_PAD_IO_CLKM_2MA);
	zigbee_clear_irq_flag();
	
	#if CPU_FREQUENCY_HZ == CPU_FREQUENCY_1_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_1MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_2_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_2MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_4_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_4MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_8_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_8MHZ);
	#elif CPU_FREQUENCY_HZ == CPU_FREQUENCY_16_MHZ
		zigbee_set_clkm_safe(ZIGBEE_CLKM_16MHZ);
	#endif
	
	sei();

	/* UART0 Init Debug */
	#if UART0_ZIGBEE_SPEC_DEBUG
		uart0_send_byte(UART_TX_ZIGBEE_PART_NUM);
		uart0_send_byte(zigbee_hw_part_num);
		uart0_send_byte(UART_TX_ZIGBEE_VER_NUM);
		uart0_send_byte(zigbee_hw_ver_num);
		uart0_send_byte(UART_TX_ZIGBEE_MAN_ID_0);
		uart0_send_byte(zigbee_hw_man_id_0);
		uart0_send_byte(UART_TX_ZIGBEE_MAN_ID_1);
		uart0_send_byte(zigbee_hw_man_id_1);
	#endif
	
	#if PROTOBOARD_LED_DEBUG
		cbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
	#endif

	while(1) {
		switch(mcu_status) {
			case MCU_STATUS_P_ON:
				mcu_status = MCU_STATUS_IDLE;
				break;
			case MCU_STATUS_IDLE:
				//if (UCSR0A & _BV(RXC0)) { 		// Process any UART data
				//	uart0_rx_parse();
				//}
				mcu_status = MCU_STATUS_RX;
				break;
			case MCU_STATUS_TX_IDLE:
				break;
			case MCU_STATUS_TX:
				switch(zigbee_get_trx_status()) {
					case ZIGBEE_TRX_STATUS_TX_ARET_ON:
						rxtx_msg = zigbee_extended_tx(rxtx_data, rxtx_data_length, tx_on_the_fly_flag);
						zigbee_enable_trx_off();
						mcu_status = MCU_STATUS_IDLE;
						break;
					case ZIGBEE_TRX_STATUS_BUSY_TX_ARET:
						break;
					case ZIGBEE_TRX_STATUS_TRX_OFF:
					case ZIGBEE_TRX_STATUS_PLL_ON:
					case ZIGBEE_TRX_STATUS_RX_ON:
					case ZIGBEE_TRX_STATUS_RX_AACK_ON:
						zigbee_set_tx_src_pan_id(BASE_NODE_PAN_ID);
						zigbee_set_tx_src_short_addr(BASE_NODE_SHORT_ADDR);
						zigbee_set_tx_dest_pan_id(RELAY_NODE_PAN_ID);
						zigbee_set_tx_dest_short_addr(RELAY_NODE_SHORT_ADDR);
						zigbee_enable_extended_tx();
						break;
				}
				break;
			case MCU_STATUS_RX:
				switch(zigbee_get_trx_status()) {
					case ZIGBEE_TRX_STATUS_RX_AACK_NOCLK:
					case ZIGBEE_TRX_STATUS_BUSY_RX_AACK_NOCLK:
					case ZIGBEE_TRX_STATUS_RX_AACK_ON:
					case ZIGBEE_TRX_STATUS_BUSY_RX_AACK:
						if (zigbee_get_irq_flag()) {
							zigbee_clear_irq_flag();
							zigbee_update_irq_status(); 				// TRX_END interrupt
						}
						if (zigbee_get_rx_ready()) {
							zigbee_clear_rx_ready();
							rxtx_msg = zigbee_extended_rx(rxtx_data, &rxtx_data_length);
							#if UART0_ZIGBEE_RX_MAC_DATA_DEBUG
								uart0_send_byte(SENTINEL_UART_TX_ZIGBEE_MAC_DATA);
								uart0_send_byte(rxtx_data_length);
								uart0_send_byte_array(rxtx_data, rxtx_data_length);
							#endif
						}
						break;
					case ZIGBEE_TRX_STATUS_TRX_OFF:
					case ZIGBEE_TRX_STATUS_RX_ON:
					case ZIGBEE_TRX_STATUS_PLL_ON:
					case ZIGBEE_TRX_STATUS_TX_ARET_ON:
						zigbee_set_addr_filter_pan_id(BASE_NODE_PAN_ID);
						zigbee_set_addr_filter_short_addr(BASE_NODE_SHORT_ADDR);	
						zigbee_enable_extended_rx();
						break;
				}
				break;
		}
	}
	
	return 0;
}
