/**
 * A Mega164P Timer Test
 *
 * Ruibing Wang
 */
 
 /**
 * Mega164P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
  
/* Constants for Timer0 */
#define TIMER_RES_US	50  

/* Constants for Timer */
#define TIMER_LATENCY_US			10000
#define TIMER_LATENCY_THRESHOLD		((uint16_t) ((TIMER_LATENCY_US / TIMER_RES_US) + 0.5))

volatile uint16_t timer_tick_count = 0;


void init_timer0(void);

ISR (TIMER0_COMPA_vect) {
	timer_tick_count++;
	
}

/**
 * Main program
 */
int main(void) {
	static uint8_t led_test = 0;
	
	DDRB = 0xFF;
	PORTB = 0x00;
	
	init_timer0();		// Initialize Timer
	
	sei();		// Enable interrupts in assembly
	
	_delay_ms(100);		// Delay to stablize initilization
	
	// Main program loop
	while(1) {
		if (timer_tick_count >= TIMER_LATENCY_THRESHOLD) {
			PORTB = led_test++;
			timer_tick_count = 0;
		}
	}
	
	return 0;
}


void init_timer0(void) {
	// A single cycle = 1/20 MHz = 50 ns
	
	TCCR0A = 0b00000010;// 7:6 - Compare Match Output A Mode
						// 5:4 - Compare Match Output B Mode
						// 3:2 - Reserved Bits
						// 1:0 - Waveform Generation Mode
	
	TCCR0B = 0b00000010;// 7 - Force Output Compare A
						// 6 - Force Output Compare B
						// 5:4 - Reserved Bits
						// 3 - Waveform Generation Mode
						// 2:0 - Clock select (8)
						//		000 - Timer stopped
						//		001 - 1 => 50 ns
						//		010 - 8 => 400 ns
						//		011 - 64 => 3.2 ms
						//		100 - 256 => 12.8 ms
						//		101 - 1024 => 51.2 ms
						//		110 - External clock source on T0 pin, falling edge
						//		111 - ", rising edge
	
	OCR0A = TIMER_RES_US * 2.5;	// OCR0 = timer resolution in us / 400 ns
	
	TIMSK0 = 0b00000010;	// 7:3 - Reserved bits for other timer interrupts
						// 2 - Timer0 Timer/Counter Output Compare Match B Interrupt Enable
						// 1 - Timer0 Timer/Counter Output Compare Match A Interrupt Enable
						// 0 - Timer/Counter0 Overflow Interrupt Enable
}
