/**
 * A Mega16L UART TX Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega16(L) DIP Pinout
 * 					___________
 * PB0 (XCK/T0)		|1		40| PA0 (ADC0)
 * PB1 (T1)			|2		39| PA1 (ADC1)
 * PB2 (INT2/AIN0)	|3		38| PA2 (ADC2)
 * PB3 (OC0/AIN1)	|4		37| PA3 (ADC3)
 * PB4 (SS)			|5		36| PA4 (ADC4)
 * PB5 (MOSI)		|6		35| PA5 (ADC5)
 * PB6 (MISO)		|7		34| PA6 (ADC6)
 * PB7 (SCK)		|8		33| PA7 (ADC7)
 * RESET			|9		32| AREF
 * VCC				|10		31| GND
 * GND				|11		30| AVCC
 * XTAL2			|12		29| PC7 (TOSC2)
 * XTAL1			|13		28| PC6 (TOSC1)
 * PD0 (RXD)		|14		27| PC5 (TDI)
 * PD1 (TXD)		|15		26| PC4 (TDO)
 * PD2 (INT0)		|16		25| PC3 (TMS)
 * PD3 (INT1)		|17		24| PC2 (TCK)
 * PD4 (OC1B)		|18		23| PC1 (SDA)
 * PD5 (OC1A)		|19		22| PC0 (SCL)
 * PD6 (ICP1)		|20		21| PD7 (OC2)
 *					-----------
 */

/* Includes */
#include <mega16.h>
#include <string.h>
#include <stdio.h>
#include <delay.h>

/* Global/System Constants */
#define true 1
#define false 0
#define UDRE 5			

/* Global Variables */
unsigned char counter;

/* Prototype Functions */
void init_uart_tx(void);
void uart_send_byte_array(char *data, unsigned char length);
void uart_send_byte(char data);                    

/**
 * UART transmission function for byte arrays
 *
 * Can be used to transmitt strings
 * Length terminated (versus null terminated)
 */
void uart_send_byte_array(unsigned char *data, unsigned char length) {
	unsigned char counter = 0;

	while (counter++ < length) {
		while ( !( UCSRA & (1<<UDRE)) );	// Wait for empty buffer
		UDR = *data++;						// Put data into buffer, sends the data 
	}
}

/**
 * UART transmission function for a single byte
 */
void uart_send_byte(unsigned char data) {
	while ( !( UCSRA & (1<<UDRE)) );	// Wait for empty buffer
 	UDR = data;							// Put data into buffer, sends the data 
}

/**
 * Main program
 */
void main(void) {
	init_uart_tx();		// Initialize UART TX
	
	delay_ms(100);		// Delay to stablize initilization
	
	counter = 0;
	
	// Main program loop
	while(1) {
		uart_send_byte(counter++);
		delay_ms(100);
	}
}

/**
 * Initialize UART RX
 */
void init_uart_tx(void){
	//UBRRL = 103;	// 4800 bps for 8 MHz Clock - No Errors
	//UBRRL = 51;		// 9600 bps " - Errors
	UBRRL = 25;		// 19200 bps " - No Errors

	UCSRA = 0b00000000;	// 7 - UART RX complete flag
						// 6 - UART TX complete flag
						// 5 - UART DRE flag
						// 4 - Frame error flag
						// 3 - Data overrun flag
						// 2 - Parity error flag
						// 1 - Double UART speed mode
						// 0 - Multi processor communication mode

	UCSRB = 0b00001000; // 7 - RX Complete Interrupt Enable
						// 6 - TX Complete Interrupt Enable
						// 5 - UART DRE Interrupt Enable
						// 4 - Enable receiver
						// 3 - Enable transmitter
						// 2 - Character size
						// 1:0 - Ninth data bit

	UCSRC = 0b10000110;	// 7 - Register Select
						// 6 - USART Mode Select
						//		0 - Asynchronous Operation
						//		1 - Synchronous Operation
						// 5:4 - Parity mode
						//		00 - Disabled
						//		01 - Reserved
						//		10 - Enabled, Even
						//		11 - Enabled, Odd
						// 3 - Select stop bit length
						//		0 - 1-bit
						//		1 - 2-bit
						// 2:1 - Character size, with UCSRB.2
						//		000 - 5 bit
						//		001 - 6 bit
						//		010 - 7 bit
						//		011 - 8 bit
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Reserved
						//		111 - 9 bit
						// 0 - Select clock polarity
						//		0 - Rising XCK edge
						//		1 - Falling XCK edge 						
}