/**
 * A Mega324P Input Debouncer Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */

 /**
 * Mega324P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */

/* Includes */
#include <mega324.h>
#include <string.h>
#include <stdio.h>
#include <delay.h>

/* Global/System Constants */
#define TRUE 	1
#define FALSE 	0
#define CPU_FREQUENCY_HZ	20000000 

/* Constants for Timer0 */
#define TIMER_RES_US	50

/* Constants for Input Debouncer */
#define DEBOUNCE_LATENCY_US	10000
#define DEBOUNCE_LATENCY_THRESHOLD	(unsigned int) ((DEBOUNCE_LATENCY_US / TIMER_RES_US) + 0.5)  

/* Global Variables */
unsigned int debounce_latency_tick_count;
unsigned char counter;

enum{DEBOUNCE_DEPUSHED, DEBOUNCE_MAYBE_PUSHED, DEBOUNCE_PUSHED, DEBOUNCE_MAYBE_DEPUSHED};
unsigned char debounce_state;

/* Prototype Functions */
void init_timer0(void);

/**
 * Triggered every tick (depends on tick_res_us)
 */
interrupt [TIM0_COMPA] void timer0_comp(void) {
	debounce_latency_tick_count++;	
}  

void main(void) {
	init_timer0();
	
	DDRB = 0xFF;
	PORTB = 0xFF;
	
	DDRA = 0x00;
	
	#asm("sei");
    
    counter = 0;
	debounce_latency_tick_count = 0;
	debounce_state = DEBOUNCE_DEPUSHED;
	
	while(1) {
		PORTB = counter;
	
    	if (debounce_latency_tick_count > DEBOUNCE_LATENCY_THRESHOLD) {
    		switch(debounce_state) {
    			case DEBOUNCE_DEPUSHED:
    				if (PINA != 0xFF) {
    					debounce_state = DEBOUNCE_MAYBE_PUSHED;
    				}
    				break;
    			case DEBOUNCE_MAYBE_PUSHED:
    			    if (PINA != 0xFF) {
    			    	debounce_state = DEBOUNCE_PUSHED;
    			    	counter++;
    			    }
    			    else {
    			    	debounce_state = DEBOUNCE_DEPUSHED;
    			    }
    				break;
    			case DEBOUNCE_PUSHED:
    				if (PINA == 0xFF) {
    			    	debounce_state = DEBOUNCE_MAYBE_DEPUSHED;
    			    }
    				break;
    			case DEBOUNCE_MAYBE_DEPUSHED:
    				if (PINA == 0xFF) {
    			    	debounce_state = DEBOUNCE_DEPUSHED;
    			    }
    			    else {
    			    	debounce_state = DEBOUNCE_PUSHED;
    			    }
    				break;
    		}
    		debounce_latency_tick_count = 0;
    	}
	}
}

/**
 * Enable timer0 to act as a timer for measuring the time
 * between rising and falling and then rising to rising.
 */
void init_timer0(void) {
	// A single cycle = 1/20 MHz = 50 ns

	TCCR0A = 0b00000010;	// 7:6 - Compare Match Output A Mode
							// 5:4 - " B "
							// 3:2 - Reserved
							// 1:0 - Waveform Generation Mode
							//	TCCR0B.3|TCCR0A.1|TCCR0A.0
							//	000 - Normal
							//	010 - CTC
	 
	TCCR0B = 0b00000010;	// 7 - Force Output Compare A
							// 6 - " B
							// 5:4 - Reserved bits
							// 3 - Waveform Generation Mode (see above)
							// 2:0 - Clock select
							//	000 - Timer stopped
							//	001 - 1
							//	010 - 8
							//	011 - 64
							//	100 - 256
							//	101 - 1024
							//	110 - Ext clock source at T0 pin, falling
							//	111 - ", rising
							
	OCR0A = TIMER_RES_US * 2.5;	// OCR0A = timer resolution in us / 400 ns 
								// => timer resolution in us * 2.5
	
	TIMSK0 = 0b00000010;	// 7:3 - Reserved bits for other timer interrupts
							// 2 - Timer0 output compare match B interrupt
							// 1 - " A "
							// 0 - Timer0 overflow interrupt
}	
