/**
 * A Mega324P PPM Capturer Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega324P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */

/* Includes */
#include <math.h>
#include <mega324.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <delay.h>

/* Global/System Constants */
#define true 1
#define false 0
#define UDRE0 5

/* Type conversions to and from 8:8 fixed point representation macros */
#define int2fix(a)   (((int)(a))<<8)            //Convert char to fix. a is a char
#define fix2int(a)   ((signed char)((a)>>8))    //Convert fix to char. a is an int
#define float2fix(a) ((int)((a)*256.0))         //Convert float to fix. a is a float
#define fix2float(a) ((float)(a)/256.0)         //Convert fix to float. a is an int 

/* Program Constants */ 
#define timer_res_us 10			// Timer resolution
								// Higher means lower cpu but lower resolution
//#define N 1					// Clock Divider (See initialization for details)
#define uart_latency_us 50000	// Delay between each UART transmission of data
#define uart_latency_tick uart_latency_us/timer_res_us	// Number of timer ticks for latency 				
 
#pragma regalloc-
int alpha, yn_1;
#pragma regalloc+

/* Global Variables */
unsigned char avg_pulse_period, pulse_period, pulse_tick_count;
unsigned int uart_latency_tick_count;
unsigned char data_ready_flag;   

/* Prototype Functions */
void init_pwm_cap(void);
void init_timer(void);
void init_uart0_tx(void);
void uart_send_byte_array(char *data, unsigned char length);
void uart_send_byte(char data);

/**
 * Triggered every tick (depends on tick_res_us)
 */
interrupt [TIM0_COMPA] void timer0_comp(void) {
	pulse_tick_count++;		// Increment a counter that keeps track
							// of how many ticks before the next rising edge
	uart_latency_tick_count++;
}

/**
 * Triggered on rising edge
 */
interrupt [INT1] void external_interrupt1(void) {
	pulse_period = pulse_tick_count;	// Pulse period is equal to the current
										// number of pulse ticks
	
	pulse_tick_count = 	0;				// Reset the tick counter for pulses
	data_ready_flag = 1;			// Raise edge detection flag to signal data ready
}

int lowpass_filter(int xx) {
    #asm
    .macro mult_acc         ;r31:r30:r24 += r23:r22 * r21:r20
        muls r23, r21		; (signed)ah * (signed)bh
    	add	r31, r0
    	mul	r22, r20		; al * bl
    	add	r24, r0
    	adc	r30, r1
    	adc	r31, r27
    	mulsu r23, r20		; (signed)ah * bl
    	add	r30, r0
    	adc	r31, r1
    	mulsu	r21, r22	; (signed)bh * al
    	add	r30, r0
    	adc	r31, r1
    .endm 
    
    push r20    ;save parameter regs
    push r21
    
	clr r27             ;permanent zero
    clr r24             ;clear 24 bit result reg; msb to lsb => r31:r30:r24
    
    ld   R30, Y         ;load input parameter xx from stack
	ldd  R31, Y+1       ; and put in result location
	
    lds  R20, _yn_1     ;load y(n-1) from  RAM
	lds  R21, _yn_1+1
	sub  r20, r30       ;form y(n-1) - xx
	sbc  r21, r31       
	lds  R22, _alpha    ;load alpha from RAM
	lds  R23, _alpha+1 
    mult_acc            ; alpha*(yn_1 - xx)) + xx
    
	sts  _yn_1, r30     ;store new output as y(n-1) to  RAM
	sts  _yn_1+1, r31 
	
    pop r21             ;restore parameter regs
	pop r20
	
	#endasm 
}

/**
 * UART transmission function for byte arrays
 *
 * Can be used to transmitt strings
 * Length terminated (versus null terminated)
 */
void uart_send_byte_array(unsigned char *data, unsigned char length) {
	unsigned char counter = 0;

	while (counter++ < length) {
		while ( !( UCSR0A & (1<<UDRE0)) );	// Wait for empty buffer
		UDR0 = *data++;						// Put data into buffer, sends the data 
	}
}

/**
 * UART transmission function for a single byte
 */
void uart_send_byte(unsigned char data) {
	while ( !( UCSR0A & (1<<UDRE0)) );	// Wait for empty buffer
 	UDR0 = data;							// Put data into buffer, sends the data 
}

/**
 * Main program
 */
void main(void) {
	init_pwm_cap();		// Initialize PWM capture
	init_timer();		// Initialize Timer
	init_uart0_tx();		// Initialize UART TX
	
	#asm("sei");		// Enable interrupts in assembly
	
	//1-pole RC lowpass, cutoff=0.0424 (1/4 Nyquist freq), gain=1 
  	alpha = float2fix(0.717);
	
	delay_ms(100);		// Delay to stablize initilization
	
	// Main program loop
	while(1) {
		// Send out data through the UART when the data and uart are ready
		if (data_ready_flag) {
			avg_pulse_period = fix2int(lowpass_filter(int2fix(pulse_period)));
			data_ready_flag = 0;
		}
		else if (uart_latency_tick_count > uart_latency_tick) {
			printf("%d ",avg_pulse_period);
			uart_latency_tick_count = 0;
		}
		else if (pulse_tick_count == 255) {
			pulse_tick_count = 0;
			avg_pulse_period = fix2int(lowpass_filter(0));
		}
	}
}

/**
 * Enable external interrupts INT1 (rising)
 * to time the periods of the pwm pulse and cycle.
 */
void init_pwm_cap(void) {
	EICRA = 0b00001100;	// 7:6 - Reserved bits
						// 5:0 - ISC2x, ISC1x, ISC0x
						//		00 - Low level
						//		01 - Any logical change
						//		10 - Falling edge
						//		11 - Rising edge
						
	EIMSK = 0b00000010;	// 7:3 - Reserved
						// 2 - INT2 interrupt request (Enabled)
						// 1 - INT1 " (Disabled)
						// 0 - INT0 " (Disabled)
							
	pulse_period = 0;	// Initialize variable
}

/**
 * Enable timer0 to act as a timer for measuring the time
 * between rising and falling and then rising to rising.
 */
void init_timer(void) {
	// A single cycle = 1/20 MHz = 50 ns

	TCCR0A = 0b00000010;	// 7:6 - Compare Match Output A Mode
							// 5:4 - " B "
							// 3:2 - Reserved
							// 1:0 - Waveform Generation Mode
							//	TCCR0B.3|TCCR0A.1|TCCR0A.0
							//	000 - Normal
							//	010 - CTC
	 
	TCCR0B = 0b00000010;	// 7 - Force Output Compare A
							// 6 - " B
							// 5:4 - Reserved bits
							// 3 - Waveform Generation Mode (see above)
							// 2:0 - Clock select
							//	000 - Timer stopped
							//	001 - 1
							//	010 - 8
							//	011 - 64
							//	100 - 256
							//	101 - 1024
							//	110 - Ext clock source at T0 pin, falling
							//	111 - ", rising
							
	OCR0A = timer_res_us * 2.5;	// OCR0A = timer resolution in us / 400 ns
								// => timer resolution in us * 2.5
	
	TIMSK0 = 0b00000010;	// 7:3 - Reserved bits for other timer interrupts
							// 2 - Timer0 output compare match B interrupt
							// 1 - " A "
							// 0 - Timer0 overflow interrupt
	
	pulse_tick_count = 0;	// Initialize variable
}


/**
 * Initialize UART RX
 */
void init_uart0_tx(void){
	UCSR0A = 0b00000000;// 7 - UART RX complete flag
						// 6 - UART TX complete flag
						// 5 - UART DRE flag
						// 4 - Frame error flag
						// 3 - Data overrun flag
						// 2 - Parity error flag
						// 1 - Double UART speed mode
						// 0 - Multi processor communication mode
						
	UCSR0B = 0b00001000;// 7 - RX Complete Interrupt
						// 6 - TX Complete Interrupt
						// 5 - UART DRE Interrupt
						// 4 - Enable receiver
						// 3 - Enable transmitter
						// 2 - Character size
						// 1:0 - Ninth data bit

	UCSR0C = 0b00000110;// 7:6 - UART Mode Select
						//	00 - Asynchronous UART
						//	01 - Synchronous UART
						//	10 - Reserved
						//	11 - Master SPI
						// 5:4 - Parity Mode
						//	00 - Disabled
						//	01 - Reserved
						//	10 - Enabled, Even
						//	11 - Enabled, Odd
						// 3 - Select stop bit length
						//	0 - 1-bit
						//	1 - 2-bit
						// 2:1 - Character size
						//	000 - 5 bit
						//	001 - 6 bit
						//	010 - 7 bit
						//	011 - 8 bit
						//	100 - Reserved
						//	101 - Reserved
						//	110 - Reserved
						//	111 - 9 bit
						// 0 - Select clock polarity
						//	0 - Rising XCK edge
						//	1 - Falling XCK edge
	
				
	UBRR0L = 129;		// UART Baud Rate
						// At 20 MHz
						//	2400 bps => 520
						//	4800 bps => 259
						//	9600 bps => 129
						//	14.4 kbps => 86
						//	19.2 kbps => 64		
}
						
