/**
 * A Mega324P PWM Fan Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */

 /**
 * Mega324P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */

/* Includes */
#include <mega324.h>
#include <string.h>
#include <stdio.h>
#include <delay.h>

/* Global/System Constants */
#define TRUE 	1
#define FALSE 	0
#define CPU_FREQUENCY_HZ	20000000	// CPU Clock Frequency (in Hz)

/* Constants for Timer0 */
#define TIMER_RES_US	50

/* Constants for PWM Fan */
#define PWM_FAN_FREQUENCY_HZ	38.147	// PWM Frequency (in Hz) = CPU_FREQUENCY/(N*(TOP+1))
										//		With CPU Frequency = 20 MHz 
										//		With N=1 => 305.176 Hz to 20 MHz
										// 		With N=8 => 38.147 Hz to 2.5 MHz
										// 		With N=64 => 4.768 Hz to 312.5 KHz
										// 		With N=256 => 1.192 Hz to 78.125 KHz
										// 		With N=1024 => 0.298 Hz to 19.531 KHz
#define CLOCK_DIVIDER 			8		// Clock divider (Lower equals higher resolution)
#define MIN_DUTY_CYCLE			0.25
#define MAX_DUTY_CYCLE			1.00  
#define	PWM_FAN_LATENCY_US		1000000
#define PWM_FAN_LATENCY_THRESHOLD	(unsigned long) ((PWM_FAN_LATENCY_US / TIMER_RES_US) + 0.5)

/* Global Variables */
unsigned int pulse_width, min_pulse_width, max_pulse_width;

/* Prototype Functions */
void init_pwm_fan(void);

void main(void) {
	init_pwm_fan();
	
	#asm("sei");
	
	while(1) {
	}
}

/**
 * Auto calculations
 *
 * Initialize OC1B and OC1A as the fast PWM output
 * from Timer 1 with a base freq set by ICR1 and pulse width set by OCR1A and OCR1B.
 *
 * OC1B currently DISABLED, OC1A currently NON-INVERTING
 */
void init_pwm_fan(void) {
	float cpu_pwm_ratio;			// Ratio of CPU clk frequency to PWM frequency
    unsigned int TOP;
	
	//DDRD.4 = 1;		// Set PORTD.4 as an output for OC1B
	DDRD.5 = 1;		// Set PORTD.5 as an output for OC1A 
	
	/*
	 * Sets the register for the correct clock divider value
	 *	000 - Timer/Counter stopped
	 *	001 - 1            
	 *	010 - 8
	 *	011 - 64
	 *	100 - 256
	 *	101 - 1024
	 *	110 - External clock source on T1 pin, falling edge
	 *	111 - ", rising edge
	 */
	switch (CLOCK_DIVIDER) {
		case 1:
			TCCR1B |= 1;
			break;
		case 8:
			TCCR1B |= 2;
			break;
		case 64:
			TCCR1B |= 3;
			break;
		case 256:
			TCCR1B |= 4;
			break;
		case 1024:
			TCCR1B |= 5;
			break;
		default:
			TCCR1B |= 1;
	}
	
	// cpu_pwm_ratio = clk_freq/pwm_freq = N*(TOP+1) = clk_freq/(pwm_freq*N) = TOP+1
	cpu_pwm_ratio = (float) CPU_FREQUENCY_HZ / ((float) PWM_FAN_FREQUENCY_HZ * (float) CLOCK_DIVIDER) - 1;
    
	TOP = (unsigned int) (cpu_pwm_ratio + 0.5);
	ICR1H = TOP >> 8;			// High byte
	ICR1L = (TOP << 8) >> 8;	// Low byte
	
	min_pulse_width = (unsigned int) ((float)TOP * (float)MIN_DUTY_CYCLE + 0.5);
	max_pulse_width = (unsigned int) ((float)TOP * (float)MAX_DUTY_CYCLE + 0.5);
	
	pulse_width = min_pulse_width;
	OCR1AH = pulse_width >> 8;
	OCR1AL = (pulse_width << 8) >> 8;
	//OCR1BH = OCR1AH;
	//OCR1BL = OCR1AL;
	
	TCCR1A = 0b10000010;	// 7:6 - OC1A's compare output mode
							//	=> 00 - Disabled (Normal Port)
							//	=> 01 - Toggle (Normal Port)
							//	=> 10 - Non-inverting
							//	=> 11 - Inverting
							// 5:4 - OC1B's "
							//	=> "
							// 3 - Reserved
							// 2 - Reserved
							// 1:0 - Waveform generation mode (Fast PWM, TOP=ICR1)
	
	TCCR1B |= 0 << 7;	// Input capture noise canceler (Disabled)
	TCCR1B |= 0 << 6;	// Input capture edge select (Disabled)
	TCCR1B |= 0 << 5;	// Reserved bit
	TCCR1B |= 1 << 4;	// Waveform generation mode (")
	TCCR1B |= 1 << 3;
	
	TIMSK1 = 0;	// Timer/Counter interrupt mask (All disabled)
				// 5 - Timer1, input capture interrupt
				// 4 - Timer1, output compare A match interrupt
				// 3 - Timer1, " B "
				// 2 - Timer1, overflow interrupt
}	
