/**
 * A Mega324P Temperature Controlled PWM Fan Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */

 /**
 * Mega324P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */

/* Includes */
#include <mega324.h>
#include <string.h>
#include <stdio.h>
#include <delay.h>

/* Global/System Constants */
#define TRUE 	1
#define FALSE 	0
#define CPU_FREQUENCY_HZ	20000000	// CPU Clock Frequency (in Hz)

/* Constants for Timer0 */
#define TIMER_RES_US	50

/* Constants for PWM Fan */
#define PWM_FAN_FREQUENCY_HZ	38.147	// PWM Frequency (in Hz) = CPU_FREQUENCY/(N*(TOP+1))
										//		With CPU Frequency = 20 MHz 
										//		With N=1 => 305.176 Hz to 20 MHz
										// 		With N=8 => 38.147 Hz to 2.5 MHz
										// 		With N=64 => 4.768 Hz to 312.5 KHz
										// 		With N=256 => 1.192 Hz to 78.125 KHz
										// 		With N=1024 => 0.298 Hz to 19.531 KHz
#define CLOCK_DIVIDER 			8		// Clock divider (Lower equals higher resolution)
#define MIN_DUTY_CYCLE			0.25
#define MAX_DUTY_CYCLE			1.00 
#define DUTY_CYCLE_STEP_SIZE	0.05 
#define	PWM_FAN_LATENCY_US		1000000
#define PWM_FAN_LATENCY_THRESHOLD	(unsigned long) ((PWM_FAN_LATENCY_US / TIMER_RES_US) + 0.5)

/* Constants for Temperature Sensor */
#define ADC_REFERENCE_VOLTAGE		5
#define ADC_TEMP_OP_AMP_SCALING		2
#define ADC_TEMP_SENSOR_SCALE_MV	10
#define ADC_TEMP_SOFT_GAIN 			1
#define ADC_TEMP_TARGET_DEGREE_F	75

/* Global Variables */
unsigned int pulse_width, min_pulse_width, max_pulse_width, pulse_width_step_size;
unsigned long pwm_fan_latency_tick_count;
unsigned char adc_complete_flag;
unsigned char adc_data, adc_temp_degree;

/* Prototype Functions */
void init_mcu(void);
void init_adc(void);
void init_pwm_fan(void);
void init_timer0(void);

/**
 * Triggered every tick (depends on tick_res_us)
 */
interrupt [TIM0_COMPA] void timer0_comp(void) {
	pwm_fan_latency_tick_count++;	
}

/**
 * ADC ISR
 * 
 * Right Adjusted: ADCH.1|ADCH.0|ADCL
 * Left Adjusted: ADCH|ADCL.7|ADCL.6
 * When reading 16-bit values (high res), the low byte must be read first
 */
interrupt [ADC_INT] void adc_done(void){
	adc_complete_flag = 1; 
}

void main(void) {
   	float voltage_per_degree_mv, value_per_degree;	// Conversion floats
    voltage_per_degree_mv = (float)ADC_TEMP_SENSOR_SCALE_MV*(float)ADC_TEMP_OP_AMP_SCALING*(float)ADC_TEMP_SOFT_GAIN;
	value_per_degree = voltage_per_degree_mv*(float)256/((float)ADC_REFERENCE_VOLTAGE*(float)1000); 
	
	DDRB = 0xFF;
	PORTB = 0xFF;
	
	init_mcu();			// Initialize MCU features
	init_adc();			// Initialize ADC
	init_timer0();
	init_pwm_fan();

	#asm("sei");		// Enable interrupts

	delay_ms(100);		// Delay for 100 ms before entering while loop
	
	while(1) {
		// Sends out the temperature in degrees whenever adc conversion finishes
		
		if (pwm_fan_latency_tick_count > PWM_FAN_LATENCY_THRESHOLD) { 
			if (adc_complete_flag) {
			    adc_data = ADCH;
			    adc_temp_degree = (unsigned char) ((float) adc_data / value_per_degree);
				PORTB = ~adc_temp_degree;
				
				if (adc_temp_degree > ADC_TEMP_TARGET_DEGREE_F) {
					pulse_width = max_pulse_width;
					OCR1AH = pulse_width >> 8;
					OCR1AL = (pulse_width << 8) >> 8;
				}
				else {
					pulse_width = min_pulse_width;
					OCR1AH = pulse_width >> 8;
					OCR1AL = (pulse_width << 8) >> 8;
				}
				adc_complete_flag = 0;
			}
			pwm_fan_latency_tick_count = 0;
		}
	}
}

void init_mcu(void) {
	SMCR = 0b00000000;	// 7:4 - Reserved bits
						// 3:1 - Sleep Mode
						//		000 - Idle
						//		001 - ADC Noise Reduction
						//		010 - Power-down
						//		011 - Power-save
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Standby
						//		111 - Extended Standby
						// 0 - Sleep Enable
						
	MCUCR |= (0 << 6);	// 6 - BOD Sleep
	MCUCR |= (0 << 5);	// 5 - BOD Sleep Enable
	
	PRR = 0b11010110;	// 7 - Power Reduction TWI
						// 6 - Power Reduction Timer2
						// 5 - Power Reduction Timer0
						// 4 - Power Reduction UART1
						// 3 - Power Reduction Timer1
						// 2 - Power Reduction SPI
						// 1 - Power Reduction UART0
						// 0 - Power Reduction ADC
}

void init_adc(void) {
	DDRA.0 = 0;		// PORTA.0 is an input for ADC0
    
	ADMUX = 0b00100000;	// 7:6 - Reference Selection Bits
						//		00 - AREF, internal Vref off
						//		01 - AVCC with external capacitor at AREF pin
						//		10 - Internal 1.1 V Ref with ext cap at AREF
						//		11 - Internal 2.56 V Ref "
						// 5 - ADC Left Adjust Result
						// 4:0 - Analog Channel and Gain Selection Bits
						//		00000 - Single Ended, ADC0
						//		00001 - ", ADC1
						//		00010 - ", ADC2
						//		00011 - ", ADC3
						//		00100 - ", ADC4
						//		00101 - ", ADC5
						//		00110 - ", ADC6
						//		00111 - ", ADC7
						//		01000 - Positive Diferential, ADC0, ADC0, 10x
						//		01001 - ", ADC1, ADC0, 10x
						//		01010 - ", ADC0, ADC0, 200x
						//		01011 - ", ADC1, ADC0, 200x
						//		01100 - ", ADC2, ADC2, 10x
						//		01101 - ", ADC3, ADC2, 10x
						//		01110 - ", ADC2, ADC2, 200x
						//		01111 - ", ADC3, ADC2, 200x
						//		10000 - ", ADC0, ADC1, 1x
						//		10001 - ", ADC1, ADC1, 1x
						//		10010 - ", ADC2, ADC1, 1x
						//		10011 - ", ADC3, ADC1, 1x
						//		10100 - ", ADC4, ADC1, 1x
						//		10101 - ", ADC5, ADC1, 1x
						//		10110 - ", ADC6, ADC1, 1x
						//		10111 - ", ADC7, ADC1, 1x
						//		11000 - ", ADC0, ADC2, 1x
						//		11001 - ", ADC1, ADC2, 1x
						//		11010 - ", ADC2, ADC2, 1x
						//		11011 - ", ADC3, ADC2, 1x
						//		11100 - ", ADC4, ADC2, 1x
						//		11101 - ", ADC5, ADC2, 1x
						//		11110 - Singled Ended Input, 1.1 V (VBG)
						//		11111 - ", 0 V (GND)
	
	ADCSRA = 0b11101111;// 7 - ADC Enable
						// 6 - ADC Start Conversion
						//	=> Single Conversion => Start each conversion
						//	=> Free Running => Start first conversion
						// 5 - ADC Auto Trigger Enable
						// 4 - ADC Interrupt Flag
						// 3 - ADC Interrupt Enable
						// 2:0 - ADC Prescaler Select Bits
						//		000 - 2 => 62.5 ns * 2 = 125 ns
						//		001 - 2 => 125 ns
						//		010 - 4 => 250 ns
						//		011 - 8 => 500 ns
						//		100 - 16 => 1 us
						//		101 - 32 => 2 us
						//		110 - 64 => 4 us
						//		111 - 128 => 8 us
	
	ADCSRB = 0b00000000;// 7:3 - Reserved Bits (6 is used elsewhere)
						// 2:0 - ADC Auto Trigger Source
						//		000 - Free Running Mode
						//		001 - Analog Comparator
						//		010 - External Interrupt Request 0
						//		011 - Timer0 Compare Match
						//		100 - Timer0 Overflow
						//		101 - Timer1 Compare Match B
						//		110 - Timer1 Overlow
						//		111 - Timer1 Capture Event
						
	DIDR0 = 0b00000001;	// 7:0 - ADC7:0 Digital Input Disable
						//	=> Disable digital input buffer to reduce power consumption 
}

/**
 * Enable timer0 to act as a timer for measuring the time
 * between rising and falling and then rising to rising.
 */
void init_timer0(void) {
	// A single cycle = 1/20 MHz = 50 ns

	TCCR0A = 0b00000010;	// 7:6 - Compare Match Output A Mode
							// 5:4 - " B "
							// 3:2 - Reserved
							// 1:0 - Waveform Generation Mode
							//	TCCR0B.3|TCCR0A.1|TCCR0A.0
							//	000 - Normal
							//	010 - CTC
	 
	TCCR0B = 0b00000010;	// 7 - Force Output Compare A
							// 6 - " B
							// 5:4 - Reserved bits
							// 3 - Waveform Generation Mode (see above)
							// 2:0 - Clock select
							//	000 - Timer stopped
							//	001 - 1
							//	010 - 8
							//	011 - 64
							//	100 - 256
							//	101 - 1024
							//	110 - Ext clock source at T0 pin, falling
							//	111 - ", rising
							
	OCR0A = TIMER_RES_US * 2.5;	// OCR0A = timer resolution in us / 400 ns
								// => timer resolution in us * 2.5
	
	TIMSK0 = 0b00000010;	// 7:3 - Reserved bits for other timer interrupts
							// 2 - Timer0 output compare match B interrupt
							// 1 - " A "
							// 0 - Timer0 overflow interrupt
}


/**
 * Auto calculations
 *
 * Initialize OC1B and OC1A as the fast PWM output
 * from Timer 1 with a base freq set by ICR1 and pulse width set by OCR1A and OCR1B.
 *
 * OC1B currently DISABLED, OC1A currently NON-INVERTING
 */
void init_pwm_fan(void) {
	float cpu_pwm_ratio;			// Ratio of CPU clk frequency to PWM frequency
    unsigned int TOP;
	
	//DDRD.4 = 1;		// Set PORTD.4 as an output for OC1B
	DDRD.5 = 1;		// Set PORTD.5 as an output for OC1A 
	
	/*
	 * Sets the register for the correct clock divider value
	 *	000 - Timer/Counter stopped
	 *	001 - 1            
	 *	010 - 8
	 *	011 - 64
	 *	100 - 256
	 *	101 - 1024
	 *	110 - External clock source on T1 pin, falling edge
	 *	111 - ", rising edge
	 */
	switch (CLOCK_DIVIDER) {
		case 1:
			TCCR1B |= 1;
			break;
		case 8:
			TCCR1B |= 2;
			break;
		case 64:
			TCCR1B |= 3;
			break;
		case 256:
			TCCR1B |= 4;
			break;
		case 1024:
			TCCR1B |= 5;
			break;
		default:
			TCCR1B |= 1;
	}
	
	// cpu_pwm_ratio = clk_freq/pwm_freq = N*(TOP+1) = clk_freq/(pwm_freq*N) = TOP+1
	cpu_pwm_ratio = (float) CPU_FREQUENCY_HZ / ((float) PWM_FAN_FREQUENCY_HZ * (float) CLOCK_DIVIDER) - 1;
    
	TOP = (unsigned int) (cpu_pwm_ratio + 0.5);
	ICR1H = TOP >> 8;			// High byte
	ICR1L = (TOP << 8) >> 8;	// Low byte
	
	min_pulse_width = (unsigned int) ((float)TOP * (float)MIN_DUTY_CYCLE + 0.5);
	max_pulse_width = (unsigned int) ((float)TOP * (float)MAX_DUTY_CYCLE + 0.5);
	pulse_width_step_size = (unsigned int) ((float)TOP * (float)DUTY_CYCLE_STEP_SIZE + 0.5);
	
	pulse_width = min_pulse_width;
	OCR1AH = pulse_width >> 8;
	OCR1AL = (pulse_width << 8) >> 8;
	//OCR1BH = OCR1AH;
	//OCR1BL = OCR1AL;
	
	TCCR1A = 0b10000010;	// 7:6 - OC1A's compare output mode
							//	=> 00 - Disabled (Normal Port)
							//	=> 01 - Toggle (Normal Port)
							//	=> 10 - Non-inverting
							//	=> 11 - Inverting
							// 5:4 - OC1B's "
							//	=> "
							// 3 - Reserved
							// 2 - Reserved
							// 1:0 - Waveform generation mode (Fast PWM, TOP=ICR1)
	
	TCCR1B |= 0 << 7;	// Input capture noise canceler (Disabled)
	TCCR1B |= 0 << 6;	// Input capture edge select (Disabled)
	TCCR1B |= 0 << 5;	// Reserved bit
	TCCR1B |= 1 << 4;	// Waveform generation mode (")
	TCCR1B |= 1 << 3;
	
	TIMSK1 = 0;	// Timer/Counter interrupt mask (All disabled)
				// 5 - Timer1, input capture interrupt
				// 4 - Timer1, output compare A match interrupt
				// 3 - Timer1, " B "
				// 2 - Timer1, overflow interrupt
}	
