/**
 * A Mega324P Temperature Sensor Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega324P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */

/* Includes */
#include <mega324.h>
#include <string.h>
#include <stdio.h>
#include <delay.h>

/* Global/System constants */
#define true 1
#define false 0
#define UDRE 5

/* Type conversions to and from 8:8 fixed point representation macros */
#define int2fix(a)	(((int)(a))<<8)            //Convert char to fix. a is a char
#define fix2int(a)	((signed char)((a)>>8))    //Convert fix to char. a is an int
#define float2fix(a)	((int)((a)*256.0))         //Convert float to fix. a is a float
#define fix2float(a)	((float)(a)/256.0)         //Convert fix to float. a is an int 

/* Multiplication and divisions for fixed point macros, twice as fast as floating mul/div */
#define multfix(a,b)	((int)((((long)(a))*((long)(b)))>>8)) //multiply two fixed #
#define divfix(a,b)  ((int)((((long)(a))<<8)/((long)(b)))) //divide two fixed #

/* User Configs */
#define reference_voltage		5
#define amplifier_factor		2
#define linear_scale_factor_mv	10
#define adc_gain				1

unsigned char adc_complete_flag;
unsigned char adc_data, temperature_degree;

void init_adc(void);
void init_mcu(void);
void init_uart_tx(void);  

/**
 * ADC ISR
 * 
 * Right Adjusted: ADCH.1|ADCH.0|ADCL
 * Left Adjusted: ADCH|ADCL.7|ADCL.6
 * When reading 16-bit values (high res), the low byte must be read first
 */
interrupt [ADC_INT] void adc_done(void){
	adc_complete_flag = 1; 
}

/**
 * UART transmission function for byte arrays
 *
 * Can be used to transmitt strings
 * Length terminated (versus null terminated)
 */
void uart_send_byte_array(unsigned char *data, unsigned char length) {
	unsigned char counter = 0;

	while (counter++ < length) {
		while ( !( UCSR0A & (1<<UDRE)) );	// Wait for empty buffer
		UDR0 = *data++;						// Put data into buffer, sends the data 
	}
}

/**
 * UART transmission function for a single byte
 */
void uart_send_byte(unsigned char data) {
	while ( !( UCSR0A & (1<<UDRE)) );	// Wait for empty buffer
 	UDR0 = data;							// Put data into buffer, sends the data 
}

/**
 * Main program loop
 */
void main(void) {
   	float voltage_per_degree_mv, value_per_degree;	// Conversion floats
    voltage_per_degree_mv = (float)linear_scale_factor_mv*(float)amplifier_factor*(float)adc_gain;
	value_per_degree = voltage_per_degree_mv*(float)256/((float)reference_voltage*(float)1000); 
	
	init_mcu();			// Initialize MCU features
	init_adc();			// Initialize ADC
	init_uart_tx();		// Initialize UART TX
	
	#asm("sei");		// Enable interrupts

	delay_ms(100);		// Delay for 100 ms before entering while loop
	
	while(1) {
		// Sends out the temperature in degrees whenever adc conversion finishes
		if (adc_complete_flag) {
		    adc_data = ADCH;
		    temperature_degree = (unsigned char) ((float) adc_data / value_per_degree);
		    uart_send_byte(temperature_degree);
			delay_ms(100);
		}
	}
}

void init_adc(void) {
	DDRA.0 = 0;		// PORTA.0 is an input for ADC0
    
	ADMUX = 0b00100000;	// 7:6 - Reference Selection Bits
						//		00 - AREF, internal Vref off
						//		01 - AVCC with external capacitor at AREF pin
						//		10 - Internal 1.1 V Ref with ext cap at AREF
						//		11 - Internal 2.56 V Ref "
						// 5 - ADC Left Adjust Result
						// 4:0 - Analog Channel and Gain Selection Bits
						//		00000 - Single Ended, ADC0
						//		00001 - ", ADC1
						//		00010 - ", ADC2
						//		00011 - ", ADC3
						//		00100 - ", ADC4
						//		00101 - ", ADC5
						//		00110 - ", ADC6
						//		00111 - ", ADC7
						//		01000 - Positive Diferential, ADC0, ADC0, 10x
						//		01001 - ", ADC1, ADC0, 10x
						//		01010 - ", ADC0, ADC0, 200x
						//		01011 - ", ADC1, ADC0, 200x
						//		01100 - ", ADC2, ADC2, 10x
						//		01101 - ", ADC3, ADC2, 10x
						//		01110 - ", ADC2, ADC2, 200x
						//		01111 - ", ADC3, ADC2, 200x
						//		10000 - ", ADC0, ADC1, 1x
						//		10001 - ", ADC1, ADC1, 1x
						//		10010 - ", ADC2, ADC1, 1x
						//		10011 - ", ADC3, ADC1, 1x
						//		10100 - ", ADC4, ADC1, 1x
						//		10101 - ", ADC5, ADC1, 1x
						//		10110 - ", ADC6, ADC1, 1x
						//		10111 - ", ADC7, ADC1, 1x
						//		11000 - ", ADC0, ADC2, 1x
						//		11001 - ", ADC1, ADC2, 1x
						//		11010 - ", ADC2, ADC2, 1x
						//		11011 - ", ADC3, ADC2, 1x
						//		11100 - ", ADC4, ADC2, 1x
						//		11101 - ", ADC5, ADC2, 1x
						//		11110 - Singled Ended Input, 1.1 V (VBG)
						//		11111 - ", 0 V (GND)
	
	ADCSRA = 0b11101111;// 7 - ADC Enable
						// 6 - ADC Start Conversion
						//	=> Single Conversion => Start each conversion
						//	=> Free Running => Start first conversion
						// 5 - ADC Auto Trigger Enable
						// 4 - ADC Interrupt Flag
						// 3 - ADC Interrupt Enable
						// 2:0 - ADC Prescaler Select Bits
						//		000 - 2 => 62.5 ns * 2 = 125 ns
						//		001 - 2 => 125 ns
						//		010 - 4 => 250 ns
						//		011 - 8 => 500 ns
						//		100 - 16 => 1 us
						//		101 - 32 => 2 us
						//		110 - 64 => 4 us
						//		111 - 128 => 8 us
	
	ADCSRB = 0b00000000;// 7:3 - Reserved Bits (6 is used elsewhere)
						// 2:0 - ADC Auto Trigger Source
						//		000 - Free Running Mode
						//		001 - Analog Comparator
						//		010 - External Interrupt Request 0
						//		011 - Timer0 Compare Match
						//		100 - Timer0 Overflow
						//		101 - Timer1 Compare Match B
						//		110 - Timer1 Overlow
						//		111 - Timer1 Capture Event
						
	DIDR0 = 0b00000001;	// 7:0 - ADC7:0 Digital Input Disable
						//	=> Disable digital input buffer to reduce power consumption 
}

/**
 * Initialize UART RX
 */
void init_uart_tx(void){
	UBRR0L = 103;		// 9600 Baud Rate for 16 MHz Clock

	UCSR0A = 0b00000000;// 7 - UART RX complete flag
						// 6 - UART TX complete flag
						// 5 - UART DRE flag
						// 4 - Frame error flag
						// 3 - Data overrun flag
						// 2 - Parity error flag
						// 1 - Double UART speed mode
						// 0 - Multi processor communication mode

	UCSR0B = 0b00001000; // 7 - RX Complete Interrupt Enable
						// 6 - TX Complete Interrupt Enable
						// 5 - UART DRE Interrupt Enable
						// 4 - Enable receiver
						// 3 - Enable transmitter
						// 2 - Character size
						// 1:0 - Ninth data bit

	UCSR0C = 0b00000110;// 7:6 - UART Mode Select
						//		00 - Asynchronous UART
						//		01 - Synchronous UART
						//		10 - Reserved
						//		11 - Master SPI
						// 5:4 - Parity mode
						//		00 - Disabled
						//		01 - Reserved
						//		10 - Enabled, Even
						//		11 - Enabled, Odd
						// 3 - Select stop bit length
						//		0 - 1-bit
						//		1 - 2-bit
						// 2:1 - Character size, with UCSRB.2
						//		000 - 5 bit
						//		001 - 6 bit
						//		010 - 7 bit
						//		011 - 8 bit
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Reserved
						//		111 - 9 bit
						// 0 - Select clock polarity
						//		0 - Rising XCK edge
						//		1 - Falling XCK edge 						
}

void init_mcu(void) {
	SMCR = 0b00000000;	// 7:4 - Reserved bits
						// 3:1 - Sleep Mode
						//		000 - Idle
						//		001 - ADC Noise Reduction
						//		010 - Power-down
						//		011 - Power-save
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Standby
						//		111 - Extended Standby
						// 0 - Sleep Enable
						
	MCUCR |= (0 << 6);	// 6 - BOD Sleep
	MCUCR |= (0 << 5);	// 5 - BOD Sleep Enable
	
	PRR = 0b11111100;	// 7 - Power Reduction TWI
						// 6 - Power Reduction Timer2
						// 5 - Power Reduction Timer0
						// 4 - Power Reduction UART1
						// 3 - Power Reduction Timer1
						// 2 - Power Reduction SPI
						// 1 - Power Reduction UART0
						// 0 - Power Reduction ADC
}
