/**
 * A Mega32 PPM Capturer Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
/**
 * Mega32 DIP Pinout
 * 					___________
 * PB0 (XCK/T0)		|1		40| PA0 (ADC0)
 * PB1 (T1)			|2		39| PA1 (ADC1)
 * PB2 (INT2/AIN0)	|3		38| PA2 (ADC2)
 * PB3 (OC0/AIN1)	|4		37| PA3 (ADC3)
 * PB4 (SS)			|5		36| PA4 (ADC4)
 * PB5 (MOSI)		|6		35| PA5 (ADC5)
 * PB6 (MISO)		|7		34| PA6 (ADC6)
 * PB7 (SCK)		|8		33| PA7 (ADC7)
 * RESET			|9		32| AREF
 * VCC				|10		31| GND
 * GND				|11		30| AVCC
 * XTAL2			|12		29| PC7 (TOSC2)
 * XTAL1			|13		28| PC6 (TOSC1)
 * PD0 (RXD)		|14		27| PC5 (TDI)
 * PD1 (TXD)		|15		26| PC4 (TDO)
 * PD2 (INT0)		|16		25| PC3 (TMS)
 * PD3 (INT1)		|17		24| PC2 (TCK)
 * PD4 (OC1B)		|18		23| PC1 (SDA)
 * PD5 (OC1A)		|19		22| PC0 (SCL)
 * PD6 (ICP1)		|20		21| PD7 (OC2)
 *					-----------
 */

/* Includes */
#include <Mega32.h>
#include <string.h>
#include <stdio.h>
#include <delay.h>

/* Global Constants */
#define TRUE 	1
#define FALSE 	0
#define UDRE 5

/* Program Constants */ 
#define timer_res_us 10			// Timer resolution
								// Higher means lower cpu but lower resolution
//#define N 1					// Clock Divider (See initialization for details)
#define uart_latency_us 5000	// Delay between each UART transmission of data
#define uart_latency_tick uart_latency_us/timer_res_us	// Number of timer ticks for latency 				

/* Global Variables */
unsigned char pulse_period, pulse_tick_count;
unsigned int uart_tick_count;
unsigned char data_ready_flag, uart_ready_flag;

/* Prototype Functions */
void init_pwm_cap(void);
void init_timer(void);
void init_uart_tx(void);
void uart_send_byte_array(char *data, unsigned char length);
void uart_send_byte(char data);

/**
 * Triggered every tick (depends on tick_res_us)
 */
interrupt [TIM0_COMP] void timer0_comp(void) {
	pulse_tick_count++;		// Increment a counter that keeps track
							// of how many ticks before the next rising edge
	
	if (uart_tick_count++ >= uart_latency_tick) {	// Counter for UART latency
		uart_ready_flag = 1;
		uart_tick_count = 0;
	}
}

/**
 * Triggered on rising edge
 */
interrupt [EXT_INT1] void external_interrupt1(void) {
	pulse_period = pulse_tick_count;	// Pulse period is equal to the current
										// number of pulse ticks
	pulse_tick_count = 	0;				// Reset the tick counter for pulses
	data_ready_flag = 1;			// Raise edge detection flag to signal data ready
}

/**
 * UART transmission function for byte arrays
 *
 * Can be used to transmitt strings
 * Length terminated (versus null terminated)
 */
void uart_send_byte_array(unsigned char *data, unsigned char length) {
	unsigned char counter = 0;

	while (counter++ < length) {
		while ( !( UCSRA & (1<<UDRE)) );	// Wait for empty buffer
		UDR = *data++;						// Put data into buffer, sends the data 
	}
}

/**
 * UART transmission function for a single byte
 */
void uart_send_byte(unsigned char data) {
	while ( !( UCSRA & (1<<UDRE)) );	// Wait for empty buffer
 	UDR = data;							// Put data into buffer, sends the data 
}

/**
 * Main program
 */
void main(void) {
	init_pwm_cap();		// Initialize PWM capture
	init_timer();		// Initialize Timer
	init_uart_tx();		// Initialize UART TX
	
	#asm("sei");		// Enable interrupts in assembly
	
	delay_ms(100);		// Delay to stablize initilization
	
	// Main program loop
	while(1) {
		// Send out data through the UART when the data and uart are ready
		if (data_ready_flag && uart_ready_flag) {
			uart_send_byte(pulse_period);
			uart_ready_flag = 0;
		}
	}
}

/**
 * Enable external interrupts INT1 (rising)
 * to time the periods of the pwm pulse and cycle.
 */
void init_pwm_cap(void) {
	MCUCR = 0b00001100;	// 7:4 - Reserved bits
						// 3:2 - INT1's sense control (Rising)
						//		00 - Low level
						//		01 - Any logical change
						//		10 - Falling edge
						//		11 - Rising edge
						// 1:0 - INT0's " (Disabled)
						//		"
						
	MCUCSR = 0 << 6;	// INT2's sense control on minimum width (Disabled)
						
	GICR = 0b10000000;	// 7 - INT1 interrupt request (Enabled)
						// 6 - INT0 " (Disabled)
						// 5 - INT2 " (Disabled)
						// 4:0 - Reserve bits
										
	pulse_period = 0;	// Initialize variable
}

/**
 * Enable timer0 to act as a timer for measuring the time
 * between rising and falling and then rising to rising.
 */
void init_timer(void) {
	// A single cycle = 1/16 MHz = 62.5 ns
	// Minimum interrupt response time = 4 + 4 = 8 cycles = 500 ns
	
	TCCR0 = 0b00001010;	// 7 - Force output compare (Disabled)
						// 6,3 - Timer Mode (!) (CTC)
						//		00 - Normal
						//		01 - Clear Timer on Compare Match
						//		10 - Phase Correct PWM
						//		11 - Fast PWM
						// 5:4 - Compare Match Output Mode (Normal)
						// 2:0 - Clock select (8)
						//		000 - Timer stopped
						//		001 - 1 => 62.5 ns (62.5 ns to 15.9 us)
						//		010 - 8 => 500 ns (500 ns to 127.5 us)
						//		011 - 64 => 4 us (4 us to 1.02 ms)
						//		100 - 256 => 16 us
						//		101 - 1024 => 64 us
						//		110 - External clock source on T0 pin, falling edge
						//		111 - ", rising edge
	
	OCR0 = timer_res_us * 2;	// OCR0 = timer resolution in us / 500 ns
								// => timer resolution in us * 2
	
	TIMSK = 0b00000010;	// 7:2 - Reserved bits for other timer interrupts
						// 1 - Timer0 output compare match interrupt (Enabled)
						// 0 - Timer0 overflow interrupt (Disabled)
	
	pulse_tick_count = 0;	// Initialize variable
}


/**
 * Initialize UART RX
 */
void init_uart_tx(void){
	UBRRL = 103;		// 9600 Baud Rate for 16 MHz Clock
	
	UCSRA = 0b00000000;	// 7 - UART RX complete flag
						// 6 - UART TX complete flag
						// 5 - UART DRE flag
						// 4 - Frame error flag
						// 3 - Data overrun flag
						// 2 - Parity error flag
						// 1 - Double UART speed mode (Disabled)
						// 0 - Multi processor communication mode (Disabled)

	UCSRB = 0b00001000; // 7 - RX Complete Interrupt (Disabled)
						// 6 - TX Complete Interrupt (Disabled)
						// 5 - UART DRE Interrupt (Disabled)
						// 4 - Enable receiver (Disabled)
						// 3 - Enable transmitter (Enabled)
						// 2 - Character size (8-bit)
						// 1:0 - Ninth data bit

	UCSRC = 0b10000110;	// 7 - Access UBRRH or UCSRC (UCSRC)
						//		0 - UBRRH
						//		1 - UCSRC
						// 6 - Select UART mode (Asynchronous)
						//		0 - Asynchronous
						//		1 - Synchronous
						// 5:4 - Enable odd parity mode (Disabled)
						//		00 - Disabled
						//		01 - Reserved
						//		10 - Enabled, Even
						//		11 - Enabled, Odd
						// 3 - Select stop bit length (1-bit)
						//		0 - 1-bit
						//		1 - 2-bit
						// 2:1 - Character size, with UCSRB.2 (8-bit)
						//		000 - 5 bit
						//		001 - 6 bit
						//		010 - 7 bit
						//		011 - 8 bit
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Reserved
						//		111 - 9 bit
						// 0 - Select clock polarity (Disabled)
						//		0 - Rising XCK edge
						//		1 - Falling XCK edge
}
						
