/**
 * A Mega32 PWM Generator Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega32 DIP Pinout
 * 					___________
 * PB0 (XCK/T0)		|1		40| PA0 (ADC0)
 * PB1 (T1)			|2		39| PA1 (ADC1)
 * PB2 (INT2/AIN0)	|3		38| PA2 (ADC2)
 * PB3 (OC0/AIN1)	|4		37| PA3 (ADC3)
 * PB4 (SS)			|5		36| PA4 (ADC4)
 * PB5 (MOSI)		|6		35| PA5 (ADC5)
 * PB6 (MISO)		|7		34| PA6 (ADC6)
 * PB7 (SCK)		|8		33| PA7 (ADC7)
 * RESET			|9		32| AREF
 * VCC				|10		31| GND
 * GND				|11		30| AVCC
 * XTAL2			|12		29| PC7 (TOSC2)
 * XTAL1			|13		28| PC6 (TOSC1)
 * PD0 (RXD)		|14		27| PC5 (TDI)
 * PD1 (TXD)		|15		26| PC4 (TDO)
 * PD2 (INT0)		|16		25| PC3 (TMS)
 * PD3 (INT1)		|17		24| PC2 (TCK)
 * PD4 (OC1B)		|18		23| PC1 (SDA)
 * PD5 (OC1A)		|19		22| PC0 (SCL)
 * PD6 (ICP1)		|20		21| PD7 (OC2)
 *					-----------
 */

/* Includes */
#include <Mega32.h>
#include <string.h>
#include <stdio.h>
#include <delay.h>

/* Global Constants */
#define TRUE 	1
#define FALSE 	0

/* Constants used for auto pwm generator */
#define cpu_clk_frequency 16000000	// CPU Clock Frequency (in Hz)
#define pwm_frequency 327000		// PWM Frequency (in Hz) = cpu_clk_frequency/(N*(TOP+1))
									//		With N=1 => 244 Hz to 16 MHz
									// 		With N=8 => 30 Hz to 2 MHz
									// 		With N=64 => 4 Hz to 250 KHz
									// 		With N=256 => 1 Hz to 62.5 KHz
									// 		With N=1024 => 0 Hz to 15.6 KHz
#define N 1							// Clock divider (Lower equals higher resolution)
//#define pulse_width_time false
//#define pulse_width_cycle true
#define initial_duty_cycle 0.5		// Initial duty cycle
//#define initial_pulse_width_us 5000	// Initial pulse width (Cannot wider than base period)

/* Global variables */
unsigned int pulse_width;

/* Function prototyes */
void man_init_pwm_gen(void);		// Manual PWM Generator Initialization (calculates values manually)
void auto_init_pwm_gen(void);		// Auto PWM Generator Initialization (calculates values automatically)

/**
 * Main program thread
 */
void main(void) {
	DDRA = 0x00;					// PORTA is an input
	
	auto_init_pwm_gen();			// Perform auto pwm generator initialization
	//man_init_pwm_gen();			// Perform manual "
	
	delay_ms(100);					// Delay to stablize initializations
	
	while(1) {
		if (PINA != 0xff) {
			if (!PINA.0) {
				pulse_width = 28;
			}
			else if (!PINA.1) {
				pulse_width = 27;
			}
			else if (!PINA.2) {
				pulse_width = 26;
			}
			else if (!PINA.3) {
				pulse_width = 25;
			}
			else if (!PINA.4) {
				pulse_width = 23;
			}
			else if (!PINA.5) {
				pulse_width = 22;
			}
			else if (!PINA.6) {
				pulse_width = 21;
			}
			else if (!PINA.7) {
				pulse_width = 20;
			}
			OCR1AH = pulse_width >> 8;		// High byte
			OCR1AL = (pulse_width << 8) >> 8;	// Low byte 
			OCR1BH = OCR1AH;
			OCR1BL = OCR1AL;
		}
	}
}

/**
 * Manual calculations
 *
 * Initialize OC1B (inverting) and OC1A (non-inverting) as the fast PWM output
 * from Timer 1 with a base freq set by ICR1 and pulse width set by OCR1A and OCR1B.
 *
 * Set N with clock select
 * Set TOP with ICR1
 * Set pulse width with OCR1A and OCR1B
 *
 * Trying to produce a freq of 327 KHz, it can be as accurate as 326.5 KHz
 */
void man_init_pwm_gen(void) {
	DDRD.4 = 1;	// Set PORTD.4 as an output for compare unit B
	DDRD.5 = 1;	// Set PORTD.5 as an output for compare unit A 
	
	// PWM base freq = 327 KHz, Clk = 16 MHz
	// PWM freq = clk/N*(TOP+1)
	// N = 1, 16000000 / 1*(48+1) = 326,530 Hz
	// N = 8, 16000000 / 8*(5+1) = 326,530 Hz
	
	TCCR1A = 0b11100010;	// 7:6 - OC1A's compare output mode (Non-inverting)
							// 5:4 - OC1B's " (Inverting)
							// 3 - OC1A's force ouput compare (Disabled)
							// 2 - OC1B's " (Disabled)
							// 1:0 - Waveform generation mode (Fast PWM, TOP=ICR1)
	
	TCCR1B = 0b00011001;	// 7 - Input capture noise canceler (Disabled)
							// 6 - Input capture edge select (Disabled)
							// 5 - Reserved bit
							// 4:3 - Waveform generation mode (")
							// 2:0 - Clock select (N)
							//	000 - Timer/Counter stopped
							//	001 - 1            
							//	010 - 8
							//	011 - 64
							//	100 - 256
							//	101 - 1024
							//	110 - External clock source on T1 pin, falling edge
							//	111 - ", rising edge
	
	pulse_width = 24;			
	OCR1AH = 0;				// BOTTOM <= OCR1A <= TOP
	OCR1AL = 24;			// OCR1A=24 => 24/48*100%=50% Duty Cycle
	
	OCR1BH = 0;				// Same as OCR1A
	OCR1BL = 24;			// "
	
	ICR1H = 0;				// ICR1 = TOP = 48
	ICR1L = 48;
	
	TIMSK = 0;	// Timer/Counter interrupt mask (All disabled)
				// 5 - Timer1, input capture interrupt
				// 4 - Timer1, output compare A match interrupt
				// 3 - Timer1, " B "
				// 2 - Timer1, overflow interrupt
}

/**
 * Auto calculations
 *
 * Initialize OC1B (inverting) and OC1A (non-inverting) as the fast PWM output
 * from Timer 1 with a base freq set by ICR1 and pulse width set by OCR1A and OCR1B.
 */
void auto_init_pwm_gen(void) {
	float cpu_pwm_ratio;			// Ratio of CPU clk frequency to PWM frequency
	float cpu_pwm_ratio_float;		// The float part of the ratio
	unsigned int cpu_pwm_ratio_int;	// The integer part of the ratio
	unsigned int TOP;				// TOP value

	DDRD.4 = 1;		// Set PORTD.4 as an output for OC1B
	DDRD.5 = 1;		// Set PORTD.5 as an output for OC1A 
	
	/*
	 * Sets the register for the correct clock divider value
	 *	000 - Timer/Counter stopped
	 *	001 - 1            
	 *	010 - 8
	 *	011 - 64
	 *	100 - 256
	 *	101 - 1024
	 *	110 - External clock source on T1 pin, falling edge
	 *	111 - ", rising edge
	 */
	if (N == 1) {
		TCCR1B |= 1;
	}
	else if (N == 8) {
		TCCR1B |= 2;
	}
	else if (N == 64) {
		TCCR1B |= 3;
	}
	else if (N == 256) {
		TCCR1B |= 4;
	}
	else if (N == 1024) {
		TCCR1B |= 5;
	}
	
	// cpu_pwm_ratio = clk_freq/pwm_freq = N*(TOP+1)
	cpu_pwm_ratio = (float) cpu_clk_frequency / (float) pwm_frequency;
	cpu_pwm_ratio_int = (unsigned int) cpu_pwm_ratio;
	cpu_pwm_ratio_float = (float) cpu_pwm_ratio - (float) cpu_pwm_ratio_int;
        
	// Since casting to a integer always floors a floating value,
	// it needs to have 0.5 added to it if we want to round up any
	// decimal equal or greater than 0.5.  
	if (cpu_pwm_ratio >= 0.5) {
		cpu_pwm_ratio += 0.5;
	}
	TOP = (unsigned int) cpu_pwm_ratio - 1;
	
	ICR1H = TOP >> 8;			// High byte
	ICR1L = (TOP << 8) >> 8;	// Low byte
	
	pulse_width = (unsigned int) ((float) TOP * (float)initial_duty_cycle);
	OCR1AH = pulse_width >> 8;			// High byte
	OCR1AL = (pulse_width << 8) >> 8;	// Low byte 
	OCR1BH = OCR1AH;
	OCR1BL = OCR1AL;
	
	TCCR1A = 0b11100010;	// 7:6 - OC1A's compare output mode (Non-inverting)
							// 5:4 - OC1B's " (Inverting)
							// 3 - OC1A's force ouput compare (Disabled)
							// 2 - OC1B's " (Disabled)
							// 1:0 - Waveform generation mode (Fast PWM, TOP=ICR1)
	
	TCCR1B |= 0 << 7;	// Input capture noise canceler (Disabled)
	TCCR1B |= 0 << 6;	// Input capture edge select (Disabled)
	TCCR1B |= 0 << 5;	// Reserved bit
	TCCR1B |= 1 << 4;	// Waveform generation mode (")
	TCCR1B |= 1 << 3;
	
	TIMSK = 0;	// Timer/Counter interrupt mask (All disabled)
				// 5 - Timer1, input capture interrupt
				// 4 - Timer1, output compare A match interrupt
				// 3 - Timer1, " B "
				// 2 - Timer1, overflow interrupt
}
	


 
 
