/**
 * Mega32 Simple RF (433 MHz RCT/P-433) Master Transceiver Lib Test
 * 
 * Ruibing Wang (rw98)
 * Based upon wireless protocal library by Meghan Desai
 */

/**
 * Mega32 DIP Pinout
 * 					___________
 * PB0 (XCK/T0)		|1		40| PA0 (ADC0)
 * PB1 (T1)			|2		39| PA1 (ADC1)
 * PB2 (INT2/AIN0)	|3		38| PA2 (ADC2)
 * PB3 (OC0/AIN1)	|4		37| PA3 (ADC3)
 * PB4 (SS)			|5		36| PA4 (ADC4)
 * PB5 (MOSI)		|6		35| PA5 (ADC5)
 * PB6 (MISO)		|7		34| PA6 (ADC6)
 * PB7 (SCK)		|8		33| PA7 (ADC7)
 * RESET			|9		32| AREF
 * VCC				|10		31| GND
 * GND				|11		30| AVCC
 * XTAL2			|12		29| PC7 (TOSC2)
 * XTAL1			|13		28| PC6 (TOSC1)
 * PD0 (RXD)		|14		27| PC5 (TDI)
 * PD1 (TXD)		|15		26| PC4 (TDO)
 * PD2 (INT0)		|16		25| PC3 (TMS)
 * PD3 (INT1)		|17		24| PC2 (TCK)
 * PD4 (OC1B)		|18		23| PC1 (SDA)
 * PD5 (OC1A)		|19		22| PC0 (SCL)
 * PD6 (ICP1)		|20		21| PD7 (OC2)
 *					-----------
 */

/* Includes */
#include <Mega32.h>
#include <lcd.h>		// LCD driver routines
#include <string.h>
#include <stdio.h>
#include <delay.h> 
#include "simple_rf_rxtx.h"

/* Global Constants */
#define TRUE 	1
#define FALSE 	0
#define id 3			// Transmitted ID
#define threshold 3		// Freq of tx
						// threshold * interrupt rate
						// 6 * 16ms = 96ms
						// 3 * 16ms = 48ms (Fastest possible for 2 date bytes) 
#define data_length 1

/* LCD Variables */
/* 
Use an 1x16 alphanumeric LCD connected
to PORTC as follows:

[LCD]   [Mega32 pin]
1 GND-  GND
2 +5V-  VCC
3 VLC 10k trimpot wiper (trimpot ends go to +5 and gnd)
4 RS -  PC0
5 RD - PC1
6 EN - PC2
11 D4 - PC4
12 D5 - PC5
13 D6 - PC6
14 D7 - PC7
*/
#define LCDwidth 16		// Number of Characters
#asm
	.equ __lcd_port=0x15
#endasm

/* Global Variables */
unsigned char lcd_buffer[17];    // LCD display buffer
unsigned char local_data[data_length];
unsigned char waiting;
unsigned char time;


/* Function Prototypes */
void init(void);

/**
 * Timer 0 compare ISR
 *
 * Interrupts every 1 ms, used to synchronize the sends/resends of the master,
 * making it push the data when it either finds that it has received no
 * acknowledgement or a bad one from the slave
 *
 * Accuracy is not important
 *
 */
interrupt [TIM0_COMP] void timer0_compare(void) {
	if(time<=threshold) 
		time++;
}                                                    

/**
 * Displays the TX data/packet
 */
void display_tx_packet(void) {
	lcd_clear();
	
	sprintf(lcd_buffer,"%-i",local_data[0]);
		lcd_gotoxy(0,0);
	 	lcd_puts(lcd_buffer);

	sprintf(lcd_buffer,"%-i",tx_data_length);
		lcd_gotoxy(0,1);
	 	lcd_puts(lcd_buffer);
	 	
	sprintf(lcd_buffer,"%-i",tx_data_hash);
		lcd_gotoxy(4,1);
	 	lcd_puts(lcd_buffer);
}

/**
 * Displays the RX data/packet
 */
void display_rx_packet(void) {
	unsigned char i, j, length, data[data_length];
	
	lcd_clear();
	
	length = get_rx_packet_length();
	get_rx_packet(data);
			
	sprintf(lcd_buffer,"%-i",data[0]);
	lcd_gotoxy(0,0);
	lcd_puts(lcd_buffer);
}

/**
 * Main function
 *
 */
void main(void) {
	// Initialization
	init();

	while(1) {
		// Whenever we go beyond threshold * 1ms,
 	  	// If it is not waiting for a response
		// and TX is ready for another transmission,
		// we can send a new packet of data
		if (!waiting && tx_is_ready()) {     
		
			// After this, we need to wait for response
			waiting = 1;
			
			//local_data[0] = readBPPByte_PCMCU();
			PORTD.7 = 0;
			lcd_clear();
			sprintf(lcd_buffer,"%-i",local_data[0]);
			lcd_gotoxy(0,0);
			lcd_puts(lcd_buffer);
 				
			// Transmit the data
 			tx_enable();	// Enable TX
			tx_byte_array(local_data, data_length, id, 0);	// Encode the data and transmit it
			while(!tx_is_ready());	// Wait for transmit
				rx_enable();	// Enable RX				
				//display_tx_packet();	// Display the transmitted data
			}
			// Waiting for a response
		else if (waiting) {
			// Received a request from the slave for the next packet
			if (rx_request_next) {
				PORTD.7 = 1;
    				waiting = 0;	// Disable wait
        			rx_request_next = 0;	// Reset variable
           	}
			// Received a request from the slave for a resend
			else if (rx_request_resend) {
				// Do nothing, as we will send it anyways in a
				// master driven system
				rx_request_resend = 0;	// Reset variable 
					// Resend packet
				tx_enable();
				tx_byte_array(local_data, data_length, id, 0);
				while(!tx_is_ready());
				rx_enable();
			}
			else {
				// Resend packet
				tx_enable();
				tx_byte_array(local_data, data_length, id, 0);
				while(!tx_is_ready());
				rx_enable();
			}
		}
		time=0;
	}
}

/**
 * Initialization
 */
void init(void) {
	// Turn all of PORTD on
	DDRD = 0xff;
	
	// Use PORTD.7 as an LED output for simple debugging and status check
	PORTD.7 = 1;
	
	// Initialize LCD
	lcd_init(LCDwidth);
	lcd_clear();
	lcd_gotoxy(0,0);
	lcd_putsf("Waiting for transmission");
	
	// Set up and initialize timer 0
	TIMSK = 2;		// Turn on timer 0 cmp match ISR 
	OCR0 = 250;  		// Set the compare re to 250 time ticks
	TCCR0=0b00001101;	// Prescalar to 1024 and turn on clear-on-match
				// 250 * 1024 / 16e6 = 16 ms
	time = 0;

	// Initialize variables
	waiting = 0;

	// Initialize RXTX library
	rxtx_init();

	// Initialize initally to be RX
	rx_enable();
	
	#asm ("sei");
}