/**
 * Mega32 Simple RF (433 MHz RCT/P-433) Slave Transceiver
 * 
 * Ruibing Wang (rw98)
 * Based upon wireless protocal library by Meghan Desai
 */

/**
 * Mega32 DIP Pinout
 * 					___________
 * PB0 (XCK/T0)		|1		40| PA0 (ADC0)
 * PB1 (T1)			|2		39| PA1 (ADC1)
 * PB2 (INT2/AIN0)	|3		38| PA2 (ADC2)
 * PB3 (OC0/AIN1)	|4		37| PA3 (ADC3)
 * PB4 (SS)			|5		36| PA4 (ADC4)
 * PB5 (MOSI)		|6		35| PA5 (ADC5)
 * PB6 (MISO)		|7		34| PA6 (ADC6)
 * PB7 (SCK)		|8		33| PA7 (ADC7)
 * RESET			|9		32| AREF
 * VCC				|10		31| GND
 * GND				|11		30| AVCC
 * XTAL2			|12		29| PC7 (TOSC2)
 * XTAL1			|13		28| PC6 (TOSC1)
 * PD0 (RXD)		|14		27| PC5 (TDI)
 * PD1 (TXD)		|15		26| PC4 (TDO)
 * PD2 (INT0)		|16		25| PC3 (TMS)
 * PD3 (INT1)		|17		24| PC2 (TCK)
 * PD4 (OC1B)		|18		23| PC1 (SDA)
 * PD5 (OC1A)		|19		22| PC0 (SCL)
 * PD6 (ICP1)		|20		21| PD7 (OC2)
 *					-----------
 */

/* Includes */
#include <Mega32.h>
#include <lcd.h>
#include <string.h>
#include <stdio.h>
#include <delay.h>
#include "simple_rf_rxtx.h"

/* Global Constants */
#define TRUE 	1
#define FALSE 	0


/* LCD Variables */
/* 
Use an 1x16 alphanumeric LCD connected
to PORTC as follows:

[LCD]   [Mega32 pin]
1 GND-  GND
2 +5V-  VCC
3 VLC 10k trimpot wiper (trimpot ends go to +5 and gnd)
4 RS -  PC0
5 RD - PC1
6 EN - PC2
11 D4 - PC4
12 D5 - PC5
13 D6 - PC6
14 D7 - PC7
*/
#define LCDwidth 16		// Number of Characters
#asm
	.equ __lcd_port=0x15
#endasm

/* Global Variables */
unsigned char local_data[1], local_length, local_id;
unsigned char lcd_buffer[17];    // LCD display buffer

/* Function Prototypes */
void init(void);

/**
 * Displays the decoded data info for the data
 * packet that was recieved
 */
void display_data_info(void) {
	lcd_clear();
	
	sprintf(lcd_buffer,"%-i", rx_data_id);
		lcd_gotoxy(0,0);
	 	lcd_puts(lcd_buffer);
	   
	sprintf(lcd_buffer,"%-i",rx_data_length);
		lcd_gotoxy(4,0);
	 	lcd_puts(lcd_buffer);
	
	sprintf(lcd_buffer,"%-i",rx_data_hash);
		lcd_gotoxy(8,0);
	 	lcd_puts(lcd_buffer);
	
	sprintf(lcd_buffer,"%-i",rx_temp_hash);
		lcd_gotoxy(12,0);
	 	lcd_puts(lcd_buffer);
}	

/**
* Main function for the RX
*
*/
void main(void) {
	// Initialize 
	init();
	
	while(1) {
		// If something was received and verified
		// we can now retrieve it and use it
		if(is_rx_ready() && tx_request_next) {
			PORTD.7 = 0;
			
			get_rx_data(local_data);
			
			sprintf(lcd_buffer,"%-i",local_data[0]);
			lcd_gotoxy(0,0);
	 		lcd_puts(lcd_buffer);


			//sendByte_MCUPrinter(local_data[0]);
			
			//display_packet();
			//display_data_info();
			
			// If the received packet was verified to be valid,
			// we enable TX to send an acknowlgement byte,
			// then we enable RX again to get ready for the new
			// byte that the master will send
			tx_enable();			// Enable TX
			tx_response(next_char, 0);	// Send response
			while(!tx_is_ready());		// Wait for tranmssion
			rx_enable();			// Enable RX
			tx_request_next = 0;		// Reset variable
			rx_reset();			// Reset RX
			PORTD.7 = 1;
		}
		// If the received packet was verified to be invalid,
		// we enable TX to sned a request for a resend,
		// then we enable RX again to get ready for the resent
		// byte that the master will send
		else if (tx_request_resend) {
			tx_enable();			// Enable TX
			tx_response(resend_char, 0);	// Send response
			while(!tx_is_ready());		// Wait for transmission
			rx_enable();			// Enable RX
			tx_request_resend = 0;		// Reset variable
			rx_reset();			// Reset RX
		}
	}
}

/**
* Initialization
*/
void init(void) {
	// Turn all of PORTD on
	DDRD = 0xff;
	
	// Use PORTD.7 as the LED output for simple debugging and status check
	PORTD.7 = 1;

	// Initialize LCD
	lcd_init(LCDwidth);
	lcd_clear();
	lcd_gotoxy(0,0);
	lcd_putsf("Waiting for reception");

	// Initialize RXTX library
	rxtx_init();
	
	// Start off with RX
	rx_enable();

	// Reset RX to fix its values
	rx_reset();

	#asm ("sei");
}