/**
 * A Mega32 Timer Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
/**
 * Mega32 DIP Pinout
 * 					___________
 * PB0 (XCK/T0)		|1		40| PA0 (ADC0)
 * PB1 (T1)			|2		39| PA1 (ADC1)
 * PB2 (INT2/AIN0)	|3		38| PA2 (ADC2)
 * PB3 (OC0/AIN1)	|4		37| PA3 (ADC3)
 * PB4 (SS)			|5		36| PA4 (ADC4)
 * PB5 (MOSI)		|6		35| PA5 (ADC5)
 * PB6 (MISO)		|7		34| PA6 (ADC6)
 * PB7 (SCK)		|8		33| PA7 (ADC7)
 * RESET			|9		32| AREF
 * VCC				|10		31| GND
 * GND				|11		30| AVCC
 * XTAL2			|12		29| PC7 (TOSC2)
 * XTAL1			|13		28| PC6 (TOSC1)
 * PD0 (RXD)		|14		27| PC5 (TDI)
 * PD1 (TXD)		|15		26| PC4 (TDO)
 * PD2 (INT0)		|16		25| PC3 (TMS)
 * PD3 (INT1)		|17		24| PC2 (TCK)
 * PD4 (OC1B)		|18		23| PC1 (SDA)
 * PD5 (OC1A)		|19		22| PC0 (SCL)
 * PD6 (ICP1)		|20		21| PD7 (OC2)
 *					-----------
 */

/* Includes */
#include <mega32.h>
#include <string.h>
#include <stdio.h>
#include <delay.h>

/* Global Constants */
#define TRUE 	1
#define FALSE 	0

/* Constants for Timer0 */
#define TIMER_RES_US	50  

/* Constants for Timer */
#define TIMER_LATENCY_US	10000
#define TIMER_LATENCY_THRESHOLD	(unsigned int) ((TIMER_LATENCY_US / TIMER_RES_US) + 0.5)  

unsigned int timer_tick_count = 0;
unsigned char led_test = 0;

void init_timer0(void);

interrupt [TIM0_COMP] void timer0_comp(void) {
	timer_tick_count++;
}

/**
 * Main program
 */
void main(void) {
	DDRB = 0xFF;
	PORTB = 0x00;
	
	init_timer0();		// Initialize Timer
	
	#asm("sei");		// Enable interrupts in assembly
	
	delay_ms(100);		// Delay to stablize initilization
	
	// Main program loop
	while(1) {
		if (timer_tick_count >= TIMER_LATENCY_THRESHOLD) {
			PORTB = led_test++;
			timer_tick_count = 0;
		}
	}
}


void init_timer0(void) {
	// A single cycle = 1/16 MHz = 62.5 ns
	// Minimum interrupt response time = 4 + 4 = 8 cycles = 500 ns
	
	TCCR0 = 0b00001010;	// 7 - Force output compare (Disabled)
						// 6,3 - Timer Mode (!) (CTC)
						//		00 - Normal
						//		01 - Clear Timer on Compare Match
						//		10 - Phase Correct PWM
						//		11 - Fast PWM
						// 5:4 - Compare Match Output Mode (Normal)
						// 2:0 - Clock select (8)
						//		000 - Timer stopped
						//		001 - 1 => 62.5 ns (62.5 ns to 15.9 us)
						//		010 - 8 => 500 ns (500 ns to 127.5 us)
						//		011 - 64 => 4 us (4 us to 1.02 ms)
						//		100 - 256 => 16 us
						//		101 - 1024 => 64 us
						//		110 - External clock source on T0 pin, falling edge
						//		111 - ", rising edge
	
	OCR0 = TIMER_RES_US * 2;	// OCR0 = timer resolution in us / 500 ns
				// => timer resolution in us * 2
	
	TIMSK = 0b00000010;	// 7:2 - Reserved bits for other timer interrupts
						// 1 - Timer0 output compare match interrupt (Enabled)
						// 0 - Timer0 overflow interrupt (Disabled)
}