/**
 * A Mega32 TWI Master Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
/**
 * Mega32 DIP Pinout
 * 					___________
 * PB0 (XCK/T0)		|1		40| PA0 (ADC0)
 * PB1 (T1)			|2		39| PA1 (ADC1)
 * PB2 (INT2/AIN0)	|3		38| PA2 (ADC2)
 * PB3 (OC0/AIN1)	|4		37| PA3 (ADC3)
 * PB4 (SS)			|5		36| PA4 (ADC4)
 * PB5 (MOSI)		|6		35| PA5 (ADC5)
 * PB6 (MISO)		|7		34| PA6 (ADC6)
 * PB7 (SCK)		|8		33| PA7 (ADC7)
 * RESET			|9		32| AREF
 * VCC				|10		31| GND
 * GND				|11		30| AVCC
 * XTAL2			|12		29| PC7 (TOSC2)
 * XTAL1			|13		28| PC6 (TOSC1)
 * PD0 (RXD)		|14		27| PC5 (TDI)
 * PD1 (TXD)		|15		26| PC4 (TDO)
 * PD2 (INT0)		|16		25| PC3 (TMS)
 * PD3 (INT1)		|17		24| PC2 (TCK)
 * PD4 (OC1B)		|18		23| PC1 (SDA)
 * PD5 (OC1A)		|19		22| PC0 (SCL)
 * PD6 (ICP1)		|20		21| PD7 (OC2)
 *					-----------
 */

/* Includes */
#include <Mega32.h>
#include <string.h>
#include <stdio.h>
#include <delay.h>

#define I2C_READ 1
#define I2C_WRITE 0

#define I2C_SUCCESS 0
#define I2C_ERROR_INVALID_STATUS 1
#define I2C_ERROR_NO_ACK 2
#define I2C_ERROR_INVALID_ADDRESS 3
#define I2C_ERROR_NO_DATA 4
#define I2C_ERROR_BUFFER_FULL 5

#define I2C_GENERAL_ADDRESS 0

//defines for TWSR
#define TWS_MASK 0xF8

//slave transmitter
#define TWS_ST_SLAR_ACK 0xA8
#define TWS_ST_ARB_LOST 0xB0
#define TWS_ST_DATA_ACK 0xB8
#define TWS_ST_DATA_NACK 0xC0
#define TWS_ST_DATA_ACK_LAST 0xC8

//slave reciever
#define TWS_SR_SLAW_ACK 0x60
#define TWS_SR_ARB_LOST 0x68
#define TWS_SR_SLAW_ACK_GEN 0x70
#define TWS_SR_ARB_LOST_GEN 0x78
#define TWS_SR_DATA_ACK 0x80
#define TWS_SR_DATA_NACK 0x88
#define TWS_SR_DATA_ACK_GEN 0x90
#define TWS_SR_DATA_NACK_GEN 0x98
#define TWS_SR_STOP 0xA0

//master transmitter
#define TWS_MT_START 0x08
#define TWS_MT_REPEATED_START 0x10
#define TWS_MT_SLAW_ACK 0x18
#define TWS_MT_SLAW_NACK 0x20
#define TWS_MT_DATA_ACK 0x28
#define TWS_MT_DATA_NACK 0x30
#define TWS_MT_ARB_LOST 0x38

//master reciever
#define TWS_MR_START 0x08
#define TWS_MR_REPEATED_START 0x10
#define TWS_MR_ARB_LOST 0x38
#define TWS_MR_SLAR_ACK 0x40
#define TWS_MR_SLAR_NACK 0x48
#define TWS_MR_DATA_ACK 0x50
#define TWS_MR_DATA_NACK 0x58

//misc
#define TWS_MISC_NONE 0xF8
#define TWS_MISC_ERROR 0x00

//defines for TWCR
#define TWINT  0b10000000
#define TWEA   0b01000000
#define TWSTA  0b00100000
#define TWSTO  0b00010000
#define TWWC   0b00001000
#define TWEN   0b00000100
#define TWIE   0b00000001

//defines for TWAR
#define TWGCE 1

#define SLAVE_ADDRESS 0x16

void sa_i2c_master_init(void);

unsigned char sa_i2c_start(unsigned char sla, unsigned char mode) {
    if (0x80 & sla) return I2C_ERROR_INVALID_ADDRESS;
    
    TWCR = TWINT | TWSTA | TWEN;
    
    while(!(TWCR & TWINT));
    
    if ( ((TWSR & TWS_MASK) != TWS_MT_START) && ((TWSR & TWS_MASK) != TWS_MT_REPEATED_START) ) {
        return I2C_ERROR_INVALID_STATUS;
    }
 
    TWDR = (sla<<1) | mode;
    TWCR = TWINT | TWEN;

    while (!(TWCR & TWINT));

    if (mode == I2C_WRITE) {
        if( (TWSR & TWS_MASK) == TWS_MT_SLAW_NACK ) {
            return I2C_ERROR_NO_ACK;
        }
        if( ((TWSR & TWS_MASK) != TWS_MT_SLAW_ACK) && ((TWSR & TWS_MASK) != TWS_MT_SLAW_NACK) ) {
            return I2C_ERROR_INVALID_STATUS;
        }
    } 
    else {
        if( (TWSR & TWS_MASK) == TWS_MR_SLAR_NACK ) {
            return I2C_ERROR_NO_ACK;
        }
        if( ((TWSR & TWS_MASK) != TWS_MR_SLAR_ACK) && ((TWSR & TWS_MASK) != TWS_MR_SLAR_NACK) ) {
            return I2C_ERROR_INVALID_STATUS;
        }
    }
    return I2C_SUCCESS;  
}

unsigned char sa_i2c_write_byte(unsigned char data) {
    TWDR = data;
    TWCR = TWINT | TWEN;    
    while(!(TWCR & TWINT));
    if( ((TWSR & TWS_MASK) != TWS_MT_DATA_ACK) && ((TWSR & TWS_MASK) != TWS_MT_DATA_NACK) ) {  
        return I2C_ERROR_INVALID_STATUS;
    }
    return I2C_SUCCESS;
}

unsigned char sa_i2c_write(unsigned char *data, unsigned char length) {
   unsigned char err, i;

   for(i=0;i<length;i++) {
      err = sa_i2c_write_byte(data[i]);
      if(err) return err;
   }
   return I2C_SUCCESS;
}

unsigned char sa_i2c_writef(unsigned char flash *data, unsigned char length) {
   unsigned char err, i;

   for(i=0;i<length;i++) {
      err = sa_i2c_write_byte(data[i]);
      if(err) return err;
   }
   return I2C_SUCCESS;
}

unsigned char sa_i2c_read_byte(char *byte, unsigned char ack) {      
    if(ack) TWCR = TWINT | TWEA | TWEN;
    else    TWCR = TWINT | TWEN;    

    while(!(TWCR & TWINT));

    if( ((TWSR & TWS_MASK) != TWS_MR_DATA_ACK) && ((TWSR & TWS_MASK) != TWS_MR_DATA_NACK) ) {
      return I2C_ERROR_INVALID_STATUS;
    }
        
    *byte = TWDR;
    return I2C_SUCCESS;
}     

unsigned char sa_i2c_read(unsigned char *data, unsigned char length) {
    unsigned char err, i;
    
    for(i = 0; i < length; i++) {
        err = sa_i2c_read_byte(data++, i!=(length-1));
        if (err) return err;
    }
    
    return I2C_SUCCESS; 
}

void sa_i2c_stop(void) {
    TWCR = TWINT | TWEN | TWSTO;
    while(TWCR & TWSTO);    
}

void main(void) {
	unsigned char counter;

	while(1) {
		sa_i2c_start(SLAVE_ADDRESS,I2C_WRITE);
		sa_i2c_write_byte(counter++);
		delay_ms(1);
		sa_i2c_stop();
		delay_ms(1);
	}
}
/* init for the i2c master
 */
void sa_i2c_master_init(void) {
    /* Bit Rates: 
       100 kHz - 0x48
       200 kHz - 0x20
       300 kHz - 0x13
       400 kHz - 0x0C
    */
    TWBR = 0x48;
    TWSR = 0x00;
}



