/**
 * A Mega32 TWI Slave Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
/**
 * Mega32 DIP Pinout
 * 					___________
 * PB0 (XCK/T0)		|1		40| PA0 (ADC0)
 * PB1 (T1)			|2		39| PA1 (ADC1)
 * PB2 (INT2/AIN0)	|3		38| PA2 (ADC2)
 * PB3 (OC0/AIN1)	|4		37| PA3 (ADC3)
 * PB4 (SS)			|5		36| PA4 (ADC4)
 * PB5 (MOSI)		|6		35| PA5 (ADC5)
 * PB6 (MISO)		|7		34| PA6 (ADC6)
 * PB7 (SCK)		|8		33| PA7 (ADC7)
 * RESET			|9		32| AREF
 * VCC				|10		31| GND
 * GND				|11		30| AVCC
 * XTAL2			|12		29| PC7 (TOSC2)
 * XTAL1			|13		28| PC6 (TOSC1)
 * PD0 (RXD)		|14		27| PC5 (TDI)
 * PD1 (TXD)		|15		26| PC4 (TDO)
 * PD2 (INT0)		|16		25| PC3 (TMS)
 * PD3 (INT1)		|17		24| PC2 (TCK)
 * PD4 (OC1B)		|18		23| PC1 (SDA)
 * PD5 (OC1A)		|19		22| PC0 (SCL)
 * PD6 (ICP1)		|20		21| PD7 (OC2)
 *					-----------
 */

/* Includes */
#include <mega32.h>
#include <string.h>
#include <stdio.h>
#include <delay.h>

/* Global/System Constants */
#define TRUE 	1
#define FALSE 	0

/* General Constants for TWI */
#define TWI_READ 			1
#define TWI_WRITE 			0
#define TWI_GENERAL_ADDRESS	0

#define TWINT		0b10000000
#define TWEA		0b01000000
#define TWSTA		0b00100000
#define TWSTO		0b00010000
#define TWWC		0b00001000
#define TWEN		0b00000100
#define TWIE		0b00000001

#define TWS_MASK 			0xF8

/* General TWI Master staus codes */                      
#define TWI_START					0x08	// START has been transmitted  
#define TWI_REP_START				0x10	// Repeated START has been transmitted
#define TWI_ARB_LOST				0x38	// Arbitration lost

/* TWI Master Transmitter staus codes */                      
#define TWI_MT_ADR_ACK				0x18  	// SLA+W has been tramsmitted and ACK received
#define TWI_MT_ADR_NACK				0x20  	// SLA+W has been tramsmitted and NACK received 
#define TWI_MT_DATA_ACK				0x28  	// Data byte has been tramsmitted and ACK received
#define TWI_MT_DATA_NACK			0x30  	// Data byte has been tramsmitted and NACK received 

/* TWI Master Receiver staus codes */  
#define TWI_MR_ADR_ACK				0x40  	// SLA+R has been tramsmitted and ACK received
#define TWI_MR_ADR_NACK				0x48  	// SLA+R has been tramsmitted and NACK received
#define TWI_MR_DATA_ACK				0x50  	// Data byte has been received and ACK tramsmitted
#define TWI_MR_DATA_NACK			0x58  	// Data byte has been received and NACK tramsmitted

/* TWI Slave Transmitter staus codes */
#define TWI_ST_ADR_ACK				0xA8  	// Own SLA+R has been received; ACK has been returned
#define TWI_ST_ADR_ACK_M_ARB_LOST	0xB0  	// Arbitration lost in SLA+R/W as Master; own SLA+R has been received; ACK has been returned
#define TWI_ST_DATA_ACK				0xB8  	// Data byte in TWDR has been transmitted; ACK has been received
#define TWI_ST_DATA_NACK			0xC0  	// Data byte in TWDR has been transmitted; NOT ACK has been received
#define TWI_ST_DATA_ACK_LAST_BYTE	0xC8  	// Last data byte in TWDR has been transmitted (TWEA = 0); ACK has been received

/* TWI Slave Receiver staus codes */
#define TWI_SR_ADR_ACK				0x60  	// Own SLA+W has been received ACK has been returned
#define TWI_SR_ADR_ACK_M_ARB_LOST 	0x68  	// Arbitration lost in SLA+R/W as Master; own SLA+W has been received; ACK has been returned
#define TWI_SR_GEN_ACK				0x70  	// General call address has been received; ACK has been returned
#define TWI_SR_GEN_ACK_M_ARB_LOST 	0x78  	// Arbitration lost in SLA+R/W as Master; General call address has been received; ACK has been returned
#define TWI_SR_ADR_DATA_ACK			0x80  	// Previously addressed with own SLA+W; data has been received; ACK has been returned
#define TWI_SR_ADR_DATA_NACK		0x88  	// Previously addressed with own SLA+W; data has been received; NOT ACK has been returned
#define TWI_SR_GEN_DATA_ACK			0x90  	// Previously addressed with general call; data has been received; ACK has been returned
#define TWI_SR_GEN_DATA_NACK		0x98  	// Previously addressed with general call; data has been received; NOT ACK has been returned
#define TWI_SR_STOP_RESTART			0xA0  	// A STOP condition or repeated START condition has been received while still addressed as Slave

/* TWI Miscellaneous status codes */
#define TWI_NO_STATE				0xF8	// No relevant state information available; TWINT = 0
#define TWI_BUS_ERROR				0x00	// Bus error due to an illegal START or STOP condition

/* Constants for TWI Functions' Return Status */
#define TWI_SUCCESS 				0
#define TWI_ERROR_INVALID_STATUS 	1
#define TWI_ERROR_NO_ACK 			2
#define TWI_ERROR_INVALID_ADDRESS 	3
#define TWI_ERROR_NO_DATA 			4
#define TWI_ERROR_BUFFER_FULL 		5

#define TWI_SLAVE_ADDRESS	2
#define TWI_GCE_ENABLE		TRUE
#define TWI_BUFFER_SIZE 	8

unsigned char twi_recieve_pos, twi_transmit_pos, twi_recieve_lost;
unsigned char twi_recieve_buffer[TWI_BUFFER_SIZE], twi_transmit_buffer[TWI_BUFFER_SIZE];
unsigned char twi_flag = 0;

unsigned char init_twi_slave(unsigned char sla, unsigned char gce);

interrupt [TWI] void twi_isr(void) {
   twi_flag = 1;
   
}

void twi_go(void) { 
    unsigned char i;
    twi_flag = 0;

	PORTB = TWSR & TWS_MASK;

    switch(TWSR & TWS_MASK) {
        case TWI_SR_ADR_ACK:
            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;
            
        case TWI_SR_ADR_ACK_M_ARB_LOST:
            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;
            
        case TWI_SR_GEN_ACK:
            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;
            
        case TWI_SR_GEN_ACK_M_ARB_LOST:
            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;
            
        case TWI_SR_ADR_DATA_ACK:
            if(twi_recieve_pos == (TWI_BUFFER_SIZE)) {
               twi_recieve_lost = 1;
            } 
            else {
               twi_recieve_buffer[twi_recieve_pos++] = TWDR;
            }
            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;
        
        case TWI_SR_ADR_DATA_NACK:
            if(twi_recieve_pos == (TWI_BUFFER_SIZE)) {
               twi_recieve_lost = 1;
            } else {
               twi_recieve_buffer[twi_recieve_pos++] = TWDR;
            }
            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;
            
        case TWI_SR_GEN_DATA_ACK:
            if(twi_recieve_pos == (TWI_BUFFER_SIZE)) {
               twi_recieve_lost = 1;
            } else {
               twi_recieve_buffer[twi_recieve_pos++] = TWDR;
            }    
            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;
        
        case TWI_SR_GEN_DATA_NACK:
            if(twi_recieve_pos == (TWI_BUFFER_SIZE)) {
               twi_recieve_lost = 1;
            } else {
               twi_recieve_buffer[twi_recieve_pos++] = TWDR;
            }            
            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;
                        
        case TWI_SR_STOP_RESTART:
            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;
            
        case TWI_ST_ADR_ACK:
   	      if(twi_transmit_pos) {
   	         TWDR = twi_transmit_buffer[0];
   	         twi_transmit_pos--;
             for(i=0; i<twi_transmit_pos; i++)
                  twi_transmit_buffer[i] = twi_transmit_buffer[(unsigned char) (i+1)];
   	      } 
   	      else {
   	            TWDR = 0xFF;
   	      }

            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;
            
        case TWI_ST_ADR_ACK_M_ARB_LOST:
   	      if(twi_transmit_pos) {
   	         TWDR = twi_transmit_buffer[0];
   	         twi_transmit_pos--;
               for(i=0; i<twi_transmit_pos; i++)
                  twi_transmit_buffer[i] = twi_transmit_buffer[(unsigned char) (i+1)];
   	      } else {
   	            TWDR = 0xFF;
   	      }

            TWCR = TWINT | TWEN | TWIE;
            break;
            
        case TWI_ST_DATA_ACK:
   	      if(twi_transmit_pos) {
   	         TWDR = twi_transmit_buffer[0];
   	         twi_transmit_pos--;
               for(i=0; i<twi_transmit_pos; i++)
                  twi_transmit_buffer[i] = twi_transmit_buffer[(unsigned char) (i+1)];
   	      } else {
   	            TWDR = 0xFF;
   	      }

            TWCR = TWINT | TWEA| TWEN | TWIE;
            break;
            
        case TWI_ST_DATA_NACK:
            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;
        
        case TWI_ST_DATA_ACK_LAST_BYTE:
            TWCR = TWINT | TWEA | TWEN | TWIE;
            break;

        case TWI_NO_STATE:
        	TWCR = TWINT | TWEA | TWEN | TWIE;
            break;
            
        case TWI_BUS_ERROR:
            TWCR = TWINT | TWSTO | TWEA | TWEN | TWIE;
            break;
            
        default:
    }
}

//set data in transmit buffer to send
unsigned char twi_write_byte(unsigned char data) {
   
   if(twi_transmit_pos==TWI_BUFFER_SIZE) {
      return TWI_ERROR_BUFFER_FULL;
   }
   
   twi_transmit_buffer[twi_transmit_pos++] = data;
   return TWI_SUCCESS;
}

//set constant data in transmit buffer to send
unsigned char twi_write(unsigned char *data, unsigned char length) {
   unsigned char err, i;

   for(i=0;i<length;i++) {
      err = twi_write_byte(*data);
      if(err) return err;
      data++;
   }
   return TWI_SUCCESS;   
}

//set constant data in transmit buffer to send
unsigned char twi_writef(unsigned char flash *data, unsigned char length) {
   unsigned char err, i;

   for(i=0;i<length;i++) {
      err = twi_write_byte(*data);
      if(err) return err;
      data++;
   }
   return TWI_SUCCESS;
}

//check how much room is in the transmit buffer
unsigned char twi_transmit_room() {
    return (TWI_BUFFER_SIZE - twi_transmit_pos);
}

//read one byte from recieve buffer
unsigned char twi_read_byte(unsigned char *byte) {
    unsigned char i;
    
    if(twi_recieve_pos) {
        *byte = twi_recieve_buffer[0];
        twi_recieve_pos--;
        for(i=0; i<twi_recieve_pos; i++)
            twi_recieve_buffer[i] = twi_recieve_buffer[(unsigned char) (i+1)];
        return TWI_SUCCESS;
    } else {
        return TWI_ERROR_NO_DATA;
    }
}

//read multiple bytes from buffer
unsigned char twi_read(unsigned char *data, unsigned char length) {
    unsigned char err, i;
    
    for(i = 0; i < length; i++) {
        // Ack is set to 1 except for last byte
        err = twi_read_byte(data++);
        if (err) return err;
    }
    
    return TWI_SUCCESS; 
}

//check if there is any data to read from the recieve buffer
unsigned char twi_receive_ready(void) {
    return twi_recieve_pos;
}

//check if data was lost due to buffer size limitation
unsigned char twi_data_lost(void) {
   unsigned char tmp;
   tmp = twi_recieve_lost;
   twi_recieve_lost = 0;
   return tmp;
}

void main(void) {
	unsigned char read_data, counter;
	counter = 0;
    
    DDRB = 0xff;
    PORTB = 0xff;
        
	init_twi_slave(TWI_SLAVE_ADDRESS, TWI_GCE_ENABLE);
    delay_ms(100);
    
    for(;;) {
    	if (twi_flag) {
    		PORTB.0 = 1;
    		twi_go();
    	}
        if (twi_receive_ready()) {
            
            twi_read_byte(&read_data); 

            while(twi_transmit_room()<1);
            
            twi_write_byte(counter++); 
            
            
            //twi_write_byte(count++);    
            //twi_next_transmit_ready();
            //twi_write_byte(data_h[read_data]);  
        }   
    }
}

unsigned char init_twi_slave(unsigned char sla, unsigned char gce){
    unsigned char i;

    if(0x80 & sla) 
    	return TWI_ERROR_INVALID_ADDRESS;

    TWAR = (sla<<1) | gce;
    TWCR = TWEA | TWEN | TWIE;
    
    
    twi_recieve_pos = 0;
    twi_transmit_pos = 0;
    twi_recieve_lost = 0;
    for(i = 0; i<TWI_BUFFER_SIZE; i++) {
        twi_transmit_buffer[i] = 0;
        twi_recieve_buffer[i] = 0;
    }
    
    //turn on ISRs
    #asm("sei");
    
    return TWI_SUCCESS;
}
					


