/**
 * A Mega644P Camera RXTX Library Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */
  
/* Includes */
#include <mega644p.h>
#include <string.h>
#include <stdio.h>
#include <delay.h>
#include "twi_master.h"  

/* Global Constants */
#define TRUE	1
#define FALSE	0
               
/* System Constants */
#define CPU_FREQUENCY_HZ				20000000        

/* UART Constants */
#define UART_BAUD_RATE_BPS				115200
#define UART_TX_ENABLE					TRUE
#define UART_RX_ENABLE					TRUE
#define UDRE	5

/* TWI Constants */ 
#define TWI_PULLUP_RESISTOR_ENABLE		TRUE
#define TWI_CAM_WRITE_ADDRESS			0xC0
#define TWI_CAM_READ_ADDRESS			0xC1
#define TWI_DATA_BUFFER_LENGTH			1

/* Camera Constants */       
#define CAM_HREF			PIND.3
#define	CAM_VSYN			PIND.2
#define	CAM_PCLK			PIND.4

/* Global Variables */
void camera_write(unsigned int reg, unsigned char value);
unsigned char camera_read(unsigned int reg);
void cam_reset(void);
void createheader(unsigned char *header, unsigned int heigh, unsigned int width);
void createinfoheader(unsigned char *infoheader, unsigned int height, unsigned int width); 
void uart_parse_rx(unsigned char temp_char);
void photo(void);
void sendtable(void);
void uart_send_byte_array(unsigned char *data, unsigned char length);
void uart_send_byte(unsigned char data);
void init_uart(void);

unsigned char cam_state;
unsigned char twi_data_buffer[TWI_DATA_BUFFER_LENGTH];
enum {CAM_STATE_IDLE, CAM_STATE_TX};


interrupt [USART0_RXC] void uart0_rxc(void) {
	unsigned char data = UDR0;
	uart_parse_rx(data);
}


void uart_send_byte_array(unsigned char *data, unsigned char length) {
	unsigned char counter = 0;

	while (counter++ < length) {
		while ( !( UCSR0A & (1<<UDRE)) );	// Wait for empty buffer
		UDR0 = *data++;						// Put data into buffer, sends the data 
	}
}

void uart_send_byte(unsigned char data) {
	while ( !( UCSR0A & (1<<UDRE)) );	// Wait for empty buffer
 	UDR0 = data;							// Put data into buffer, sends the data 
}

void camera_write(unsigned int reg, unsigned char value) {
	twi_data_buffer[0] = value; 
	twi_write_register(TWI_CAM_WRITE_ADDRESS, reg, 1, twi_data_buffer);
}

unsigned char camera_read(unsigned int reg) {
	twi_read_register(TWI_CAM_WRITE_ADDRESS, TWI_CAM_READ_ADDRESS, reg, 1, twi_data_buffer);
	return twi_data_buffer[0];
}

void createheader(unsigned char *header, unsigned int heigh, unsigned int width){
	unsigned char bytelow;
	unsigned char bytehigh;
	unsigned int sizefile = (unsigned int) (1078 + (heigh*width));
	bytelow = (sizefile << 8) >> 8;
	bytehigh = sizefile >> 8;
	
	//2 Bytes --BM Starting
	header[0] = 'B';
	header[1] = 'M';
	//4 Bytes -- Size of file in bytes = 14 + 40 +1024 + HEIGH * WIDTH
	header[2] = bytelow;
	header[3] = bytehigh;
	header[4] = 0;
	header[5] = 0;
	//4 Bytes of reserved (= 0)
	header[6] = 0;
	header[7] = 0;
	header[8] = 0;
	header[9] = 0;
	//4 Bytes of offset to the init of the data
	header[10] = 0x36;
	header[11] = 0x04;
	header[12] = 0;
	header[13] = 0;
}

void createinfoheader(unsigned char *infoheader, unsigned int height, unsigned int width){
	unsigned char height_low;
	unsigned char height_high;
	unsigned char width_low;
	unsigned char width_high;

	height_low = (height << 8) >> 8;
	height_high = height >> 8;
	width_low = (width << 8) >> 8;
	width_high = width >>8;

	//4 Bytes -- Size of InfoHeader =40 
	infoheader[0] = 40; 
	infoheader[1] = 0;
	infoheader[2] = 0;
	infoheader[3] = 0;
	//4 Bytes -- specifies the width of the image, in pixels.
	infoheader[4] = width_low;
	infoheader[5] = width_high;
	infoheader[6] = 0;
	infoheader[7] = 0;
	//4 Bytes -- specifies the heigth of the image, in pixels.
	infoheader[8] = height_low;
	infoheader[9] = height_high;
	infoheader[10] = 0;
	infoheader[11] = 0;
	//2 Bytes -- Number of planes of the image
	infoheader[12] = 1;
	infoheader[13] = 0;
	//2 Bytes Bits per Pixel  -- In our case 8.
	infoheader[14] = 8;
	infoheader[15] = 0;
	//4 bytes -- Type of Compression 0 = BI_RGB   no compression   
	infoheader[16] = 0;
	infoheader[17] = 0;
	infoheader[18] = 0;
	infoheader[19] = 0;
	//4 bytes -- ImageSize  (compressed) It is valid to set this =0 
	infoheader[20] = 0;
	infoheader[21] = 0;
	infoheader[22] = 0;
	infoheader[23] = 0;
	//XpixelsPerM 4 bytes horizontal resolution: Pixels/meter 
	infoheader[24] = 0;
	infoheader[25] = 0;
	infoheader[26] = 0;
	infoheader[27] = 0;
	//YpixelsPerM 4 bytes vertical resolution: Pixels/meter 
	infoheader[28] = 0;
	infoheader[29] = 0;
	infoheader[30] = 0;
	infoheader[31] = 0;
	//ColorsUsed 4 bytes Number of actually used colors =256
	infoheader[32] = 0;
	infoheader[33] = 1;
	infoheader[34] = 0;
	infoheader[35] = 0;
	//ColorsImportant 4 bytes Number of important colors  0 = all 
	infoheader[36] = 0;
	infoheader[37] = 0;
	infoheader[38] = 0;
	infoheader[39] = 0;
}

void sendtable(void){
	unsigned int i;
	for(i=0; i<256;i++){
		uart_send_byte(i);
		uart_send_byte(i);
		uart_send_byte(i);
		uart_send_byte(0);
	}
}

void photo(void){
	unsigned int i, j, k;
 	for (i = 0; i<352; i++) {
	  while (CAM_VSYN);
	  while (!CAM_VSYN);
	  for (j = 0; j<244; j++){
	    while (!CAM_HREF);
		for (k = 0; k<i; k++){ 
			while (CAM_PCLK);
			while (!CAM_PCLK);
	    }
		uart_send_byte(PINA);
		while (CAM_HREF);
		}
	}
}

void uart_parse_rx(unsigned char temp_char) {
	
	if (cam_state == CAM_STATE_IDLE) {
		if (temp_char == 0xAA) {
			cam_state = CAM_STATE_TX;
	    }
    } 
    else {
    	cam_reset();
    }
}

void cam_reset(void) {
	twi_read_register(TWI_CAM_WRITE_ADDRESS, TWI_CAM_READ_ADDRESS, 0x12, 1, twi_data_buffer);
	twi_data_buffer[0] |= 0b10000000;
	twi_write_register(TWI_CAM_WRITE_ADDRESS, 0x12, 1, twi_data_buffer);

	cam_state = CAM_STATE_IDLE;
}
              
void main(void) {
	// Declare local variables
	unsigned char preamble[4];
	unsigned char header[14];
	unsigned char infoheader[40];
	unsigned long size;
	
	// Initialize local variables
	size = (unsigned long) (14 + 40 + 1024 + (352*244));
	preamble[0] = size >> 24;
	preamble[1] = (size << 8) >> 24;
	preamble[2] = (size << 16) >> 24;
	preamble[3] = (size << 24) >> 24;
	createheader(header,352,244);
	createinfoheader(infoheader,352,244);
    
	// Initialize MCU modules
	init_uart(); 
	init_twi_master(CPU_FREQUENCY_HZ, TWI_PULLUP_RESISTOR_ENABLE);
	DDRA = 0x00;
	DDRD = (DDRD & 0xE3);
	
	#asm("sei");
	
	// Protoboard LED
	DDRD.7 = 1;
	PORTD.7 = 1;
	
	// Reset Camera
    cam_reset();
    
	while(1) {
		switch (cam_state) {
			case CAM_STATE_IDLE:
				//uart_send_byte(0x00);
				//delay_ms(500); 
				break;
			case CAM_STATE_TX:    
				PORTD.7 = 0;
				uart_send_byte_array(preamble,4);
				uart_send_byte_array(header,14);
		  		uart_send_byte_array(infoheader,40); 
		  	
				sendtable();   
				
				camera_write(0x11, 0x10);
				//#asm("cli");
				photo();
				//#asm("sei");
				camera_write(0x11, 0x00);    
				
				cam_state = CAM_STATE_IDLE;
				PORTD.7 = 1;
				break;
		}
	}
}

void init_uart(void){
	UBRR0L = 10;		// 38400 Baud Rate for 20 MHz Clock

	UCSR0A = 0b00000000;// 7 - UART RX complete flag
						// 6 - UART TX complete flag
						// 5 - UART DRE flag
						// 4 - Frame error flag
						// 3 - Data overrun flag
						// 2 - Parity error flag
						// 1 - Double UART speed mode
						// 0 - Multi processor communication mode

	UCSR0B = 0b10011000; // 7 - RX Complete Interrupt Enable
						// 6 - TX Complete Interrupt Enable
						// 5 - UART DRE Interrupt Enable
						// 4 - Enable receiver
						// 3 - Enable transmitter
						// 2 - Character size
						// 1:0 - Ninth data bit

	UCSR0C = 0b00000110;// 7:6 - UART Mode Select
						//		00 - Asynchronous UART
						//		01 - Synchronous UART
						//		10 - Reserved
						//		11 - Master SPI
						// 5:4 - Parity mode
						//		00 - Disabled
						//		01 - Reserved
						//		10 - Enabled, Even
						//		11 - Enabled, Odd
						// 3 - Select stop bit length
						//		0 - 1-bit
						//		1 - 2-bit
						// 2:1 - Character size, with UCSRB.2
						//		000 - 5 bit
						//		001 - 6 bit
						//		010 - 7 bit
						//		011 - 8 bit
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Reserved
						//		111 - 9 bit
						// 0 - Select clock polarity
						//		0 - Rising XCK edge
						//		1 - Falling XCK edge 						
}