/**
 * A Mega644P Program Space Flash w/ Internal Oscillator Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */

/* Includes */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <string.h>
#include <stdio.h>

/* Global/System Constants */
#define TRUE	1
#define FALSE	0

#define sbi(data,bit)			(data |= (1<<bit))	// Set Bit
#define cbi(data,bit) 			(data &= ~(1<<bit))	// Clear Bit
#define gbi(data,bit)			(data & (1<<bit))	// Get Bit
#define tbi(data,bit)			(data ^= (1<<bit))	// Toggle Bit

/* FLASH Variables */
uint8_t flash_const_data[11][10] PROGMEM =
{
        {0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09},
        {0x0A,0x0B,0x0C,0x0D,0x0E,0x0F,0x10,0x11,0x12,0x13},
        {0x14,0x15,0x16,0x17,0x18,0x19,0x1A,0x1B,0x1C,0x1D},
        {0x1E,0x1F,0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27},
        {0x28,0x29,0x2A,0x2B,0x2C,0x2D,0x2E,0x2F,0x30,0x31},
        {0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x3A,0x3B},
        {0x3C,0x3D,0x3E,0x3F,0x40,0x41,0x42,0x43,0x44,0x45},
        {0x46,0x47,0x48,0x49,0x4A,0x4B,0x4C,0x4D,0x4E,0x4F},
        {0x50,0x51,0x52,0x53,0x54,0x55,0x56,0x57,0x58,0x59},
        {0x5A,0x5B,0x5C,0x5D,0x5E,0x5F,0x60,0x61,0x62,0x63},
        {0x64,0x65,0x66,0x67,0x68,0x69,0x6A,0x6B,0x6C,0x6D}
};

/* Prototype Functions */
void init_mcu(void);          

/**
 * Main program
 */
int main(void) {
	DDRB = 0xFF;
	PORTB = 0xFF;

	// Main program loop
	while(1) {
		for (uint8_t i=0; i<11; i++) {
			for (uint8_t j=0; j<10; j++) {
				PORTB = ~pgm_read_byte(&(flash_const_data[i][j]));
				_delay_ms(500);
			}
		}
	}
}

void init_mcu(void) {
	SMCR = 0b00000000;	// 7:4 - Reserved bits
						// 3:1 - Sleep Mode
						//		000 - Idle
						//		001 - ADC Noise Reduction
						//		010 - Power-down
						//		011 - Power-save
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Standby
						//		111 - Extended Standby
						// 0 - Sleep Enable
						
	MCUCR |= (0 << 6);	// 6 - BOD Sleep
	MCUCR |= (0 << 5);	// 5 - BOD Sleep Enable
	
	PRR = 0b11111111;	// 7 - Power Reduction TWI
						// 6 - Power Reduction Timer2
						// 5 - Power Reduction Timer0
						// 4 - Power Reduction UART1
						// 3 - Power Reduction Timer1
						// 2 - Power Reduction SPI
						// 1 - Power Reduction UART0
						// 0 - Power Reduction ADC
}
