/**
 * A Mega644P PIR External Interrupt w/ Internal Oscillator Test
 *
 * - Low Level interrupt does not work without ISR
 * - Cannot wake from sleep without ISR
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/sleep.h>		// Include AVR-GCC sleep lib

/* Global Constants */
#define TRUE		1
#define FALSE		0

#define CPU_FREQUENCY_HZ	8000000

#define sbi(data,bit)		(data |= (1<<bit))	// Set Bit
#define cbi(data,bit) 		(data &= ~(1<<bit))	// Clear Bit
#define gbi(data,bit)		(data & (1<<bit))	// Get Bit
#define tbi(data,bit)		(data ^= (1<<bit))	// Toggle Bit

#define INT0_DDR	DDRD
#define INT0_PORT	PORTD
#define INT0_PIN	PD2

#define UART0_BAUD_RATE		9600
#define UART0_UBBR			((uint16_t) (((float)CPU_FREQUENCY_HZ) / (16*(float)UART0_BAUD_RATE) - 0.5))

void init_ext_int0(void);
void init_uart0(void);

/**
 * UART transmission function for a single byte
 */
void uart0_send_byte(uint8_t data) {
	while ( !( UCSR0A & _BV(UDRE0)) );	// Wait for empty buffer
	UDR0 = data;							// Put data into buffer, sends the data 
}

/**
 * Main program
 */
int main(void) {
	uint8_t int0_counter = 0;
	
	init_ext_int0();
	init_uart0();
	
	_delay_ms(100);		// Delay to stablize initilization
	
	// Main program loop
	while(1) {
		set_sleep_mode(SLEEP_MODE_IDLE);
		sleep_mode();
		_delay_ms(1);
		uart0_send_byte(int0_counter++);
		//if (EIFR & _BV(INTF0)) {
		//	EIFR = (EIFR & ~_BV(INTF0)) | _BV(INTF0);	// Clear the flag to make sure next interrupt is not missed
		//	uart0_send_byte(int0_counter++);
		//}
	}
	
	return 0;
}


void init_ext_int0(void) {
	EICRA = 0b00000010;	// 7:6 - Reserved Bits
						// 5:4 - ISC21, ISC20 - External Int 2 Sense Control Bits
						// 3:2 - ISC11, ISC10 - External Int 1 "
						// 1:0 - ISC01, ISC00 - External Int 0 "
						//	00 - Low level
						//	01 - Any edge
						//	10 - Falling edge
						//	11 - Rising edge
						
	EIMSK = 0b00000000;	// 7:3 - Reserved Bits
						// 2:0  INT2:0: External Interrupt Request 2 - 0 Enable
	
	EIFR = 0b00000000;	// 7:3 - Reserved Bits
						// 2:0  INTF2:0: External Interrupt Flags 2 - 0
						
	PCICR = 0b00000000;	// 7:4 - Reserved Bits
						// 3  PCIE3: Pin Change Interrupt Enable 3
						// 2  PCIE2: Pin Change Interrupt Enable 2
						// 1  PCIE1: Pin Change Interrupt Enable 1
						// 0  PCIE0: Pin Change Interrupt Enable 0
						
	PCIFR = 0b00000000;	// 7:4 - Reserved Bits
						// 3  PCIF3: Pin Change Interrupt Flag 3
						// 2  PCIF2: Pin Change Interrupt Flag 2
						// 1  PCIF1: Pin Change Interrupt Flag 1
						// 0  PCIF0: Pin Change Interrupt Flag 0
						
	PCMSK3 = 0b00000000;	// 7:0  PCINT31:24: Pin Change Enable Mask 31:24
	
	PCMSK2 = 0b00000000;	// 7:0  PCINT23:16: Pin Change Enable Mask 23:16
	
	PCMSK1 = 0b00000000;	// 7:0  PCINT15:8: Pin Change Enable Mask 15:8
	
	PCMSK0 = 0b00000000;	// 7:0  PCINT7:0: Pin Change Enable Mask 7:0
}

/**
 * Initialize UART0
 *
 * For thread-based operation
 */
void init_uart0(void) {	
	UBRR0H = (UART0_UBBR >> 8) & 0xFF;
	UBRR0L = UART0_UBBR & 0xFF;

	UCSR0A = 0b00000000;	// 7 - RXCn: UART RX complete flag
							// 6 - TXCn: UART TX complete flag
							// 5 - UDREn: UART DRE flag
							// 4 - FEn: Frame error flag
							// 3 - DORn: Data overrun flag
							// 2 - UPEn: Parity error flag
							// 1 - U2Xn: Double UART speed mode
							// 0 - MPCMn: Multi processor communication mode

	UCSR0B = 0b00001000; 	// 7 - RXCIEn: RX Complete Interrupt Enable
							// 6 - TXCIEn: TX Complete Interrupt Enable
							// 5 - UDRIEn: UART DRE Interrupt Enable
							// 4 - RXENn: Enable receiver
							// 3 - TXENn: Enable transmitter
							// 2 - UCSZn2: Character size
							// 1 - RXB8n: Receive Data Bit 8
							// 0 - TXB8n: Transmit Data Bit 8

	UCSR0C = 0b00000110;	// 7:6 - UMSELn1:0: UART Mode Select
							//		00 - Asynchronous UART
							//		01 - Synchronous UART
							//		10 - Reserved
							//		11 - Master SPI
							// 5:4 - UPMn1:0: Parity mode
							//		00 - Disabled
							//		01 - Reserved
							//		10 - Enabled, Even
							//		11 - Enabled, Odd
							// 3 - USBSn: Select stop bit length
							//		0 - 1-bit
							//		1 - 2-bit
							// 2:1 - UCSZn1:0: Character size, with UCSRB.2
							//		000 - 5 bit
							//		001 - 6 bit
							//		010 - 7 bit
							//		011 - 8 bit
							//		100 - Reserved
							//		101 - Reserved
							//		110 - Reserved
							//		111 - 9 bit
							// 0 - UCPOLn: Select clock polarity
							//		0 - Rising XCK edge
							//		1 - Falling XCK edge 						
}