/**
 * A Mega644P PIR External Interrupt w/ Internal Oscillator Test
 *
 * - Int0 wakes from power down sleep through ISR
 * - Low level interrupt is continuous
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/sleep.h>		// Include AVR-GCC sleep lib

/* Global Constants */
#define TRUE		1
#define FALSE		0

#define CPU_FREQUENCY_HZ	8000000

#define sbi(data,bit)		(data |= (1<<bit))	// Set Bit
#define cbi(data,bit) 		(data &= ~(1<<bit))	// Clear Bit
#define gbi(data,bit)		(data & (1<<bit))	// Get Bit
#define tbi(data,bit)		(data ^= (1<<bit))	// Toggle Bit

#define TIMER0_RES_US		1000
#define TIMER0_DIV			64
#define TIMER0_DIV_MASK		0b00000111
#define TIMER0_OCR0A		(uint8_t) (((float)CPU_FREQUENCY_HZ/(float)TIMER0_DIV) / ((float)1000000/(float)TIMER0_RES_US) + 0.5)

#define INT0_TRIGGER_LATENCY_THRESHOLD	200

#define INT0_DDR	DDRD
#define INT0_PORT	PORTD
#define INT0_PIN	PIND
#define INT0_BIT	PD2

#define UART0_BAUD_RATE		9600
#define UART0_UBBR			((uint16_t) (((float)CPU_FREQUENCY_HZ) / (16*(float)UART0_BAUD_RATE) - 0.5))

enum{MCU_IDLE, MCU_DETECT};
uint8_t mcu_state;

volatile uint8_t int0_trigger_enable_flag;
volatile uint8_t int0_counter;

void init_ext_int0(void);
void init_uart0(void);
void init_timer0(void);
void uart0_send_byte(uint8_t data);

ISR (INT0_vect) {
	cli();
}

//ISR (TIMER0_COMPA_vect) {
//	int0_trigger_latency_tick_count++;
//}

/**
 * UART transmission function for a single byte
 */
void uart0_send_byte(uint8_t data) {
	while ( !( UCSR0A & _BV(UDRE0)) );	// Wait for empty buffer
	UDR0 = data;							// Put data into buffer, sends the data 
}

/**
 * Main program
 */
int main(void) {
	int0_trigger_enable_flag = FALSE;
	int0_counter = 0;
	
	mcu_state = MCU_IDLE;
	
	//init_timer0();
	init_ext_int0();
	init_uart0();
	
	sei();		// Enable interrupts in assembly
	
	_delay_ms(100);		// Delay to stablize initilization
	
	/**
	 * If an enabled interrupt occurs while the MCU is in a sleep mode, the MCU wakes up. The MCU
	 * is then halted for four cycles in addition to the start-up time, executes the interrupt routine, 
	 * and resumes execution from the instruction following SLEEP. The contents of the Register File and
	 * SRAM are unaltered when the device wakes up from sleep. If a reset occurs during sleep mode,
	 * the MCU wakes up and executes from the Reset Vector.
	 */
	while(1) {
		/*
		switch(mcu_state) {
			case MCU_IDLE:
				uart0_send_byte(0xAA);
				set_sleep_mode(SLEEP_MODE_PWR_DOWN);
				sleep_mode();
				

				 
				mcu_state = MCU_DETECT;
				int0_trigger_enable_flag = TRUE;
				break;
			case MCU_DETECT:
				if (int0_trigger_enable_flag) {
					int0_trigger_enable_flag = FALSE;
					uart0_send_byte(int0_counter++);
				}
				mcu_state = MCU_IDLE;
				break;
		}
		*/
		
		/*
		uart0_send_byte(0xAA);
		set_sleep_mode(SLEEP_MODE_PWR_DOWN);
		sleep_mode();
		
		while(1) {
			uart0_send_byte(int0_counter++);
			_delay_ms(100);
		}
		*/
		
		/*
		sbi(EIMSK, 0);
		set_sleep_mode(SLEEP_MODE_PWR_DOWN);
		sleep_mode();
		cbi(EIMSK, 0);

		uart0_send_byte(int0_counter++);
		_delay_ms(100);
		*/
		
		sei();
		set_sleep_mode(SLEEP_MODE_PWR_DOWN);
		sleep_mode();

		uart0_send_byte(int0_counter++);
		_delay_ms(100);
	}
	
	return 0;
}

void init_ext_int0(void) {
	EICRA = 0b00000000;	// 7:6 - Reserved Bits
						// 5:4 - ISC21, ISC20 - External Int 2 Sense Control Bits
						// 3:2 - ISC11, ISC10 - External Int 1 "
						// 1:0 - ISC01, ISC00 - External Int 0 "
						//	00 - Low level
						//	01 - Any edge
						//	10 - Falling edge
						//	11 - Rising edge
						
	EIMSK = 0b00000001;	// 7:3 - Reserved Bits
						// 2:0  INT2:0: External Interrupt Request 2 - 0 Enable
	
	EIFR = 0b00000000;	// 7:3 - Reserved Bits
						// 2:0  INTF2:0: External Interrupt Flags 2 - 0
						
	PCICR = 0b00000000;	// 7:4 - Reserved Bits
						// 3  PCIE3: Pin Change Interrupt Enable 3
						// 2  PCIE2: Pin Change Interrupt Enable 2
						// 1  PCIE1: Pin Change Interrupt Enable 1
						// 0  PCIE0: Pin Change Interrupt Enable 0
						
	PCIFR = 0b00000000;	// 7:4 - Reserved Bits
						// 3  PCIF3: Pin Change Interrupt Flag 3
						// 2  PCIF2: Pin Change Interrupt Flag 2
						// 1  PCIF1: Pin Change Interrupt Flag 1
						// 0  PCIF0: Pin Change Interrupt Flag 0
						
	PCMSK3 = 0b00000000;	// 7:0  PCINT31:24: Pin Change Enable Mask 31:24
	
	PCMSK2 = 0b00000000;	// 7:0  PCINT23:16: Pin Change Enable Mask 23:16
	
	PCMSK1 = 0b00000000;	// 7:0  PCINT15:8: Pin Change Enable Mask 15:8
	
	PCMSK0 = 0b00000000;	// 7:0  PCINT7:0: Pin Change Enable Mask 7:0
}

/**
 * Initialize Timer0
 *
 * For interrupt-based CTC mode
 */
void init_timer0(void) {
	TCCR0A = 0b00000010;	// 7:6 - COM0A1:0: Compare Match Output A Mode
							// 5:4 - COM0B1:0: Compare Match Output B Mode
							// 3:2 - Reserved Bits
							// 1:0 - WGM01:0: Waveform Generation Mode
							//	000 - Normal (0xFF)
							//	001 - PWM, Phase Correct (0xFF)
							//	010 - CTC (OCRA)
							//	011 - Fast PWM (0xFF)
							//	100 - Reserved
							//	101 - PWM, Phase Correct (OCRA)
							//	110 - Reserved
							//	111 - Fast PWM (OCRA)
	
	TCCR0B = 0b00000100;	// 7 - FOC0A: Force Output Compare A
							// 6 - FOC0B: Force Output Compare B
							// 5:4 - Reserved Bits
							// 3 - WGM02: Waveform Generation Mode
							// 2:0 - CS02:0: Clock select
							//	000 - Timer stopped
							//	001 - 1
							//	010 - 8
							//	011 - 64
							//	100 - 256
							//	101 - 1024
							//	110 - External clock source on T0 pin, falling edge
							//	111 - ", rising edge
	
	TIMSK0 = 0b00000010;	// 7:3 - Reserved bits for other timer interrupts
							// 2 - OCIE0B: Timer0 Timer/Counter Output Compare Match B Interrupt Enable
							// 1 - OCIE0A: Timer0 Timer/Counter Output Compare Match A Interrupt Enable
							// 0 - TOIE0: Timer/Counter0 Overflow Interrupt Enable
							
	OCR0A = TIMER0_OCR0A;
}

/**
 * Initialize UART0
 *
 * For thread-based operation
 */
void init_uart0(void) {	
	UBRR0H = (UART0_UBBR >> 8) & 0xFF;
	UBRR0L = UART0_UBBR & 0xFF;

	UCSR0A = 0b00000000;	// 7 - RXCn: UART RX complete flag
							// 6 - TXCn: UART TX complete flag
							// 5 - UDREn: UART DRE flag
							// 4 - FEn: Frame error flag
							// 3 - DORn: Data overrun flag
							// 2 - UPEn: Parity error flag
							// 1 - U2Xn: Double UART speed mode
							// 0 - MPCMn: Multi processor communication mode

	UCSR0B = 0b00001000; 	// 7 - RXCIEn: RX Complete Interrupt Enable
							// 6 - TXCIEn: TX Complete Interrupt Enable
							// 5 - UDRIEn: UART DRE Interrupt Enable
							// 4 - RXENn: Enable receiver
							// 3 - TXENn: Enable transmitter
							// 2 - UCSZn2: Character size
							// 1 - RXB8n: Receive Data Bit 8
							// 0 - TXB8n: Transmit Data Bit 8

	UCSR0C = 0b00000110;	// 7:6 - UMSELn1:0: UART Mode Select
							//		00 - Asynchronous UART
							//		01 - Synchronous UART
							//		10 - Reserved
							//		11 - Master SPI
							// 5:4 - UPMn1:0: Parity mode
							//		00 - Disabled
							//		01 - Reserved
							//		10 - Enabled, Even
							//		11 - Enabled, Odd
							// 3 - USBSn: Select stop bit length
							//		0 - 1-bit
							//		1 - 2-bit
							// 2:1 - UCSZn1:0: Character size, with UCSRB.2
							//		000 - 5 bit
							//		001 - 6 bit
							//		010 - 7 bit
							//		011 - 8 bit
							//		100 - Reserved
							//		101 - Reserved
							//		110 - Reserved
							//		111 - 9 bit
							// 0 - UCPOLn: Select clock polarity
							//		0 - Rising XCK edge
							//		1 - Falling XCK edge 						
}