/**
 * A Mega644P PIR External Interrupt Timer2 Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/sleep.h>		// Include AVR-GCC sleep lib

/* Global Constants */
#define TRUE		1
#define FALSE		0

#define CPU_FREQUENCY_HZ	1000000

#define sbi(data,bit)		(data |= (1<<bit))	// Set Bit
#define cbi(data,bit) 		(data &= ~(1<<bit))	// Clear Bit
#define gbi(data,bit)		(data & (1<<bit))	// Get Bit
#define tbi(data,bit)		(data ^= (1<<bit))	// Toggle Bit

#define PROTOBOARD_LED_DDR	DDRD
#define PROTOBOARD_LED_PORT	PORTD
#define PROTOBOARD_LED_0	PD4
#define PROTOBOARD_LED_1	PD5

#define INT0_DDR	DDRD
#define INT0_PORT	PORTD
#define INT0_PIN	PIND
#define INT0_BIT	PD2

#define UART0_BAUD_RATE		4800
#define UART0_UBBR			((uint16_t) (((float)CPU_FREQUENCY_HZ) / (16*(float)UART0_BAUD_RATE) - 0.5))

/* Timer Constants */
#define TIMER_DIV_0			0
#define TIMER_DIV_1			1
#define TIMER_DIV_8			8
#define TIMER_DIV_32		32	// Timer2 Only
#define TIMER_DIV_64		64
#define TIMER_DIV_128		128	// Timer2 Only
#define TIMER_DIV_256		256
#define TIMER_DIV_1024		1024

/* Timer2 Constants */
#define TIMER2A_ISR_ENABLE	TRUE
#define TIMER2B_ISR_ENABLE	FALSE
#define TIMER2A_RES_US		200000
#define TIMER2B_RES_US		32
#define TIMER2_DIV			TIMER_DIV_1024	// Set Timer2 divider
#define TIMER2_DIV_MASK		0b00000111	// Define Timer2 divider register mask
#define TIMER2_OCR2A		(uint8_t) (((float)CPU_FREQUENCY_HZ/(float)TIMER2_DIV) / ((float)1000000/(float)TIMER2A_RES_US) + 0.5)
#define TIMER2_OCR2B		(uint8_t) (((float)CPU_FREQUENCY_HZ/(float)TIMER2_DIV) / ((float)1000000/(float)TIMER2B_RES_US) + 0.5)
#define TIMER2_DDR			DDRD
#define TIMER2_PIN			PIND
#define TIMER2A_OUTPUT		PD7
#define TIMER2B_OUTPUT		PD6

#define TIMER0A_RES_MS					200
#define SENSOR_LATENCY_MS				10000
#define SENSOR_LATENCY_THRESHOLD		(uint16_t) (SENSOR_LATENCY_MS / TIMER0A_RES_MS)

volatile uint8_t sensor_detection_flag;
volatile uint16_t sensor_latency_tick_count;
uint8_t sensor_latency_enable_flag;

void init_ext_int0(void);
void init_uart0(void);
void uart0_send_byte(uint8_t data);
void init_timer2(void);

ISR (INT1_vect) {
	cbi(EIMSK, 1);
	sensor_detection_flag = TRUE;
}

ISR (TIMER2_COMPA_vect) {
	sensor_latency_tick_count++;
}

/**
 * UART transmission function for a single byte
 */
void uart0_send_byte(uint8_t data) {
	while ( !( UCSR0A & _BV(UDRE0)) );	// Wait for empty buffer
	UDR0 = data;							// Put data into buffer, sends the data 
}

/**
 * Main program
 */
int main(void) {
	sensor_latency_tick_count = 0;
	sensor_latency_enable_flag = FALSE;

	sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_0);
	sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_1);
	sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
	sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);

	init_ext_int0();
	//init_uart0();
	init_timer2();
	
	sei();		// Enable interrupts in assembly
	
	_delay_ms(100);		// Delay to stablize initilization
	
	cbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
	cbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);
	
	/**
	 * If an enabled interrupt occurs while the MCU is in a sleep mode, the MCU wakes up. The MCU
	 * is then halted for four cycles in addition to the start-up time, executes the interrupt routine, 
	 * and resumes execution from the instruction following SLEEP. The contents of the Register File and
	 * SRAM are unaltered when the device wakes up from sleep. If a reset occurs during sleep mode,
	 * the MCU wakes up and executes from the Reset Vector.
	 */
	while(1) {
		// Works
		/*
		set_sleep_mode(SLEEP_MODE_PWR_SAVE);
		sleep_mode();
		
		if (sensor_latency_tick_count > SENSOR_LATENCY_THRESHOLD) {
			sensor_latency_tick_count = 0;

			tbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
		}
		*/
		
		// Works
		/*
		sbi(EIMSK, 1);
		set_sleep_mode(SLEEP_MODE_PWR_SAVE);
		sleep_mode();
		
		if (sensor_detection_flag) {
			sensor_detection_flag = FALSE;
			
			tbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);
			
			if (sensor_latency_tick_count > SENSOR_LATENCY_THRESHOLD) {
				sensor_latency_tick_count = 0;

				tbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
			}
		}
		*/
		
		// Works
		/*
		if (sensor_latency_enable_flag) {
			set_sleep_mode(SLEEP_MODE_PWR_SAVE);	
			
		}
		else {
			sbi(EIMSK, 1);
			set_sleep_mode(SLEEP_MODE_PWR_DOWN);
		}
		sleep_mode();
		
		if (sensor_detection_flag & !sensor_latency_enable_flag) {
			sensor_detection_flag = FALSE;
			
			tbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
			
			sensor_latency_enable_flag = TRUE;
			sensor_latency_tick_count = 0;
			sbi(TIMSK2, OCIE2A);
		}
		
		if (sensor_latency_tick_count > SENSOR_LATENCY_THRESHOLD) {
			sensor_latency_tick_count = 0;
			
			tbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);
			
			cbi(TIMSK2, OCIE2A);
			sensor_latency_enable_flag = FALSE;
			sbi(EIMSK, 1);
		}
		*/
		
		if (sensor_latency_enable_flag) {
			sbi(TIMSK2, OCIE2A);
			set_sleep_mode(SLEEP_MODE_PWR_SAVE);	
			
		}
		else {
			sbi(EIMSK, 1);
			set_sleep_mode(SLEEP_MODE_PWR_DOWN);
		}
		sleep_mode();
		
		if (sensor_detection_flag) {
			sensor_detection_flag = FALSE;
			
			tbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
			
			sensor_latency_enable_flag = TRUE;
		}
		
		if (sensor_latency_tick_count > SENSOR_LATENCY_THRESHOLD) {
			sensor_latency_tick_count = 0;
			
			tbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);
			
			sensor_latency_enable_flag = FALSE;
			cbi(TIMSK2, OCIE2A);
			sbi(EIMSK, 1);
		}
		
	}
	
	return 0;
}

void init_ext_int0(void) {
	EICRA = 0b00001000;	// 7:6 - Reserved Bits
						// 5:4 - ISC21, ISC20 - External Int 2 Sense Control Bits
						// 3:2 - ISC11, ISC10 - External Int 1 "
						// 1:0 - ISC01, ISC00 - External Int 0 "
						//	00 - Low level
						//	01 - Any edge
						//	10 - Falling edge
						//	11 - Rising edge
						
	EIMSK = 0b00000010;	// 7:3 - Reserved Bits
						// 2:0  INT2:0: External Interrupt Request 2 - 0 Enable
	
	EIFR = 0b00000000;	// 7:3 - Reserved Bits
						// 2:0  INTF2:0: External Interrupt Flags 2 - 0
						
	PCICR = 0b00000000;	// 7:4 - Reserved Bits
						// 3  PCIE3: Pin Change Interrupt Enable 3
						// 2  PCIE2: Pin Change Interrupt Enable 2
						// 1  PCIE1: Pin Change Interrupt Enable 1
						// 0  PCIE0: Pin Change Interrupt Enable 0
						
	PCIFR = 0b00000000;	// 7:4 - Reserved Bits
						// 3  PCIF3: Pin Change Interrupt Flag 3
						// 2  PCIF2: Pin Change Interrupt Flag 2
						// 1  PCIF1: Pin Change Interrupt Flag 1
						// 0  PCIF0: Pin Change Interrupt Flag 0
						
	PCMSK3 = 0b00000000;	// 7:0  PCINT31:24: Pin Change Enable Mask 31:24
	
	PCMSK2 = 0b00000000;	// 7:0  PCINT23:16: Pin Change Enable Mask 23:16
	
	PCMSK1 = 0b00000000;	// 7:0  PCINT15:8: Pin Change Enable Mask 15:8
	
	PCMSK0 = 0b00000000;	// 7:0  PCINT7:0: Pin Change Enable Mask 7:0
}

/**
 * Initialize UART0
 *
 * For thread-based operation
 */
void init_uart0(void) {	
	UBRR0H = (UART0_UBBR >> 8) & 0xFF;
	UBRR0L = UART0_UBBR & 0xFF;

	UCSR0A = 0b00000000;	// 7 - RXCn: UART RX complete flag
							// 6 - TXCn: UART TX complete flag
							// 5 - UDREn: UART DRE flag
							// 4 - FEn: Frame error flag
							// 3 - DORn: Data overrun flag
							// 2 - UPEn: Parity error flag
							// 1 - U2Xn: Double UART speed mode
							// 0 - MPCMn: Multi processor communication mode

	UCSR0B = 0b00001000; 	// 7 - RXCIEn: RX Complete Interrupt Enable
							// 6 - TXCIEn: TX Complete Interrupt Enable
							// 5 - UDRIEn: UART DRE Interrupt Enable
							// 4 - RXENn: Enable receiver
							// 3 - TXENn: Enable transmitter
							// 2 - UCSZn2: Character size
							// 1 - RXB8n: Receive Data Bit 8
							// 0 - TXB8n: Transmit Data Bit 8

	UCSR0C = 0b00000110;	// 7:6 - UMSELn1:0: UART Mode Select
							//		00 - Asynchronous UART
							//		01 - Synchronous UART
							//		10 - Reserved
							//		11 - Master SPI
							// 5:4 - UPMn1:0: Parity mode
							//		00 - Disabled
							//		01 - Reserved
							//		10 - Enabled, Even
							//		11 - Enabled, Odd
							// 3 - USBSn: Select stop bit length
							//		0 - 1-bit
							//		1 - 2-bit
							// 2:1 - UCSZn1:0: Character size, with UCSRB.2
							//		000 - 5 bit
							//		001 - 6 bit
							//		010 - 7 bit
							//		011 - 8 bit
							//		100 - Reserved
							//		101 - Reserved
							//		110 - Reserved
							//		111 - 9 bit
							// 0 - UCPOLn: Select clock polarity
							//		0 - Rising XCK edge
							//		1 - Falling XCK edge 						
}

/**
 * Initialize Timer2
 *
 * For interrupt-based CTC mode
 */
void init_timer2(void) {
	cbi(PRR, PRTIM2);

	TCCR2A = 0b00000010;	// 7:6  COM2A1:0: Compare Match Output A Mode
							// 5:4  COM2B1:0: Compare Match Output B Mode
							// 3:2  Reserved Bits
							// 1:0  WGM21:0: Waveform Generation Mode
	/*
	sbi(TIMER2_DDR,TIMER2A_OUTPUT);
	sbi(TCCR2A,6);
	*/
		
	TCCR2B = 0b00000000;	// 7  FOC2A: Force Output Compare A
							// 6  FOC2B: Force Output Compare B
							// 5:4  Reserved Bits
							// 3  WGM22: Waveform Generation Mode
							// 2:0  CS22:0: Clock Select
							//	000 - No clk
							//	001 - 1
							//	010 - 8
							//	011 - 32
							//	100 - 64
							//	101 - 128
							//	110 - 256
							//	111 - 1024
	
	#if TIMER2_DIV == TIMER_DIV_0
		TCCR2B = (TCCR2B & ~TIMER2_DIV_MASK) | 0;
	#elif TIMER2_DIV == TIMER_DIV_1
		TCCR2B = (TCCR2B & ~TIMER2_DIV_MASK) | 1;
	#elif TIMER2_DIV == TIMER_DIV_8
		TCCR2B = (TCCR2B & ~TIMER2_DIV_MASK) | 2;
	#elif TIMER2_DIV == TIMER_DIV_32
		TCCR2B = (TCCR2B & ~TIMER2_DIV_MASK) | 3;
	#elif TIMER2_DIV == TIMER_DIV_64
		TCCR2B = (TCCR2B & ~TIMER2_DIV_MASK) | 4;
	#elif TIMER2_DIV == TIMER_DIV_128
		TCCR2B = (TCCR2B & ~TIMER2_DIV_MASK) | 5;
	#elif TIMER2_DIV == TIMER_DIV_256
		TCCR2B = (TCCR2B & ~TIMER2_DIV_MASK) | 6;
	#elif TIMER2_DIV == TIMER_DIV_1024
		TCCR2B = (TCCR2B & ~TIMER2_DIV_MASK) | 7;
	#endif
	
	
	ASSR = 0b00000000;		// 6  EXCLK: Enable External Clock Input
							// 5  AS2: Asynchronous Timer/Counter2
							// 4  TCN2UB: Timer/Counter2 Update Busy
							// 3  OCR2AUB: Output Compare Register2 Update Busy
							// 2  OCR2BUB: Output Compare Register2 Update Busy
							// 1  TCR2AUB: Timer/Counter Control Register2 Update Busy
							// 0  TCR2BUB: Timer/Counter Control Register2 Update Busy
	
	TIMSK2 = 0b00000000;	// 7:3 - Reserved Bits
							// 2 - OCIE2B: Timer/Counter2 Output Compare Match B Interrupt Enable
							// 1  OCIE2A: Timer/Counter2 Output Compare Match A Interrupt Enable
							// 0  TOIE2: Timer/Counter2 Overflow Interrupt Enable
	
	OCR2A = TIMER2_OCR2A;
	OCR2B = TIMER2_OCR2B;
	
	#if TIMER2A_ISR_ENABLE
		TIMSK2 |= _BV(OCIE2A);
	#endif
	
	#if TIMER2B_ISR_ENABLE
		TIMSK2 |= _BV(OCIE2B);
	#endif		
}