/**
 * A Mega644P UART RXTX w/ Internal Oscillator Alternative Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */

/* Includes */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>
#include <stdio.h>

/* Global/System Constants */
#define TRUE	1
#define FALSE	0

#define sbi(data,bit)			(data |= (1<<bit))	// Set Bit
#define cbi(data,bit) 			(data &= ~(1<<bit))	// Clear Bit
#define gbi(data,bit)			(data & (1<<bit))	// Get Bit
#define tbi(data,bit)			(data ^= (1<<bit))	// Toggle Bit

/* PC/MCU Communication Constants */
#define PC_DISCOVER_REQUEST	0x1A
#define MCU_DISCOVER_REPLY	0x2A	

/* Global Variables */
uint8_t mcu_state;
enum {MCU_STATE_IDLE, MCU_STATE_TX};

/* Prototype Functions */
void init_mcu(void);
void init_uart0_rxtx(void); 
void uart0_send_byte_array(uint8_t *data, uint8_t length);
void uart0_send_byte(uint8_t data);
void uart0_parse_rx(uint8_t rx_data);            

/**
 * UART transmission function for byte arrays
 *
 * Can be used to transmitt strings
 * Length terminated (versus null terminated)
 */
void uart0_send_byte_array(uint8_t *data, uint8_t length) {
	for (uint8_t i=0; i<length; i++) {
		while ( !( UCSR0A & (1<<UDRE0)) );	// Wait for empty buffer
		UDR0 = *data++;						// Put data into buffer, sends the data 
	}
}

/**
 * UART transmission function for a single byte
 */
void uart0_send_byte(uint8_t data) {
	while ( !( UCSR0A & (1<<UDRE0)) );	// Wait for empty buffer
 	UDR0 = data;							// Put data into buffer, sends the data 
}

void uart0_parse_rx(uint8_t rx_data) {
	if (mcu_state == MCU_STATE_IDLE) {
		if (rx_data == PC_DISCOVER_REQUEST) {
			mcu_state = MCU_STATE_TX;
	    }
    }
}

/**
 * Main program
 */
int main(void) {
	mcu_state = MCU_STATE_IDLE;
	
	init_mcu();
	init_uart0_rxtx();		// Initialize UART TX
	
	_delay_ms(100);		// Delay to stablize initilization

	// Main program loop
	while(1) {
		switch (mcu_state) {
			case MCU_STATE_IDLE:
				if (UCSR0A & (1<<RXC0)) {
					uart0_parse_rx(UDR0);
				}
				break;
			case MCU_STATE_TX:
				uart0_send_byte(MCU_DISCOVER_REPLY);
				mcu_state = MCU_STATE_IDLE;
				break;
		}
		
	}
}

/**
 * Initialize UART RX
 */
void init_uart0_rxtx(void){
	UBRR0L = 12;		// 38400 Baud Rate for 8 MHz Clock

	UCSR0A = 0b00000000;// 7 - UART RX complete flag
						// 6 - UART TX complete flag
						// 5 - UART DRE flag
						// 4 - Frame error flag
						// 3 - Data overrun flag
						// 2 - Parity error flag
						// 1 - Double UART speed mode
						// 0 - Multi processor communication mode

	UCSR0B = 0b00011000; // 7 - RX Complete Interrupt Enable
						// 6 - TX Complete Interrupt Enable
						// 5 - UART DRE Interrupt Enable
						// 4 - Enable receiver
						// 3 - Enable transmitter
						// 2 - Character size
						// 1:0 - Ninth data bit

	UCSR0C = 0b00000110;// 7:6 - UART Mode Select
						//		00 - Asynchronous UART
						//		01 - Synchronous UART
						//		10 - Reserved
						//		11 - Master SPI
						// 5:4 - Parity mode
						//		00 - Disabled
						//		01 - Reserved
						//		10 - Enabled, Even
						//		11 - Enabled, Odd
						// 3 - Select stop bit length
						//		0 - 1-bit
						//		1 - 2-bit
						// 2:1 - Character size, with UCSRB.2
						//		000 - 5 bit
						//		001 - 6 bit
						//		010 - 7 bit
						//		011 - 8 bit
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Reserved
						//		111 - 9 bit
						// 0 - Select clock polarity
						//		0 - Rising XCK edge
						//		1 - Falling XCK edge 						
}

void init_mcu(void) {
	SMCR = 0b00000000;	// 7:4 - Reserved bits
						// 3:1 - Sleep Mode
						//		000 - Idle
						//		001 - ADC Noise Reduction
						//		010 - Power-down
						//		011 - Power-save
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Standby
						//		111 - Extended Standby
						// 0 - Sleep Enable
						
	MCUCR |= (0 << 6);	// 6 - BOD Sleep
	MCUCR |= (0 << 5);	// 5 - BOD Sleep Enable
	
	PRR = 0b11111101;	// 7 - Power Reduction TWI
						// 6 - Power Reduction Timer2
						// 5 - Power Reduction Timer0
						// 4 - Power Reduction UART1
						// 3 - Power Reduction Timer1
						// 2 - Power Reduction SPI
						// 1 - Power Reduction UART0
						// 0 - Power Reduction ADC
}
