/**
 * A Mega644P UART TX SPI-based ADC w/ Internal Oscillator Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 * MCP3204
 * 		_______
 * CH0	|1	14|	VDD
 * CH1	|2	13|	VREF
 * CH2	|3	12|	AGND
 * CH3	|4	11|	CLK
 * NC	|5	10|	DOUT
 * NC	|6	 9|	DIN
 * DGND	|7	 8|	/CS/SHDN
 *		-------
 */

/* Includes */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>
#include <stdio.h>

/* Global/System Constants */
#define TRUE	1
#define FALSE	0	

#define sbi(data,bit)			(data |= (1<<bit))	// Set Bit
#define cbi(data,bit) 			(data &= ~(1<<bit))	// Clear Bit
#define gbi(data,bit)			(data & (1<<bit))	// Get Bit
#define tbi(data,bit)			(data ^= (1<<bit))	// Toggle Bit

/* SPI Constants */
#define SPI_PORT	PORTB
#define SPI_SEL		4

/* Global Variables */
uint8_t counter;

/* Prototype Functions */
void init_mcu(void);
void init_uart_tx(void);
void init_adc_spi(void);                 

/**
 * UART transmission function for byte arrays
 *
 * Can be used to transmitt strings
 * Length terminated (versus null terminated)
 */
void uart_send_byte_array(uint8_t *data, uint8_t length) {
	for (uint8_t i=0; i<length; i++) {
		while ( !( UCSR0A & (1<<UDRE0)) );	// Wait for empty buffer
		UDR0 = *data++;						// Put data into buffer, sends the data 
	}
}

/**
 * UART transmission function for a single byte
 */
void uart_send_byte(uint8_t data) {
	while ( !( UCSR0A & (1<<UDRE0)) );	// Wait for empty buffer
 	UDR0 = data;							// Put data into buffer, sends the data 
}

uint8_t spi_send_receive_byte(uint8_t data) {
	SPDR = data;
	
	while(!(SPSR & _BV(SPIF)));

	return SPDR;
}

/**
 *
 *
 */
uint16_t spi_adc(uint8_t chan) {
	uint16_t data;
	
	cbi(SPI_PORT, SPI_SEL);
	spi_send_receive_byte(0b00000110);	// Sending Start | Sgl/Diff
	data = (spi_send_receive_byte(chan << 6) & 0b00001111) << 8;
	data |= spi_send_receive_byte(0b00000000);
	sbi(SPI_PORT, SPI_SEL);
	
	return data;
}

/**
 * Main program
 */
int main(void) {
	uint8_t adc_chan = 0;
	uint16_t adc_data = 0;

	init_mcu();
	init_uart_tx();		// Initialize UART TX
	init_adc_spi();
	
	_delay_ms(100);		// Delay to stablize initilization
	
	counter = 0;
	
	// Main program loop
	while(1) {
		for (adc_chan = 0; adc_chan < 3; adc_chan++) {
			adc_data = spi_adc(adc_chan);
			uart_send_byte(adc_data >> 8);
			uart_send_byte((adc_data << 8) >> 8);
			_delay_ms(100);
		}
	}
}

/**
 * Initialize UART RX
 */
void init_uart_tx(void){
	UBRR0L = 25;		// 19200 Baud Rate for 8 MHz Clock

	UCSR0A = 0b00000000;// 7 - UART RX complete flag
						// 6 - UART TX complete flag
						// 5 - UART DRE flag
						// 4 - Frame error flag
						// 3 - Data overrun flag
						// 2 - Parity error flag
						// 1 - Double UART speed mode
						// 0 - Multi processor communication mode

	UCSR0B = 0b00001000; // 7 - RX Complete Interrupt Enable
						// 6 - TX Complete Interrupt Enable
						// 5 - UART DRE Interrupt Enable
						// 4 - Enable receiver
						// 3 - Enable transmitter
						// 2 - Character size
						// 1:0 - Ninth data bit

	UCSR0C = 0b00000110;// 7:6 - UART Mode Select
						//		00 - Asynchronous UART
						//		01 - Synchronous UART
						//		10 - Reserved
						//		11 - Master SPI
						// 5:4 - Parity mode
						//		00 - Disabled
						//		01 - Reserved
						//		10 - Enabled, Even
						//		11 - Enabled, Odd
						// 3 - Select stop bit length
						//		0 - 1-bit
						//		1 - 2-bit
						// 2:1 - Character size, with UCSRB.2
						//		000 - 5 bit
						//		001 - 6 bit
						//		010 - 7 bit
						//		011 - 8 bit
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Reserved
						//		111 - 9 bit
						// 0 - Select clock polarity
						//		0 - Rising XCK edge
						//		1 - Falling XCK edge 						
}

/**
 * It is also required that the microcontroller 
 * SPI port be configured to clock out data 
 * on the falling edge of clock and latch data 
 * in on the rising edge.
 *	=> Can operate in either 1,1 or 0,0 mode
 *
 * Provide more clocks than are required by 
 * sending leading zeros before the start bit.
 */
void init_adc_spi(void) {
	sbi(DDRB,7);
	cbi(DDRB,6);
	sbi(DDRB,5);
	sbi(DDRB,4);
	
	sbi(SPI_PORT, SPI_SEL);

	SPCR = 0b01011100;	// 7  SPIE: SPI Interrupt Enable
						// 6  SPE: SPI Enable
						// 5  DORD: Data Order
						// 4  MSTR: Master/Slave Select
						// 3  CPOL: Clock Polarity
						// 2  CPHA: Clock Phase
						// 1:0  SPR1, SPR0: SPI Clock Rate Select 1 and 0
						//	2X	SPR	SCK Freq
						//	0	00	4
						//	0	01	16
						//	0	10	64
						//	0	11	128
						//	1	00	2
						//	1	01	8
						//	1	10	32
						//	1	11	64
						
	SPSR = 0b00000000;	// 7  SPIF: SPI Interrupt Flag
						// 6  WCOL: Write COLlision Flag
						// 5:1  Reserved Bits
						// 0  SPI2X: Double SPI Speed Bit
}

void init_mcu(void) {
	SMCR = 0b00000000;	// 7:4 - Reserved bits
						// 3:1 - Sleep Mode
						//		000 - Idle
						//		001 - ADC Noise Reduction
						//		010 - Power-down
						//		011 - Power-save
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Standby
						//		111 - Extended Standby
						// 0 - Sleep Enable
						
	cbi(MCUCR,6);	// 6 - BOD Sleep
	cbi(MCUCR,5);	// 5 - BOD Sleep Enable
	
	PRR = 0b11111001;	// 7 - Power Reduction TWI
						// 6 - Power Reduction Timer2
						// 5 - Power Reduction Timer0
						// 4 - Power Reduction UART1
						// 3 - Power Reduction Timer1
						// 2 - Power Reduction SPI
						// 1 - Power Reduction UART0
						// 0 - Power Reduction ADC
}
