/**
 * A Mega644P ZigBee SPI w/ Ext Oscillator Library Alternative Test
 * -Remove use of ISR for detecting Zigbee interrupts
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */

/* Global/System Constants */
#define TRUE 				1
#define FALSE 				0

#define CPU_FREQUENCY_HZ	8000000

#define sbi(data,bit)		(data |= (1<<bit))	// Set Bit
#define cbi(data,bit) 		(data &= ~(1<<bit))	// Clear Bit
#define gbi(data,bit)		(data & (1<<bit))	// Get Bit
#define tbi(data,bit)		(data ^= (1<<bit))	// Toggle Bit

/* Zigbee Constants */
#define ZIGBEE_TX			FALSE
#define ZIGBEE_RX			TRUE

/* Debug Constants */
#define PROTOBOARD_LED_DEBUG	TRUE
#define STK500_LED_DEBUG		FALSE

#define UART0_PREPROCESSOR_DEBUG		FALSE
#define UART0_ZIGBEE_SPEC_DEBUG			FALSE
#define UART0_ZIGBEE_IRQ_STATUS_DEBUG	TRUE
#define UART0_ZIGBEE_STATE_DEBUG		FALSE
#define UART0_ZIGBEE_RX_DATA_DEBUG		TRUE
#define UART0_ZIGBEE_RX_LQI_DEBUG		TRUE
#define UART0_ZIGBEE_CCA_DEBUG			TRUE

/* UART Protocol */
#define UART_TX_MCU_PREPROCESSOR		0x10
#define UART_TX_ZIGBEE_STATE			0x20
#define UART_TX_ZIGBEE_RX_DATA			0x30
#define UART_TX_ZIGBEE_RX_LQI			0x31
#define UART_TX_ZIGBEE_IRQ_STATUS		0x40

#define UART_TX_ZIGBEE_CCA_MODE			0x50
#define UART_TX_ZIGBEE_CCA_ED_THRES		0x51
#define UART_TX_ZIGBEE_CCA_VALUE		0x52

#define UART_TX_ZIGBEE_PART_NUM			0xE0
#define UART_TX_ZIGBEE_VER_NUM			0xE1
#define UART_TX_ZIGBEE_MAN_ID_0			0xE2
#define UART_TX_ZIGBEE_MAN_ID_1			0xE3

/* Protoboard Constants */
#define PROTOBOARD_LED_DDR	DDRD
#define PROTOBOARD_LED_PORT	PORTD
#define PROTOBOARD_LED_0	PD4
#define PROTOBOARD_LED_1	PD5

/* External Interrupt Constants */
#define INT0_FLAG_MASK		0b11111110
#define INT0_DDR			DDRD
#define INT0_PORT			PORTD
#define INT0_PIN			PD2

/* Includes */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>
#include <stdio.h>
#include "uart_tx.h"
#include "spi.h"
#include "zigbee.h"

/* Global Variables */

/* Volatile Constants */

/* Function Prototypes */
void init_mcu(void);
void init_ext_int0(void);
void init_adc(void);

uint8_t get_sw_crc_gen(uint8_t *tx_data, uint8_t tx_data_length) {
	uint8_t tx_crc_value = 0;
	for (uint8_t i=0; i<tx_data_length; i++) {
		tx_crc_value ^= tx_data[i];
	}
	return tx_crc_value;
}

uint8_t sw_crc_check(uint8_t *rx_data, uint8_t rx_data_length, uint8_t rx_data_crc) {
	uint8_t rx_crc_value = 0;
	for (uint8_t i=0; i<rx_data_length; i++) {
		rx_crc_value ^= rx_data[i];
	}
	if (rx_crc_value == rx_data_crc) {
		return TRUE;
	}
	return FALSE;
}

/**
 *
 */
int main(void) {
	/* Local Variables */
	
	/* Global Variable Initialization */
	zigbee_status = ZIGBEE_TRX_STATUS_P_ON;
	zigbee_clkm = ZIGBEE_CLKM_1MHZ;
	zigbee_channel = ZIGBEE_CHANNEL_11;
	
	zigbee_crc_enable_flag = FALSE;
	
	zigbee_irq_pll_lock_flag = FALSE;
	zigbee_irq_trx_end_flag = FALSE;
	zigbee_irq_rx_start_flag = FALSE;
	zigbee_irq_trx_ur_flag = FALSE;
	zigbee_irq_bat_low_flag = FALSE;

	/* Set PORT for LED Functionality */
	#if PROTOBOARD_LED_DEBUG
		sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_0);
		sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_1);
		sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
		sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);
	#endif

	/* MCU Initialization */
	init_mcu();			// Initialize MCU
	init_uart0_tx();	// Initialize UART
	init_spi_master();	// Initialize SPI Master
	init_ext_int0();	// Initialize External Int0
    
    /* Enable Interrupts */
    //sei();
	
    /**
	 * ZigBee Hardware Initialization 
	 */
	sbi(ZIGBEE_DDR, ZIGBEE_SLP_TR);		// Set PORT for ZigBee Functionality
	sbi(ZIGBEE_DDR, ZIGBEE_RST);
	zigbee_init();		// Hardware initialization after power-on event
   	zigbee_set_clkm(ZIGBEE_CLKM_8MHZ, ZIGBEE_CLKM_SHA_IMM);	// Set the Master CLK from the Zigbee Transceiver at the proper freq
	zigbee_clkm = ZIGBEE_CLKM_8MHZ;
	
	//zigbee_set_crc(TRUE);
	
	/* UART0 Init Debug */
	#if UART0_ZIGBEE_STATE_DEBUG
		uart0_send_byte(UART_TX_ZIGBEE_STATE);
		uart0_send_byte(zigbee_state);				// Should be in TRX_OFF state
	#endif
	
	#if UART0_PREPROCESSOR_DEBUG
		uart0_send_byte(UART_TX_MCU_PREPROCESSOR);
		uart0_send_byte(2);
		uart0_send_byte(UART0_UBBR0>>8);
		uart0_send_byte((UART0_UBBR0<<8)>>8);
	#endif
	
	#if UART0_ZIGBEE_SPEC_DEBUG
		uart0_send_byte(UART_TX_ZIGBEE_PART_NUM);
		uart0_send_byte(zigbee_hw_part_num);
		uart0_send_byte(UART_TX_ZIGBEE_VER_NUM);
		uart0_send_byte(zigbee_hw_ver_num);
		uart0_send_byte(UART_TX_ZIGBEE_MAN_ID_0);
		uart0_send_byte(zigbee_hw_man_id_0);
		uart0_send_byte(UART_TX_ZIGBEE_MAN_ID_1);
		uart0_send_byte(zigbee_hw_man_id_1);
	#endif
	
	#if PROTOBOARD_LED_DEBUG
		cbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
	#endif
	
	
	
	while(1) {
		zigbee_status = zigbee_get_trx_status();
		
		#if ZIGBEE_TX
			if (zigbee_man_cca()) {
				zigbee_tx_data[0]++;
				zigbee_auto_basic_tx(zigbee_tx_data, zigbee_tx_data_length);
				_delay_ms(200);
			}
			
			/*
			zigbee_enable_basic_tx();
			zigbee_tx_data[0]++;
			zigbee_man_basic_tx(zigbee_tx_data, zigbee_tx_data_length);
			zigbee_enable_trx_off();
			_delay_ms(200);
			*/
		#elif ZIGBEE_RX
			/*
			zigbee_auto_basic_rx(zigbee_rx_data, &zigbee_rx_data_length, &zigbee_rx_data_lqi);
			if (zigbee_rx_data_lqi > 0) {
				#if UART0_ZIGBEE_RX_LQI_DEBUG
					uart0_send_byte(UART_TX_ZIGBEE_RX_LQI);
					uart0_send_byte(zigbee_rx_data_lqi);
				#endif
				#if UART0_ZIGBEE_RX_DATA_DEBUG
					uart0_send_byte(UART_TX_ZIGBEE_RX_DATA);
					uart0_send_byte(zigbee_rx_data_length);
					uart0_send_byte_array(zigbee_rx_data, zigbee_rx_data_length);
				#endif
			}
			*/
			
			if (zigbee_status != ZIGBEE_TRX_STATUS_RX_ON && zigbee_status != ZIGBEE_TRX_STATUS_BUSY_RX) {
				zigbee_enable_basic_rx();
			}
			if (EIFR & _BV(INTF0)) {
				EIFR = (EIFR & INT0_FLAG_MASK) | _BV(INTF0);	// Clear the flag to make sure next interrupt is not missed
				zigbee_update_irq_status(); 					// ZIGBEE_IRQ_MASK_RX_START
																// ZIGBEE_IRQ_MASK_TRX_END
			}
			if (zigbee_irq_rx_start_flag && zigbee_irq_trx_end_flag) {
				zigbee_irq_rx_start_flag = FALSE;
				zigbee_irq_trx_end_flag = FALSE;
				zigbee_man_basic_rx(zigbee_rx_data, &zigbee_rx_data_length, &zigbee_rx_data_lqi);
				//ZIGBEE_TRX_STATUS_RX_ON
				
				//ZigBee networks require to determine the best route between two nodes. Both, the
				//	LQI and the RSSI/ED can be used for this, depending on the optimization criteria. As a
				//	rule of thumb RSSI/ED is useful to differentiate between links with high LQI values.
				//	Transmission links with low LQI values should be discarded for routing decisions even if
				//	the RSSI/ED values are high.
				
				//if (zigbee_rx_data_lqi > 0) {
					
				
					#if UART0_ZIGBEE_RX_LQI_DEBUG
						uart0_send_byte(UART_TX_ZIGBEE_RX_LQI);
						uart0_send_byte(zigbee_rx_data_lqi);
					#endif
					#if UART0_ZIGBEE_RX_DATA_DEBUG
						uart0_send_byte(UART_TX_ZIGBEE_RX_DATA);
						uart0_send_byte(zigbee_rx_data_length);
						uart0_send_byte_array(zigbee_rx_data, zigbee_rx_data_length);
					#endif
				//}
			}
		#endif
	}
	
	return 0;
}



/**
 * Initialize ADC
 *
 * For thread-based, single conversion ADC mode
 */
void init_adc(void) {
	ADMUX = 0b01100000;		// 7:6  REFS1:0: Reference Selection Bits
							// 5  ADLAR: ADC Left Adjust Result
							// 4:0  MUX4:0: Analog Channel and Gain Selection Bits
							
	ADCSRA = 0b11000101;	// 7  ADEN: ADC Enable
							// 6  ADSC: ADC Start Conversion
							// 5  ADATE: ADC Auto Trigger Enable
							// 4  ADIF: ADC Interrupt Flag
							// 3  ADIE: ADC Interrupt Enable
							// 2:0  ADPS2:0: ADC Prescaler Select Bits
							
	ADCSRB = 0b00000000;	// 7, 5:3 - Reserved Bits
							// 2:0  ADTS2:0: ADC Auto Trigger Source

	DIDR0 = 0b00000111;		// 7:0 - ADC7:0 Digital Input Disable
							//	=> Disable digital input buffer to reduce power consumption 
}

void init_ext_int0(void) {
	EICRA = 0b00000011;	// 7:6 - Reserved Bits
						// 5:4 - ISC21, ISC20 - External Int 2 Sense Control Bits
						// 3:2 - ISC11, ISC10 - External Int 1 "
						// 1:0 - ISC01, ISC00 - External Int 0 "
						//	00 - Low level
						//	01 - Any edge
						//	10 - Falling edge
						//	11 - Rising edge
						
	EIMSK = 0b00000000;	// 7:3 - Reserved Bits
						// 2:0  INT2:0: External Interrupt Request 2 - 0 Enable
	
	EIFR = 0b00000000;	// 7:3 - Reserved Bits
						// 2:0  INTF2:0: External Interrupt Flags 2 - 0
						
	PCICR = 0b00000000;	// 7:4 - Reserved Bits
						// 3  PCIE3: Pin Change Interrupt Enable 3
						// 2  PCIE2: Pin Change Interrupt Enable 2
						// 1  PCIE1: Pin Change Interrupt Enable 1
						// 0  PCIE0: Pin Change Interrupt Enable 0
						
	PCIFR = 0b00000000;	// 7:4 - Reserved Bits
						// 3  PCIF3: Pin Change Interrupt Flag 3
						// 2  PCIF2: Pin Change Interrupt Flag 2
						// 1  PCIF1: Pin Change Interrupt Flag 1
						// 0  PCIF0: Pin Change Interrupt Flag 0
						
	PCMSK3 = 0b00000000;	// 7:0  PCINT31:24: Pin Change Enable Mask 31:24
	
	PCMSK2 = 0b00000000;	// 7:0  PCINT23:16: Pin Change Enable Mask 23:16
	
	PCMSK1 = 0b00000000;	// 7:0  PCINT15:8: Pin Change Enable Mask 15:8
	
	PCMSK0 = 0b00000000;	// 7:0  PCINT7:0: Pin Change Enable Mask 7:0
}

/**
 *
 */
void init_mcu(void) {
	SMCR = 0b00000000;	// 7:4 - Reserved bits
						// 3:1 - Sleep Mode
						//		000 - Idle
						//		001 - ADC Noise Reduction
						//		010 - Power-down
						//		011 - Power-save
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Standby
						//		111 - Extended Standby
						// 0 - Sleep Enable
						
	MCUCR |= (0 << 6);	// 6 - BOD Sleep
	MCUCR |= (0 << 5);	// 5 - BOD Sleep Enable
	
	PRR = 0b11111001;	// 7 - Power Reduction TWI
						// 6 - Power Reduction Timer2
						// 5 - Power Reduction Timer0
						// 4 - Power Reduction UART1
						// 3 - Power Reduction Timer1
						// 2 - Power Reduction SPI
						// 1 - Power Reduction UART0
						// 0 - Power Reduction ADC
}
