/* SPI Constants */
#define SPI_DDR				DDRB
#define SPI_PORT			PORTB
#define SPI_SCK				PB7
#define SPI_MISO			PB6
#define SPI_MOSI			PB5
#define SPI_SEL				PB4
#define SPI_GARBAGE_DATA	0xAA

/**
 * SPI Protocol (for the ZigBee Transceiver)
 * Command Byte
 *	B7	B6	B5	B4	B3	B2	B1	B0	Mode
 *	1	0	Register address [5:0]	Register Access Mode  Read Access
 *	1	1	Register address [5:0]	Register Access Mode  Write Access
 *	0	0	1	Reserved			Frame Buffer Access Mode  Read Access
 *	0	1	1	Reserved			Frame Buffer Access Mode  Write Access
 *	0	0	0	Reserved			SRAM Access Mode  Read Access
 *	0	1	0	Reserved			SRAM Access Mode  Write Access
 * 
 *			|--------------------------	|--------------	|----------	|----------	|----------	|----------
 * Register Write Access
 *	MOSI	1	1	address[5:0]		write data[7:0]
 *	MISO	XX							XX
 * Register Read Access
 *	MOSI	1	0	address[5:0]		XX
 *	MISO	XX							read data[7:0]
 * Frame Buffer Write Access
 *	MOSI	0	1	1	reserved[4:0]	PHR[7:0]		PSDU[7:0]	PSDU[7:0]	...			PSDU[7:0]
 *	MISO	XX							XX				XX			XX			...			XX
 * Frame Buffer Read Access
 *  MOSI	0	0	1	reserved[4:0]	XX				XX			...			XX			XX
 *	MISO	XX							PHR[7:0]		PSDU[7:0]	...			PSDU[7:0]	LQI[7:0]
 * SRAM Write Access
 *	MOSI	0	1	0	reserved[4:0]	address[7:0]	data[7:0]	...			data[7:0]
 *	MISO	XX							XX				XX			...			XX	
 * SRAM Read Access
 *	MOSI	0	0	0	reserved[4:0]	address[7:0]	XX			...			XX
 *	MISO	XX							XX				data[7:0]	...			data[7:0]	 
 *
 * Note: Logic values stated with XX on MOSI are ignored by the radio
 * transceiver, but need to have a valid level. Return values on MISO 
 * stated as XX shall be ignored by the microcontroller.
 * Note: Maximum frame length (# of PSDU bytes) is 127.
 */
 
 #define SPI_REGISTER_WRITE_CMD		0b11000000
 #define SPI_REGISTER_READ_CMD		0b10000000
 #define SPI_FRAME_BUFFER_WRITE_CMD	0b01100000
 #define SPI_FRAME_BUFFER_READ_CMD	0b00100000
 #define SPI_SRAM_WRITE_CMD			0b01000000
 #define SPI_SRAM_READ_CMD			0b00000000
 
void init_spi_master(void);
uint8_t spi_transfer_byte(uint8_t tx_data);
void spi_write_register(uint8_t reg_address, uint8_t reg_value);
uint8_t spi_read_register(uint8_t reg_address);
void spi_write_frame_buffer(uint8_t phr, uint8_t *psdu);
void spi_read_frame_buffer(uint8_t *phr, uint8_t *psdu, uint8_t *lqi);
void spi_write_sram(uint8_t sram_address, uint8_t *sram_data, uint8_t sram_data_length);
void spi_read_sram(uint8_t sram_address, uint8_t *sram_data, uint8_t sram_data_length);

/**
 * Transfer (send & receive) data across the SPI
 *
 * Writes a byte of data onto the SPI Data Register
 * Wait for transmission to complete
 * Reads a byte of data from the SPI Data Register
 *
 * When only writing data onto another device, ignore rx data
 * When only reading data from another device, send junk data
 */
uint8_t spi_transfer_byte(uint8_t tx_data) {
	/* Send data */
	SPDR = tx_data;
	
	/* Wait for transmission complete */
	while(!(SPSR & _BV(SPIF)));
	
	// return the received data
	return SPDR;
}

/**
 * Register Write Access
 *
 * Write a byte of register value onto the specified register
 */
void spi_write_register(uint8_t reg_address, uint8_t reg_value) {
	cbi(SPI_PORT, SPI_SEL);
	
	spi_transfer_byte(SPI_REGISTER_WRITE_CMD | reg_address);
	spi_transfer_byte(reg_value);
	
	sbi(SPI_PORT, SPI_SEL);
}

/**
 * Register Read Access
 *
 * Read a byte of register value from the specified register
 */
uint8_t spi_read_register(uint8_t reg_address) {
	cbi(SPI_PORT, SPI_SEL);
	
	spi_transfer_byte(SPI_REGISTER_READ_CMD | reg_address);
	uint8_t reg_value = spi_transfer_byte(SPI_GARBAGE_DATA);
	
	sbi(SPI_PORT, SPI_SEL);
	return reg_value;
}

/**
 * Frame Buffer Write Access
 *
 * IEEE 802.15.4-2003 PHY Layer Frame Structure
 *
 * Synchronization Header (SHR)	| PHY Header (PHR)	| PHY Service (PSDU)
 * Preamble Seq | SFD			| Frame Length		| PHY Payload
 * 4 Bytes		| 1 Byte 		| 1 Byte			| Max 127 Bytes
 */
void spi_write_frame_buffer(uint8_t phr, uint8_t *psdu) {
	cbi(SPI_PORT, SPI_SEL);
	
    spi_transfer_byte(SPI_FRAME_BUFFER_WRITE_CMD);
    spi_transfer_byte(phr);
    for (uint8_t i=0; i<phr; i++) {
    	spi_transfer_byte(psdu[i]);
    } 
	
	sbi(SPI_PORT, SPI_SEL);
}

/**
 * Frame Buffer Read Access
 *
 * See above
 */
void spi_read_frame_buffer(uint8_t *phr, uint8_t *psdu, uint8_t *lqi) {
	cbi(SPI_PORT, SPI_SEL);
	
    spi_transfer_byte(SPI_FRAME_BUFFER_READ_CMD);
    *phr = spi_transfer_byte(SPI_GARBAGE_DATA);
    for (uint8_t i=0; i<(*phr); i++) {
    	psdu[i] = spi_transfer_byte(SPI_GARBAGE_DATA);
    }
    *lqi = spi_transfer_byte(SPI_GARBAGE_DATA); 
	
	sbi(SPI_PORT, SPI_SEL);
}

/**
 * SRAM Write Access
 *
 * Write an array of sram value onto the specified sram register
 */
void spi_write_sram(uint8_t sram_address, uint8_t *sram_data, uint8_t sram_data_length) {
	cbi(SPI_PORT, SPI_SEL);
	
    spi_transfer_byte(SPI_SRAM_WRITE_CMD);
    spi_transfer_byte(sram_address);
    for (uint8_t i=0; i<sram_data_length; i++) {
    	spi_transfer_byte(sram_data[i]);
    } 
	
	sbi(SPI_PORT, SPI_SEL);
}

/**
 * SRAM Read Access
 *
 * Read an array of sram value from the specified sram register
 */
void spi_read_sram(uint8_t sram_address, uint8_t *sram_data, uint8_t sram_data_length) {
	cbi(SPI_PORT, SPI_SEL);
    
	spi_transfer_byte(SPI_SRAM_READ_CMD);
    spi_transfer_byte(sram_address);
    for (uint8_t i=0; i<sram_data_length; i++) {
    	sram_data[i] = spi_transfer_byte(SPI_GARBAGE_DATA);
    } 
	
	sbi(SPI_PORT, SPI_SEL);
}

/**
 *
 */
void init_spi_master(void) {
	/**
	 * If SS is configured as an output, the pin is a general 
	 * output pin which does not affect the SPI system. 
	 * Typically, the pin will be driving the SS pin of the SPI Slave. 
	 *
	 * If SS is configured as an input, it must be held high 
	 * to ensure Master SPI operation. If the SS pin is driven 
	 * low by peripheral circuitry when the SPI is configured 
	 * as a Master with the SS pin defined as an input, the SPI 
	 * system interprets this as another master selecting the SPI as a
	 * slave and starting to send data to it.
	 */

	sbi(SPI_DDR, SPI_SEL);
	sbi(SPI_DDR, SPI_MOSI);
	cbi(SPI_DDR, SPI_MISO);
	sbi(SPI_DDR, SPI_SCK);
	
	sbi(SPI_PORT, SPI_SEL);
	
	SPCR = 0b01010000;	// 7 - SPI Interrupt Enable
						// 6 - SPI Enable
						// 5 - Data Order
						// 4 - Master/Slave Select
						// 3 - Clock Polarity
						// 2 - Clock Phase
						// 1:0 - SPI Clock Rate
						//	SPI2X = 0
						//		00 - 1/4
						//		01 - 1/16
						//		10 - 1/64
						//		11 - 1/128
						//	SPI2X = 1
						//		00 - 1/2
						//		01 - 1/8
						//		10 - 1/32
						//		11 - 1/64
						
	SPSR = 0b00000000;	// 7 - SPI Interrupt Flag
						// 6 - Write COLlision Flag
						// 5:1 - Reserved Bits
						// 0 - Double SPI Speed Bit
}
