/**
 * A Mega644P ZigBee SPI w/ Ext Oscillator Library Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */

/* Global/System Constants */
#define TRUE 				1
#define FALSE 				0

#define CPU_FREQUENCY_HZ	8000000

#define sbi(data,bit)			(data |= (1<<bit))	// Set Bit
#define cbi(data,bit) 			(data &= ~(1<<bit))	// Clear Bit
#define gbi(data,bit)			(data & (1<<bit))	// Get Bit
#define tbi(data,bit)			(data ^= (1<<bit))	// Toggle Bit

/* Zigbee Constants */
#define ZIGBEE_TX			FALSE
#define ZIGBEE_RX			TRUE

/* Debug Constants */
#define PROTOBOARD_DEBUG	TRUE
#define STK500_DEBUG		FALSE
#define PREPROCESSOR_DEBUG	FALSE

#define UART0_DEBUG			FALSE
#define ZIGBEE_UART0_DEBUG			FALSE

#define PROTOBOARD_LED_DDR	DDRD
#define PROTOBOARD_LED_PORT	PORTD
#define PROTOBOARD_LED_0	PD4
#define PROTOBOARD_LED_1	PD5

/* Includes */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>
#include <stdio.h>
#include "uart_tx.h"
#include "spi.h"
#include "zigbee.h"

/* External Interrupt Constants */
#define INT0_FLAG_MASK		0b11111110
#define INT0_DDR	DDRD
#define INT0_PORT	PORTD
#define INT0_PIN	PD2

/* Volatile Constants */

ISR (INT0_vect) {
	zigbee_update_irq_status();
}

/* Function Prototypes */
void init_mcu(void);
void init_ext_int0(void);
void init_adc(void);

/**
 *
 */
int main(void) {
	/* Local Variables */
	#if ZIGBEE_TX
	uint8_t tx_data_length = 1;
	uint8_t tx_data[1] = {0};
	#elif ZIGBEE_RX
	uint8_t rx_data_length = 1;
	uint8_t rx_data[1] = {0};
	uint8_t rx_data_lqi = 0;
	#endif
	
	/* Global Variable Initialization */

	/* Set PORT for LED Functionality */
	#if PROTOBOARD_DEBUG
	sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_0);
	sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_1);
	sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
	sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);
	#endif
    
	/* Set PORT for ZigBee Functionality */
	sbi(ZIGBEE_DDR, ZIGBEE_SLP_TR);
	sbi(ZIGBEE_DDR, ZIGBEE_RST);

	/* MCU Initialization */
	init_mcu();		// Initialize MCU
	init_uart0_tx();	// Initialize UART
	init_spi_master();	// Initialize SPI Master
	init_ext_int0();
    
    /* Enable Interrupts */
    sei();
	
    /* ZigBee Hardware Initialization */
    zigbee_init();	// Hardware initialization after power-on event                        

   	zigbee_set_clkm(ZIGBEE_CLKM_8MHZ, ZIGBEE_CLKM_SHA_IMM);
	
	#if UART0_DEBUG
	uart0_send_byte(0xA0);
	uart0_send_byte(zigbee_state);				// Should be in TRX_OFF state
	#endif
	
	#if PROTOBOARD_DEBUG
	cbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);
	#endif
	
	#if PREPROCESSOR_DEBUG
	uart0_send_byte(UART0_UBBR0>>8);
	uart0_send_byte((UART0_UBBR0<<8)>>8);
	#endif
	
	while(1) {
		#if ZIGBEE_TX
		switch(zigbee_state) {
			case ZIGBEE_TRX_STATUS_P_ON:	// Should not be in P_ON state
				#if UART0_DEBUG 
				uart0_send_byte(0xF0);
				uart0_send_byte(zigbee_state);
				#endif
				while(1);
				break;
			case ZIGBEE_TRX_STATUS_TRX_OFF:
				#if PROTOBOARD_DEBUG
				cbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
				#endif
				
				zigbee_irq_trx_end_flag = TRUE;
				
				zigbee_set_state_trx_cmd(ZIGBEE_TRX_STATUS_PLL_ON);	
				
				#if UART0_DEBUG
				uart0_send_byte(0xA2);
				uart0_send_byte(zigbee_state);		// Should be in PLL_ON state
				#endif
				break;
			case ZIGBEE_TRX_STATUS_SLEEP:
				#if UART0_DEBUG
				uart0_send_byte(0xF1);
				uart0_send_byte(zigbee_state);
				#endif
				while(1);
				break;
			case ZIGBEE_TRX_STATUS_PLL_ON:
				if (zigbee_irq_trx_end_flag) {
					zigbee_irq_trx_end_flag = FALSE;
			
					// Write the entire frame before the transmission is initiated
					// to prevent a buffer underrun interrupt
					tx_data[0]++;
					spi_write_frame_buffer(tx_data_length, tx_data);
					
					zigbee_tx_start_pin();			// Start transmission
					
					#if UART0_DEBUG
					uart0_send_byte(0xA3);
					uart0_send_byte(zigbee_state);	// Should be in BUSY_TX state
					#endif
				}
				break;
			case ZIGBEE_TRX_STATUS_BUSY_TX:
				#if PROTOBOARD_DEBUG
				cbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);
				#endif
				
				#if UART0_DEBUG
				uart0_send_byte(0xA4);
				zigbee_state = zigbee_get_trx_status();	
				uart0_send_byte(zigbee_state);		// Should be in BUSY_TX state
				#endif
				
				while(!zigbee_irq_trx_end_flag);			// Wait for IRQ_MASK_TRX_END
				zigbee_state = zigbee_get_trx_status();		
				
				#if UART0_DEBUG
				uart0_send_byte(0xA5);
				uart0_send_byte(zigbee_state);	// Should be in PLL_ON state
				#endif
				
				#if PROTOBOARD_DEBUG
				sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);
				#endif
				_delay_ms(100);					// Delay 100 ms before next send
				break;
			case ZIGBEE_TRX_STATUS_RX_ON:
				break;
			case ZIGBEE_TRX_STATUS_BUSY_RX:
				break;
			case ZIGBEE_TRX_STATUS_RX_ON_NOCLK: 
			case ZIGBEE_TRX_STATUS_BUSY_RX_AACK:
			case ZIGBEE_TRX_STATUS_BUSY_TX_ARET:
			case ZIGBEE_TRX_STATUS_RX_AACK_ON:
			case ZIGBEE_TRX_STATUS_TX_ARET_ON:
			case ZIGBEE_TRX_STATUS_RX_AACK_NOCLK:
			case ZIGBEE_TRX_STATUS_BUSY_RX_AACK_NOCLK:
			case ZIGBEE_TRX_STATUS_TRANSITION:
				break;
			default:
				break;
		}
		#elif ZIGBEE_RX
		switch(zigbee_state) {
			case ZIGBEE_TRX_STATUS_P_ON:
				#if UART0_DEBUG 
				uart0_send_byte(0xF0);
				uart0_send_byte(zigbee_state);
				#endif
				while(1);
				break;
			case ZIGBEE_TRX_STATUS_TRX_OFF:
				#if PROTOBOARD_DEBUG
				cbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
				#endif
				
				zigbee_set_state_trx_cmd(ZIGBEE_TRX_STATUS_RX_ON);
				
				#if UART0_DEBUG
				uart0_send_byte(0xA2);
				uart0_send_byte(zigbee_state);	// Should be in RX_ON state
				#endif
				break;
			case ZIGBEE_TRX_STATUS_SLEEP:
				#if UART0_DEBUG
				uart0_send_byte(0xF1);
				uart0_send_byte(zigbee_state);
				#endif
				while(1);
				break;
			case ZIGBEE_TRX_STATUS_PLL_ON:
				break;
			case ZIGBEE_TRX_STATUS_BUSY_TX:
				break;
			case ZIGBEE_TRX_STATUS_RX_ON:
			case ZIGBEE_TRX_STATUS_BUSY_RX:
				if (zigbee_irq_rx_start_flag) {
					#if PROTOBOARD_DEBUG
					cbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);
					#endif
					
					if (zigbee_irq_trx_end_flag) {
						zigbee_irq_rx_start_flag = FALSE;
						zigbee_irq_trx_end_flag = FALSE;
						
						zigbee_state = zigbee_get_trx_status();	
						
						#if UART0_DEBUG
						uart0_send_byte(0xA3);
						uart0_send_byte(zigbee_state);	// Should be in RX_ON
						#endif
						
						spi_read_frame_buffer(&rx_data_length, rx_data, &rx_data_lqi);
						#if UART0_DEBUG
						uart0_send_byte(0xA4);
						uart0_send_byte(rx_data_lqi);
						#endif
						uart0_send_byte(rx_data[0]);
						
						#if PROTOBOARD_DEBUG
						sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);
						#endif
					}
				}
				break;
			case ZIGBEE_TRX_STATUS_RX_ON_NOCLK: 
			case ZIGBEE_TRX_STATUS_BUSY_RX_AACK:
			case ZIGBEE_TRX_STATUS_BUSY_TX_ARET:
			case ZIGBEE_TRX_STATUS_RX_AACK_ON:
			case ZIGBEE_TRX_STATUS_TX_ARET_ON:
			case ZIGBEE_TRX_STATUS_RX_AACK_NOCLK:
			case ZIGBEE_TRX_STATUS_BUSY_RX_AACK_NOCLK:
			case ZIGBEE_TRX_STATUS_TRANSITION:
				break;
			default:
				break;
		}
		#endif
	}
	
	return 0;
}



/**
 * Initialize ADC
 *
 * For thread-based, single conversion ADC mode
 */
void init_adc(void) {
	ADMUX = 0b01100000;		// 7:6  REFS1:0: Reference Selection Bits
							// 5  ADLAR: ADC Left Adjust Result
							// 4:0  MUX4:0: Analog Channel and Gain Selection Bits
							
	ADCSRA = 0b11000101;	// 7  ADEN: ADC Enable
							// 6  ADSC: ADC Start Conversion
							// 5  ADATE: ADC Auto Trigger Enable
							// 4  ADIF: ADC Interrupt Flag
							// 3  ADIE: ADC Interrupt Enable
							// 2:0  ADPS2:0: ADC Prescaler Select Bits
							
	ADCSRB = 0b00000000;	// 7, 5:3 - Reserved Bits
							// 2:0  ADTS2:0: ADC Auto Trigger Source

	DIDR0 = 0b00000111;		// 7:0 - ADC7:0 Digital Input Disable
							//	=> Disable digital input buffer to reduce power consumption 
}

void init_ext_int0(void) {
	EICRA = 0b00000011;	// 7:6 - Reserved Bits
						// 5:4 - ISC21, ISC20 - External Int 2 Sense Control Bits
						// 3:2 - ISC11, ISC10 - External Int 1 "
						// 1:0 - ISC01, ISC00 - External Int 0 "
						//	00 - Low level
						//	01 - Any edge
						//	10 - Falling edge
						//	11 - Rising edge
						
	EIMSK = 0b00000001;	// 7:3 - Reserved Bits
						// 2:0  INT2:0: External Interrupt Request 2 - 0 Enable
	
	EIFR = 0b00000000;	// 7:3 - Reserved Bits
						// 2:0  INTF2:0: External Interrupt Flags 2 - 0
						
	PCICR = 0b00000000;	// 7:4 - Reserved Bits
						// 3  PCIE3: Pin Change Interrupt Enable 3
						// 2  PCIE2: Pin Change Interrupt Enable 2
						// 1  PCIE1: Pin Change Interrupt Enable 1
						// 0  PCIE0: Pin Change Interrupt Enable 0
						
	PCIFR = 0b00000000;	// 7:4 - Reserved Bits
						// 3  PCIF3: Pin Change Interrupt Flag 3
						// 2  PCIF2: Pin Change Interrupt Flag 2
						// 1  PCIF1: Pin Change Interrupt Flag 1
						// 0  PCIF0: Pin Change Interrupt Flag 0
						
	PCMSK3 = 0b00000000;	// 7:0  PCINT31:24: Pin Change Enable Mask 31:24
	
	PCMSK2 = 0b00000000;	// 7:0  PCINT23:16: Pin Change Enable Mask 23:16
	
	PCMSK1 = 0b00000000;	// 7:0  PCINT15:8: Pin Change Enable Mask 15:8
	
	PCMSK0 = 0b00000000;	// 7:0  PCINT7:0: Pin Change Enable Mask 7:0
}

/**
 *
 */
void init_mcu(void) {
	SMCR = 0b00000000;	// 7:4 - Reserved bits
						// 3:1 - Sleep Mode
						//		000 - Idle
						//		001 - ADC Noise Reduction
						//		010 - Power-down
						//		011 - Power-save
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Standby
						//		111 - Extended Standby
						// 0 - Sleep Enable
						
	MCUCR |= (0 << 6);	// 6 - BOD Sleep
	MCUCR |= (0 << 5);	// 5 - BOD Sleep Enable
	
	PRR = 0b11111001;	// 7 - Power Reduction TWI
						// 6 - Power Reduction Timer2
						// 5 - Power Reduction Timer0
						// 4 - Power Reduction UART1
						// 3 - Power Reduction Timer1
						// 2 - Power Reduction SPI
						// 1 - Power Reduction UART0
						// 0 - Power Reduction ADC
}
