/**
 * A Mega644P ZigBee SPI w/ Ext Oscillator Library Alternative Test
 * -Remove use of ISR for detecting Zigbee interrupts
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2A/PCINT31)
 *							-----------
 */

/* Global/System Constants */
#define TRUE 				1
#define FALSE 				0

#define CPU_FREQUENCY_HZ	1000000

#define sbi(data,bit)		(data |= (1<<bit))	// Set Bit
#define cbi(data,bit) 		(data &= ~(1<<bit))	// Clear Bit
#define gbi(data,bit)		(data & (1<<bit))	// Get Bit
#define tbi(data,bit)		(data ^= (1<<bit))	// Toggle Bit

/* Zigbee Constants */
#define MCU_TX						FALSE
#define MCU_RX						TRUE

#define ZIGBEE_TX_SHORT_ADDR		0xABCD
#define ZIGBEE_TX_DEST_SHORT_ADDR	0xDCBA
#define ZIGBEE_TX_PAN_ID			0xBEEF
#define ZIGBEE_TX_DEST_PAN_ID		0xBEEF
#define ZIGBEE_RX_SHORT_ADDR		0xDCBA
#define ZIGBEE_RX_DEST_SHORT_ADDR	0xABCD
#define ZIGBEE_RX_PAN_ID			0xBEEF
#define ZIGBEE_RX_DEST_PAN_ID		0xBEEF

/* Debug Constants */
#define PROTOBOARD_LED_DEBUG	TRUE
#define STK500_LED_DEBUG		FALSE

#define UART0_ZIGBEE_SPEC_DEBUG			FALSE
#define UART0_ZIGBEE_IRQ_STATUS_DEBUG	TRUE
#define UART0_ZIGBEE_TRAC_STATUS_DEBUG	TRUE
#define UART0_ZIGBEE_STATE_DEBUG		FALSE
#define UART0_ZIGBEE_RX_PHY_DATA_DEBUG	TRUE
#define UART0_ZIGBEE_RX_PHY_LQI_DEBUG	TRUE
#define UART0_ZIGBEE_RX_MAC_DATA_DEBUG	TRUE
#define UART0_ZIGBEE_RX_MAC_FRAME_DEBUG	TRUE
#define UART0_ZIGBEE_CCA_DEBUG			FALSE

/* UART Protocol */
#define UART_TX_ZIGBEE_STATE			0x20
#define UART_TX_ZIGBEE_IRQ_STATUS		0x21
#define UART_TX_ZIGBEE_TRAC_STATUS		0x22

#define UART_TX_ZIGBEE_PHY_DATA			0x30
#define UART_TX_ZIGBEE_PHY_LQI			0x31

#define UART_TX_ZIGBEE_MAC_DATA			0x40
#define UART_TX_ZIGBEE_MAC_DEST_PAN_ID	0x41
#define UART_TX_ZIGBEE_MAC_DEST_ADDR	0x42
#define UART_TX_ZIGBEE_MAC_SRC_PAN_ID	0x43
#define UART_TX_ZIGBEE_MAC_SRC_ADDR		0x44

#define UART_TX_ZIGBEE_CCA_MODE			0x50
#define UART_TX_ZIGBEE_CCA_ED_THRES		0x51
#define UART_TX_ZIGBEE_CCA_VALUE		0x52

#define UART_TX_ZIGBEE_PART_NUM			0xE0
#define UART_TX_ZIGBEE_VER_NUM			0xE1
#define UART_TX_ZIGBEE_MAN_ID_0			0xE2
#define UART_TX_ZIGBEE_MAN_ID_1			0xE3

/* Protoboard Constants */
#define PROTOBOARD_LED_DDR	DDRD
#define PROTOBOARD_LED_PORT	PORTD
#define PROTOBOARD_LED_0	PD4
#define PROTOBOARD_LED_1	PD5

/* External Interrupt Constants */
#define INT0_DDR			DDRD
#define INT0_PORT			PORTD
#define INT0_INPUT			PD2

/* Includes */
#include <avr/io.h>			// Include AVR-GCC Mega644P specific I/O, register lib
#include <avr/interrupt.h>	// Include AVR-GCC ISR lib
#include <avr/sleep.h>		// Include AVR-GCC sleep lib
#include <avr/eeprom.h>		// Include AVR-GCC eeprom lib
#include <util/delay.h>		// Include AVR-GCC delay lib
#include <string.h>			// Include AVR-GCC string lib
#include <stdio.h>			// Include AVR-GCC stdio lib
#include "uart_tx.h"		// Include user UART TX lib
#include "spi.h"			// Include user SPI lib
#include "zigbee.h"			// Include user Zigbee lib
#include "timer.h"			// Include user timer lib

/* Global Variables */
enum{MCU_STATUS_P_ON,
	MCU_STATUS_IDLE,
	MCU_STATUS_TX,
	MCU_STATUS_RX};
uint8_t mcu_status;

/* Volatile Constants */

/* Function Prototypes */
void init_mcu(void);
void init_ext_int0(void);
void init_adc(void);

uint8_t get_sw_crc(uint8_t *data, uint8_t data_length);

uint8_t get_sw_crc(uint8_t *data, uint8_t data_length) {
	uint8_t crc_value = 0;
	for (uint8_t i=0; i<data_length; i++) {
		crc_value ^= data[i];
	}
	return crc_value;
}

/**
 *
 */
int main(void) {
	/* Local Variables */
	uint8_t rxtx_data[ZIGBEE_PHY_DATA_LENGTH];
	uint8_t rxtx_data_length;
	uint8_t rxtx_msg;
	
	/* Global Variable Initialization */
	mcu_status = MCU_STATUS_P_ON;

	/* Set PORT for LED Functionality */
	#if PROTOBOARD_LED_DEBUG
		sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_0);
		sbi(PROTOBOARD_LED_DDR, PROTOBOARD_LED_1);
		sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
		sbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_1);
	#endif

	/* MCU Initialization */
	init_mcu();			// Initialize MCU
	init_uart0_tx();	// Initialize UART
	init_spi_master();	// Initialize SPI Master
	init_ext_int0();	// Initialize External Int0
    
    /* Enable Interrupts */
    //sei();
	
    /* ZigBee Hardware Initialization */
	zigbee_init();							// Hardware initialization after power-on event
   	zigbee_clear_irq_status();
	zigbee_set_extended_mode(TRUE);
	//zigbee_set_clkm_safe(ZIGBEE_CLKM_8MHZ);	// Set the Master CLK from the Zigbee Transceiver at the proper freq
	
	/* UART0 Init Debug */
	#if UART0_ZIGBEE_SPEC_DEBUG
		uart0_send_byte(UART_TX_ZIGBEE_PART_NUM);
		uart0_send_byte(zigbee_hw_part_num);
		uart0_send_byte(UART_TX_ZIGBEE_VER_NUM);
		uart0_send_byte(zigbee_hw_ver_num);
		uart0_send_byte(UART_TX_ZIGBEE_MAN_ID_0);
		uart0_send_byte(zigbee_hw_man_id_0);
		uart0_send_byte(UART_TX_ZIGBEE_MAN_ID_1);
		uart0_send_byte(zigbee_hw_man_id_1);
	#endif
	
	#if PROTOBOARD_LED_DEBUG
		cbi(PROTOBOARD_LED_PORT, PROTOBOARD_LED_0);
	#endif

	while(1) {
		switch(mcu_status) {
			case MCU_STATUS_P_ON:
				mcu_status = MCU_STATUS_IDLE;
				break;
			case MCU_STATUS_IDLE:
				#if MCU_TX
					rxtx_data[0]++;
					rxtx_data_length = 1;
					mcu_status = MCU_STATUS_TX;
				#elif MCU_RX
					mcu_status = MCU_STATUS_RX;
				#endif
				break;
			case MCU_STATUS_TX:
				switch(zigbee_get_trx_status()) {
					case ZIGBEE_TRX_STATUS_PLL_ON:
						rxtx_msg = zigbee_basic_tx(rxtx_data, rxtx_data_length, TRUE);
						zigbee_trx_off();
						_delay_ms(1000);
						mcu_status = MCU_STATUS_IDLE;
						break;
					case ZIGBEE_TRX_STATUS_BUSY_TX:
						break;
					case ZIGBEE_TRX_STATUS_TX_ARET_ON:
						zigbee_set_tx_ack_request(TRUE);
						zigbee_inc_tx_sequence_number();
						rxtx_msg = zigbee_extended_tx(rxtx_data, rxtx_data_length, TRUE, TRUE);
						zigbee_trx_off();
						_delay_ms(1000);
						mcu_status = MCU_STATUS_IDLE;
						break;
					case ZIGBEE_TRX_STATUS_BUSY_TX_ARET:
						break;
					case ZIGBEE_TRX_STATUS_TRX_OFF:
						if (zigbee_get_extended_mode()) {
							zigbee_set_tx_src_pan_id(ZIGBEE_TX_PAN_ID);
							zigbee_set_tx_src_short_addr(ZIGBEE_TX_SHORT_ADDR);
							zigbee_set_tx_dest_short_addr(ZIGBEE_TX_DEST_SHORT_ADDR);
							zigbee_set_tx_dest_pan_id(ZIGBEE_TX_DEST_PAN_ID);
							zigbee_enable_extended_tx();
						}
						else {
							zigbee_enable_basic_tx();
						}
						break;
				}
				break;
			case MCU_STATUS_RX:
				switch(zigbee_get_trx_status()) {
					case ZIGBEE_TRX_STATUS_RX_ON:	// RX_START interrupt can occur in either RX_ON or BUSY_RX state
					case ZIGBEE_TRX_STATUS_BUSY_RX:
						if (EIFR & _BV(INTF0)) {
							EIFR = (EIFR & ~_BV(INTF0)) | _BV(INTF0);	// Clear the flag to make sure next interrupt is not missed
							zigbee_update_irq_status(); 				// RX_START and/or TRX_END interrupt
						}
						if (zigbee_get_rx_ready()) {
							zigbee_clear_rx_ready();
							
							rxtx_msg = zigbee_basic_rx(rxtx_data, &rxtx_data_length);
							/* RX_ON state */
							#if UART0_ZIGBEE_RX_PHY_DATA_DEBUG
								uart0_send_byte(UART_TX_ZIGBEE_PHY_DATA);
								uart0_send_byte(rxtx_data_length);
								uart0_send_byte_array(rxtx_data, rxtx_data_length);
							#endif
						}
						break;
					case ZIGBEE_TRX_STATUS_RX_AACK_ON:
						break;
					case ZIGBEE_TRX_STATUS_BUSY_RX_AACK:
						if (EIFR & _BV(INTF0)) {
							EIFR = (EIFR & ~_BV(INTF0)) | _BV(INTF0);	// Clear the flag to make sure next interrupt is not missed
							zigbee_update_irq_status(); 					// TRX_END interrupt
						}
						if (zigbee_get_rx_ready()) {
							zigbee_clear_rx_ready();
							rxtx_msg = zigbee_extended_rx(rxtx_data, &rxtx_data_length);
							#if UART0_ZIGBEE_RX_MAC_DATA_DEBUG
								uart0_send_byte(UART_TX_ZIGBEE_MAC_DATA);
								uart0_send_byte(rxtx_data_length);
								uart0_send_byte_array(rxtx_data, rxtx_data_length);
							#endif
						}
						break;
					case ZIGBEE_TRX_STATUS_TRX_OFF:
						if (zigbee_get_extended_mode()) {
							zigbee_set_addr_filter_pan_id(ZIGBEE_RX_PAN_ID);
							zigbee_set_addr_filter_short_addr(ZIGBEE_RX_SHORT_ADDR);	
							zigbee_enable_extended_rx();
						}
						else {
							zigbee_enable_basic_rx();
						}
						break;
				}
				break;
		}
	}
	
	return 0;
}

/**
 * Initialize ADC
 *
 * For thread-based, single conversion ADC mode
 */
void init_adc(void) {
	ADMUX = 0b01100000;		// 7:6  REFS1:0: Reference Selection Bits
							// 5  ADLAR: ADC Left Adjust Result
							// 4:0  MUX4:0: Analog Channel and Gain Selection Bits
							
	ADCSRA = 0b11000101;	// 7  ADEN: ADC Enable
							// 6  ADSC: ADC Start Conversion
							// 5  ADATE: ADC Auto Trigger Enable
							// 4  ADIF: ADC Interrupt Flag
							// 3  ADIE: ADC Interrupt Enable
							// 2:0  ADPS2:0: ADC Prescaler Select Bits
							
	ADCSRB = 0b00000000;	// 7, 5:3 - Reserved Bits
							// 2:0  ADTS2:0: ADC Auto Trigger Source

	DIDR0 = 0b11111111;		// 7:0 - ADC7:0 Digital Input Disable
							//	=> Disable digital input buffer to reduce power consumption 
}

/**
 * Initialize External Int0
 *
 * For thread-based interrupt flag detection
 */
void init_ext_int0(void) {
	EICRA = 0b00000011;	// 7:6 - Reserved Bits
						// 5:4 - ISC21, ISC20 - External Int 2 Sense Control Bits
						// 3:2 - ISC11, ISC10 - External Int 1 "
						// 1:0 - ISC01, ISC00 - External Int 0 "
						//	00 - Low level
						//	01 - Any edge
						//	10 - Falling edge
						//	11 - Rising edge
						
	EIMSK = 0b00000000;	// 7:3 - Reserved Bits
						// 2:0  INT2:0: External Interrupt Request 2 - 0 Enable
	
	EIFR = 0b00000000;	// 7:3 - Reserved Bits
						// 2:0  INTF2:0: External Interrupt Flags 2 - 0
						
	PCICR = 0b00000000;	// 7:4 - Reserved Bits
						// 3  PCIE3: Pin Change Interrupt Enable 3
						// 2  PCIE2: Pin Change Interrupt Enable 2
						// 1  PCIE1: Pin Change Interrupt Enable 1
						// 0  PCIE0: Pin Change Interrupt Enable 0
						
	PCIFR = 0b00000000;	// 7:4 - Reserved Bits
						// 3  PCIF3: Pin Change Interrupt Flag 3
						// 2  PCIF2: Pin Change Interrupt Flag 2
						// 1  PCIF1: Pin Change Interrupt Flag 1
						// 0  PCIF0: Pin Change Interrupt Flag 0
						
	PCMSK3 = 0b00000000;	// 7:0  PCINT31:24: Pin Change Enable Mask 31:24
	
	PCMSK2 = 0b00000000;	// 7:0  PCINT23:16: Pin Change Enable Mask 23:16
	
	PCMSK1 = 0b00000000;	// 7:0  PCINT15:8: Pin Change Enable Mask 15:8
	
	PCMSK0 = 0b00000000;	// 7:0  PCINT7:0: Pin Change Enable Mask 7:0
}

/**
 * Initialize External Int0
 *
 * Enable Power Reduction registers for unused peripherals
 */
void init_mcu(void) {
	SMCR = 0b00000000;	// 7:4 - Reserved bits
						// 3:1 - SM2:0: Sleep Mode
						//		000 - Idle
						//		001 - ADC Noise Reduction
						//		010 - Power-down
						//		011 - Power-save
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Standby
						//		111 - Extended Standby
						// 0 - SE: Sleep Enable
			
	MCUCR |= (0 << 6);	// 6 - BODS: BOD Sleep
	MCUCR |= (0 << 5);	// 5 - BODSE: BOD Sleep Enable
	
	PRR = 0b11111001;	// 7 - PRTWI: Power Reduction TWI
						// 6 - PRTIM2: Power Reduction Timer2
						// 5 - PRTIM0: Power Reduction Timer0
						// 4 - PRUSART1: Power Reduction UART1
						// 3 - PRTIM1: Power Reduction Timer1
						// 2 - PRSPI: Power Reduction SPI
						// 1 - PRUSART0: Power Reduction UART0
						// 0 - PRADC: Power Reduction ADC
}
