/**
 * A Mega644P ZigBee SPI w/ Ext Oscillator Library Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */

/* Global Constants */
#define TRUE 	1
#define FALSE 	0

/* Includes */
#include <mega644p.h>
#include <string.h>
#include <stdio.h>
#include <delay.h>
#include "..\Libraries\uart_tx.h"
#include "..\Libraries\spi.h"
#include "..\Libraries\zigbee.h"

/* System Constants */
#define CPU_FREQUENCY_HZ_0	1000000
#define CPU_FREQUENCY_HZ_1	8000000
#define CPU_RATIO			CPU_FREQUENCY_HZ_1 / CPU_FREQUENCY_HZ_0
#define CYCLES_PER_US_0		CPU_FREQUENCY_HZ_0 / 1000000
#define	CYCLES_PER_US_1		CPU_FREQUENCY_HZ_1 / 1000000	

#define LED_0				PORTD.4
#define LED_1				PORTD.5

/* UART Constants */
#define UART0_BAUD_RATE		19200

/* Function Prototypes */
void init_mcu(void);

/**
 *
 */
void main() {
	/* Local Variables */
	unsigned char data;
	
	/* Set PORT for LED Functionality */
    DDRD.4 = 1;		// LED_0
    DDRD.5 = 1;		// LED_1
    LED_0 = 1;
    LED_1 = 1;
    
    LED_0 = 0;
    
	/* Set PORT for ZigBee Functionality */
	DDRB.0 = 1;		// SLP_TR
	DDRB.1 = 1;		// RST

	/* MCU Initialization */
	init_mcu();		// Initialize MCU
	init_uart0_tx(CPU_FREQUENCY_HZ_1, UART0_BAUD_RATE);	// Initialize UART at 1 MHz
	init_spi_master();	// Initialize SPI Master
    
    /* Enable Interrupts */
    #asm("sei");
    
    /* ZigBee Hardware Initialization */
    zigbee_init();	// Hardware initialization after power-on event                        
    spi_write_register(ZIGBEE_TRX_CTRL_0, 0b00010100);	// Set CLKM clock to 8 MHz immediately
   	zigbee_clkm = ZIGBEE_CLKM_8MHZ;
   	
   	LED_1 = 0;
   	
	while(1) {
		/* Read register from Transceiver */
		spi_read_register(ZIGBEE_TRX_CTRL_0, &data);
		uart0_send_byte(data);
		delay_ms(100);
	}
}



/**
 *
 */
void init_mcu(void) {
	SMCR = 0b00000000;	// 7:4 - Reserved bits
						// 3:1 - Sleep Mode
						//		000 - Idle
						//		001 - ADC Noise Reduction
						//		010 - Power-down
						//		011 - Power-save
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Standby
						//		111 - Extended Standby
						// 0 - Sleep Enable
						
	MCUCR |= (0 << 6);	// 6 - BOD Sleep
	MCUCR |= (0 << 5);	// 5 - BOD Sleep Enable
	
	//PRR = 0b11111001;	// 7 - Power Reduction TWI
						// 6 - Power Reduction Timer2
						// 5 - Power Reduction Timer0
						// 4 - Power Reduction UART1
						// 3 - Power Reduction Timer1
						// 2 - Power Reduction SPI
						// 1 - Power Reduction UART0
						// 0 - Power Reduction ADC
}