/**
 * A Mega644P ZigBee SPI w/ Ext Oscillator Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
 /**
 * Mega164P/324P/644P DIP Pinout
 * 							___________
 * PB0 (PCINT8/XCK0/T0)		|1		40| PA0 (ADC0/PCINT0)
 * PB1 (PCINT9/CLK0/T1)		|2		39| PA1 (ADC1/PCINT1)
 * PB2 (PCINT10/INT2/AIN0)	|3		38| PA2 (ADC2/PCINT2)
 * PB3 (PCINT11/OC0A/AIN1)	|4		37| PA3 (ADC3/PCINT3)
 * PB4 (PCINT12/OCOB/SS)	|5		36| PA4 (ADC4/PCINT4)
 * PB5 (PCINT13/MOSI)		|6		35| PA5 (ADC5/PCINT5)
 * PB6 (PCINT14/MISO)		|7		34| PA6 (ADC6/PCINT6)
 * PB7 (PCINT15/SCK)		|8		33| PA7 (ADC7/PCINT7)
 * RESET					|9		32| AREF
 * VCC						|10		31| GND
 * GND						|11		30| AVCC
 * XTAL2					|12		29| PC7 (TOSC2/PCINT23)
 * XTAL1					|13		28| PC6 (TOSC1/PCINT22)
 * PD0 (PCINT24/RXD0)		|14		27| PC5 (TDI/PCINT21)
 * PD1 (PCINT25/TXD0)		|15		26| PC4 (TDO/PCINT20)
 * PD2 (PCINT26/RXD1/INT0)	|16		25| PC3 (TMS/PCINT19)
 * PD3 (PCINT27/TXD1/INT1)	|17		24| PC2 (TCK/PCINT18)
 * PD4 (PCINT28/XCK1/OC1B)	|18		23| PC1 (SDA/PCINT17)
 * PD5 (PCINT29/OC1A)		|19		22| PC0 (SCL/PCINT16)
 * PD6 (PCINT30/OC2B/ICP)	|20		21| PD7 (OC2/PCINT31)
 *							-----------
 */

/* Includes */
#include <mega644p.h>
#include <string.h>
#include <stdio.h>
#include <delay.h>

/* Global Constants */
#define TRUE 	1
#define FALSE 	0

/* System Constants */
#define CPU_INIT_FREQUENCY_HZ	1000000
#define CPU_FINAL_FREQUENCY_HZ	8000000

/* SPI Constants */
#define SPI_GARBAGE_DATA	0xAA
#define SPIF				7
#define SPI_SLP				PORTB.0
#define SPI_RST				PORTB.1
#define SPI_SEL				PORTB.4
// #define SPI_MOSI			PORTB.5
// #define SPI_MISO			PORTB.6
// #define SPI_SCK			PORTB.7

/* ZigBee Registers */
#define ZIGBEE_TRX_STATUS	0x01
#define ZIGBEE_TRX_STATE	0x02
#define ZIGBEE_TRX_CTRL_0	0x03
#define ZIGBEE_XOSC_CTRL	0x12 

#define ZIGBEE_PART_NUM		0x1C
#define ZIGBEE_VER_NUM		0x1D
#define ZIGBEE_MAN_ID		0x1E

/* Zigbee State Values */
#define ZIGBEE_NOP			0x00
#define ZIGBEE_TX_START		0x02
#define ZIGBEE_FORCE_TRX_OFF	0x03
#define ZIGBEE_RX_ON		0x06
#define ZIGBEE_TRX_OFF		0x08
#define ZIGBEE_PLL_ON		0x09
#define ZIGBEE_RX_AACK_ON	0x16
#define ZIGBEE_TX_ARET_ON	0x19

/* UART Constants */
#define UART0_BAUD_RATE		19200
#define	UBRR0_INIT			(CPU_INIT_FREQUENCY_HZ / (16 * UART0_BAUD_RATE) - 1)
#define UBRR0_FINAL			(CPU_FINAL_FREQUENCY_HZ / (16 * UART0_BAUD_RATE) - 1)
#define UDRE				5

/* Function Prototypes */
void init_mcu(void);

void init_spi_master(void);
unsigned char spi_write_read_byte(unsigned char tx_data);
void spi_write_register(unsigned char address, unsigned char data);
void spi_read_register(unsigned char address, unsigned char *data);

void init_uart0_tx(unsigned int baud_rate);
void uart0_send_byte_array(char *data, unsigned char length);
void uart0_send_byte(char data);   

/* Global Variables */
unsigned char spi_complete;

/**
 *
 */
interrupt [SPI_STC] void spi_int(void) {
	spi_complete = TRUE;
}

/**
 *
 */
unsigned char spi_write_read_byte(unsigned char tx_data) {
	unsigned char rx_data;
	
	/* Send data */
	SPDR = tx_data;
	
	/* Wait for transmission complete */
	while(!spi_complete);
	spi_complete = FALSE;
	
	// return the received data
	rx_data = SPDR;
	return rx_data;
}

/**
 *
 */
void spi_write_register(unsigned char address, unsigned char data) {
	SPI_SEL = 0;
	spi_write_read_byte((0b11000000 | address));
	spi_write_read_byte(data);
	SPI_SEL = 1;
}

/**
 *
 */
void spi_read_register(unsigned char address, unsigned char *data) {
	SPI_SEL = 0;
	spi_write_read_byte((0b10000000 | address));
	*data = spi_write_read_byte(SPI_GARBAGE_DATA);
	SPI_SEL = 1;
}

/**
 * UART transmission function for byte arrays
 *
 * Can be used to transmitt strings
 * Length terminated (versus null terminated)
 */
void uart0_send_byte_array(unsigned char *data, unsigned char length) {
	unsigned char counter = 0;

	while (counter++ < length) {
		while ( !( UCSR0A & (1<<UDRE)) );	// Wait for empty buffer
		UDR0 = *data++;						// Put data into buffer, sends the data 
	}
}

/**
 * UART transmission function for a single byte
 */
void uart0_send_byte(unsigned char data) {
	while ( !( UCSR0A & (1<<UDRE)) );	// Wait for empty buffer
 	UDR0 = data;							// Put data into buffer, sends the data 
}

/**
 * Brings the radio transceiver into the TRX_OFF state 
 * after a power-on event.
 */
void zigbee_init(void) {
   	/* Transceiver Initialization */
	delay_us(510);	// Delay 510us for transition to P_ON
	
	/* Transceiver Reset */
	SPI_SLP = 0;	// Pull SLP low
	SPI_RST = 0;	// Pull RST low
	delay_us(6);	// Delay 6us for hardware reset
	SPI_RST = 1;	// Pull RST high
    
	/* Send TRX_OFF Command */
	spi_write_register(ZIGBEE_TRX_STATE, ZIGBEE_TRX_OFF);
	delay_us(510);	// Delay 510us for transition from P_ON to TRX_OFF
}

/**
 * Resets the Zigbee Transceiver from any state except P_ON
 *
 * The programming sequence will typically be used to recover 
 * from fatal errors and bring the radio transceiver to a known state.
 * All registers will have their default values and the state machine 
 * will be in TRX_OFF state.
 */
void zigbee_hardware_reset(void) {
	SPI_SLP = 0;	// Pull SLP low
	SPI_RST = 0;	// Pull RST low
	delay_us(6);	// Delay 6us for hardware reset
	SPI_RST = 1;	// Pull RST high
}

/**
 * Resets the state machine of the radio transceiver to the TRX_OFF state.
 */
void zigbee_fsm_reset(void) {
	SPI_SLP = 0;	// Pull SLP low
	delay_us(6);	// Delay 6us for hardware reset
	spi_write_register(ZIGBEE_TRX_STATE, ZIGBEE_FORCE_TRX_OFF);
	delay_us(6);	// Delay 6us for hardware reset
}	

/**
 *
 */
void main() {
	/* Local Variables */
	unsigned char counter, data;
	
	/* Set PORT for LED Functionality */
    DDRD.4 = 1;
    DDRD.5 = 1;
    PORTD.4 = 0;
    PORTD.5 = 1;
    
	/* Set PORT for ZigBee Functionality */
	DDRB.0 = 1;		// SLP_TR
	DDRB.1 = 1;		// SLP_RST

	/* MCU Initialization */
	init_mcu();		// Initialize MCU
	init_uart0_tx(UBRR0_FINAL);	// Initialize UART at 1 MHz
	init_spi_master();	// Initialize SPI Master
    
    /* Enable Interrupts */
    #asm("sei");
    
    /* ZigBee Hardware Initialization */
    zigbee_init();	// Hardware initialization after power-on event                        
    spi_write_register(ZIGBEE_TRX_CTRL_0, 0b00010100);	// Set CLKM clock to 8 MHz immediately
   	
   	PORTD.5 = 0;
   	
	while(1) {
		/* Read register from Transceiver */
		spi_read_register(ZIGBEE_TRX_CTRL_0, &data);
		uart0_send_byte(data);
		delay_ms(100);
	}
}

/**
 * Initialize UART TX
 */
void init_uart0_tx(unsigned int baud_rate){
	UBRR0H = baud_rate >> 8;
	UBRR0L = (baud_rate << 8) >> 8;

	UCSR0A = 0b00000000;// 7 - UART RX complete flag
						// 6 - UART TX complete flag
						// 5 - UART DRE flag
						// 4 - Frame error flag
						// 3 - Data overrun flag
						// 2 - Parity error flag
						// 1 - Double UART speed mode
						// 0 - Multi processor communication mode

	UCSR0B = 0b00001000; // 7 - RX Complete Interrupt Enable
						// 6 - TX Complete Interrupt Enable
						// 5 - UART DRE Interrupt Enable
						// 4 - Enable receiver
						// 3 - Enable transmitter
						// 2 - Character size
						// 1:0 - Ninth data bit

	UCSR0C = 0b00000110;// 7:6 - UART Mode Select
						//		00 - Asynchronous UART
						//		01 - Synchronous UART
						//		10 - Reserved
						//		11 - Master SPI
						// 5:4 - Parity mode
						//		00 - Disabled
						//		01 - Reserved
						//		10 - Enabled, Even
						//		11 - Enabled, Odd
						// 3 - Select stop bit length
						//		0 - 1-bit
						//		1 - 2-bit
						// 2:1 - Character size, with UCSRB.2
						//		000 - 5 bit
						//		001 - 6 bit
						//		010 - 7 bit
						//		011 - 8 bit
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Reserved
						//		111 - 9 bit
						// 0 - Select clock polarity
						//		0 - Rising XCK edge
						//		1 - Falling XCK edge 						
}

/**
 *
 */
void init_spi_master(void) {
	/**
	 * If SS is configured as an output, the pin is a general 
	 * output pin which does not affect the SPI system. 
	 * Typically, the pin will be driving the SS pin of the SPI Slave. 
	 *
	 * If SS is configured as an input, it must be held high 
	 * to ensure Master SPI operation. If the SS pin is driven 
	 * low by peripheral circuitry when the SPI is configured 
	 * as a Master with the SS pin defined as an input, the SPI 
	 * system interprets this as another master selecting the SPI as a
	 * slave and starting to send data to it.
	 */

	DDRB.4 = 1; // SS as Output
	DDRB.5 = 1;	// MOSI as Output
	DDRB.6 = 0;	// MISO as Input
	DDRB.7 = 1;	// SCK as Output
	SPI_SEL = 1;	// Set SS High
	
	SPCR = 0b11010000;	// 7 - SPI Interrupt Enable
						// 6 - SPI Enable
						// 5 - Data Order
						// 4 - Master/Slave Select
						// 3 - Clock Polarity
						// 2 - Clock Phase
						// 1:0 - SPI Clock Rate
						//	SPI2X = 0
						//		00 - 1/4
						//		01 - 1/16
						//		10 - 1/64
						//		11 - 1/128
						//	SPI2X = 1
						//		00 - 1/2
						//		01 - 1/8
						//		10 - 1/32
						//		11 - 1/64
						
	SPSR = 0b00000000;	// 7 - SPI Interrupt Flag
						// 6 - Write COLlision Flag
						// 5:1 - Reserved Bits
						// 0 - Double SPI Speed Bit
						
	spi_complete = TRUE;
} 

/**
 *
 */
void init_mcu(void) {
	SMCR = 0b00000000;	// 7:4 - Reserved bits
						// 3:1 - Sleep Mode
						//		000 - Idle
						//		001 - ADC Noise Reduction
						//		010 - Power-down
						//		011 - Power-save
						//		100 - Reserved
						//		101 - Reserved
						//		110 - Standby
						//		111 - Extended Standby
						// 0 - Sleep Enable
						
	MCUCR |= (0 << 6);	// 6 - BOD Sleep
	MCUCR |= (0 << 5);	// 5 - BOD Sleep Enable
	
	//PRR = 0b11111001;	// 7 - Power Reduction TWI
						// 6 - Power Reduction Timer2
						// 5 - Power Reduction Timer0
						// 4 - Power Reduction UART1
						// 3 - Power Reduction Timer1
						// 2 - Power Reduction SPI
						// 1 - Power Reduction UART0
						// 0 - Power Reduction ADC
}