/**
 * A Tiny26 USI TWI Slave Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
/**
 * Tiny26 DIP Pinout
 * 							___________
 * PB0 (MOSI/DI/SDA/OC1A)	|1		20| PA0 (ADC0)
 * PB1 (MISO/DO/OC1A)		|2		19| PA1 (ADC1)
 * PB2 (SCK/SCL/OC1B)		|3		18| PA2 (ADC2)
 * PB3 (OC1B)				|4		17| PA3 (AREF)
 * VCC                      |5		16| GND
 * GND 						|6		15| AVCC
 * PB4 (ADC7/XTAL1)			|7		14| PA4 (ADC3)
 * PB5 (ADC8/XTAL2)			|8		13| PA5 (ADC4)
 * PB6 (ADC9/INT0/T0)		|9		12| PA6 (ADC5/AIN0)
 * PB7 (ADC10/RESET)		|10		11| PA7 (ADC6/AIN1)
 *							-----------
 */

/* Includes */
#include <tiny26.h>
#include <delay.h>

/* System/Settings Constants */
#define TRUE			1
#define FALSE			0

#define TWI_SLAVE_ADDRESS	2
#define TWI_BUFFER_SIZE		32

/* Constants for TWI */
#define TWI_READ		1
#define TWI_WRITE		0
#define TWI_GENERAL_ADDRESS 0

/* USICR */
#define USISIE  0b10000000
#define USIOIE  0b01000000
#define USIWM1  0b00100000
#define USIWM0  0b00010000
#define USICS1  0b00001000
#define USICS0  0b00000100
#define USICLK  0b00000010
#define USITC   0b00000001

/* USISR */
#define USISIF   0b10000000
#define USIOIF   0b01000000
#define USIPF    0b00100000
#define USIDC    0b00010000
#define USICNT3  0b00001000
#define USICNT2  0b00000100
#define USICNT1  0b00000010
#define USICNT0  0b00000001

/* Tiny26 Specific Constants for TWI */
#define SDA 0b00000001
#define SCL 0b00000100

/* Status States */
#define TWI_ADDRESS  		0
#define TWI_ACK_RECIEVE  	1
#define TWI_ACK_TRANSMIT 	2
#define TWI_RECIEVE 		3
#define TWI_ACK 			4
#define TWI_TRANSMIT 		5

/* Function Return Status */
#define TWI_SUCCESS 				0
#define TWI_ERROR_INVALID_STATUS 	1
#define TWI_ERROR_NO_ACK 			2
#define TWI_ERROR_INVALID_ADDRESS 	3
#define TWI_ERROR_NO_DATA 			4
#define TWI_ERROR_BUFFER_FULL 		5

/* Global Variables */
unsigned char twi_status, twi_rx_buffer_index, twi_tx_buffer_index, twi_rx_lost_flag;
unsigned char twi_rx_buffer[TWI_BUFFER_SIZE], twi_tx_buffer[TWI_BUFFER_SIZE];

/* Prototype Functions */
void init_usi_twi_slave(void);

interrupt [USI_STRT] void usi_start_isr(void) {
	USISR = USISIF | USIOIF | USIPF | USIDC;	//clear twi_status register, write 1 to clear flags 
	USICR |= USIOIE;							//turn on the overflow interrupt
	DDRB &= ~SCL;								//release scl
	twi_status = TWI_ADDRESS;					//reset twi_status
}

interrupt [USI_OVF] void usi_overflow_isr(void) {
	unsigned char i;
   
	while(PINB & SCL);							//wait for scl to go low

	PORTB &= ~SCL;								//now hold the line low
	DDRB |= SCL;
	
	switch(twi_status) {
		case TWI_ADDRESS:						//slave address is received
			if ((USIDR >> 1) == TWI_SLAVE_ADDRESS) {
				if(USIDR & 0x01 == TWI_READ) {		//if master wants to read data slave goes into transmit mode
	            	twi_status = TWI_ACK_TRANSMIT;
	   			} 
	   			else {								//if master wants to send data slave goes into receive mode
	            	twi_status = TWI_ACK_RECIEVE;
				}
				USISR = USICNT3 | USICNT2 | USICNT1;//set twi_status counter to 0b1110 to wait for ack (high and back low)
				USIDR = 0;							//data to 0 for ack, must change only when  SCL is low
				DDRB |= SDA;						//set sda to output for ack
	      	} 
	      	else if(((USIDR >> 1) == TWI_GENERAL_ADDRESS) && (USIDR & 0x01 == TWI_WRITE)) {
				twi_status = TWI_ACK_RECIEVE;			//general address received, must be write mode because multiple slaves cannot reply at once
				USISR = USICNT3 | USICNT2 | USICNT1;	//set twi_status counter to 0b1110 to wait for ack (high and back low)
				USIDR = 0;								//data to 0 for ack, must change only when  SCL is low
				DDRB |= SDA;							//set sda to output for ack
			}
			else {
				USICR &= ~USIOIE;					//if not our address then ignore data, set overflow interrupt off
			}
			break;
		case TWI_ACK_RECIEVE:						//receive acknowledged, prepare to receive data
			twi_status = TWI_RECIEVE;
			USISR = 0;								//clear twi_status counter to prepare for data
			DDRB &= ~SDA;							//change sda to input, ack complete
			break;
	      
		case TWI_RECIEVE:							//wait for another ack to receive more data
			twi_status = TWI_ACK_RECIEVE;
			if (twi_rx_buffer_index >= (TWI_BUFFER_SIZE)) {		//store received data
            	twi_rx_lost_flag = 1;
			} 
			else {
				twi_rx_buffer[twi_rx_buffer_index++] = USIDR;
			}
			USISR = USICNT3 | USICNT2 | USICNT1;	//set twi_status counter to 0b1110 to wait for ack (high and back low)
			USIDR = 0;								//load 0 for ack
			DDRB |= SDA;							//set sda output for ack
			break;

	   case TWI_ACK:								//prepare to transmit data
			twi_status = TWI_TRANSMIT;				
			DDRB &= ~SDA;							//set sda to input
			USISR = USICNT3 | USICNT2 | USICNT1;	//set twi_status counter to 0b1110 to wait for ack (high and back low)
			USIDR = 0xFF;							//setup for ack receive
			break;
	      
	   case TWI_ACK_TRANSMIT:						//transmit acknowledged, prepare to send data
			twi_status = TWI_ACK;
			if (twi_tx_buffer_index) {					//load data to send
				twi_tx_buffer_index--;
				USIDR = twi_tx_buffer[0];
				for(i=0; i<twi_tx_buffer_index; i++) {
					twi_tx_buffer[i] = twi_tx_buffer[i+1];
				}
			}
			else {
				USIDR = 0xFF;
			}
			USISR = 0;						//clear twi_status counter to prepare for data
			PORTB |= SDA;					//set sda to high
			break;
	      
	   case TWI_TRANSMIT:					//check if ack received   
			if (~USIDR & 0x01) {			//wait for next ack
				twi_status = TWI_ACK;
				USISR = 0;					//clear counter
				if(twi_tx_buffer_index) {		//load data to send
					twi_tx_buffer_index--;
					USIDR = twi_tx_buffer[0];	         
					for(i=0; i<twi_tx_buffer_index; i++)
						twi_tx_buffer[i] = twi_tx_buffer[i+1];
				} 
				else {
					USIDR = 0xFF;
				}
				DDRB |= SDA;				// set sda to output for ack
			} 
			else {							//nack received (also can be used for no more data to send)
				USICR &= ~USIOIE;			//turn off interrupt on overflow
			}
			break;
	}

	USISR |= USIOIF;						//release SCL and keep it released
	DDRB &= ~SCL;

}

unsigned char twi_write_byte(unsigned char data) {
   
   if(twi_tx_buffer_index >= TWI_BUFFER_SIZE) {
      return TWI_ERROR_BUFFER_FULL;
   }
   
   twi_tx_buffer[twi_tx_buffer_index++] = data;
   return TWI_SUCCESS;
}

unsigned char twi_write(unsigned char *data, unsigned char length) {
   unsigned char err, i;

   for(i=0;i<length;i++) {
      err = twi_write_byte(*data);
      if(err) return err;
      data++;
   }
   return TWI_SUCCESS;   
}

unsigned char twi_writef(unsigned char flash *data, unsigned char length) {
   unsigned char err, i;

   for(i=0;i<length;i++) {
      err = twi_write_byte(*data);
      if(err) return err;
      data++;
   }
   return TWI_SUCCESS;
}

unsigned char twi_transmit_room() {
    return (TWI_BUFFER_SIZE - twi_tx_buffer_index);
}

unsigned char twi_read_byte(unsigned char *byte) {
    unsigned char i;
    
    if(twi_rx_buffer_index) {
        *byte = twi_rx_buffer[0];
        twi_rx_buffer_index--;
        for(i=0; i<twi_rx_buffer_index; i++)
            twi_rx_buffer[i] = twi_rx_buffer[i+1];
        return TWI_SUCCESS;
    } else {
        return TWI_ERROR_NO_DATA;
    }
}

unsigned char twi_read(unsigned char *data, unsigned char length) {
    unsigned char err, i;
    
    for(i = 0; i < length; i++) {
        // Ack is set to 1 except for last byte
        err = twi_read_byte(data++);
        if (err) return err;
    }
    
    return TWI_SUCCESS; 
}

unsigned char twi_receive_ready(void) {
    return twi_rx_buffer_index;
}

unsigned char twi_data_lost(void) {
   unsigned char tmp;
   tmp = twi_rx_lost_flag;
   twi_rx_lost_flag = 0;
   return tmp;
}

void main(void) {
	unsigned char read_data, counter;
	counter = 0;
        
	init_usi_twi_slave(); 
    delay_ms(1000);
    
    for(;;) {
        if (twi_receive_ready()) {
            twi_read_byte(&read_data); 

            while(twi_transmit_room()<1);
            
            twi_write_byte(counter++); 
            
            
            //twi_write_byte(count++);    
            //twi_next_transmit_ready();
            //twi_write_byte(data_h[read_data]);  
        }   
    }
}

void init_usi_twi_slave(void) {
    unsigned char i;

    twi_status = TWI_ADDRESS;
    twi_rx_buffer_index = 0;
    twi_tx_buffer_index = 0;
    twi_rx_lost_flag = 0;
    for(i = 0; i < TWI_BUFFER_SIZE; i++) {
        twi_tx_buffer[i] = 0;
    }

    USICR = USISIE | USIWM1 | USIWM0 | USICS1;		// Initialize USI
    												// Enable interrupt on start
    												// Set wire mode to TWI 
    												// Select appropriate clock source (see manual for description)  
    												
    DDRB &= ~(SDA | SCL);							// Makes sda and scl inputs
    PORTB &= ~(SDA | SCL);

    #asm("sei");
}
