/**
 * A Tiny45 Audible Alarm Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
/**
 * Tiny45 DIP Pinout
 * 									___________
 * PB5 (PCINT5/RESET/ADC0/dW)		|1		40| VCC
 * PB3 (PCINT3/XTAL1/CLKI/OC1B/ADC3)|2		39| PB2 (SCK/USCK/SCL/ADC1/T0/INT0/PCINT2)
 * PB4 (PCINT4/XTAL2/CLKO/OC1B/ADC2)|3		38| PB1 (MISO/DO/AIN1/OC0B/OC1A/PCINT1)
 * GND 								|4		37| PB0 (MOSI/DI/SDA/AIN0/0C0A/OC1A/AREF/PCINT0)
 *									-----------
 */

/* Includes */
#include <TINY45.H>
#include <string.h>
#include <stdio.h>
#include <delay.h>

/* Constants for System */
#define TRUE				1
#define FALSE				0
#define CPU_FREQUENCY_HZ	2000000

/* Constants for Timer */
#define TIMER_CLOCK_DIVIDER		1		// With a 8 MHz CPU
										//	Clock divider = 1 => 125 ns to 31.875 us
										//	" = 8 => 1 us to 255 us
										//	" = 64 => 8 us to 2.04 ms
										//	" = 256 => 32 us to 8.16 ms
										//	" = 1024 => 128 us to 32.64 ms
										// With a 1 MHz CPU
										//	" = 1 => 1 us to 255 us
										//	" = 8 => 8 us to 2.04 s
										//	" = 64 => 64 us to 8.16 ms
										//	" = 256 => 256 us to 32.64 ms
										//	" = 1024 => 1024 us to 261.12 ms
#define TIMER_RES_US			50		
#define TIMER_COMPARE_VALUE		(TIMER_RES_US * CPU_FREQUENCY_HZ) / (1000000 * TIMER_CLOCK_DIVIDER)

/* Constants for Audible Alarm */
#define AUDIBLE_ALARM_CALIBRATION_MODE		FALSE 
#define AUDIBLE_ALARM_FREQUENCY_HZ			1230
#define AUDIBLE_ALARM_TOGGLE_PERIOD_US		1000000 / (2 * AUDIBLE_ALARM_FREQUENCY_HZ)
#define AUDIBLE_ALARM_TOGGLE_THRESHOLD_OFFSET	-1
#define AUDIBLE_ALARM_TOGGLE_THRESHOLD		AUDIBLE_ALARM_TOGGLE_PERIOD_US / TIMER_RES_US + AUDIBLE_ALARM_TOGGLE_THRESHOLD_OFFSET
#define AUDIBLE_ALARM_ENABLE_US				50000
#define AUDIBLE_ALARM_ENABLE_THRESHOLD		AUDIBLE_ALARM_ENABLE_US / TIMER_RES_US

#define AUDIBLE_ALARM_ACTIVE_FINAL_LATENCY_US		50000
#define AUDIBLE_ALARM_ACTIVE_FINAL_THRESHOLD		AUDIBLE_ALARM_ACTIVE_FINAL_LATENCY_US / TIMER_RES_US
#define AUDIBLE_ALARM_ACTIVE_INIT_LATENCY_US		250000
#define AUDIBLE_ALARM_ACTIVE_INIT_THRESHOLD			AUDIBLE_ALARM_ACTIVE_INIT_LATENCY_US / TIMER_RES_US
#define AUDIBLE_ALARM_ACTIVE_MOD_US					25000
#define AUDIBLE_ALARM_ACTIVE_MOD_THRESHOLD			AUDIBLE_ALARM_ACTIVE_MOD_US / TIMER_RES_US
#define AUDIBLE_ALARM_ACTIVE_MOD_LATENCY_US			2000000
#define AUDIBLE_ALARM_ACTIVE_MOD_LATENCY_THRESHOLD	AUDIBLE_ALARM_ACTIVE_MOD_LATENCY_US / TIMER_RES_US		

#define AUDIBLE_ALARM_PAUSE_LATENCY_MS		10000  
#define AUDIBLE_ALARM_PAUSE_LATENCY_ONE_SEC	1000000
#define AUDIBLE_ALARM_PAUSE_THRESHOLD		AUDIBLE_ALARM_PAUSE_LATENCY_ONE_SEC / TIMER_RES_US
#define AUDIBLE_ALARM_PAUSE_COUNT_THRESHOLD	AUDIBLE_ALARM_PAUSE_LATENCY_MS * 1000 / AUDIBLE_ALARM_PAUSE_LATENCY_ONE_SEC

/* Constants for Pushbutton Input */
#define DEBOUNCE_PUSH_COUNT_THRESHOLD		4
#define DEBOUNCE_LATENCY_US					50000
#define DEBOUNCE_LATENCY_THRESHOLD			DEBOUNCE_LATENCY_US / TIMER_RES_US

/* Global Variables */
unsigned int audible_alarm_active_tick_count, audible_alarm_toggle_tick_count, audible_alarm_enable_tick_count;
unsigned int audible_alarm_pause_tick_count, audible_alarm_pause_count;
unsigned int audible_alarm_active_threshold, audible_alarm_active_mod_tick_count;

unsigned int debounce_latency_tick_count;
unsigned char debounce_push_count;
enum{DEBOUNCE_DEPUSHED, DEBOUNCE_MAYBE_PUSHED, DEBOUNCE_PUSHED, DEBOUNCE_MAYBE_DEPUSHED};
unsigned char debounce_state;

unsigned char audible_alarm_enable_flag, audible_alarm_active_flag, audible_alarm_toggle_flag;
unsigned char audible_alarm_pause_flag;

/* Prototype Functions */
void init_system(void);
void init_timer0(void);
void init_external_int(void);

interrupt [TIM0_COMPA] void timer0_comp_a(void) {
	audible_alarm_active_tick_count++;
	audible_alarm_toggle_tick_count++;
	audible_alarm_enable_tick_count++;
	audible_alarm_pause_tick_count++;
	audible_alarm_active_mod_tick_count++;
	debounce_latency_tick_count++;	
} 

interrupt [EXT_INT0] void pin_change(void) {
	audible_alarm_enable_tick_count = 0;
	audible_alarm_enable_flag = 0;
	audible_alarm_pause_flag = 0;
}

void main(void) {
	DDRB.0 = 1;
	DDRB.1 = 0;

	init_system();
	init_timer0();
	init_external_int();
	
	#asm("sei");
	
	delay_ms(5000);
    
    audible_alarm_active_threshold = AUDIBLE_ALARM_ACTIVE_INIT_THRESHOLD;
	
	audible_alarm_active_tick_count = 0;
	audible_alarm_toggle_tick_count = 0;
	audible_alarm_enable_tick_count = 0;
	audible_alarm_pause_tick_count = 0;
	audible_alarm_active_mod_tick_count = 0;
	
	audible_alarm_enable_flag = 0;
	audible_alarm_active_flag = 0;
	audible_alarm_toggle_flag = 0;
	audible_alarm_pause_flag = 0;
	
	debounce_latency_tick_count = 0;
	debounce_state = DEBOUNCE_DEPUSHED;
	
	while(1) {
		if (audible_alarm_enable_flag) {
			if (audible_alarm_toggle_tick_count > AUDIBLE_ALARM_TOGGLE_THRESHOLD) {
				if (!audible_alarm_active_flag || audible_alarm_toggle_flag) {
					PORTB.0 = 0;
					audible_alarm_toggle_flag = 0;
				}
				else {
					PORTB.0 = 1;
					audible_alarm_toggle_flag = 1;
				}
				audible_alarm_toggle_tick_count = 0;
			}
			if (audible_alarm_active_tick_count > audible_alarm_active_threshold) {
				if (AUDIBLE_ALARM_CALIBRATION_MODE || !audible_alarm_active_flag) {
					audible_alarm_active_flag = 1;
				}
				else {
					audible_alarm_active_flag = 0;
				}
				audible_alarm_active_tick_count = 0;
			}
			if (audible_alarm_active_mod_tick_count > AUDIBLE_ALARM_ACTIVE_MOD_LATENCY_THRESHOLD) {
				if (audible_alarm_active_threshold > AUDIBLE_ALARM_ACTIVE_FINAL_THRESHOLD) {
					audible_alarm_active_threshold -= AUDIBLE_ALARM_ACTIVE_MOD_THRESHOLD;
				}
			   	audible_alarm_active_mod_tick_count = 0;
			}
			if (debounce_latency_tick_count > DEBOUNCE_LATENCY_THRESHOLD) {
	    		switch(debounce_state) {
	    			case DEBOUNCE_DEPUSHED:
	    				if (PINB.1 != 1) {
	    					debounce_state = DEBOUNCE_MAYBE_PUSHED;
	    				}
	    				break;
	    			case DEBOUNCE_MAYBE_PUSHED:
	    			    if (PINB.1 != 1) {
	    			    	debounce_state = DEBOUNCE_PUSHED;
	    			    	debounce_push_count++;
	    			    	if (debounce_push_count >= DEBOUNCE_PUSH_COUNT_THRESHOLD) {
		    			    	PORTB.0 = 0;
								audible_alarm_pause_flag = 1;
								audible_alarm_enable_flag = 0;
								audible_alarm_pause_count = 0;
								audible_alarm_pause_tick_count = 0;
								debounce_push_count = 0;
							}
	    			    }
	    			    else {
	    			    	debounce_state = DEBOUNCE_DEPUSHED;
	    			    }
	    				break;
	    			case DEBOUNCE_PUSHED:
	    				if (PINB.1 == 1) {
	    			    	debounce_state = DEBOUNCE_MAYBE_DEPUSHED;
	    			    }
	    				break;
	    			case DEBOUNCE_MAYBE_DEPUSHED:
	    				if (PINB.1 == 1) {
	    			    	debounce_state = DEBOUNCE_DEPUSHED;
	    			    }
	    			    else {
	    			    	debounce_state = DEBOUNCE_PUSHED;
	    			    }
	    				break;
	    		}
	    		debounce_latency_tick_count = 0;
    		}
		}
		else if (audible_alarm_pause_flag && audible_alarm_pause_tick_count > AUDIBLE_ALARM_PAUSE_THRESHOLD) {
			audible_alarm_pause_count++;
			if (audible_alarm_pause_count >= AUDIBLE_ALARM_PAUSE_COUNT_THRESHOLD) {
				audible_alarm_pause_flag = 0;
				audible_alarm_enable_tick_count = 0;
			}
			audible_alarm_pause_tick_count = 0;
		}
		else if (!audible_alarm_pause_flag && audible_alarm_enable_tick_count > AUDIBLE_ALARM_ENABLE_THRESHOLD) {
			audible_alarm_enable_flag = 1;
			audible_alarm_toggle_flag = 0;
			audible_alarm_active_flag = 0;
            
			audible_alarm_active_threshold = AUDIBLE_ALARM_ACTIVE_INIT_THRESHOLD;
			
			audible_alarm_toggle_tick_count = 0;
			audible_alarm_active_tick_count = 0;
			audible_alarm_active_mod_tick_count = 0;
		}
		else {
 			PORTB.0 = 0;
		}	
	}
}

void init_system(void) {
	MCUCR = 0b00000000;		// 7 - BODS: BOD Sleep
							// 6 - Reserved
							// 5 - SE: Sleep Enable
							// 4:3 - SM1:0: Sleep Mode
							//		00 - Idle
							//		01 - ADC Noise Reduction
							//		10 - Power-down
							//		11 - Reserved
							// 2 - BODSE: BOD Sleep Enable
							// 1:0 - Reserved
							
	CLKPR = 0b10000000;		// 7 - CLKPCE: Clock Prescaler Change Enable
							// 6:4 - Reserved
							// 3:0 - CLKPS3:0: Clock Prescaler Select
							//		0000 - 1
							//		0001 - 2
							//		0010 - 4
							//		0011 - 8
							//		0100 - 16
							//		0101 - 32
							//		0110 - 64
							//		0111 - 128
							//		1000 - 256
							//		.... - Reserved
	CLKPR = 0b00000010;
	
	PRR = 0b00001011;		// 7:4 - Reserved
							// 3 - PRTIM1: Power Reduction Timer1
							// 2 - PRTIM0: Power Reduction Timer0
							// 1 - PRUSI: Power Reduction USI
							// 0 - PRADC: Power Reduction ADC
}

void init_timer0(void) {
	TCCR0A = 0b00000010;	// 7:6 - COM0A1:0: Compare Match Output A Mode
							//		00 - Normal, OC0A disconnected
							//		01 - Toggle OC0A
							//		10 - Clear OC0A
							//		11 - Set OC0A
							// 5:4 - COM0B1:0: " B "
							// 3:2 - Reserved
							// 1:0 - WGM01:0: Waveform Generation Mode
							//		WGM2|WGM1|WGM0
							//		000 - Normal
							//		010 - CTC
	
	switch(TIMER_CLOCK_DIVIDER) {
		case 1:
			TCCR0B = 0b001;
			break;
		case 8:
			TCCR0B = 0b010;
			break;
		case 64:
			TCCR0B = 0b011;
			break;
		case 256:
			TCCR0B = 0b100;
			break;
		case 1024:
			TCCR0B = 0b101;
			break;
	}					
	TCCR0B |= 0b00000000;// 7 - FOC0A: Force Output Compare A
						// 6 - FOC0B: " B
						// 5:4 - Reserved
						// 3 - WGM02: Waveform Generation Mode (see above)
						// 2:0 - CS02:0: Clock select
						//		000 - None
						//		001 - 1
						//		010 - 8
						//		011 - 64
						//		100 - 256
						//		101 - 1024
						//		110 - External clock, T0 pin, falling edge
						//		111 - ", ", rising edge	
	
	// TIMER_RES_US = 1 / CPU_FREQUENCY_HZ * TIMER_CLOCK_DIVIDER * TIMER_COMPARE_VALUE
	// TIMER_COMPARE_VALUE = TIMER_RES_US * CPU_FREQUENCY_HZ / TIMER_CLOCK_DIVIDER			
	OCR0A = TIMER_COMPARE_VALUE;
	
	TIMSK = 0b00010000;	// 7 - Reserved
						// 6:5 - Reserved for Timer1
						// 4 - Timer0 Output Comapre Match A Interrupt Enable
						// 3 - " B "
						// 2 - Reserved for Timer1
						// 1 - Timer0 Overflow Interrupt Enable
						// 0 - Reserved	
}

void init_external_int(void) {
	MCUCR |= 0b00000001;	// 7:2 - Reserved for other functions
							// 2:0 - ISC01:0: Interrupt Sense Control
							//		00 - Low level of INT0
							//		01 - Logical change of INT0
							//		10 - Falling edge of INT0
							//		11 - Rising edge of INT0
							
	GIMSK = 0b01000000;		// 7 - Reserved
							// 6 - INT0: External Interrupt 0 Enable
							// 5 - PCIE: Pin Change Interrupt Enable
							// 4:0 - Reserved
	
	GIFR = 0b000000000;		// 7 - Reserved
							// 6 - INTF0: External Interrupt 0 Flag
							// 5 - PCIF: Pin Change Interrupt Flag
							// 4:0 - Reserved
							
	PCMSK = 0b00000000;		// 7:6 - Reserved
							// 5:0 - PCINT5:0: Pin Change Enable Mask
}
	
