/**
 * A Tiny45 PWM Hardware Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
/**
 * Tiny45 DIP Pinout
 * 									___________
 * PB5 (PCINT5/RESET/ADC0/dW)		|1		40| VCC
 * PB3 (PCINT3/XTAL1/CLKI/OC1B/ADC3)|2		39| PB2 (SCK/USCK/SCL/ADC1/T0/INT0/PCINT2)
 * PB4 (PCINT4/XTAL2/CLKO/OC1B/ADC2)|3		38| PB1 (MISO/DO/AIN1/OC0B/OC1A/PCINT1)
 * GND 								|4		37| PB0 (MOSI/DI/SDA/AIN0/0C0A/OC1A/AREF/PCINT0)
 *									-----------
 */

/* Includes */
#include <TINY45.H>
#include <string.h>
#include <stdio.h>
#include <delay.h>

/* Constants for System */
#define CPU_FREQUENCY_HZ	8000000

/* Constants for PWM */
#define PWM_CLOCK_DIVIDER		64			// With a 8 MHz CPU
										//	Clock divider = 1 => 125 ns to 31.875 us
										//	" = 2 => 250 ns to 63.75 us
										//	" = 4 => 500 ns to 127.5 us
										//	" = 8 => 1 us to 255 us
										//	...
#define PWM_TOGGLE_FREQUENCY_HZ		1230
#define PWM_COMPARE_VALUE			CPU_FREQUENCY_HZ / (2 * PWM_TOGGLE_FREQUENCY_HZ * PWM_CLOCK_DIVIDER)

/* Global Variables */

/* Prototype Functions */
void init_pwm(void);

void main(void) {
	init_pwm();

	while(1) {
	}
}

void init_system(void) {
	MCUCR = 0b00000000;		// 7 - BODS: BOD Sleep
							// 6 - Reserved
							// 5 - SE: Sleep Enable
							// 4:3 - SM1:0: Sleep Mode
							//		00 - Idle
							//		01 - ADC Noise Reduction
							//		10 - Power-down
							//		11 - Reserved
							// 2 - BODSE: BOD Sleep Enable
							// 1:0 - Reserved
	
	PRR = 0b00000000;		// 7:4 - Reserved
							// 3 - PRTIM1: Power Reduction Timer1
							// 2 - PRTIM0: Power Reduction Timer0
							// 1 - PRUSI: Power Reduction USI
							// 0 - PRADC: Power Reduction ADC
}

void init_pwm(void) {
	DDRB.0 = 1;
	
	TCCR0A = 0b01000010;// 7:6 - COM0A1:0: Compare Match Output A Mode
						//		00 - Normal, OC0A disconnected
						//		01 - Toggle OC0A
						//		10 - Clear OC0A
						//		11 - Set OC0A
						// 5:4 - COM0B1:0: " B "
						// 3:2 - Reserved
						// 1:0 - WGM01:0: Waveform Generation Mode
						//		WGM2|WGM1|WGM0
						//		000 - Normal
						//		010 - CTC
	
	switch(PWM_CLOCK_DIVIDER) {
		case 1:
			TCCR0B = 0b001;
			break;
		case 8:
			TCCR0B = 0b010;
			break;
		case 64:
			TCCR0B = 0b011;
			break;
		case 256:
			TCCR0B = 0b100;
			break;
		case 1024:
			TCCR0B = 0b101;
			break;
	}
						
	TCCR0B |= 0b00000000;// 7 - FOC0A: Force Output Compare A
						// 6 - FOC0B: " B
						// 5:4 - Reserved
						// 3 - WGM02: Waveform Generation Mode (see above)
						// 2:0 - CS02:0: Clock select
						//		000 - None
						//		001 - 1
						//		010 - 8
						//		011 - 64
						//		100 - 256
						//		101 - 1024
						//		110 - External clock, T0 pin, falling edge
						//		111 - ", ", rising edge	
	
	
	// Timer Resolution = 1 / CPU Frequency * OCR0A
	// OCR0A = Timer Resolution * CPU Frequency = TIMER_RES_US * CPU_FREQUENCY_HZ / 1E6					
	OCR0A = PWM_COMPARE_VALUE;
	
	TIMSK = 0b00010000;	// 7 - Reserved
						// 6:5 - Reserved for Timer1
						// 4 - Timer0 Output Comapre Match A Interrupt Enable
						// 3 - " B "
						// 2 - Reserved for Timer1
						// 1 - Timer0 Overflow Interrupt Enable
						// 0 - Reserved	
}
	
