/**
 * A Tiny45 PWM Software Test
 *
 * Ruibing Wang (rw98@cornell.edu)
 */
 
/**
 * Tiny45 DIP Pinout
 * 									___________
 * PB5 (PCINT5/RESET/ADC0/dW)		|1		40| VCC
 * PB3 (PCINT3/XTAL1/CLKI/OC1B/ADC3)|2		39| PB2 (SCK/USCK/SCL/ADC1/T0/INT0/PCINT2)
 * PB4 (PCINT4/XTAL2/CLKO/OC1B/ADC2)|3		38| PB1 (MISO/DO/AIN1/OC0B/OC1A/PCINT1)
 * GND 								|4		37| PB0 (MOSI/DI/SDA/AIN0/0C0A/OC1A/AREF/PCINT0)
 *									-----------
 */

/* Includes */
#include <TINY45.H>
#include <string.h>
#include <stdio.h>
#include <delay.h>

/* Constants */
#define CPU_FREQUENCY_HZ	8000000
#define CLOCK_DIVIDER		8			// With a 20 MHz CPU
										//	Clock divider = 1 => 50 ns to 12.75 us
										//	" = 8 => 400 ns to 102 us
										//	" = 64 => 3.2 us to 816 us
										//	" = 256 => 12.8 us to 3.264 ms
										//	" = 1024 => 51.2 us to 13.056 ms
										// With a 8 MHz CPU
										//	Clock divider = 1 => 125 ns to 31.875 us
										//	" = 8 => 1 us to 255 us
										//	" = 64 => 8 us to 2.04 ms
										//	" = 256 => 32 us to 8.16 ms
										//	" = 1024 => 128 us to 32.64 ms
#define TIMER_RES_US		50
#define TIMER0_COMPARE_VALUE	(TIMER_RES_US * CPU_FREQUENCY_HZ) / (1000000 * CLOCK_DIVIDER)

unsigned char led;

void init_timer0(void);

interrupt [TIM0_COMPA] void timer0_comp_a(void) {
	led = ~led;
	PORTB.0 = led;
}


void main(void) {
	DDRB.0 = 1;
	PORTB.0 = 1;
	
	init_timer0();
	
	#asm("sei");
	
	led = 0;
	
	while(1) {
	}
}

/**
 * Timer0 Initialization
 * 
 * Timer interrupt period = 1 / CPU * Clock Divider * OCR0A
 *
 */
void init_timer0(void) {
	//GTCCR = 0x00;		// 7 - Timer/Counter Synchronization Mode
						// 6:1 - Reserved for Timer1
						// 0 - Prescaler Reset Timer/Counter0
						
	TCCR0A = 0b00000010;// 7:6 - Compare Match Output A Mode
						//		00 - Normal, OC0A disconnected
						//		01 - Toggle OC0A
						//		10 - Clear OC0A
						//		11 - Set OC0A
						// 5:4 - " B "
						// 3:2 - Reserved
						// 1:0 - Waveform Generation Mode
						//		TCCR0B.3|TCCR0A.1|TCCR0A.0
						//		000 - Normal
						//		010 - CTC
	
	switch(CLOCK_DIVIDER) {
		case 1:
			TCCR0B = 0b001;
			break;
		case 8:
			TCCR0B = 0b010;
			break;
		case 64:
			TCCR0B = 0b011;
			break;
		case 256:
			TCCR0B = 0b100;
			break;
		case 1024:
			TCCR0B = 0b101;
			break;
	}
						
	TCCR0B |= 0b00000000;// 7 - Force Output Compare A
						// 6 - " B
						// 5:4 - Reserved
						// 3 - Waveform Generation Mode (see above)
						// 2:0 - Clock select
						//		000 - None
						//		001 - 1
						//		010 - 8
						//		011 - 64
						//		100 - 256
						//		101 - 1024
						//		110 - External clock, T0 pin, falling edge
						//		111 - ", ", rising edge	
	
	
	// Timer Resolution = 1 / CPU Frequency * OCR0A
	// OCR0A = Timer Resolution * CPU Frequency = TIMER_RES_US * CPU_FREQUENCY_HZ / 1E6					
	OCR0A = TIMER0_COMPARE_VALUE;
	
	TIMSK = 0b00010000;	// 7 - Reserved
						// 6:5 - Reserved for Timer1
						// 4 - Timer0 Output Comapre Match A Interrupt Enable
						// 3 - " B "
						// 2 - Reserved for Timer1
						// 1 - Timer0 Overflow Interrupt Enable
						// 0 - Reserved	
}
	