import java.util.concurrent.TimeUnit;

/**
 * The communication class that acts as a bridge between the high levle Sentinel UI class 
 * and the low level Serial communication class. This class is responsible for polling the FIFO
 * buffer in the Serial class performing the parsing.
 * 
 * @author Ruibing Wang (rw98@cornell.edu)
 *
 */
public class Communicator extends Thread {
	/* UART RXTX Protocol */
	final private int UART_RX_SENTINEL_STATUS 			=	0x00;
	final private int UART_RX_SENTINEL_LOCAL_DHCP		= 	0x01;
	
	final private int UART_RX_SENTINEL_SENSOR_DETECTION_RELAY	=	0x10;
	final private int UART_RX_SENTINEL_SENSOR_DETECTION_REPORT	=	0x11;
	
	final private int UART_TX_SENTINEL_DISCOVER			=	0x20;
	final private int UART_RX_SENTINEL_DISCOVER			=	0x21;
	final private int UART_TX_SENTINEL_INFO				=	0x22;
	final private int UART_RX_SENTINEL_INFO				=	0x23;
	
	final private int UART_TX_SENTINEL_CLEAR_CLIENT_LIST=	0x30;
	final private int UART_TX_SENTINEL_CLEAR_RELAY_LIST	=	0x31;
	final private int UART_TX_SENTINEL_CLEAR_SENSOR_LIST=	0x32;
	final private int UART_TX_SENTINEL_CLEAR_LOCAL_DHCP	=	0x33;
	
	final private int UART_RX_ZIGBEE_STATE 				=	0xA0;
	final private int UART_RX_ZIGBEE_IRQ_STATUS 		=	0xA1;
	final private int UART_RX_ZIGBEE_TRAC_STATUS 		=	0xA2;
	
	final private int UART_RX_ZIGBEE_PHY_DATA 			=	0xB0;	// UART_TX_ZIGBEE_RX_DATA | length | data
	final private int UART_RX_ZIGBEE_PHY_LQI 			=	0xB1;
	
	final private int UART_RX_ZIGBEE_MAC_DATA 			=	0xB2;
	final private int UART_RX_ZIGBEE_MAC_DEST_PAN_ID 	=	0xB3;
	final private int UART_RX_ZIGBEE_MAC_DEST_ADDR 		= 	0xB4;
	final private int UART_RX_ZIGBEE_MAC_SRC_PAN_ID 	=	0xB5;
	final private int UART_RX_ZIGBEE_MAC_SRC_ADDR 		=	0xB6;
	
	final private int UART_RX_ZIGBEE_ED_LEVEL 			= 	0xC3;
	
	final private int UART_RX_ZIGBEE_PART_NUM 			=	0xD0;
	final private int UART_RX_ZIGBEE_VER_NUM 			=	0xD1;
	final private int UART_RX_ZIGBEE_MAN_ID_0 			= 	0xD2;
	final private int UART_RX_ZIGBEE_MAN_ID_1 			=	0xD3;
	
	final private int SENTINEL_BASE_NODE				=	0x00;
	final private int SENTINEL_RELAY_NODE				=	0x01;
	final private int SENTINEL_SENSOR_NODE				=	0x02;
	
	/* Zigbee Status */
	final private int ZIGBEE_TRX_STATUS_P_ON 				= 	0x00;	// Power-on after VDD (XOSC=On, Pull-up=On)
	final private int ZIGBEE_TRX_STATUS_TRX_OFF 			= 	0x08;	// Clock state (XOSC=On, Pull-up=Off)
	final private int ZIGBEE_TRX_STATUS_SLEEP 				= 	0x0F;	// Sleep State (XOSC=Off, Pull-up=Off)
	final private int ZIGBEE_TRX_STATUS_PLL_ON 				=	0x09;	// PLL State
	final private int ZIGBEE_TRX_STATUS_BUSY_TX 			=	0x02;	// Transmit State
	final private int ZIGBEE_TRX_STATUS_RX_ON 				=	0x06;	// Rx Listen State
	final private int ZIGBEE_TRX_STATUS_BUSY_RX 			=	0x01;	// Receive State
	final private int ZIGBEE_TRX_STATUS_RX_ON_NOCLK 		=	0x1C;	// Rx Listen State (CLKM=Off)
	final private int ZIGBEE_TRX_STATUS_BUSY_RX_AACK 		= 	0x11;
	final private int ZIGBEE_TRX_STATUS_BUSY_TX_ARET 		=	0x12;
	final private int ZIGBEE_TRX_STATUS_RX_AACK_ON 			=	0x16;
	final private int ZIGBEE_TRX_STATUS_TX_ARET_ON 			=	0x19;
	final private int ZIGBEE_TRX_STATUS_RX_AACK_NOCLK 		=	0x1D;
	final private int ZIGBEE_TRX_STATUS_BUSY_RX_AACK_NOCLK 	=	0x1E;
	final private int ZIGBEE_TRX_STATUS_TRANSITION 			=	0x1F;
	
	/* Sentinel Status Protocol */
	final private int SENTINEL_STATUS_P_ON 				=	0x00;
	final private int SENTINEL_STATUS_IDLE 				=	0x01;
	final private int SENTINEL_STATUS_SLEEP 			= 	0x02;
	final private int SENTINEL_STATUS_TRX_OFF 			=	0x10;
	final private int SENTINEL_STATUS_TX 				= 	0x20;
	final private int SENTINEL_STATUS_RX_LISTEN 		= 	0x30;
	final private int SENTINEL_STATUS_RX_LISTEN_SLEEP 	= 	0x31;
	final private int SENTINEL_STATUS_RX_BUSY 			= 	0x32;
	final private int SENTINEL_STATUS_RX_SUCCESS		= 	0x33;
	final private int SENTINEL_STATUS_RX_LOW_LQI		= 	0x34;
	final private int SENTINEL_STATUS_RX_DUPLICATE_FRAME= 	0x35;
	
	/* Sentinel Local DHCP Protocol */
	final private int SENTINEL_LOCAL_DHCP_TX_DISCOVERY				=	0x00;
	final private int SENTINEL_LOCAL_DHCP_RX_DISCOVERY				=	0x01;
	final private int SENTINEL_LOCAL_DHCP_RX_DISCOVERY_MAX_RELAY	=	0x02;
	final private int SENTINEL_LOCAL_DHCP_RX_DISCOVERY_MAX_SENSOR	=	0x03;
	final private int SENTINEL_LOCAL_DHCP_TX_OFFER					=	0x10;
	final private int SENTINEL_LOCAL_DHCP_RX_OFFER_MODE				=	0x11;
	final private int SENTINEL_LOCAL_DHCP_RX_OFFER					=	0x12;
	final private int SENTINEL_LOCAL_DHCP_TX_REQUEST				=	0x20;
	final private int SENTINEL_LOCAL_DHCP_RX_REQUEST				=	0x21;
	final private int SENTINEL_LOCAL_DHCP_RX_REQUEST_NO_MATCH		=	0x22;
	final private int SENTINEL_LOCAL_DHCP_TX_ACK					=	0x30;
	final private int SENTINEL_LOCAL_DHCP_RX_ACK					=	0x31;
	final private int SENTINEL_LOCAL_DHCP_CONFIG					=	0x40;
	final private int SENTINEL_LOCAL_DHCP_COMPLETE					=	0x41;
	final private int SENTINEL_LOCAL_DHCP_RX_ACK_TIMEOUT			=	0x50;
	final private int SENTINEL_LOCAL_DHCP_RX_OFFER_TIMEOUT			=	0x51;
	
	private Serial serial = null;
	private Sentinel sentinel = null;
	
	private boolean serialDebug = false;
	
	private Integer rxHeader, rxLength, rxData;
	private String rxOutput;

	/**
	 * This method constructs the Communicator class by passing the Sentinel initialization class
	 * It also calls the lower Serial class to handle the low level serial/uart transmission
	 * 
	 * @param sentinel
	 */
	public Communicator(Sentinel sentinel) {
		this.sentinel = sentinel;
		serial = new Serial(this);
	}
	
	/**
	 * This method calls the Thread superclass for the JVM to start the separate thread
	 */
	public void start() {
		super.start();
	}
	
	/**
	 * This method is initialized and run by the JVM through the start function
	 * It is intented to run and continue parsing of the received data from a FIFO LinkedBlockQueue in the Serial class
	 */
	public void run() {
		while (true) {
			rxHeader = pollRxQueue();
			if (rxHeader != null) {
				switch (rxHeader.intValue()) {
					case UART_RX_SENTINEL_STATUS:
						rxOutput = "sentinelStatus: ";
						rxData = pollRxQueue();
						if (rxData != null) {
							switch(rxData.intValue()) {
								case SENTINEL_STATUS_P_ON:
									rxOutput += "Power On - ";
									rxData = pollRxQueue();
									if (rxData != null) {
										switch(rxData.intValue()) {
											case SENTINEL_BASE_NODE:
												rxOutput += "Base Node - LocalDepth: [" + pollByteAsInt() + "], LocalAddr: [" + pollShortAsHex() + "] LocalPanID: [" + pollShortAsHex() + "]";
												break;
											case SENTINEL_RELAY_NODE:
												rxOutput += "Relay Node";
												break;
											case SENTINEL_SENSOR_NODE:
												rxOutput += "Sensor Node";
												break;
										}
									}
									else {
										rxOutput += "Error";
									}
									break;
								case SENTINEL_STATUS_IDLE:
									rxOutput += "Idle";
									break;
								case SENTINEL_STATUS_SLEEP:
									rxOutput += "Sleep";
									break;
								case SENTINEL_STATUS_TRX_OFF:
									rxOutput += "TRX Off";
									break;
								case SENTINEL_STATUS_TX:
									rxOutput += "TX";
									rxLength = pollRxQueue();
									if (rxLength != null) {
										rxOutput += "[" + rxLength.intValue() + "]: " + pollByteArrayAsHex(rxLength.intValue());
										rxOutput += " - DestShortAddr: [" + pollShortAsHex() + "], DestPanID: [" + pollShortAsHex() + "]";
									}
									else {
										rxOutput += ": ERROR";
									}
									break;
								case SENTINEL_STATUS_RX_LISTEN:
									rxOutput += "RX Listen";
									break;
								case SENTINEL_STATUS_RX_LISTEN_SLEEP:
									rxOutput += "RX Listen Sleep";
									break;
								case SENTINEL_STATUS_RX_BUSY:
									rxOutput += "RX Busy";
									break;
								case SENTINEL_STATUS_RX_SUCCESS:
									rxOutput += "RX";
									rxLength = pollRxQueue();
									if (rxLength != null) {
										rxOutput += "[" + rxLength.intValue() + "]: " + pollByteArrayAsHex(rxLength.intValue());
										rxOutput += " - SrcShortAddr: [" + pollShortAsHex() + "], SrcPanID: [" + pollShortAsHex() + "]";
									}
									else {
										rxOutput += ": ERROR";
									}
									break;
								case SENTINEL_STATUS_RX_LOW_LQI:
									rxOutput += "RX Low LQI";
									break;
								case SENTINEL_STATUS_RX_DUPLICATE_FRAME:
									rxOutput += "RX Duplicate Frame";
									break;
								default:
									rxOutput += "Error";
									break;
							}
						}
						else {
							rxOutput += " Error";
						}
						break;
					case UART_RX_SENTINEL_LOCAL_DHCP:
						rxOutput = "localDHCP: ";
						rxData = pollRxQueue();
						if (rxData != null) {
							switch(rxData.intValue()) {
								case SENTINEL_LOCAL_DHCP_TX_DISCOVERY:
									rxOutput += "TX Discovery - ";
									rxData = pollRxQueue();
									if (rxData != null) {
										switch(rxData.intValue()) {
											case SENTINEL_BASE_NODE:
												rxOutput += "Base Node Pan ID";
												break;
											case SENTINEL_RELAY_NODE:
												rxOutput += "Relay Node Pan ID";
												break;
											case SENTINEL_SENSOR_NODE:
												rxOutput += "Sensor Node Pan ID";
												break;
										}
									}
									else {
										rxOutput += "Error";
									}
									break;
								case SENTINEL_LOCAL_DHCP_RX_DISCOVERY:
									rxOutput += "RX Discovery - ";
									rxData = pollRxQueue();
									if (rxData != null) {
										switch(rxData.intValue()) {
											case SENTINEL_BASE_NODE:
												rxOutput += "Base Node";
												break;
											case SENTINEL_RELAY_NODE:
												rxOutput += "Relay Node";
												break;
											case SENTINEL_SENSOR_NODE:
												rxOutput += "Sensor Node";
												break;
										}
									}
									else {
										rxOutput += "Error";
									}
									break;
								case SENTINEL_LOCAL_DHCP_RX_DISCOVERY_MAX_RELAY:
									rxOutput += "RX Discovery Max Relay";
									break;
								case SENTINEL_LOCAL_DHCP_RX_DISCOVERY_MAX_SENSOR:
									rxOutput += "RX Discovery Max Sensor";
									break;
								case SENTINEL_LOCAL_DHCP_TX_OFFER:
									rxOutput += "TX Offer - OfferedAddr: [" + pollShortAsHex() + "]";
									break;
								case SENTINEL_LOCAL_DHCP_RX_OFFER_MODE:
									rxOutput += "RX Offer Mode";
									break;
								case SENTINEL_LOCAL_DHCP_RX_OFFER:
									rxOutput += "RX Offer - OfferNo. " + pollByteAsInt() + ", ServerType: ";
									rxData = pollRxQueue();
									if (rxData != null) {
										switch(rxData.intValue()) {
											case SENTINEL_BASE_NODE:
												rxOutput += "Base Node";
												break;
											case SENTINEL_RELAY_NODE:
												rxOutput += "Relay Node";
												break;
											case SENTINEL_SENSOR_NODE:
												rxOutput += "Sensor Node";
												break;
										}
									}
									else {
										rxOutput += "Error";
									}
									rxOutput += ", ServerDepth: [" + pollByteAsInt() + "], ED Level: [" + pollByteAsInt() + "], OfferedAddr: [" + pollShortAsHex() + "] ServerAddr: [" + pollShortAsHex() + "]";
									break;
								case SENTINEL_LOCAL_DHCP_TX_REQUEST:
									rxOutput += "TX Request - OfferNo. " + pollByteAsInt() + ", ServerType: ";
									rxData = pollRxQueue();
									if (rxData != null) {
										switch(rxData.intValue()) {
											case SENTINEL_BASE_NODE:
												rxOutput += "Base Node";
												break;
											case SENTINEL_RELAY_NODE:
												rxOutput += "Relay Node";
												break;
											case SENTINEL_SENSOR_NODE:
												rxOutput += "Sensor Node";
												break;
										}
									}
									else {
										rxOutput += "Error";
									}
									rxOutput += ", ServerDepth: [" + pollByteAsInt() + "], ED Level: [" + pollByteAsInt() + "], ServerAddr: [" + pollShortAsHex() + "], OfferedAddr: [" + pollShortAsHex() + "]";
									break;
								case SENTINEL_LOCAL_DHCP_RX_REQUEST:
									rxOutput += "RX Request - ServerAddr: [" + pollShortAsHex() + "]";
									break;
								case SENTINEL_LOCAL_DHCP_RX_REQUEST_NO_MATCH:
									rxOutput += "RX Request No Match - LocalAddr: [" + pollShortAsHex() + "]";
									break;
								case SENTINEL_LOCAL_DHCP_TX_ACK:
									rxOutput += "TX Acknowledgement - OfferedDepth: [" + pollByteAsInt() + "], OfferedAddr: [" + pollShortAsHex() + "]";
									break;
								case SENTINEL_LOCAL_DHCP_RX_ACK:
									rxOutput += "RX Acknowledgement - ServerType: "; 
									rxData = pollRxQueue();
									if (rxData != null) {
										switch(rxData.intValue()) {
											case SENTINEL_BASE_NODE:
												rxOutput += "Base Node";
												break;
											case SENTINEL_RELAY_NODE:
												rxOutput += "Relay Node";
												break;
											case SENTINEL_SENSOR_NODE:
												rxOutput += "Sensor Node";
												break;
										}
									}
									else {
										rxOutput += "Error";
									}
									rxOutput += ", OfferedAddr: [" + pollShortAsHex() + "], ServerAddr: [" + pollShortAsHex() + "]";
									break;
								case SENTINEL_LOCAL_DHCP_CONFIG:
									rxOutput += "Configuration";
									break;
								case SENTINEL_LOCAL_DHCP_COMPLETE:
									rxOutput += "Complete - Depth: [" + pollByteAsInt() + "], LocalAddr: [" + pollShortAsHex() + "], LocalPanID: [" + pollShortAsHex() + "], ServerAddr: [" + pollShortAsHex() + "], ServerPanID: [" + pollShortAsHex() + "]";
									break;
								case SENTINEL_LOCAL_DHCP_RX_ACK_TIMEOUT:
									rxOutput += "RX Timeout";
									break;
								case SENTINEL_LOCAL_DHCP_RX_OFFER_TIMEOUT:
									rxOutput += "Offer Timeout";
									break; 
								default:
									rxOutput += "Error";
									break;
							}
						}
						else {
							rxOutput += " Error";
						}
						break;
					case UART_RX_SENTINEL_SENSOR_DETECTION_RELAY:
						rxOutput = "sensorDetectionRelay - SensorAddr: [" + pollShortAsHex() + "], SrcAddr: [" + pollShortAsHex() + "], SrcPanID: [" + pollShortAsHex() + "]";
						break;
					case UART_RX_SENTINEL_SENSOR_DETECTION_REPORT:
						rxOutput = pollShortAsHex();
						sentinel.sensorDetection(rxOutput);
						rxOutput = "sensorDetectionReport - SensorAddr: [" + rxOutput + "]";
						break;	
					case UART_RX_SENTINEL_DISCOVER:
						rxOutput = "discover: ";
						rxData = pollRxQueue();
						if (rxData != null) {
							switch(rxData.intValue()) {
								case SENTINEL_BASE_NODE:
									rxOutput += "Base Node";
									break;
								case SENTINEL_RELAY_NODE:
									rxOutput += "Relay Node";
									break;
								case SENTINEL_SENSOR_NODE:
									rxOutput += "Sensor Node";
									break;
							}
						}
						else {
							rxOutput += "Error";
						}
						break;
					case UART_RX_SENTINEL_INFO:
						rxOutput = "info: ";
						rxData = pollRxQueue();
						if (rxData != null) {
							switch(rxData.intValue()) {
								case SENTINEL_BASE_NODE:
									rxOutput += "Base Node - ";
									rxOutput += "OwnAddr: [" + pollShortAsHex() + "], OwnPanID: [" + pollShortAsHex() + "], ";
									rxOutput += "#ofRelayClients: [" + pollByteAsInt() + "], #ofSensorClients: [" + pollByteAsInt() + "]";
									break;
								case SENTINEL_RELAY_NODE:
									rxOutput += "Relay Node - ";
									rxOutput += "OwnDepth: [" + pollByteAsInt() + "], OwnAddr: [" + pollShortAsHex() + "], OwnPanID: [" + pollShortAsHex() + "], ";
									rxOutput += "ServerAddr: [" + pollShortAsHex() + "], ServerPanID: [" + pollShortAsHex() + "], ";
									rxOutput += "#ofRelayClients: [" + pollByteAsInt() + "], #ofSensorClients: [" + pollByteAsInt() + "]";
									break;
								case SENTINEL_SENSOR_NODE:
									rxOutput += "Sensor Node - ";
									rxOutput += "OwnDepth: [" + pollByteAsInt() + "], OwnAddr: [" + pollShortAsHex() + "], OwnPanID: [" + pollShortAsHex() +"], ";
									rxOutput += "ServerAddr: [" + pollShortAsHex() + "], ServerPanID: [" + pollShortAsHex() +"]";
									break;
							}
						}
						else {
							rxOutput += "Error";
						}
						break;
					case UART_RX_ZIGBEE_STATE:							// ZIGBEE_STATE
						rxOutput = "state: ";
						rxData = pollRxQueue();
						if (rxData != null) {
							switch(rxData.intValue()) {
								case ZIGBEE_TRX_STATUS_P_ON:
									rxOutput += "P_ON";
									break;
								case ZIGBEE_TRX_STATUS_TRX_OFF:
									rxOutput += "TRX_OFF";
									break;
								case ZIGBEE_TRX_STATUS_SLEEP:
									rxOutput += "SLEEP";
									break;
								case ZIGBEE_TRX_STATUS_PLL_ON:
									rxOutput += "PLL_ON";
									break;
								case ZIGBEE_TRX_STATUS_BUSY_TX:
									rxOutput += "BUSY_TX";
									break;
								case ZIGBEE_TRX_STATUS_RX_ON:
									rxOutput += "RX_ON";
									break;
								case ZIGBEE_TRX_STATUS_BUSY_RX:
									rxOutput += "BUSY_RX";
									break;
								case ZIGBEE_TRX_STATUS_RX_ON_NOCLK:
									rxOutput += "RX_ON_NOCLK";
									break;
								case ZIGBEE_TRX_STATUS_BUSY_RX_AACK:
									rxOutput += "BUSY_RX_AACK";
									break;
								case ZIGBEE_TRX_STATUS_BUSY_TX_ARET:
									rxOutput += "BUSY_TX_ARET";
									break;
								case ZIGBEE_TRX_STATUS_RX_AACK_ON:
									rxOutput += "RX_AACK_ON";
									break;
								case ZIGBEE_TRX_STATUS_TX_ARET_ON:
									rxOutput += "TX_ARET_ON";
									break;
								case ZIGBEE_TRX_STATUS_RX_AACK_NOCLK:
									rxOutput += "RX_AACK_NOCLK";
									break;
								case ZIGBEE_TRX_STATUS_BUSY_RX_AACK_NOCLK:
									rxOutput += "BUSY_RX_AACK_NOCLK";
									break;
								case ZIGBEE_TRX_STATUS_TRANSITION:
									rxOutput += "TRANSITION";
									break;
								default:
									rxOutput += "Unknown";
									break;
							}
						}
						else {
							rxOutput += "Error";
						}
						break;
					case UART_RX_ZIGBEE_IRQ_STATUS:						// IRQ_STATUS
						rxOutput = "irqStatus: ";
						rxData = pollRxQueue();
						if (rxData != null) {
							rxOutput += "0b" + Data.intToBinary(rxData.intValue()) + " ";
							if (Data.intToBinary(rxData.intValue()).charAt(7) == '1') {
								rxOutput += "- PLL Lock";
							}
							if (Data.intToBinary(rxData.intValue()).charAt(6) == '1') {
								rxOutput += "- PLL Unlock";
							}
							if (Data.intToBinary(rxData.intValue()).charAt(5) == '1') {
								rxOutput += "- RX Start";
							}
							if (Data.intToBinary(rxData.intValue()).charAt(4) == '1') {
								rxOutput += "- TRX End";
							}
							if (Data.intToBinary(rxData.intValue()).charAt(1) == '1') {
								rxOutput += "- TRX UR";
							}
							if (Data.intToBinary(rxData.intValue()).charAt(0) == '1') {
								rxOutput += "- BAT Low";
							}
						}
						else {
							rxOutput += "Error";
						}
						break;
					case UART_RX_ZIGBEE_TRAC_STATUS:					// TRAC_STATUS
						rxOutput = "tracStatus: ";
						rxData = pollRxQueue();
						if (rxData != null) {
							switch(rxData.intValue()) {
								case 0:
									rxOutput += "Success";
									break;
								case 1:
									rxOutput += "Success Data Pending";
									break;
								case 3:
									rxOutput += "Success Access Failure";
									break;
								case 5:
									rxOutput += "No Ack";
									break;
								case 7:
									rxOutput += "Invalid";
									break;
								default:
									rxOutput += "Unknown";
									break;
							}
						}
						else {
							rxOutput += "Error";
						}
						break;
					case UART_RX_ZIGBEE_PHY_DATA:						// PHY_DATA
						rxOutput = "phyData";
						rxLength = pollRxQueue();
						if (rxLength != null) {
							rxOutput += "[" + rxLength.intValue() + "]: " + pollByteArrayAsHex(rxLength.intValue());
						}
						else {
							rxOutput += ": ERROR";
						}
						break;
					case UART_RX_ZIGBEE_PHY_LQI:					// PHY_LQI
						rxOutput = "phyLQI: " + pollByteAsInt();
						break;
					case UART_RX_ZIGBEE_MAC_DATA:
						rxOutput = "macData";
						rxLength = pollRxQueue();
						if (rxLength != null) {
							rxOutput += "[" + rxLength.intValue() + "]: " + pollByteArrayAsHex(rxLength.intValue());
						}
						else {
							rxOutput += ": ERROR";
						}
						break;
					case UART_RX_ZIGBEE_MAC_DEST_PAN_ID:
						rxOutput = "macDestPanID: " + pollShortAsHex();
						break;
					case UART_RX_ZIGBEE_MAC_DEST_ADDR:
						rxOutput = "macDestAddr: ";
						rxLength = pollRxQueue();
						if (rxLength != null) {
							switch (rxLength.intValue()) {
								case 2:
									rxOutput += pollShortAsHex();
									break;
								case 8:
									rxOutput += pollExtendedAsHex();
									break;
								default:
									rxOutput += "ERROR";
									break;
							}
						}
						break;
					case UART_RX_ZIGBEE_MAC_SRC_PAN_ID:
						rxOutput = "macSrcPanID: " + pollShortAsHex();
						break;
					case UART_RX_ZIGBEE_MAC_SRC_ADDR:
						rxOutput = "macSrcAddr: ";
						rxLength = pollRxQueue();
						if (rxLength != null) {
							switch (rxLength.intValue()) {
								case 2:
									rxOutput += pollShortAsHex();
									break;
								case 8:
									rxOutput += pollExtendedAsHex();
									break;
								default:
									rxOutput += "ERROR";
									break;
							}
						}
						break;
					case UART_RX_ZIGBEE_ED_LEVEL:
						rxOutput = "edLevel: " + pollByteAsInt();
						break;
					case UART_RX_ZIGBEE_PART_NUM:
						rxOutput = "partNum: " + pollByteAsHex();
						break;
					case UART_RX_ZIGBEE_VER_NUM:
						rxOutput = "verNum: " + pollByteAsHex();
						break;
					case UART_RX_ZIGBEE_MAN_ID_0:
						rxOutput = "manID0: " + pollByteAsHex();
						break;
					case UART_RX_ZIGBEE_MAN_ID_1:
						rxOutput = "manID1: " + pollByteAsHex();
						break;
					default:
						rxOutput = "Unknown";
						break;
				}
				println(rxOutput);
			}
		}
	}
	
	/**
	 * This method polls the Serial queue for a byte array, starting the with the length,
	 * and then outputting it as a String of hex
	 * 
	 * @return String
	 */
	private String pollByteArrayAsHex() {
		Integer rxLength = pollRxQueue();
		if (rxLength != null) {
			return pollByteArrayAsHex(rxLength.intValue());
		}
		else {
			return " ERROR";
		}
	}
	
	/**
	 * This method polls the Serial queue for a byte array of the given length,
	 * and then outputting it as a String of hex
	 * 
	 * @param rxLength
	 * @return String
	 */
	private String pollByteArrayAsHex(int rxLength) {
		Integer rxData;
		String rxOutput = "";
		for (int i=0; i<rxLength; i++) {
			rxData = pollRxQueue();
			if (rxData != null) {
				rxOutput += "0x" + Data.intToHex(rxData.intValue());
				if (i < (rxLength-1)) {
					rxOutput += " ";
				}
			}
			else {
				rxOutput += "ERROR";
				i = rxLength;
			}
		}
		return rxOutput;
	}
	
	/**
	 * This method polls the Serial queue for a byte and then output it in decimal form in String
	 * 
	 * @return String
	 */
	private String pollByteAsInt() {
		Integer rxData = pollRxQueue();
		String rxOutput = null;
		if (rxData != null) {
			rxOutput = "" + rxData.intValue();
		}
		else {
			rxOutput = "Error";
		}
		return rxOutput;	
	}
	
	/**
	 * This method polls the Serial queue for a byte and then output it in hex form in String
	 * 
	 * @return String
	 */
	private String pollByteAsHex() {
		Integer rxData = pollRxQueue();
		String rxOutput = null;
		if (rxData != null) {
			rxOutput = "0x" + Data.intToHex(rxData.intValue());
		}
		else {
			rxOutput = "Error";
		}
		return rxOutput;	
	}
	
	/**
	 * This method polls the Serial queue for a short (16-bit) and then output it in hex form in String
	 * 
	 * @return String
	 */
	private String pollShortAsHex() {
		Integer rxData = pollRxQueue();
		String rxOutput;
		if (rxData != null) {
			rxOutput = Data.intToHex(rxData.intValue());
			rxData = pollRxQueue();
			if (rxData != null) {
				rxOutput = "0x" + Data.intToHex(rxData.intValue()) + rxOutput;
			}
			else {
				rxOutput = "ERROR";
			}
		}
		else {
			rxOutput = "Error";
		}
		return rxOutput;	
	}
	
	/**
	 * This method polls the Serial queue for an extended (64-bit) and then output it in hex form in String
	 * 
	 * @return String
	 */
	private String pollExtendedAsHex() {
		Integer rxData;
		String rxOutput = "";
		for (int i=0; i<8; i++) {
			rxData = pollRxQueue();
			if (rxData != null) {
				rxOutput = Data.intToHex(rxData.intValue()) + rxOutput;
			}
			else {
				rxOutput += " ERROR";
				i = 8;
			}
		}
		rxOutput = "0x" + rxOutput;
		return rxOutput;
	}
	
	/**
	 * This method polls the Serial queue for a byte and then output it as an integer
	 * 
	 * @return Integer
	 */
	public Integer pollRxQueue() {
		Integer data = null;
		try {
			data = (Integer) serial.getRxQueue().poll(1000, TimeUnit.MILLISECONDS);
		}
		catch (InterruptedException e) {
			System.err.println(e);
		}
		if (data != null && serialDebug) {
			println("RX: " + Data.intToHex(data.intValue()));
		}
		return data;
	}
	
	/**
	 * This method peeks the Serial queue for a byte and then output it as an integer
	 * 
	 * @return Integer
	 */
	public Integer peekRxQueue() {
		Integer data = null;
		data = (Integer) serial.getRxQueue().peek();
		if (data != null && serialDebug) {
			println("RX: " + Data.intToHex(data.intValue()));
		}
		return data;
	}
	
	/**
	 * This method sends a discover request to the attached network node
	 *
	 */
	public void requestDiscover() {
		serial.sendByte((byte)UART_TX_SENTINEL_DISCOVER);
		println("requestDiscover");
	}
	
	/**
	 * This method sends a info request to the attached network node
	 *
	 */
	public void requestInfo() {
		serial.sendByte((byte)UART_TX_SENTINEL_INFO);
		println("requestInfo");
	}
	
	/**
	 * This method sends a request of clearing the all lease list to the attached network node
	 *
	 */
	public void requestClearClientList() {
		serial.sendByte((byte)UART_TX_SENTINEL_CLEAR_CLIENT_LIST);
		println("requestClearClientList");
	}
	
	/**
	 * This method sends a request of clearing relay lease list to the attached network node
	 *
	 */
	public void requestClearRelayList() {
		serial.sendByte((byte)UART_TX_SENTINEL_CLEAR_RELAY_LIST);
		println("requestClearRelayList");
	}
	
	/**
	 * This method sends a request of clearing sensor lease list to the attached network node
	 *
	 */
	public void requestClearSensorList() {
		serial.sendByte((byte)UART_TX_SENTINEL_CLEAR_SENSOR_LIST);
		println("requestClearSensorList");
	}
	
	/**
	 * This method sends a request of reset the local DHCP to the attached network node
	 *
	 */
	public void requestClearLocalDHCP() {
		serial.sendByte((byte)UART_TX_SENTINEL_CLEAR_LOCAL_DHCP);
		println("requestClearLocalDHCP");
	}
	
	/**
	 * This method enables serial debug for the communicator
	 *
	 */
	public void enableSerialDebug() {
		serialDebug = true;
	}
	
	/**
	 * This method disables serial debug for the communicator
	 *
	 */
	public void disableSerialDebug() {
		serialDebug = false;
	}
	
	/**
	 * This method returns whether the serial debug has been enabled
	 * 
	 * @return boolean
	 */
	public boolean getSerialDebug() {
		return serialDebug;
	}
	
	/**
	 * This method prints out a line of String either on the Sentinel UI textArea or the Java
	 * console terminal if the Sentinel class is unavailable
	 * 
	 * @param str
	 */
	public void println(String str) {
		if (sentinel != null) {
			sentinel.println(str);
		}
		else {
			System.out.println(str);
		}
	}
	
	/**
	 * This method initializes the Communicator class without a Sentinel class
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		Communicator communicator = new Communicator(null);	
		communicator.start();
	}
	
}