import gnu.io.CommPortIdentifier;
import gnu.io.CommPort;
import gnu.io.PortInUseException;
import gnu.io.SerialPort;
import gnu.io.SerialPortEvent;
import gnu.io.SerialPortEventListener;
import gnu.io.UnsupportedCommOperationException;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.TooManyListenersException;
import java.util.Vector;

import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import javax.swing.JOptionPane;

/**
 * The class reponsible for the local serial port, such as listening to it and
 * sending data to it.
 * 
 * @author Ruibing Wang
 * 
 */
public class Serial implements SerialPortEventListener {
	private LinkedBlockingQueue rxQueue = null;

	/**
	 * Serial Port Config along with all the possible options possible
	 */
	private SerialPort serialPort = null;
	// Linux/Mac/Unix - /dev/tty*
	// Win - COM*
	private String comPort = null;
	// SerialPort.PARITY_NONE
	// SerialPort.PARITY_ODD
	// SerialPort.PARITY_EVEN
	// SerialPort.PARITY_MARK
	// SerialPort.PARITY_SPACE
	private int parity = SerialPort.PARITY_NONE;
	// SerialPort.STOPBITS_1
	// SerialPort.STOPBITS_2
	private int stopBits = SerialPort.STOPBITS_1;
	// 2400
	// 9600
	// 19200
	// 38400
	// 57600
	// 115200
	private int baudRate = 4800;
	// SerialPort.DATABITS_8
	// SerialPort.DATABITS_7
	// SerialPort.DATABITS_6
	// SerialPort.DATABITS_5
	private int dataBits = SerialPort.DATABITS_8;
	// Declaring local variable
	private InputStream inputStream = null;
	private OutputStream outputStream = null;

	/**
	 * Main constructor responsible for selecting the local com port, taking
	 * over it, and starting the queue.
	 * 
	 */
	public Serial() {
		System.out.println("Starting Serial Module...");

		rxQueue = new LinkedBlockingQueue();
		
		comPort = selectSerialPort("Select Serial Port");
		updateSerialPort();

		System.out.println("...Serial Module Ready\n");
	}

	/**
	 * Serial event listener responsible for decrypting serial port evetns. Only
	 * really care about when data becomes available.
	 * 
	 * @param event - the event that was just detected
	 */
	public void serialEvent(SerialPortEvent event) {
		switch (event.getEventType()) {

		case SerialPortEvent.BI:
		case SerialPortEvent.OE:
		case SerialPortEvent.FE:
		case SerialPortEvent.PE:
		case SerialPortEvent.CD:
		case SerialPortEvent.CTS:
		case SerialPortEvent.DSR:
		case SerialPortEvent.RI:
			System.out.println("Bad data received. Ignoring the most recent received data packet");
			break;
		case SerialPortEvent.OUTPUT_BUFFER_EMPTY:
			break;
		case SerialPortEvent.DATA_AVAILABLE:
			try {
				while (inputStream.available() > 0) {
					int i = inputStream.read();
					//System.out.println("RX: " + Data.intToHex(i));
					try {
						rxQueue.put(new Integer(i));
					}
					catch (InterruptedException e) {
						System.err.println(e);
					}
				}
			} 
			catch (IOException e) {
				System.err.println("IO Exception occurred while trying to read the data from inputStream.");
				System.err.println(e);
			}
			break;
		}
	}

	/**
	 * Send a string to the serial port
	 * 
	 * @param str - a text String
	 */
	public void sendString(String str) {
		try {
			outputStream.write(str.getBytes());
		} catch (IOException e) {
			System.err.println(e);
		}

	}

	/**
	 * Send a byte to the serial port
	 * 
	 * @param b - a byte
	 */
	public void sendByte(byte b) {
		try {
			outputStream.write(b);
		} catch (IOException e) {
			System.err.println(e);
		}

	}

	/**
	 * Send a byte array to the serial port using a new instance everything
	 *
	 * @param b - a byte array
	 */
	public void sendByteArray(byte[] b) {
		try {
			outputStream.write(b);
		} catch (IOException e) {
			System.out.println("Error writing to outputStream");
			System.err.println(e);
		}
	}

	/**
	 * Sets the current com port and then updates the program to reflect that change.
	 * 
	 * @param newCPort - a string of the name of the new com port
	 */
	public void setComPort(String newCPort) {
		comPort = newCPort;
		updateSerialPort();
	}

	/**
	 * Returns the current com port
	 * 
	 * @return the name of current com port as a String
	 */
	public String getComPort() {
		return comPort;
	}

	/**
	 * Sets the current parity and then updates the program to reflect that change.
	 * 
	 * @param newParity - the new parity
	 */
	public void setParity(int newParity) {
		parity = newParity;
		updateSerialPort();
	}

	/**
	 * Returns the current parity of the com port
	 * 
	 * @return the parity of com port
	 */
	public int getParity() {
		return parity;
	}

	/**
	 * Sets the current stopbits and then updates the program to reflect that change.
	 * 
	 * @param newStopBits - the new stopbits
	 */
	public void setStopBits(int newStopBits) {
		stopBits = newStopBits;
		updateSerialPort();
	}
	
	/**
	 * Returns the current stopbits of the com port
	 * 
	 * @return the stopbits of com port
	 */
	public int getStopBits() {
		return stopBits;
	}

	/**
	 * Sets the current baudrate and then updates the program to reflect that change.
	 * 
	 * @param newBRate - the new baudrate
	 */
	public void setBaudRate(int newBRate) {
		baudRate = newBRate;
		updateSerialPort();
	}

	/**
	 * Returns the current baudrate of the com port
	 * 
	 * @return the baudrate of com port
	 */
	public int getBaudRate() {
		return baudRate;
	}

	/**
	 * Sets the current databits and then updates the program to reflect that change.
	 * 
	 * @param newDBits - the new databits
	 */
	public void setDataBits(int newDBits) {
		dataBits = newDBits;
		updateSerialPort();
	}

	/**
	 * Returns the current databits of the com port
	 * 
	 * @return the databits of com port
	 */
	public int getDataBits() {
		return dataBits;
	}

	/**
	 * Updates the serial port with these specific settings and then calls to
	 * updateSerialPort()
	 * 
	 * @param cPort - string representation of com port
	 * @param parity - parity
	 * @param sBits - stopbits
	 * @param bRate - baudrate
	 * @param dBits - databits
	 */
	public void updateSerialPort(String cPort, int parity, int sBits,
			int bRate, int dBits) {
		this.comPort = cPort;
		this.parity = parity;
		this.stopBits = sBits;
		this.baudRate = bRate;
		this.dataBits = dBits;
		updateSerialPort();
	}

	/**
	 * Updates the serial port with the current settings and retrieves
	 * appropriate information about the port
	 * 
	 */
	public void updateSerialPort() {

		CommPortIdentifier portId = null;
		Enumeration portList = null;
		boolean portFound = false;

		if (serialPort != null) {
			serialPort.close();
		}
		
		// Retrieve a list of the ports and loop through them until the com port
		// we want is found
		// TODO: This might be simplified since we already know the name of the
		// one we want and we know it is already available
		portList = CommPortIdentifier.getPortIdentifiers();
		while (portList.hasMoreElements() && !portFound) {
			portId = (CommPortIdentifier) portList.nextElement();
			
			if (portId.getPortType() == CommPortIdentifier.PORT_SERIAL) {
				// Tests whether this is the com port we want
				if (portId.getName().equals(comPort)) {
					portFound = true;

					/* now put a thread to listen to the Serial port's events */
					try {
						if (serialPort != null) {
							serialPort.close();
						}
						serialPort = (SerialPort) portId.open("topobocom",
								10000); // wait up to 10 sec for the port to open

						// set the Serial Port parameters
						serialPort.setSerialPortParams(baudRate, dataBits, stopBits, parity);

						// set teh flow control to hardware - none
						serialPort.setFlowControlMode(SerialPort.FLOWCONTROL_NONE);

						// get the streams associated with the serial port
						inputStream = serialPort.getInputStream();
						outputStream = serialPort.getOutputStream();

						// start listening for events
						serialPort.addEventListener(this);
						serialPort.notifyOnDataAvailable(true);

					} catch (PortInUseException e) {
						System.out.println("Serial Port already in use.");
						System.err.println(e);
					} catch (IOException e) {
						System.out.println("An IO Exception occurred.");
						System.err.println(e);
					} catch (TooManyListenersException e) {
						System.out.println("too many listeners.");
						System.err.println(e);
					} catch (UnsupportedCommOperationException e) {
						System.out.println("Some Unknown Comm Exception Occurred.");
						System.err.println(e);
					}
				}
			}
		}
		if (!portFound) {
			System.out.println("Port " + comPort + " was not found!");
		}

	}

	/**
	 * Returns an array of String (though it is uncasted as an Object[]) that
	 * contains a list of comports that are serial.
	 * TODO: Filter through tty.cu, bluetooth, etc
	 * 
	 * @param available -
	 *            a boolean that determines whether we only return a list
	 *            containing available ports or not.
	 * @return a list of ports
	 */
	public Object[] listPorts(boolean available) {
		Vector result = new Vector();
		Enumeration portEnum = CommPortIdentifier.getPortIdentifiers();
		while (portEnum.hasMoreElements()) {
			CommPortIdentifier com = (CommPortIdentifier) portEnum
					.nextElement();
			if (available) {
				switch (com.getPortType()) {
				case CommPortIdentifier.PORT_SERIAL:
					try {
						// Test open the port to see if it's available for 50ms
						CommPort thePort = com.open("commtest", 50);
						thePort.close();
						result.add(new String(com.getName() + " - "
								+ getPortTypeName(com.getPortType())));
					} 
					catch (PortInUseException e) {
						System.err.println(e);
					} 
					catch (Exception e) {
						System.err.println(e);
					}
				}
			} else {
				result.add(new String(com.getName() + " - "
						+ getPortTypeName(com.getPortType())));
			}
		}
		return result.toArray();
	}

	/**
	 * Returns the port type of the com port
	 * TODO: Can be removed.
	 * 
	 * @param portType - an int representation of the port type
	 * @return - a String representation of the port type (serial, i2c, etc)
	 */
	public String getPortTypeName(int portType) {
		switch (portType) {
		case CommPortIdentifier.PORT_I2C:
			return "I2C";
		case CommPortIdentifier.PORT_PARALLEL:
			return "Parallel";
		case CommPortIdentifier.PORT_RAW:
			return "Raw";
		case CommPortIdentifier.PORT_RS485:
			return "RS485";
		case CommPortIdentifier.PORT_SERIAL:
			return "Serial";
		default:
			return "unknown type";
		}
	}

	/**
	 * Allows the user to select a serial com port from a list of available ones
	 * 
	 * @param msg - message to be displayed to the user when asking them to select
	 * @return the String name of the serial com port
	 */
	private String selectSerialPort(String msg) {
		Object[] list = listPorts(true);
		String port = null;
		if (list.length > 0) {
			port = (String) JOptionPane.showInputDialog(null, msg, "Input",
				JOptionPane.INFORMATION_MESSAGE, null, list, list[0]);
		}
		else {
			System.out.println("Must connect a serial port");
			System.exit(0);
		}
		if (port == null) {
			System.out.println("Exiting");
			System.exit(0);
		}
		port = port.substring(0, port.indexOf(" - "));
		System.out.println(port);
		return port;
	}
	
	public LinkedBlockingQueue getRxQueue() {
		return rxQueue;
	}
	
	public static void main(String[] args) {
		Serial serial = new Serial();	
	}
}
